/* vi: set sw=4 ts=4: */
/*
 * Mini syslogd implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Copyright (C) 2000 by Karl M. Hegbloom <karlheg@debian.org>
 *
 * "circular buffer" Copyright (C) 2001 by Gennady Feldman <gfeldman@gena01.com>
 *
 * Maintainer: Gennady Feldman <gfeldman@gena01.com> as of Mar 12, 2001
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */

//usage:#define syslogd_trivial_usage
//usage:       "[OPTIONS]"
//usage:#define syslogd_full_usage "\n\n"
//usage:       "System logging utility\n"
//usage:	IF_NOT_FEATURE_SYSLOGD_CFG(
//usage:       "(this version of syslogd ignores /etc/syslog.conf)\n"
//usage:	)
//usage:     "\n	-n		Run in foreground"
//usage:	IF_FEATURE_REMOTE_LOG(
//usage:     "\n	-R HOST[:PORT]	Log to HOST:PORT (default PORT:514)"
//usage:     "\n	-L		Log locally and via network (default is network only if -R)"
//usage:	)
//usage:	IF_FEATURE_IPC_SYSLOG(
/* NB: -Csize shouldn't have space (because size is optional) */
//usage:     "\n	-C[size_kb]	Log to shared mem buffer (use logread to read it)"
//usage:	)
//usage:	IF_FEATURE_KMSG_SYSLOG(
//usage:     "\n	-K		Log to kernel printk buffer (use dmesg to read it)"
//usage:	)
//usage:     "\n	-O FILE		Log to FILE (default:/var/log/messages, stdout if -)"
//usage:	IF_FEATURE_ROTATE_LOGFILE(
//usage:     "\n	-s SIZE		Max size (KB) before rotation (default:200KB, 0=off)"
//usage:     "\n	-b N		N rotated logs to keep (default:1, max=99, 0=purge)"
//usage:	)
//usage:     "\n	-l N		Log only messages more urgent than prio N (1-8)"
//usage:     "\n	-S		Smaller output"
//usage:	IF_FEATURE_SYSLOGD_DUP(
//usage:     "\n	-D		Drop duplicates"
//usage:	)
//usage:	IF_FEATURE_SYSLOGD_CFG(
//usage:     "\n	-f FILE		Use FILE as config (default:/etc/syslog.conf)"
//usage:	)
/* //usage:  "\n	-m MIN		Minutes between MARK lines (default:20, 0=off)" */
//usage:
//usage:#define syslogd_example_usage
//usage:       "$ syslogd -R masterlog:514\n"
//usage:       "$ syslogd -R 192.168.1.1:601\n"

/*
 * Done in syslogd_and_logger.c:
#include "libbb.h"
#define SYSLOG_NAMES
#define SYSLOG_NAMES_CONST
#include <syslog.h>
*/
#ifndef _PATH_LOG
#define _PATH_LOG	"/dev/log"
#endif

#include <sys/un.h>
#include <sys/uio.h>

#if ENABLE_FEATURE_REMOTE_LOG
#include <netinet/in.h>
#endif

#if ENABLE_FEATURE_IPC_SYSLOG
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#endif

#include <net/if.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <arpa/inet.h>

#include <pthread.h>
#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <nvram.h>
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
#include <dlfcn.h>
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */


#define SYSLOG_PORT_TLS         6514
#define SYSLOG_PORT_UDP         514
#define SSL_CONNECT_TIME_OUT    10
#define DELAY_MULTIPLYER        30
#define ENABLE_BB_MSG           1

#define PVALUE_LOAD_CA_CERT         "8502"
#define PVALUE_VERIFY_SERVER        "8463"
#define PVALUE_DISABLE_WEAK_CIPHER  "8536"
#define PVALUE_WAN_DEVICE           "wan_if"
#define TLS_CA_DIR                  "/etc/pki/tls/"
#define TLS_CA_FILE                 "/tmp/ssl/ca_certs.pem"
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
#define LIBSSL                      "/usr/lib/libssl.so"
#define LIBCRYPTO                   "/usr/lib/libcrypto.so"
#define LIBPTHREAD                  "/lib/libpthread.so"
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
typedef int (*SSL_library_init_t)(void);
typedef int (*SSL_set_fd_t)(SSL *s, int fd);
typedef int (*SSL_connect_t)(SSL *ssl);
typedef SSL *(*SSL_new_t)(SSL_CTX *);
typedef int (*SSL_shutdown_t)(SSL *);
typedef void (*SSL_free_t)(SSL *);
typedef int (*SSL_read_t)(SSL *, void *, int);
typedef int (*SSL_write_t)(SSL *, const void *, int);
typedef int (*SSL_get_error_t)(const SSL *, int);
typedef void (*SSL_CTX_free_t)(SSL_CTX *);
typedef SSL_CTX *(*SSL_CTX_new_t)(const SSL_METHOD *);
typedef const SSL_METHOD *(*SSLv23_method_t)(void);
typedef int (*SSL_CTX_load_verify_locations_t)(SSL_CTX *, const char *, const char *);
typedef void (*SSL_CTX_set_verify_t)(SSL_CTX *, int, int (*) (int, X509_STORE_CTX *));
typedef long (*SSL_CTX_ctrl_t)(SSL_CTX *ctx, int cmd, long larg, void *parg);
typedef void (*OpenSSL_add_all_algorithms_t)(void);
typedef int (*pthread_mutexattr_init_t)(pthread_mutexattr_t *);
typedef int (*pthread_mutexattr_destroy_t)(pthread_mutexattr_t *);
typedef int (*pthread_mutexattr_settype_t)(pthread_mutexattr_t *, int);
typedef int (*pthread_mutex_lock_t)(pthread_mutex_t *);
typedef int (*pthread_mutex_unlock_t)(pthread_mutex_t *);
typedef int (*pthread_mutex_init_t)(pthread_mutex_t *, const pthread_mutexattr_t *);
typedef int (*pthread_mutex_destroy_t)(pthread_mutex_t *);
typedef int (*pthread_create_t)(pthread_t *, const pthread_attr_t *, void *(*) (void *), void *);

static void *ssl_handle = NULL;
static void *crypto_handle = NULL;
static void *pthread_handle = NULL;
static SSL_library_init_t SSL_library_init_p = NULL;
static SSL_set_fd_t SSL_set_fd_p = NULL;
static SSL_connect_t SSL_connect_p = NULL;
static SSL_new_t SSL_new_p = NULL;
static SSL_shutdown_t SSL_shutdown_p = NULL;
static SSL_free_t SSL_free_p = NULL;
static SSL_read_t SSL_read_p = NULL;
static SSL_write_t SSL_write_p = NULL;
static SSL_get_error_t SSL_get_error_p = NULL;
static SSL_CTX_free_t SSL_CTX_free_p = NULL;
static SSL_CTX_new_t SSL_CTX_new_p = NULL;
static SSLv23_method_t SSLv23_method_p = NULL;
static SSL_CTX_load_verify_locations_t SSL_CTX_load_verify_locations_p = NULL;
static SSL_CTX_set_verify_t SSL_CTX_set_verify_p = NULL;
static SSL_CTX_ctrl_t SSL_CTX_ctrl_p = NULL;
#define SSL_CTX_set_cb_cipher_get_disabled_p(ctx,op) \
        if ( SSL_CTX_ctrl_p ) SSL_CTX_ctrl_p((ctx),SSL_CTRL_SET_CB_CIPHER_GET_DISABLED,(op),NULL)
#define SSL_CTX_rebuild_cipher_list_p(ctx) \
        if ( SSL_CTX_ctrl_p ) SSL_CTX_ctrl_p((ctx),SSL_CTRL_REBUILD_CIPHER_LIST,0,NULL)
static OpenSSL_add_all_algorithms_t OpenSSL_add_all_algorithms_p = NULL;
static pthread_mutexattr_init_t pthread_mutexattr_init_p = NULL;
static pthread_mutexattr_destroy_t pthread_mutexattr_destroy_p = NULL;
static pthread_mutexattr_settype_t pthread_mutexattr_settype_p = NULL;
static pthread_mutex_lock_t pthread_mutex_lock_p = NULL;
static pthread_mutex_unlock_t pthread_mutex_unlock_p = NULL;
static pthread_mutex_init_t pthread_mutex_init_p = NULL;
static pthread_mutex_destroy_t pthread_mutex_destroy_p = NULL;
static pthread_create_t pthread_create_p = NULL;
#else
static SSL_library_init_t SSL_library_init_p = SSL_library_init;
static SSL_set_fd_t SSL_set_fd_p = SSL_set_fd;
static SSL_connect_t SSL_connect_p = SSL_connect;
static SSL_new_t SSL_new_p = SSL_new;
static SSL_shutdown_t SSL_shutdown_p = SSL_shutdown;
static SSL_free_t SSL_free_p = SSL_free;
static SSL_read_t SSL_read_p = SSL_read;
static SSL_write_t SSL_write_p = SSL_write;
static SSL_get_error_t SSL_get_error_p = SSL_get_error;
static SSL_CTX_free_t SSL_CTX_free_p = SSL_CTX_free;
static SSL_CTX_new_t SSL_CTX_new_p = SSL_CTX_new;
static SSLv23_method_t SSLv23_method_p = SSLv23_method;
static SSL_CTX_load_verify_locations_t SSL_CTX_load_verify_locations_p = SSL_CTX_load_verify_locations;
static SSL_CTX_set_verify_t SSL_CTX_set_verify_p = SSL_CTX_set_verify;
#define SSL_CTX_set_cb_cipher_get_disabled_p(ctx, op) SSL_CTX_set_cb_cipher_get_disabled(ctx, op)
#define SSL_CTX_rebuild_cipher_list_p(ctx) SSL_CTX_rebuild_cipher_list(ctx)
#define OpenSSL_add_all_algorithms_p() OpenSSL_add_all_algorithms()
static pthread_mutexattr_init_t pthread_mutexattr_init_p = pthread_mutexattr_init;
static pthread_mutexattr_destroy_t pthread_mutexattr_destroy_p = pthread_mutexattr_destroy;
static pthread_mutexattr_settype_t pthread_mutexattr_settype_p = pthread_mutexattr_settype;
static pthread_mutex_lock_t pthread_mutex_lock_p = pthread_mutex_lock;
static pthread_mutex_unlock_t pthread_mutex_unlock_p = pthread_mutex_unlock;
static pthread_mutex_init_t pthread_mutex_init_p = pthread_mutex_init;
static pthread_mutex_destroy_t pthread_mutex_destroy_p = pthread_mutex_destroy;
static pthread_create_t pthread_create_p = pthread_create;
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

#define DEBUG 0

/* MARK code is not very useful, is bloat, and broken:
 * can deadlock if alarmed to make MARK while writing to IPC buffer
 * (semaphores are down but do_mark routine tries to down them again) */
#undef SYSLOGD_MARK

/* Write locking does not seem to be useful either */
#undef SYSLOGD_WRLOCK

enum {
	MAX_READ = CONFIG_FEATURE_SYSLOGD_READ_BUFFER_SIZE,
	DNS_WAIT_SEC = 2 * 60,
};

/* Semaphore operation structures */
struct shbuf_ds {
	int32_t size;   /* size of data - 1 */
	int32_t tail;   /* end of message list */
	char data[1];   /* data/messages */
};

#if ENABLE_FEATURE_REMOTE_LOG
typedef struct {
	int remoteFD;
	unsigned last_dns_resolve;
	len_and_sockaddr *remoteAddr;
	const char *remoteHostname;
	/* bug 141499 Support SSL/TLS */
	int tls_con_status;
	int tls_retry;
	SSL_CTX* tls_ctx;
	SSL* tls_ssl;
	pthread_mutex_t cleanup_lock;
} remoteHost_t;
#endif

typedef struct logFile_t {
	const char *path;
	int fd;
	time_t last_log_time;
#if ENABLE_FEATURE_ROTATE_LOGFILE
	unsigned size;
	uint8_t isRegular;
#endif
} logFile_t;

#if ENABLE_FEATURE_SYSLOGD_CFG
typedef struct logRule_t {
	uint8_t enabled_facility_priomap[LOG_NFACILITIES];
	struct logFile_t *file;
	struct logRule_t *next;
} logRule_t;
#endif

/* Allows us to have smaller initializer. Ugly. */
#define GLOBALS \
	logFile_t logFile;                      \
	/* interval between marks in seconds */ \
	/*int markInterval;*/                   \
	/* level of messages to be logged */    \
	int logLevel;                           \
	/*bug 10817 Vonage */               \
	int user ;							\
	/* bug 141499 Support SSL/TLS */        \
	int tls;                                \
	pthread_t tls_disconnect_thread;        \
	pthread_t tls_reconnect_thread;         \
	pthread_t tls_linkstatus_thread;        \
IF_FEATURE_ROTATE_LOGFILE( \
	/* max size of file before rotation */  \
	unsigned logFileSize;                   \
	/* number of rotated message files */   \
	unsigned logFileRotate;                 \
) \
IF_FEATURE_IPC_SYSLOG( \
	int shmid; /* ipc shared memory id */   \
	int s_semid; /* ipc semaphore id */     \
	int shm_size;                           \
	struct sembuf SMwup[1];                 \
	struct sembuf SMwdn[3];                 \
) \
IF_FEATURE_SYSLOGD_CFG( \
	logRule_t *log_rules; \
) \
IF_FEATURE_KMSG_SYSLOG( \
	int kmsgfd; \
	int primask; \
)

struct init_globals {
	GLOBALS
};

struct globals {
	GLOBALS

#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *remoteHosts;
#endif
#if ENABLE_FEATURE_IPC_SYSLOG
	struct shbuf_ds *shbuf;
#endif
	/* localhost's name. We print only first 64 chars */
	char *hostname;

	/* We recv into recvbuf... */
	char recvbuf[MAX_READ * (1 + ENABLE_FEATURE_SYSLOGD_DUP)];
	/* ...then copy to parsebuf, escaping control chars */
	/* (can grow x2 max) */
	char parsebuf[MAX_READ*2];
	/* ...then sprintf into printbuf, adding timestamp (15 chars),
	 * host (64), fac.prio (20) to the message */
	/* (growth by: 15 + 64 + 20 + delims = ~110) */
	char printbuf[MAX_READ*2 + 128];
};

static const struct init_globals init_data = {
	.logFile = {
		.path = "/var/log/messages",
		.fd = -1,
	},
#ifdef SYSLOGD_MARK
	.markInterval = 20 * 60,
#endif
	.logLevel = 8,
#if ENABLE_FEATURE_ROTATE_LOGFILE
	.logFileSize = 200 * 1024,
	.logFileRotate = 1,
#endif
#if ENABLE_FEATURE_IPC_SYSLOG
	.shmid = -1,
	.s_semid = -1,
	.shm_size = ((CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE)*1024), /* default shm size */
	.SMwup = { {1, -1, IPC_NOWAIT} },
	.SMwdn = { {0, 0}, {1, 0}, {1, +1} },
#endif
	.user = 0,		/*Bug 10817 Vonage */
	.tls = 0, /* Bug 141499 */
};

#define G (*ptr_to_globals)
#define INIT_G() do { \
	SET_PTR_TO_GLOBALS(memcpy(xzalloc(sizeof(G)), &init_data, sizeof(init_data))); \
} while (0)


/* Options */
enum {
	OPTBIT_mark = 0, // -m
	OPTBIT_nofork, // -n
	OPTBIT_outfile, // -O
	OPTBIT_loglevel, // -l
	OPTBIT_small, // -S
	OPTBIT_user, // -U  Bug 10817 Vonage
	OPTBIT_TLS, // -T  Bug 141499
	IF_FEATURE_ROTATE_LOGFILE(OPTBIT_filesize   ,)	// -s
	IF_FEATURE_ROTATE_LOGFILE(OPTBIT_rotatecnt  ,)	// -b
	IF_FEATURE_REMOTE_LOG(    OPTBIT_remotelog  ,)	// -R
	IF_FEATURE_REMOTE_LOG(    OPTBIT_locallog   ,)	// -L
	IF_FEATURE_IPC_SYSLOG(    OPTBIT_circularlog,)	// -C
	IF_FEATURE_SYSLOGD_DUP(   OPTBIT_dup        ,)	// -D
	IF_FEATURE_SYSLOGD_CFG(   OPTBIT_cfg        ,)	// -f
	IF_FEATURE_KMSG_SYSLOG(   OPTBIT_kmsg       ,)	// -K

	OPT_mark        = 1 << OPTBIT_mark    ,
	OPT_nofork      = 1 << OPTBIT_nofork  ,
	OPT_outfile     = 1 << OPTBIT_outfile ,
	OPT_loglevel    = 1 << OPTBIT_loglevel,
	OPT_small       = 1 << OPTBIT_small   ,
	OPT_user        = 1 << OPTBIT_user,
	OPT_TLS         = 1 << OPTBIT_TLS   ,
	OPT_filesize    = IF_FEATURE_ROTATE_LOGFILE((1 << OPTBIT_filesize   )) + 0,
	OPT_rotatecnt   = IF_FEATURE_ROTATE_LOGFILE((1 << OPTBIT_rotatecnt  )) + 0,
	OPT_remotelog   = IF_FEATURE_REMOTE_LOG(    (1 << OPTBIT_remotelog  )) + 0,
	OPT_locallog    = IF_FEATURE_REMOTE_LOG(    (1 << OPTBIT_locallog   )) + 0,
	OPT_circularlog = IF_FEATURE_IPC_SYSLOG(    (1 << OPTBIT_circularlog)) + 0,
	OPT_dup         = IF_FEATURE_SYSLOGD_DUP(   (1 << OPTBIT_dup        )) + 0,
	OPT_cfg         = IF_FEATURE_SYSLOGD_CFG(   (1 << OPTBIT_cfg        )) + 0,
	OPT_kmsg        = IF_FEATURE_KMSG_SYSLOG(   (1 << OPTBIT_kmsg       )) + 0,
};
#define OPTION_STR "m:nO:l:SUT" \
	IF_FEATURE_ROTATE_LOGFILE("s:" ) \
	IF_FEATURE_ROTATE_LOGFILE("b:" ) \
	IF_FEATURE_REMOTE_LOG(    "R:" ) \
	IF_FEATURE_REMOTE_LOG(    "L"  ) \
	IF_FEATURE_IPC_SYSLOG(    "C::") \
	IF_FEATURE_SYSLOGD_DUP(   "D"  ) \
	IF_FEATURE_SYSLOGD_CFG(   "f:" ) \
	IF_FEATURE_KMSG_SYSLOG(   "K"  )
#define OPTION_DECL *opt_m, *opt_l \
	IF_FEATURE_ROTATE_LOGFILE(,*opt_s) \
	IF_FEATURE_ROTATE_LOGFILE(,*opt_b) \
	IF_FEATURE_IPC_SYSLOG(    ,*opt_C = NULL) \
	IF_FEATURE_SYSLOGD_CFG(   ,*opt_f = NULL)
#define OPTION_PARAM &opt_m, &(G.logFile.path), &opt_l \
	IF_FEATURE_ROTATE_LOGFILE(,&opt_s) \
	IF_FEATURE_ROTATE_LOGFILE(,&opt_b) \
	IF_FEATURE_REMOTE_LOG(    ,&remoteAddrList) \
	IF_FEATURE_IPC_SYSLOG(    ,&opt_C) \
	IF_FEATURE_SYSLOGD_CFG(   ,&opt_f)


#if ENABLE_BB_MSG
#define print_msg(s, ...) bb_error_msg(s, ##__VA_ARGS__);
#else
static void print_msg(const char *s, ...)
{
#define MSG_BUF_LEN 8192
	char buf[MSG_BUF_LEN] = {0};
	va_list p;

	va_start(p, s);
	vsnprintf(buf, MSG_BUF_LEN - 1, s, p);
	va_end(p);
	fprintf(stdout, "%s", buf);
}
#endif

#if ENABLE_FEATURE_SYSLOGD_CFG
static const CODE* find_by_name(char *name, const CODE* c_set)
{
	for (; c_set->c_name; c_set++) {
		if (strcmp(name, c_set->c_name) == 0)
			return c_set;
	}
	return NULL;
}
#endif
static const CODE* find_by_val(int val, const CODE* c_set)
{
	for (; c_set->c_name; c_set++) {
		if (c_set->c_val == val)
			return c_set;
	}
	return NULL;
}

#if ENABLE_FEATURE_SYSLOGD_CFG
static void parse_syslogdcfg(const char *file)
{
	char *t;
	logRule_t **pp_rule;
	/* tok[0] set of selectors */
	/* tok[1] file name */
	/* tok[2] has to be NULL */
	char *tok[3];
	parser_t *parser;

	parser = config_open2(file ? file : "/etc/syslog.conf",
				file ? xfopen_for_read : fopen_for_read);
	if (!parser)
		/* didn't find default /etc/syslog.conf */
		/* proceed as if we built busybox without config support */
		return;

	/* use ptr to ptr to avoid checking whether head was initialized */
	pp_rule = &G.log_rules;
	/* iterate through lines of config, skipping comments */
	while (config_read(parser, tok, 3, 2, "# \t", PARSE_NORMAL | PARSE_MIN_DIE)) {
		char *cur_selector;
		logRule_t *cur_rule;

		/* unexpected trailing token? */
		if (tok[2])
			goto cfgerr;

		cur_rule = *pp_rule = xzalloc(sizeof(*cur_rule));

		cur_selector = tok[0];
		/* iterate through selectors: "kern.info;kern.!err;..." */
		do {
			const CODE *code;
			char *next_selector;
			uint8_t negated_prio; /* "kern.!err" */
			uint8_t single_prio;  /* "kern.=err" */
			uint32_t facmap; /* bitmap of enabled facilities */
			uint8_t primap;  /* bitmap of enabled priorities */
			unsigned i;

			next_selector = strchr(cur_selector, ';');
			if (next_selector)
				*next_selector++ = '\0';

			t = strchr(cur_selector, '.');
			if (!t)
				goto cfgerr;
			*t++ = '\0'; /* separate facility from priority */

			negated_prio = 0;
			single_prio = 0;
			if (*t == '!') {
				negated_prio = 1;
				++t;
			}
			if (*t == '=') {
				single_prio = 1;
				++t;
			}

			/* parse priority */
			if (*t == '*')
				primap = 0xff; /* all 8 log levels enabled */
			else {
				uint8_t priority;
				code = find_by_name(t, prioritynames);
				if (!code)
					goto cfgerr;
				primap = 0;
				priority = code->c_val;
				if (priority == INTERNAL_NOPRI) {
					/* ensure we take "enabled_facility_priomap[fac] &= 0" branch below */
					negated_prio = 1;
				} else {
					priority = 1 << priority;
					do {
						primap |= priority;
						if (single_prio)
							break;
						priority >>= 1;
					} while (priority);
					if (negated_prio)
						primap = ~primap;
				}
			}

			/* parse facility */
			if (*cur_selector == '*')
				facmap = (1<<LOG_NFACILITIES) - 1;
			else {
				char *next_facility;
				facmap = 0;
				t = cur_selector;
				/* iterate through facilities: "kern,daemon.<priospec>" */
				do {
					next_facility = strchr(t, ',');
					if (next_facility)
						*next_facility++ = '\0';
					code = find_by_name(t, facilitynames);
					if (!code)
						goto cfgerr;
					/* "mark" is not a real facility, skip it */
					if (code->c_val != INTERNAL_MARK)
						facmap |= 1<<(LOG_FAC(code->c_val));
					t = next_facility;
				} while (t);
			}

			/* merge result with previous selectors */
			for (i = 0; i < LOG_NFACILITIES; ++i) {
				if (!(facmap & (1<<i)))
					continue;
				if (negated_prio)
					cur_rule->enabled_facility_priomap[i] &= primap;
				else
					cur_rule->enabled_facility_priomap[i] |= primap;
			}

			cur_selector = next_selector;
		} while (cur_selector);

		/* check whether current file name was mentioned in previous rules or
		 * as global logfile (G.logFile).
		 */
		if (strcmp(G.logFile.path, tok[1]) == 0) {
			cur_rule->file = &G.logFile;
			goto found;
		}
		/* temporarily use cur_rule as iterator, but *pp_rule still points
		 * to currently processing rule entry.
		 * NOTE: *pp_rule points to the current (and last in the list) rule.
		 */
		for (cur_rule = G.log_rules; cur_rule != *pp_rule; cur_rule = cur_rule->next) {
			if (strcmp(cur_rule->file->path, tok[1]) == 0) {
				/* found - reuse the same file structure */
				(*pp_rule)->file = cur_rule->file;
				cur_rule = *pp_rule;
				goto found;
			}
		}
		cur_rule->file = xzalloc(sizeof(*cur_rule->file));
		cur_rule->file->fd = -1;
		cur_rule->file->path = xstrdup(tok[1]);
 found:
		pp_rule = &cur_rule->next;
	}
	config_close(parser);
	return;

 cfgerr:
	bb_error_msg_and_die("error in '%s' at line %d",
			file ? file : "/etc/syslog.conf",
			parser->lineno);
}
#endif

/* circular buffer variables/structures */
#if ENABLE_FEATURE_IPC_SYSLOG

#if CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE < 4
#error Sorry, you must set the syslogd buffer size to at least 4KB.
#error Please check CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE
#endif

/* our shared key (syslogd.c and logread.c must be in sync) */
enum { KEY_ID = 0x414e4547 }; /* "GENA" */

static void ipcsyslog_cleanup(void)
{
	if (G.shmid != -1) {
		shmdt(G.shbuf);
	}
	if (G.shmid != -1) {
		shmctl(G.shmid, IPC_RMID, NULL);
	}
	if (G.s_semid != -1) {
		semctl(G.s_semid, 0, IPC_RMID, 0);
	}
}

static void ipcsyslog_init(void)
{
	if (DEBUG)
		printf("shmget(%x, %d,...)\n", (int)KEY_ID, G.shm_size);

	G.shmid = shmget(KEY_ID, G.shm_size, IPC_CREAT | 0644);
	if (G.shmid == -1) {
		bb_perror_msg_and_die("shmget");
	}

	G.shbuf = shmat(G.shmid, NULL, 0);
	if (G.shbuf == (void*) -1L) { /* shmat has bizarre error return */
		bb_perror_msg_and_die("shmat");
	}

	memset(G.shbuf, 0, G.shm_size);
	G.shbuf->size = G.shm_size - offsetof(struct shbuf_ds, data) - 1;
	/*G.shbuf->tail = 0;*/

	/* we'll trust the OS to set initial semval to 0 (let's hope) */
	G.s_semid = semget(KEY_ID, 2, IPC_CREAT | IPC_EXCL | 1023);
	if (G.s_semid == -1) {
		if (errno == EEXIST) {
			G.s_semid = semget(KEY_ID, 2, 0);
			if (G.s_semid != -1)
				return;
		}
		bb_perror_msg_and_die("semget");
	}
}

/* Write message to shared mem buffer */
static void log_to_shmem(const char *msg)
{
	int old_tail, new_tail;
	int len;

	if (semop(G.s_semid, G.SMwdn, 3) == -1) {
		bb_perror_msg_and_die("SMwdn");
	}

	/* Circular Buffer Algorithm:
	 * --------------------------
	 * tail == position where to store next syslog message.
	 * tail's max value is (shbuf->size - 1)
	 * Last byte of buffer is never used and remains NUL.
	 */
	len = strlen(msg) + 1; /* length with NUL included */
 again:
	old_tail = G.shbuf->tail;
	new_tail = old_tail + len;
	if (new_tail < G.shbuf->size) {
		/* store message, set new tail */
		memcpy(G.shbuf->data + old_tail, msg, len);
		G.shbuf->tail = new_tail;
	} else {
		/* k == available buffer space ahead of old tail */
		int k = G.shbuf->size - old_tail;
		/* copy what fits to the end of buffer, and repeat */
		memcpy(G.shbuf->data + old_tail, msg, k);
		msg += k;
		len -= k;
		G.shbuf->tail = 0;
		goto again;
	}
	if (semop(G.s_semid, G.SMwup, 1) == -1) {
		bb_perror_msg_and_die("SMwup");
	}
	if (DEBUG)
		printf("tail:%d\n", G.shbuf->tail);
}
#else
static void ipcsyslog_cleanup(void) {}
static void ipcsyslog_init(void) {}
void log_to_shmem(const char *msg);
#endif /* FEATURE_IPC_SYSLOG */

#if ENABLE_FEATURE_KMSG_SYSLOG
static void kmsg_init(void)
{
	G.kmsgfd = xopen("/dev/kmsg", O_WRONLY);

	/*
	 * kernel < 3.5 expects single char printk KERN_* priority prefix,
	 * from 3.5 onwards the full syslog facility/priority format is supported
	 */
	if (get_linux_version_code() < KERNEL_VERSION(3,5,0))
		G.primask = LOG_PRIMASK;
	else
		G.primask = -1;
}

static void kmsg_cleanup(void)
{
	if (ENABLE_FEATURE_CLEAN_UP)
		close(G.kmsgfd);
}

/* Write message to /dev/kmsg */
static void log_to_kmsg(int pri, const char *msg)
{
	/*
	 * kernel < 3.5 expects single char printk KERN_* priority prefix,
	 * from 3.5 onwards the full syslog facility/priority format is supported
	 */
	pri &= G.primask;

	full_write(G.kmsgfd, G.printbuf, sprintf(G.printbuf, "<%d>%s\n", pri, msg));
}
#else
static void kmsg_init(void) {}
static void kmsg_cleanup(void) {}
static void log_to_kmsg(int pri UNUSED_PARAM, const char *msg UNUSED_PARAM) {}
#endif /* FEATURE_KMSG_SYSLOG */

/* Print a message to the log file. */
static void log_locally(time_t now, char *msg, logFile_t *log_file)
{
#ifdef SYSLOGD_WRLOCK
	struct flock fl;
#endif
	int len = strlen(msg);

	/* fd can't be 0 (we connect fd 0 to /dev/log socket) */
	/* fd is 1 if "-O -" is in use */
	if (log_file->fd > 1) {
		/* Reopen log files every second. This allows admin
		 * to delete the files and not worry about restarting us.
		 * This costs almost nothing since it happens
		 * _at most_ once a second for each file, and happens
		 * only when each file is actually written.
		 */
		if (!now)
			now = time(NULL);
		if (log_file->last_log_time != now) {
			log_file->last_log_time = now;
			close(log_file->fd);
			goto reopen;
		}
	}
	else if (log_file->fd == 1) {
		/* We are logging to stdout: do nothing */
	}
	else {
		if (LONE_DASH(log_file->path)) {
			log_file->fd = 1;
			/* log_file->isRegular = 0; - already is */
		} else {
 reopen:
			log_file->fd = open(log_file->path, O_WRONLY | O_CREAT
					| O_NOCTTY | O_APPEND | O_NONBLOCK,
					0666);
			if (log_file->fd < 0) {
				/* cannot open logfile? - print to /dev/console then */
				int fd = device_open(DEV_CONSOLE, O_WRONLY | O_NOCTTY | O_NONBLOCK);
				if (fd < 0)
					fd = 2; /* then stderr, dammit */
				full_write(fd, msg, len);
				if (fd != 2)
					close(fd);
				return;
			}
#if ENABLE_FEATURE_ROTATE_LOGFILE
			{
				struct stat statf;
				log_file->isRegular = (fstat(log_file->fd, &statf) == 0 && S_ISREG(statf.st_mode));
				/* bug (mostly harmless): can wrap around if file > 4gb */
				log_file->size = statf.st_size;
			}
#endif
		}
	}

#ifdef SYSLOGD_WRLOCK
	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 1;
	fl.l_type = F_WRLCK;
	fcntl(log_file->fd, F_SETLKW, &fl);
#endif

#if ENABLE_FEATURE_ROTATE_LOGFILE
	if (G.logFileSize && log_file->isRegular && log_file->size > G.logFileSize) {
		if (G.logFileRotate) { /* always 0..99 */
			int i = strlen(log_file->path) + 3 + 1;
			char oldFile[i];
			char newFile[i];
			i = G.logFileRotate - 1;
			/* rename: f.8 -> f.9; f.7 -> f.8; ... */
			while (1) {
				sprintf(newFile, "%s.%d", log_file->path, i);
				if (i == 0) break;
				sprintf(oldFile, "%s.%d", log_file->path, --i);
				/* ignore errors - file might be missing */
				rename(oldFile, newFile);
			}
			/* newFile == "f.0" now */
			rename(log_file->path, newFile);
		}

		/* We may or may not have just renamed the file away;
		 * if we didn't rename because we aren't keeping any backlog,
		 * then it's time to clobber the file. If we did rename it...,
		 * incredibly, if F and F.0 are hardlinks, POSIX _demands_
		 * that rename returns 0 but does not remove F!!!
		 * (hardlinked F/F.0 pair was observed after
		 * power failure during rename()).
		 * So ensure old file is gone in any case:
		 */
		unlink(log_file->path);
#ifdef SYSLOGD_WRLOCK
		fl.l_type = F_UNLCK;
		fcntl(log_file->fd, F_SETLKW, &fl);
#endif
		close(log_file->fd);
		goto reopen;
	}
/* TODO: what to do on write errors ("disk full")? */
	len = full_write(log_file->fd, msg, len);
	if (len > 0)
		log_file->size += len;
#else
	full_write(log_file->fd, msg, len);
#endif

#ifdef SYSLOGD_WRLOCK
	fl.l_type = F_UNLCK;
	fcntl(log_file->fd, F_SETLKW, &fl);
#endif
}

static void parse_fac_prio_20(int pri, char *res20)
{
	const CODE *c_pri, *c_fac;

	c_fac = find_by_val(LOG_FAC(pri) << 3, facilitynames);
	if (c_fac) {
		c_pri = find_by_val(LOG_PRI(pri), prioritynames);
		if (c_pri) {
			snprintf(res20, 20, "%s.%s", c_fac->c_name, c_pri->c_name);
			return;
		}
	}
	snprintf(res20, 20, "<%d>", pri);
}

/* len parameter is used only for "is there a timestamp?" check.
 * NB: some callers cheat and supply len==0 when they know
 * that there is no timestamp, short-circuiting the test. */
static void timestamp_and_log(int pri, char *msg, int len)
{
	char *timestamp;
	time_t now;

	/* Jan 18 00:11:22 msg... */
	/* 01234567890123456 */
	if (len < 16 || msg[3] != ' ' || msg[6] != ' '
	 || msg[9] != ':' || msg[12] != ':' || msg[15] != ' '
	) {
		time(&now);
		timestamp = ctime(&now) + 4; /* skip day of week */
	} else {
		now = 0;
		timestamp = msg;
		msg += 16;
	}
	timestamp[15] = '\0';

	if (option_mask32 & OPT_kmsg) {
		log_to_kmsg(pri, msg);
		return;
	}

	if (option_mask32 & OPT_small)
		sprintf(G.printbuf, "%s %s\n", timestamp, msg);
	else {
		char res[20];
		parse_fac_prio_20(pri, res);
		sprintf(G.printbuf, "%s %.64s %s %s\n", timestamp, G.hostname, res, msg);
	}

	/* Log message locally (to file or shared mem) */
#if ENABLE_FEATURE_SYSLOGD_CFG
	{
		bool match = 0;
		logRule_t *rule;
		uint8_t facility = LOG_FAC(pri);
		uint8_t prio_bit = 1 << LOG_PRI(pri);

		for (rule = G.log_rules; rule; rule = rule->next) {
			if (rule->enabled_facility_priomap[facility] & prio_bit) {
				log_locally(now, G.printbuf, rule->file);
				match = 1;
			}
		}
		if (match)
			return;
	}
#endif
	if (LOG_PRI(pri) < G.logLevel) {
#if ENABLE_FEATURE_IPC_SYSLOG
		if ((option_mask32 & OPT_circularlog) && G.shbuf) {
			log_to_shmem(G.printbuf);
			return;
		}
#endif
		log_locally(now, G.printbuf, &G.logFile);
	}
}

static void timestamp_and_log_internal(const char *msg)
{
	/* -L, or no -R */
	if (ENABLE_FEATURE_REMOTE_LOG && !(option_mask32 & OPT_locallog))
		return;
	timestamp_and_log(LOG_SYSLOG | LOG_INFO, (char*)msg, 0);
}

/* tmpbuf[len] is a NUL byte (set by caller), but there can be other,
 * embedded NULs. Split messages on each of these NULs, parse prio,
 * escape control chars and log each locally. */
static void split_escape_and_log(char *tmpbuf, int len)
{
	char *p = tmpbuf;

	tmpbuf += len;
	while (p < tmpbuf) {
		char c;
		char *q = G.parsebuf;
		int pri = (LOG_USER | LOG_NOTICE);

		if (*p == '<') {
			/* Parse the magic priority number */
			pri = bb_strtou(p + 1, &p, 10);
			if (*p == '>')
				p++;
			if (pri & ~(LOG_FACMASK | LOG_PRIMASK))
				pri = (LOG_USER | LOG_NOTICE);
		}

		while ((c = *p++)) {
			if (c == '\n')
				c = ' ';
			if (!(c & ~0x1f) && c != '\t') {
				*q++ = '^';
				c += '@'; /* ^@, ^A, ^B... */
			}
			*q++ = c;
		}
		*q = '\0';

		/* Now log it */
		timestamp_and_log(pri, G.parsebuf, q - G.parsebuf);
	}
}

#ifdef SYSLOGD_MARK
static void do_mark(int sig)
{
	if (G.markInterval) {
		timestamp_and_log_internal("-- MARK --");
		alarm(G.markInterval);
	}
}
#endif

/* Don't inline: prevent struct sockaddr_un to take up space on stack
 * permanently */
static NOINLINE int create_socket(void)
{
	struct sockaddr_un sunx;
	int sock_fd;
	char *dev_log_name;

#if ENABLE_FEATURE_SYSTEMD
	if (sd_listen_fds() == 1)
		return SD_LISTEN_FDS_START;
#endif

	memset(&sunx, 0, sizeof(sunx));
	sunx.sun_family = AF_UNIX;

	/* Unlink old /dev/log or object it points to. */
	/* (if it exists, bind will fail) */
	strcpy(sunx.sun_path, _PATH_LOG);
	/*
	dev_log_name = xmalloc_follow_symlinks(_PATH_LOG);
	if (dev_log_name) {
		safe_strncpy(sunx.sun_path, dev_log_name, sizeof(sunx.sun_path));
		free(dev_log_name);
	}
	*/
	unlink(sunx.sun_path);

	sock_fd = xsocket(AF_UNIX, SOCK_DGRAM, 0);
	xbind(sock_fd, (struct sockaddr *) &sunx, sizeof(sunx));
	chmod(_PATH_LOG, 0666);

	return sock_fd;
}

#if ENABLE_FEATURE_REMOTE_LOG
static int try_to_resolve_remote(remoteHost_t *rh)
{
	int port, socktype;
	if ( G.tls )
	{
		port = SYSLOG_PORT_TLS;
		socktype = SOCK_STREAM;
	}
	else
	{
		port = SYSLOG_PORT_UDP;
		socktype = SOCK_DGRAM;
	}
	if (!rh->remoteAddr) {
		unsigned now = monotonic_sec();

		/* Don't resolve name too often - DNS timeouts can be big */
		if ((now - rh->last_dns_resolve) < DNS_WAIT_SEC)
			return -1;
		rh->last_dns_resolve = now;
		rh->remoteAddr = host2sockaddr(rh->remoteHostname, port);
		if (!rh->remoteAddr)
			return -1;
	}
	return xsocket(rh->remoteAddr->u.sa.sa_family, socktype, 0);
}
#endif

/* Bug 136852/140696 */
enum WEAK_CIPHER_LEVEL
{
    WCL_ENABLE = 0,
    WCL_ENC_DES_3DES_RC4,
    WCL_ENC_SEED,
    WCL_ENC_ALL, /* Disable all weak enc. */
    WCL_MAC_MD5, 
    WCL_MAC_ALL, /* Disable all weak mac. */
    WCL_SSL_SSLV2_SSLV3,
    WCL_SSL_ALL, /* Disable all weak ssl. */
    WCL_DISABLE_ALL, /* Disable all weak cipher. */
    /* All other options can only be added following WCL_DISABLE_ALL. */
};

static void
getDisabledCipherMask(unsigned long *mkey, unsigned long *auth, unsigned long *enc, unsigned long *mac, unsigned long *ssl)
{
	char *pDisWeakCipher = nvram_get(PVALUE_DISABLE_WEAK_CIPHER);
	int nDisWeakCipher = 0;
	if ( pDisWeakCipher )
	{
		nDisWeakCipher = atoi(pDisWeakCipher);
	}

	switch ( nDisWeakCipher )
	{
	case WCL_ENC_DES_3DES_RC4:
	{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
		}
		break;
	case WCL_ENC_SEED:
		{
			*enc |= SSL_SEED;
		}
		break;
	case WCL_ENC_ALL:
		{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
			*enc |= SSL_SEED;
		}
		break;
	case WCL_MAC_MD5:
		{
			*mac |= SSL_MD5;
		}
		break;
	case WCL_MAC_ALL:
		{
			*mac |= SSL_MD5;
		}
		break;
	case WCL_SSL_SSLV2_SSLV3:
		{
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_SSL_ALL:
		{
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_DISABLE_ALL:
		{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
			*enc |= SSL_SEED;
			*mac |= SSL_MD5;
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_ENABLE:
	default:
		{
			return;
		}
		break;
	}
}

static void tls_cleanup( remoteHost_t *rh )
{
	print_msg( "%s start\n", __FUNCTION__ );
	pthread_mutex_lock_p( &rh->cleanup_lock );
	print_msg( "%s locked\n", __FUNCTION__ );

	if ( rh->tls_ssl != NULL )
	{
		if ( rh->tls_con_status )
		{
			print_msg( "%s TLS shutdown\n", __FUNCTION__ );
			SSL_shutdown_p( rh->tls_ssl );
		}
		SSL_free_p( rh->tls_ssl );
	}

	if ( rh->tls_ctx != NULL )
	{
		SSL_CTX_free_p( rh->tls_ctx );
	}

	if ( -1 != rh->remoteFD )
	{
		close( rh->remoteFD );
		rh->remoteFD = -1;
	}
	rh->tls_ssl = NULL;
	rh->tls_ctx = NULL;
	rh->tls_con_status = 0;

	pthread_mutex_unlock_p( &rh->cleanup_lock );
	print_msg( " %s done\n", __FUNCTION__ );
}

void *tls_disconnect( void* dummy )
{
#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *item;
#endif
	char buf[ MAX_READ ];

	print_msg( "%s start monitor\n", __FUNCTION__ );
	while ( !bb_got_signal )
	{
#if ENABLE_FEATURE_REMOTE_LOG
		for (item = G.remoteHosts; item != NULL; item = item->link) {
			remoteHost_t *rh = (remoteHost_t *)item->data;

			pthread_mutex_lock_p( &rh->cleanup_lock );
			if ( rh->tls_con_status && rh->tls_ssl != NULL )
			{
				int read_res = SSL_read_p( rh->tls_ssl, buf, MAX_READ );

				if ( read_res == 0 ) // tls disconnect case
				{
					print_msg( "%s TLS disconnected\n", __FUNCTION__ );
					tls_cleanup( rh );
				}
				else if ( read_res < 0 )
				{
					int ssl_error = SSL_get_error_p (rh->tls_ssl, read_res);

					switch ( ssl_error )
					{
						case SSL_ERROR_WANT_READ:
						case SSL_ERROR_WANT_WRITE:
							break;
						default:
							print_msg( "%s TLS disconnect error, %d\n", __FUNCTION__, ssl_error );
							tls_cleanup( rh );
							break;
					}
				}
			}
			pthread_mutex_unlock_p( &rh->cleanup_lock );
		}
#endif

		sleep( 3 );
	}
	return NULL;
}

static int get_link_status( void )
{
	struct ifreq ifr;
	struct ethtool_value edata;
	int fd;
	int rc;
	char buf[ MAX_READ ] = {0};

	if ( ( fd = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
	{
		return 0;
	}

	memset( &ifr, 0, sizeof( ifr ) );
	edata.cmd = ETHTOOL_GLINK;

	nvram_get_safe( PVALUE_WAN_DEVICE, buf, MAX_READ );
	if ( buf[0] == 0 )
	{
		strcpy(buf, "eth0");
	}
	strncpy( ifr.ifr_name, buf, sizeof( ifr.ifr_name ) - 1 );
	ifr.ifr_data = ( char* )&edata;

	rc = ioctl( fd, SIOCETHTOOL, &ifr );
	close( fd );
	if ( rc == -1 )
	{
		return 0;
	}
	return edata.data != 0 ? 1 : 0;
}

void *tls_linkstatus( void* dummy )
{
#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *item;
#endif
	print_msg( "%s start monitor\n", __FUNCTION__ );

	while ( !bb_got_signal )
	{
#if ENABLE_FEATURE_REMOTE_LOG
		if ( get_link_status() == 0 )
		{
			print_msg( "%s link down\n", __FUNCTION__ );
			for (item = G.remoteHosts; item != NULL; item = item->link) {
				remoteHost_t *rh = (remoteHost_t *)item->data;

				if ( rh->tls_con_status )
				{
					tls_cleanup( rh );
				}
			}
		}
#endif
		sleep( 1 );
	}
	return NULL;
}

static int tls_connect( remoteHost_t *rh )
{
	char ip_addr[ INET_ADDRSTRLEN + INET6_ADDRSTRLEN ];
	char buf[MAX_READ] = {0};
	int flags;
	fd_set set;
	int width;
	struct timeval tv;
	int con_ret;
	void *addr;

	print_msg("%s start\n", __FUNCTION__);

	if ( rh->remoteHostname )
	{
		if ( -1 == rh->remoteFD )
		{
			rh->remoteFD = try_to_resolve_remote(rh);
			if ( -1 == rh->remoteFD )
			{
				return -1;
			}
		}
	}
	else
	{
		return -1;
	}

	if ( NULL == rh->tls_ctx )
	{
		rh->tls_ctx = SSL_CTX_new_p( SSLv23_method_p() );
	}
	if ( NULL == rh->tls_ctx )
	{
		return -1;
	}
	/* Bug 136852 */
	SSL_CTX_set_cb_cipher_get_disabled_p(rh->tls_ctx, (long)getDisabledCipherMask);
	SSL_CTX_rebuild_cipher_list_p(rh->tls_ctx);

	/* enforce CA verification in TLS mode
	 *              rfc5425
	 * 4.2.1.  Certificate-Based Authentication
	 *
	 * Both syslog transport sender (TLS client) and syslog transport
	 * receiver (TLS server) MUST implement certificate-based
	 * authentication.
	 */
	nvram_get_safe( PVALUE_VERIFY_SERVER, buf, sizeof( buf ) );
	if ( buf[0] == '1' )
	{
		buf[0] = 0;
		nvram_get_safe( PVALUE_LOAD_CA_CERT, buf, sizeof( buf ) );
		if ( SSL_CTX_load_verify_locations_p( rh->tls_ctx, TLS_CA_FILE, NULL ) < 1 )
		{
			print_msg( "%s load CA failed\n", __FUNCTION__ );
			goto connect_failed;
			/* SSL_CTX_set_default_verify_paths( rh->tls_ctx ); */
		}
		SSL_CTX_set_verify_p( rh->tls_ctx, SSL_VERIFY_PEER, NULL );
	}

	if ( connect( rh->remoteFD, &rh->remoteAddr->u.sa, rh->remoteAddr->len ) == -1 )
	{
		switch ( rh->remoteAddr->u.sa.sa_family )
		{
		case AF_INET:
			addr = &((struct sockaddr_in*)&rh->remoteAddr->u.sa)->sin_addr;
			break;
		case AF_INET6:
			addr = &((struct sockaddr_in6*)&rh->remoteAddr->u.sa)->sin6_addr;
			break;
		default:
			print_msg( "%s unsupported sa family %d\n", __FUNCTION__, rh->remoteAddr->u.sa.sa_family );
			goto connect_failed;
			break;
		}
		inet_ntop(rh->remoteAddr->u.sa.sa_family, addr, ip_addr, INET_ADDRSTRLEN + INET6_ADDRSTRLEN);
		print_msg( "%s sock %x Cannot connect to host %s [%s] on port %d, err %d - %s\n",
					__FUNCTION__,
					rh->remoteFD,
					rh->remoteHostname,
					ip_addr,
					ntohs( get_nport( &rh->remoteAddr->u.sa ) ),
					errno, strerror(errno) );
		goto connect_failed;
	}
	if ( NULL == rh->tls_ssl )
	{
		rh->tls_ssl = SSL_new_p( rh->tls_ctx );
	}
	if ( NULL == rh->tls_ssl )
	{
		goto connect_failed;
	}

	SSL_set_fd_p( rh->tls_ssl, rh->remoteFD );

#if 1
	if ( ( flags = fcntl( rh->remoteFD, F_GETFL ) ) >= 0 )
	{
		if ( ( flags & O_NONBLOCK ) == 0 )
		{
			fcntl( rh->remoteFD, F_SETFL, flags | O_NONBLOCK );
		}
	}
#endif

	while( (con_ret = SSL_connect_p( rh->tls_ssl )) != 1 )
	{
		FD_ZERO(&set);
		FD_SET( rh->remoteFD, &set );
		width = rh->remoteFD;

		tv.tv_sec = SSL_CONNECT_TIME_OUT;
		tv.tv_usec = 0;

		switch( SSL_get_error_p( rh->tls_ssl, con_ret ) )
		{
		case SSL_ERROR_WANT_READ:
			print_msg( "%s SSL connection get error SSL_ERROR_WANT_READ\n", __FUNCTION__ );
			if ( select( width + 1, &set, NULL, NULL, &tv ) <= 0 )
			{
				goto connect_failed;
			}
			break;

		case SSL_ERROR_WANT_WRITE:
			print_msg( "%s SSL connection get error SSL_ERROR_WANT_WRITE\n", __FUNCTION__ );
			break;

		default:
			goto connect_failed;
			break;
		}
	}

	print_msg( "%s TLS connected\n", __FUNCTION__ );
	rh->tls_con_status = 1;
	rh->tls_retry = 0;
	return 0;

connect_failed:
	tls_cleanup( rh );
	return -1;
}

void *tls_reconnect( void* dummy )
{
#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *item;
#endif
	int tls_retry;
	print_msg( "%s start monitor\n", __FUNCTION__ );
	while ( !bb_got_signal )
	{
#if ENABLE_FEATURE_REMOTE_LOG
		tls_retry = 0;
		for (item = G.remoteHosts; item != NULL; item = item->link) {
			remoteHost_t *rh = (remoteHost_t *)item->data;
			if ( !rh->tls_con_status && tls_connect(rh) < 0 )
			{
				tls_retry = rh->tls_retry > tls_retry ? rh->tls_retry : tls_retry;
				rh->tls_retry++;
			}
		}
#endif
		if ( tls_retry > 0 )
		{
			print_msg( "TLS connect failed max retry %d wait %d\n", tls_retry, tls_retry * DELAY_MULTIPLYER );
			sleep( tls_retry * DELAY_MULTIPLYER );
		}
		else
		{
			sleep( 1 );
		}
	}
	return NULL;
}

static void do_syslogd(void) NORETURN;
static void do_syslogd(void)
{
	int sock_fd;
	int i, ret;
	char prefix_fd;
	char filter_prefix[16];
	int filter_len;
#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *item;
#endif
#if ENABLE_FEATURE_SYSLOGD_DUP
	int last_sz = -1;
	char *last_buf;
	char *recvbuf = G.recvbuf;
#else
#define recvbuf (G.recvbuf)
#endif

	memset(filter_prefix, 0, sizeof(filter_prefix));
	if (G.user) {
		char *log_mask = getenv("syslog_id");
		if (log_mask) {
			strncpy(filter_prefix,log_mask,16);
			filter_len = strlen(filter_prefix);
			goto filter_prefix_done;
		} else {
			prefix_fd = open("/proc/gxp/dev_info/dev_alias", O_RDONLY);
			if (prefix_fd < 0) {
				goto read_filter_prefix_failed;
			} else {
				char log_msg[] = "syslogd: loaded filter prefix                ";
				if (read(prefix_fd, filter_prefix, 16) <= 0) {
					goto read_filter_prefix_failed;
				}
				filter_len = strlen(filter_prefix);
				sprintf(log_msg + 31, "%s", filter_prefix);
				timestamp_and_log_internal(log_msg);
				goto filter_prefix_done;
			}
		}
	} else {
		prefix_fd = open("/proc/gxp/dev_info/dev_alias", O_RDONLY);
		if (prefix_fd < 0) {
read_filter_prefix_failed:
			snprintf(filter_prefix, 12, "GRANDSTREAM");
			filter_len = strlen(filter_prefix);
			timestamp_and_log_internal("syslogd: falling back to default filter prefix GRANDSTREAM");
		} else {
			if (read(prefix_fd, filter_prefix, 16) <= 0)
				goto read_filter_prefix_failed;
			close(prefix_fd);
			filter_len = strlen(filter_prefix);
			timestamp_and_log_internal("syslogd started: BusyBox v" BB_VER);
		}
	}
filter_prefix_done:
	/* Set up signal handlers (so that they interrupt read()) */
	signal_no_SA_RESTART_empty_mask(SIGTERM, record_signo);
	signal_no_SA_RESTART_empty_mask(SIGINT, record_signo);
	//signal_no_SA_RESTART_empty_mask(SIGQUIT, record_signo);
	signal(SIGHUP, SIG_IGN);
#ifdef SYSLOGD_MARK
	signal(SIGALRM, do_mark);
	alarm(G.markInterval);
#endif
	xmove_fd(create_socket(), STDIN_FILENO);

	if (option_mask32 & OPT_circularlog)
		ipcsyslog_init();

	if (option_mask32 & OPT_kmsg)
		kmsg_init();

	while (!bb_got_signal) {
		ssize_t sz;

#if ENABLE_FEATURE_SYSLOGD_DUP
		last_buf = recvbuf;
		if (recvbuf == G.recvbuf)
			recvbuf = G.recvbuf + MAX_READ;
		else
			recvbuf = G.recvbuf;
#endif
 read_again:
		sz = read(STDIN_FILENO, recvbuf, MAX_READ - 1);
		if (sz < 0) {
			if (!bb_got_signal)
				bb_perror_msg("read from %s", _PATH_LOG);
			break;
		}

		/* Drop trailing '\n' and NULs (typically there is one NUL) */
		while (1) {
			if (sz == 0)
				goto read_again;
			/* man 3 syslog says: "A trailing newline is added when needed".
			 * However, neither glibc nor uclibc do this:
			 * syslog(prio, "test")   sends "test\0" to /dev/log,
			 * syslog(prio, "test\n") sends "test\n\0".
			 * IOW: newline is passed verbatim!
			 * I take it to mean that it's syslogd's job
			 * to make those look identical in the log files. */
			if (recvbuf[sz-1] != '\0' && G.user)
				break;
			if (recvbuf[sz-1] != '\0' && recvbuf[sz-1] != '\n')
				break;
			sz--;
		}

		/* Drop anything preceding the log prefix */
		for (i=0; i<(sz-filter_len); i++) {
			int j;
			for (j=0; j<filter_len; j++) {
				if (recvbuf[i+j] != filter_prefix[j])
					break;
			}
			if (j == filter_len) {
				int k;

				for (k=0; k<(sz-i); k++) {
					recvbuf[k] = recvbuf[i+k];
				}
				sz = sz-i;
				break;
			}
		}
		/* Drop if log prefix not matched */
		if (i == (sz-filter_len))
			goto read_again;
#if ENABLE_FEATURE_SYSLOGD_DUP
		if ((option_mask32 & OPT_dup) && (sz == last_sz))
			if (memcmp(last_buf, recvbuf, sz) == 0)
				continue;
		last_sz = sz;
#endif
#if ENABLE_FEATURE_REMOTE_LOG
		/* Stock syslogd sends it '\n'-terminated
		 * over network, mimic that */
		recvbuf[sz++] = '\n';

		/* We are not modifying log messages in any way before send */
		/* Remote site cannot trust _us_ anyway and need to do validation again */
		for (item = G.remoteHosts; item != NULL; item = item->link) {
			remoteHost_t *rh = (remoteHost_t *)item->data;

			if (rh->remoteFD == -1) {
				rh->remoteFD = try_to_resolve_remote(rh);
				if (rh->remoteFD == -1)
					continue;
			}
			if ( G.tls )
			{
				if ( !rh->tls_con_status )
				{
					continue;
				}
			}
			/* Send message to remote logger.
			 * On some errors, close and set remoteFD to -1
			 * so that DNS resolution is retried.
			 */
			if (G.user && strstr(recvbuf, filter_prefix)) {
				char *tmpbuf = recvbuf;
				if(recvbuf[0] == ':' && recvbuf[1] == '<') {
					tmpbuf= &recvbuf[1] ;
					sz-- ;
				}
				if ( G.tls )
				{
					ret = SSL_write_p( rh->tls_ssl, tmpbuf, sz );
					if ( ret < 0 )
					{
						if ( rh->tls_ssl != NULL )
						{
							print_msg( "SSL_write error:%d\n", SSL_get_error_p( rh->tls_ssl, ret ) );
						}
						else
						{
							print_msg( "SSL_write error: TLS closed\n" );
						}
						tls_cleanup( rh );
					}
				}
				else
				{
					if (sendto(rh->remoteFD, tmpbuf, sz,
							MSG_DONTWAIT | MSG_NOSIGNAL,
							&(rh->remoteAddr->u.sa), rh->remoteAddr->len) == -1
					) {
						switch (errno) {
						case ECONNRESET:
						case ENOTCONN: /* paranoia */
						case EPIPE:
							close(rh->remoteFD);
							rh->remoteFD = -1;
							free(rh->remoteAddr);
							rh->remoteAddr = NULL;
						}
					}
				}
			} else if (strncmp( recvbuf, filter_prefix, filter_len) == 0 ) {
				if ( G.tls )
				{
					ret = SSL_write_p( rh->tls_ssl, recvbuf, sz );
					if ( ret < 0 )
					{
						if ( rh->tls_ssl != NULL )
						{
							print_msg( "SSL_write error:%d\n", SSL_get_error_p( rh->tls_ssl, ret ) );
						}
						else
						{
							print_msg( "SSL_write error: TLS closed\n" );
						}
						tls_cleanup( rh );
					}
				}
				else
				{
					if (sendto(rh->remoteFD, recvbuf, sz,
						MSG_DONTWAIT | MSG_NOSIGNAL,
						&(rh->remoteAddr->u.sa), rh->remoteAddr->len) == -1
						) {
						switch (errno) {
						case ECONNRESET:
						case ENOTCONN: /* paranoia */
						case EPIPE:
							close(rh->remoteFD);
							rh->remoteFD = -1;
							free(rh->remoteAddr);
							rh->remoteAddr = NULL;
						}
					}
				}
		}
	}
#endif
		if (!ENABLE_FEATURE_REMOTE_LOG || (option_mask32 & OPT_locallog)) {
			recvbuf[sz] = '\0'; /* ensure it *is* NUL terminated */
			split_escape_and_log(recvbuf, sz);
		}
	} /* while (!bb_got_signal) */

	timestamp_and_log_internal("syslogd exiting");
	remove_pidfile(CONFIG_PID_FILE_PATH "/syslogd.pid");
	ipcsyslog_cleanup();
	if (option_mask32 & OPT_kmsg)
		kmsg_cleanup();

#if ENABLE_FEATURE_REMOTE_LOG
	if ( G.tls )
	{
		for (item = G.remoteHosts; item != NULL; item = item->link) {
			remoteHost_t *rh = (remoteHost_t *)item->data;

			tls_cleanup(rh);
			pthread_mutex_destroy(&rh->cleanup_lock);
		}
	}
#endif
	kill_myself_with_sig(bb_got_signal);
#undef recvbuf
}

#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
static int open_libssl(void)
{
	char *error;

	ssl_handle = dlopen(LIBSSL, RTLD_LAZY);
	if (!ssl_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}
	crypto_handle = dlopen(LIBCRYPTO, RTLD_LAZY);
	if (!crypto_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}
	pthread_handle = dlopen(LIBPTHREAD, RTLD_LAZY);
	if (!pthread_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}

	SSL_library_init_p = (SSL_library_init_t)dlsym(ssl_handle, "SSL_library_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_set_fd_p = (SSL_set_fd_t)dlsym(ssl_handle, "SSL_set_fd");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_connect_p = (SSL_connect_t)dlsym(ssl_handle, "SSL_connect");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_new_p = (SSL_new_t)dlsym(ssl_handle, "SSL_new");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_shutdown_p = (SSL_new_t)dlsym(ssl_handle, "SSL_shutdown");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_free_p = (SSL_free_t)dlsym(ssl_handle, "SSL_free");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_read_p = (SSL_read_t)dlsym(ssl_handle, "SSL_read");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_write_p = (SSL_write_t)dlsym(ssl_handle, "SSL_write");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_get_error_p = (SSL_get_error_t)dlsym(ssl_handle, "SSL_get_error");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_free_p = (SSL_CTX_free_t)dlsym(ssl_handle, "SSL_CTX_free");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_new_p = (SSL_new_t)dlsym(ssl_handle, "SSL_CTX_new");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSLv23_method_p = (SSLv23_method_t)dlsym(ssl_handle, "SSLv23_method");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_load_verify_locations_p =
		(SSL_CTX_load_verify_locations_t)dlsym(ssl_handle, "SSL_CTX_load_verify_locations");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_set_verify_p = (SSL_CTX_set_verify_t)dlsym(ssl_handle, "SSL_CTX_set_verify");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_ctrl_p = (SSL_CTX_ctrl_t)dlsym(ssl_handle, "SSL_CTX_ctrl");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
#ifdef OPENSSL_LOAD_CONF
	OpenSSL_add_all_algorithms_p =
		(OpenSSL_add_all_algorithms_t)dlsym(crypto_handle, "OPENSSL_add_all_algorithms_conf");
#else
	OpenSSL_add_all_algorithms_p =
		(OpenSSL_add_all_algorithms_t)dlsym(crypto_handle, "OPENSSL_add_all_algorithms_noconf");
#endif
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_init_p = (pthread_mutexattr_init_t)dlsym(pthread_handle, "pthread_mutexattr_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_destroy_p = (pthread_mutexattr_destroy_t)dlsym(pthread_handle, "pthread_mutexattr_destroy");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_settype_p = (pthread_mutexattr_settype_t)dlsym(pthread_handle, "pthread_mutexattr_settype");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_lock_p = (pthread_mutex_lock_t)dlsym(pthread_handle, "pthread_mutex_lock");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_unlock_p = (pthread_mutex_unlock_t)dlsym(pthread_handle, "pthread_mutex_unlock");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_init_p = (pthread_mutex_init_t)dlsym(pthread_handle, "pthread_mutex_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_destroy_p = (pthread_mutex_destroy_t)dlsym(pthread_handle, "pthread_mutex_destroy");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_create_p = (pthread_create_t)dlsym(pthread_handle, "pthread_create");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}

	return 0;
HANDLE_ERROR:
	if ( ssl_handle )
	{
		dlclose(ssl_handle);
		ssl_handle = NULL;
	}
	if ( crypto_handle )
	{
		dlclose(crypto_handle);
		crypto_handle = NULL;
	}
	if ( pthread_handle )
	{
		dlclose(pthread_handle);
		pthread_handle = NULL;
	}
	return -1;
}

static void close_libssl(void)
{
	if ( ssl_handle )
	{
		dlclose(ssl_handle);
		ssl_handle = NULL;
	}
	if ( crypto_handle )
	{
		dlclose(crypto_handle);
		crypto_handle = NULL;
	}
	if ( pthread_handle )
	{
		dlclose(pthread_handle);
		pthread_handle = NULL;
	}
}
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

int syslogd_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
int syslogd_main(int argc UNUSED_PARAM, char **argv)
{
	int opts;
	char OPTION_DECL;
#if ENABLE_FEATURE_REMOTE_LOG
	llist_t *remoteAddrList = NULL;
	llist_t *item;
#endif
	pthread_mutexattr_t attr;

	INIT_G();

	/* No non-option params, -R can occur multiple times */
	opt_complementary = "=0" IF_FEATURE_REMOTE_LOG(":R::");
	opts = getopt32(argv, OPTION_STR, OPTION_PARAM);
#if ENABLE_FEATURE_REMOTE_LOG
	while (remoteAddrList) {
		remoteHost_t *rh = xzalloc(sizeof(*rh));
		rh->remoteHostname = llist_pop(&remoteAddrList);
		rh->remoteFD = -1;
		rh->last_dns_resolve = monotonic_sec() - DNS_WAIT_SEC - 1;
		llist_add_to(&G.remoteHosts, rh);
	}
#endif

#ifdef SYSLOGD_MARK
	if (opts & OPT_mark) // -m
		G.markInterval = xatou_range(opt_m, 0, INT_MAX/60) * 60;
#endif
	//if (opts & OPT_nofork) // -n
	//if (opts & OPT_outfile) // -O
	if (opts & OPT_loglevel) // -l
		G.logLevel = xatou_range(opt_l, 1, 8);
	//if (opts & OPT_small) // -S
	if (opts & OPT_user)			/* -U  Bug 10817 Vonage */
		G.user = 1 ;
#if ENABLE_FEATURE_ROTATE_LOGFILE
	if (opts & OPT_filesize) // -s
		G.logFileSize = xatou_range(opt_s, 0, INT_MAX/1024) * 1024;
	if (opts & OPT_rotatecnt) // -b
		G.logFileRotate = xatou_range(opt_b, 0, 99);
#endif
#if ENABLE_FEATURE_IPC_SYSLOG
	if (opt_C) // -Cn
		G.shm_size = xatoul_range(opt_C, 4, INT_MAX/1024) * 1024;
#endif
	if (opts & OPT_TLS) // -T  Bug 141499
	{
		G.tls = 1;
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
		if ( open_libssl() )
		{
			exit(-1);
		}
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
	}

	/* If they have not specified remote logging, then log locally */
	if (ENABLE_FEATURE_REMOTE_LOG && !(opts & OPT_remotelog)) // -R
		option_mask32 |= OPT_locallog;
#if ENABLE_FEATURE_SYSLOGD_CFG
	parse_syslogdcfg(opt_f);
#endif

	/* Store away localhost's name before the fork */
	G.hostname = safe_gethostname();
	*strchrnul(G.hostname, '.') = '\0';

	if (!(opts & OPT_nofork)) {
		bb_daemonize_or_rexec(DAEMON_CHDIR_ROOT, argv);
	}

	//umask(0); - why??
	write_pidfile(CONFIG_PID_FILE_PATH "/syslogd.pid");

#if ENABLE_FEATURE_REMOTE_LOG
	/* bug 141499 */
	if ( G.tls )
	{
		if ( SSL_library_init_p() < 0 )
		{
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
			close_libssl();
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
			bb_perror_msg_and_die( "Could not initialize the OpenSSL library!\n" );
		}
		/* tls connection */
		OpenSSL_add_all_algorithms_p();
		//ERR_load_BIO_strings();
		//ERR_load_crypto_strings();
		//SSL_load_error_strings();

		pthread_mutexattr_init_p( &attr );
		pthread_mutexattr_settype_p( &attr, PTHREAD_MUTEX_RECURSIVE );

		for (item = G.remoteHosts; item != NULL; item = item->link) {
			remoteHost_t *rh = (remoteHost_t *)item->data;
			pthread_mutex_init_p( &rh->cleanup_lock, &attr );
		}
		pthread_mutexattr_destroy_p( &attr );

		pthread_create_p( &G.tls_disconnect_thread, NULL, tls_disconnect, NULL );
		pthread_create_p( &G.tls_reconnect_thread, NULL, tls_reconnect, NULL );
		pthread_create_p( &G.tls_linkstatus_thread, NULL, tls_linkstatus, NULL );
	}
#endif

	do_syslogd();
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
	close_libssl();
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
	/* return EXIT_SUCCESS; */
}

/* Clean up. Needed because we are included from syslogd_and_logger.c */
#undef DEBUG
#undef SYSLOGD_MARK
#undef SYSLOGD_WRLOCK
#undef G
#undef GLOBALS
#undef INIT_G
#undef OPTION_STR
#undef OPTION_DECL
#undef OPTION_PARAM
