/*
 *  linux/arch/arm/mach-dvf/time.c
 *
 *  Copyright (C) 2016 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/err.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/time64.h>
#include <linux/clksrc-dspg-rtc.h>
#include <linux/clksrc-dspg-timer.h>

#include <asm/mach/time.h>

static struct dspg_rtc *rtc;
static struct dspg_tmr *tmr_clockevent, *tmr_clocksource;

static void dspg_read_persistent_clock(struct timespec64 *ts)
{
	if (rtc) {
		dspg_rtc_persistent_clock(rtc, ts);
	} else if (tmr_clocksource) {
		dspg_tmr_persistent_clock(tmr_clocksource, ts);
	} else {
		ts->tv_sec = 0;
		ts->tv_nsec = 0;
	}
}

static const struct of_device_id dvf_rtc_ids[] __initconst = {
	{ .compatible = "dspg,rtc" },
	{ /* Sentinel */ },
};

static const struct of_device_id dvf_tmr_clockevent_ids[] __initconst = {
	{ .compatible = "dspg,clockevent" },
	{ /* Sentinel */ },
};

static const struct of_device_id dvf_tmr_clocksource_ids[] __initconst = {
	{ .compatible = "dspg,clocksource" },
	{ /* Sentinel */ },
};

void __init dvf_timer_init(void)
{
	struct device_node *rtc_node;
	struct device_node *tmr_clockevent_node, *tmr_clocksource_node = NULL;

	rtc_node = of_find_matching_node(NULL, dvf_rtc_ids);
	if (!rtc_node) {
		tmr_clocksource_node = of_find_matching_node(NULL,
						dvf_tmr_clocksource_ids);
		if (!tmr_clocksource_node)
			panic("No RTC or timer for sched clock to use");
	}

	tmr_clockevent_node = of_find_matching_node(NULL,
						    dvf_tmr_clockevent_ids);
	if (!tmr_clockevent_node)
		panic("No timer for clockevents to use");

	if (rtc_node) {
		rtc = dspg_rtc_init(rtc_node);
		if (IS_ERR(rtc))
			panic("dspg_rtc_init failed: %ld!", PTR_ERR(rtc));
	} else {
		tmr_clocksource =
			dspg_tmr_clocksource_init(tmr_clocksource_node);
		if (IS_ERR(tmr_clocksource))
			panic("dspg_tmr_clocksource_init failed: %ld!",
			      PTR_ERR(tmr_clocksource));
	}

	tmr_clockevent = dspg_tmr_clockevent_init(tmr_clockevent_node);
	if (IS_ERR(tmr_clockevent))
		panic("dspg_tmr_clockevent_init failed: %ld!",
		      PTR_ERR(tmr_clockevent));

	if (rtc_node)
		of_node_put(rtc_node);
	else
		of_node_put(tmr_clocksource_node);
	of_node_put(tmr_clockevent_node);

	if (rtc_node)
		dspg_rtc_register_clocksource(rtc);
	else
		dspg_tmr_register_clocksource(tmr_clocksource);
	dspg_tmr_register_clockevent(tmr_clockevent);
	/* if (rtc_node)
		setup_sched_clock(dvf_read_sched_clock, 32,
				  dspg_rtc_rate(rtc));
	else
		setup_sched_clock(dvf_read_sched_clock, 32,
				  dspg_tmr_rate(tmr_clocksource)); */
	register_persistent_clock(NULL, dspg_read_persistent_clock);
}
