/****************************************************************************
 *
 * FILENAME:            factory_proc.c
 *
 * LAST MODIFIED:   Date: 2017/05/16
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2017 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/mfd/dvf9a.h>
#include <linux/gpio.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/seq_file.h>
#include <video/dspgfb.h>
#include <mach/factory.h>
#include <provision_funcs.h>

#define GS_FACT_PROC_DEBUG

#ifdef GS_FACT_PROC_DEBUG
	#define FACT_PROC_DBG(args...)	\
		do { \
			printk("FACT_PROC: %s: ", __FUNCTION__);\
			printk(args);\
		} while (0)
#else
#define FACT_PROC_DBG(args...)
#endif

#define DISABLE_SECURITY_PROCFS_WRITE

enum {
	iPN        = 0,
	idev_alias = 1,
	idev_data  = 2,
	idev_id    = 3,
	idev_mac   = 4,
	idev_rev   = 5,
	irequired_bits = 6,
	imodel_key     = 7,
	idev_prefix    = 8,
	ifact_ver  = 9
};

struct gs_procfs_entry {
	char *name;
	const struct file_operations *fops;
	int data;
};

static int erase_helper(struct mtd_info *mtd, loff_t offset, size_t len);
static void erase_callback(struct erase_info *info);

static struct {
	int enabled;
	unsigned int value;
} oem_override;

#if defined(CONFIG_HT)
static unsigned int reset_gpio;
#endif

static unsigned int cfg_reload_count;

static char syslog_id[64];

static int erase_helper(struct mtd_info *mtd, loff_t offset, size_t len)
{
	struct erase_info info;
	DECLARE_COMPLETION_ONSTACK(complete);
	int ret;

	memset(&info, 0, sizeof(info));
	info.mtd = mtd;
	info.addr = offset;
	info.len = len;
	info.callback = erase_callback;
	info.priv = (long)&complete;

	ret = mtd_erase(mtd, &info);
	if (ret)
		return -EIO;

	wait_for_completion(&complete);
	if (info.state != MTD_ERASE_DONE)
		return -EIO;

	return 0;
}

static void erase_callback(struct erase_info *info)
{
	complete((struct completion *)info->priv);
}

/************************
* reset_button_read_proc
*************************/
static int reset_button_proc_show(struct seq_file *m, void *v)
{
	int gpio_val = !gpio_get_value(reset_gpio);

	seq_printf(m, "%d", gpio_val);
	return 0;
}

static int reset_button_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, reset_button_proc_show, PDE_DATA(inode));
}

static const struct file_operations reset_button_proc_fops = {
	.open = reset_button_proc_open,
	.read = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

/************************
* oem_override_proc
*************************/
static int oem_override_proc_show(struct seq_file *m, void *v)
{
	seq_printf(m, "%u", oem_override.value);
	return 0;
}

static ssize_t oem_override_proc_write(struct file *file,
		const char __user *buffer,
		size_t len,
		loff_t *offset)
{
	int i;

	for (i = 0; i < len - 1; i++) {
		if (!(buffer[i] >= '0' && buffer[i] <= '9')) {
			printk("%s: value must be a positive integer\n", __func__);
			return -EINVAL;
		}
	}

	oem_override.value = (unsigned int)simple_strtoul(buffer, NULL, 10);
	oem_override.enabled = 1;

	return len;

}

static int oem_override_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, oem_override_proc_show, PDE_DATA(inode));
}

static const struct file_operations oem_override_proc_fops = {
	.open = oem_override_proc_open,
	.read = seq_read,
	.write = oem_override_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

/************************
* hw_feature_proc
*************************/
hw_info_t *hw_feature_get(void)
{
	static hw_info_t *hw_info;

	if (hw_info == NULL) {
		GS_FACTORYPARA *factory_para;

		factory_para = fact_para_get();
		if (factory_para->checksum !=
			fact_para_checksum(factory_para)) {
			printk("INVALID FACTORY CHECKSUM\n");
		}

		hw_info = hw_get_hw_feature();
	}

	return hw_info;
}

static int hw_feature_proc_show(struct seq_file *m, void *v)
{
	hw_info_t *hw_info = hw_feature_get();
	void *data = (void *)m->private;

	switch ((int)data) {
	case HW_FEATURE_VENDOR:
		seq_printf(m, "%s", hw_info->vendor);
		break;
	case HW_FEATURE_VENDOR_FULLNAME:
		seq_printf(m, "%s", hw_info->vendor_fullname);
		break;
	case HW_FEATURE_MODELSTR:
		seq_printf(m, "%s", hw_info->modelstr);
		break;
	case HW_FEATURE_MODELALIAS:
		seq_printf(m, "%s", hw_info->modelalias);
		break;
	case HW_FEATURE_MODEL_BASE:
		seq_printf(m, "%s", hw_info->gs_model);
		break;
	case HW_FEATURE_OEM_ID:
		seq_printf(m, "%x", hw_info->oem_id);
		break;
	case HW_FEATURE_NUV_DEV:
		seq_printf(m, "%x", hw_info->nuv_dev);
		break;
	case HW_FEATURE_SI3217X_DEV:
		seq_printf(m, "%x", hw_info->si3217x_dev);
		break;
	case HW_FEATURE_NUV_NEON:
		seq_printf(m, "%x", hw_info->nuv_neon);
		break;
	case HW_FEATURE_FWKEY:
		if (hw_info->fwkey[0] == 0)
			seq_printf(m, "nokey");
		else
			seq_printf(m, "%s", hw_info->fwkey);
		break;
	case HW_FEATURE_ETH:
		switch (hw_info->eth) {
		case ETH_TYPE_KSZ8873:
			seq_printf(m, "KSZ8873");
			break;
		case ETH_TYPE_KSZ8863:
			seq_printf(m, "KSZ8863");
			break;
		case ETH_TYPE_ICPLUS:
			seq_printf(m, "IC+");
			break;
		case ETH_TYPE_KSZ8051:
			seq_printf(m, "KSZ8051");
			break;
		case ETH_TYPE_KSZ8031:
			seq_printf(m, "KSZ8031");
			break;
		case ETH_TYPE_RTL8363SB:
			seq_printf(m, "RTL8363SB");
			break;
		case ETH_TYPE_KSZ8081:
			seq_printf(m, "KSZ8081");
			break;
		case ETH_TYPE_QCA8334:
			seq_printf(m, "QCA8334");
			break;
		case ETH_TYPE_RTL8304MB:
			seq_printf(m, "RTL8304MB");
			break;
		case ETH_TYPE_YT8512H:
			seq_printf(m, "YT8512H");
			break;
		case ETH_TYPE_RTL8363NB:
			seq_printf(m, "RTL8363NB");
			break;
		default:
			seq_printf(m, "Unknown");
			break;
		}
		break;
	case HW_FEATURE_ETH_INTERFACE:
		switch (hw_info->eth_interface) {
		case ETH_IF_RGMII:
			seq_printf(m, "RGMII");
			break;
		case ETH_IF_RMII:
			seq_printf(m, "RMII");
			break;
		default:
			seq_printf(m, "Unknown");
			break;
		}
		break;
	case HW_FEATURE_ETH_CLOCK_SOURCE:
		switch (hw_info->eth_clock_source) {
		case ETH_CLOCK_SOURCE_EXTERNAL:
			seq_printf(m, "External");
			break;
		case ETH_CLOCK_SOURCE_INTERNAL:
			seq_printf(m, "Internal");
			break;
		default:
			seq_printf(m, "Unknown");
			break;
		}
		break;
	case HW_FEATURE_HANDSET:
		switch (hw_info->handset) {
		case HANDSET_WIDEBAND:
			seq_printf(m, "wideband");
			break;
		case HANDSET_NARROWBAND:
			seq_printf(m, "narrowband");
			break;
		}
		break;
	case HW_FEATURE_NUM_ACCTS:
		seq_printf(m, "%d", hw_info->num_accts);
		break;
	case HW_FEATURE_EXT:
		seq_printf(m, "%d", hw_info->ext);
		break;
	case HW_FEATURE_BLUETOOTH:
		seq_printf(m, "%d", hw_info->bluetooth);
		break;
	case HW_FEATURE_NUM_MPK:
		seq_printf(m, "%d", hw_info->mpk);
		break;
	case HW_FEATURE_NUM_SOFTKEYS:
		seq_printf(m, "%d", hw_info->num_softkeys);
		break;
	case HW_FEATURE_NO_USB:
		seq_printf(m, "%d", hw_info->usb_unsupported);
		break;
	case HW_FEATURE_CPU:
		if (hw_info->cpu[0] == 0)
			seq_printf(m, "9918");
		else
			seq_printf(m, hw_info->cpu);
		break;
	case HW_FEATURE_FLASH:
		switch (hw_info->flash_device) {
		case FLASH_K9F1G08U0D:
			seq_printf(m, "K9F1G08U0D");
			break;
		case FLASH_K9F1G08U0E:
			seq_printf(m, "K9F1G08U0E");
			break;
		case FLASH_N25Q128A:
			seq_printf(m, "N25Q128A");
			break;
		default:
			seq_printf(m, "Unknown");
		break;
		}
		break;
	case HW_FEATURE_FLASH_SIZE:
		if (hw_info->flash_size[0] == 0)
			seq_printf(m, "16");
		else
			seq_printf(m, "%s", hw_info->flash_size);
		break;

	case HW_FEATURE_DECT:
		seq_printf(m, "%d", hw_info->dect);
		break;
	case HW_FEATURE_DECT_HS:
		seq_printf(m, "%d", hw_info->dect_hs);
		break;
	case HW_FEATURE_DECT_REGION:
		seq_printf(m, "%s", hw_info->dect_region);
		break;
	case HW_FEATURE_POE_CLASS:
		seq_printf(m, "%d", hw_info->poe_class);
		break;
	case HW_FEATURE_RTL8363_SMI:
		seq_printf(m, "%d", hw_info->rtl8363_smi);
		break;
	case HW_FEATURE_DAP1912:
		seq_printf(m, "%d", hw_info->dap1912);
		break;
	case HW_FEATURE_NUM_LINEKEYS:
		seq_printf(m, "%d", hw_info->num_linekeys);
		break;
	case HW_FEATURE_HEADSET_MIC_DETECT:
		seq_printf(m, "%d", hw_info->headset_mic_detect);
		break;
	case HW_FEATURE_MANUFACT_ID:
		seq_printf(m, "%d", hw_info->hw_manufact_id);
		break;
	case HW_FEATURE_MANUFACT_OPT_1:
		seq_printf(m, "%d", hw_info->hw_manufact_opt_1);
		break;
	case HW_FEATURE_MANUFACT_OPT_2:
		seq_printf(m, "%d", hw_info->hw_manufact_opt_2);
		break;
	case HW_FEATURE_SLIC_VENDOR:
		seq_printf(m, "%d", hw_info->slic_vendor);
		break;
	case HW_FEATURE_SLIC_DCDC_TYPE:
		seq_printf(m, "%d", hw_info->slic_dcdc_type);
		break;
	case HW_FEATURE_SLIC_DCDC_VIN:
		seq_printf(m, "%d", hw_info->slic_dcdc_vin);
		break;
	case HW_FEATURE_SLIC_HW_PROT:
		seq_printf(m, "%d", hw_info->slic_hw_prot);
		break;
	case HW_FEATURE_NUM_SLIC_DEV:
		seq_printf(m, "%d", hw_info->num_slic_dev);
		break;
	case HW_FEATURE_NUM_CH_PER_SLIC:
		seq_printf(m, "%d", hw_info->num_ch_per_slic);
		break;
	case HW_FEATURE_NUM_DAA_DEV:
		seq_printf(m, "%d", hw_info->num_daa_dev);
		break;
	case HW_FEATURE_NUM_CH_PER_DAA:
		seq_printf(m, "%d", hw_info->num_ch_per_daa);
		break;
	case HW_FEATURE_NEON_SUPPORT:
		seq_printf(m, "%d", hw_info->neon_support);
		break;
	case HW_FEATURE_MAX_REN_PER_CH:
		seq_printf(m, "%d", hw_info->max_ren_per_ch);
		break;
	case HW_FEATURE_LCD_BACKLIGHT:
		/* If lcd_backlight is not set in Flash, assume that the backlight is there.*/
		seq_printf(m, "%d", (hw_info->lcd_backlight == 0xFF) ?
			1 : hw_info->lcd_backlight);
		break;
	case HW_FEATURE_REQUIRE_PEM:
		seq_printf(m, "%d", hw_info->require_pem);
		break;
	case HW_FEATURE_REQUIRE_SN:
		seq_printf(m, "%d", hw_info->require_sn);
		break;
	case HW_FEATURE_REQUIRE_ADMIN_PWD:
		seq_printf(m, "%d", hw_info->require_admin_pwd);
		break;
	case HW_FEATURE_REQUIRE_XML_KEY:
		seq_printf(m, "%d", hw_info->require_xml_key);
		break;
	case HW_FEATURE_REQUIRE_HASH:
		seq_printf(m, "%d", hw_info->require_hash);
		break;
	case HW_FEATURE_ETH_RST:
		seq_printf(m, "%x", hw_info->eth_rst);
		break;
	default:
		printk("Unknown hw feature\n");
		break;
	}

	return 0;
}

static int hw_feature_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, hw_feature_proc_show, PDE_DATA(inode));
}

static const struct file_operations hw_feature_proc_fops = {
	.open = hw_feature_proc_open,
	.read = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

/******************************************
* sec_feature_proc
* For security table item certificate and private key,
* their length may exceed one page, use seq_file
******************************************/
static void *sec_feature_seq_start(
	struct seq_file *s,
	loff_t *pos)
{
	int len = 0;
	void *data = (void *)s->private;
	sec_info_t *sec_info = sec_info_get();
	void *seq_data = NULL;
	int seq_len = 0;

	switch ((int)data) {
	case SEC_FEATURE_SN:
		if (sec_info->sn.len == 0) {
			printk("serial number not set\n");
			break;
		}
		len = strlen(sec_info->sn.data);
		if (len > sec_info->sn.len)
			printk("sn length is greater than the stored length\n");
		seq_data = (void *)(sec_info->sn.data);
		seq_len = len;
		break;
	case SEC_FEATURE_PRIVATE_KEY:
		if (sec_info->private_key.len == 0) {
			printk("private key not set\n");
			break;
		}
		len = strlen(sec_info->private_key.data);
		if (len > sec_info->private_key.len)
			printk("private key length is greater than the stored length\n");
		seq_data = (void *)(sec_info->private_key.data);
		seq_len = len;
		break;
	case SEC_FEATURE_CERTIFICATE:
		if (sec_info->certificate.len == 0) {
			printk("certificate not set\n");
			break;
		}
		len = strlen(sec_info->certificate.data);
		if (len > sec_info->certificate.len)
			printk("certificate length is greater than the stored length\n");
		seq_data = (void *)(sec_info->certificate.data);
		seq_len = len;
		break;
	case SEC_FEATURE_CHECKSUM:
		if (sec_info->checksum.len == 0) {
			printk("checksum not set\n");
			break;
		}
		len = strlen(sec_info->checksum.data);
		if (len > sec_info->checksum.len)
			printk("checksum length is greater than the stored length\n");
		seq_data = (void *)(sec_info->checksum.data);
		seq_len = len;
		break;
	case SEC_FEATURE_VONAGE_HASH:
		if (sec_info->hash.len == 0) {
			printk("hash not set\n");
			break;
		}
		len = strlen(sec_info->hash.data);
		if (len > sec_info->hash.len)
		printk("hash length is greater than the stored length\n");
		seq_data = (void *)(sec_info->hash.data);
		seq_len = len;
		break;
	case SEC_FEATURE_VONAGE_PASSWORD:
		if (sec_info->password.len == 0) {
			printk("password not set\n");
			break;
		}
		len = strlen(sec_info->password.data);
		if (len > sec_info->password.len)
			printk("password length is greater than the stored length\n");
		seq_data = (void *)(sec_info->password.data);
		seq_len = len;
		break;
	case SEC_FEATURE_VONAGE_XML_KEY:
		if (sec_info->xml_key.len == 0) {
			printk("xml_key not set\n");
			break;
		}
		len = strlen(sec_info->xml_key.data);
		if (len > sec_info->xml_key.len)
			printk("xml key length is greater than the stored length\n");
		seq_data = (void *)(sec_info->xml_key.data);
		seq_len = len;
		break;
	default:
		break;
	}

	if (*pos >= seq_len)
		return NULL;
	else
		return seq_data + *pos;
}

static void *sec_feature_seq_next(struct seq_file *s,
	void *v, loff_t *pos)
{
	(*pos)++;
	return sec_feature_seq_start(s, pos);
}

static void sec_feature_seq_stop(struct seq_file *s, void *v)
{
	return;
}

static int sec_feature_seq_show(struct seq_file *s, void *v)
{
	seq_printf(s, "%c", *(char *)v);
	return 0;
}

static struct seq_operations sec_feature_seq_ops = {
	.start = sec_feature_seq_start,
	.next = sec_feature_seq_next,
	.stop = sec_feature_seq_stop,
	.show = sec_feature_seq_show
};

#ifndef DISABLE_SECURITY_PROCFS_WRITE
static ssize_t sec_feature_proc_write(struct file *file,
		const char __user *buffer,
		size_t len,
		loff_t *offset)
{
	void *data = PDE_DATA(file_inode(file));
	factory_param_sector_t *factory_para;
	GS_FACTORYPARA *fact_para;
	struct mtd_info *mtd;
	size_t mtdread_len;
	uint8_t *factory_buf;
	hw_feature_table_header_t *hw_table;
	hw_feature_attrib_t *hw_feature;
	hw_feature_attrib_t *existing_features[SEC_FEATURE_MAX];
	int existing_feature_count;
	int i;
	int sec_table_size, sec_table_entries;
	char *sec_table;
	int factpara_len = sizeof(GS_FACTORYPARA);
	unsigned int gs_factory_cfg_size_flash = GS_FACTORY_CFG_SIZE_FLASH;

	factory_para = (factory_param_sector_t *)fact_para_get();
	fact_para = fact_para_get();
	if (fact_para->checksum != fact_para_checksum(fact_para)) {
		printk("sec_feature_proc_write: checksum fail\n");
		return -EINVAL;
	}

	/* printk("sec_feature_proc_write: got %lu bytes\n", count); */

	mtd = get_mtd_device_nm("bastic_fact");
	if (mtd <= 0) {
		printk("sec_feature_proc_write: can't get mtd device\n");
		return -EFAULT;
	}

#if (defined(CONFIG_MTD_NAND) && defined(CONFIG_HT))
	gs_factory_cfg_size_flash = mtd->size;
#endif

	factory_buf = vmalloc(gs_factory_cfg_size_flash);
	if (factory_buf == NULL) {
		printk("sec_feature_proc_write: can't allocate buffer\n");
		put_mtd_device(mtd);
		return -EFAULT;
	}

	/* See if there is a security table already */
	hw_table = &factory_para->hw_feature_table_header;
	while (hw_table->hw_feature_table_version !=
		HW_SECURITY_TABLE_VERSION) {
		/* printk("factory_para: %08x hw_table: %08x version: %08x\n",
				factory_para, hw_table, hw_table->hw_feature_table_version);
		*/
		factpara_len += hw_table->hw_feature_table_len * sizeof(uint16_t);
		if (hw_table->hw_feature_table_version == 0) {
			/* create security table */
			hw_table->hw_feature_table_version = HW_SECURITY_TABLE_VERSION;
			hw_table->hw_feature_attrib_count = 0;

			break;
		}
		hw_table = (hw_feature_table_header_t *)
			((uint16_t *)hw_table + hw_table->hw_feature_table_len);
	}

	/* locate any existing entries */
	sec_table_size = sizeof(hw_feature_table_header_t);
	sec_table_entries = 0;
	existing_feature_count = 0;
	if (hw_table->hw_feature_attrib_count > 0) {
		/* printk("found existing security table size %d entries %d\n",
			hw_table->hw_feature_table_len, hw_table->hw_feature_attrib_count);
		*/
		memset(existing_features, 0, sizeof(existing_features));
		hw_feature = (uint8_t *)hw_table + sizeof(hw_feature_table_header_t);
		for (i = 0; i < hw_table->hw_feature_attrib_count; i++) {
			if ((int)data != hw_feature->hw_feature_type) {
				/* printk("found existing feature type %d len %d\n",
					hw_feature->hw_feature_type, hw_feature->hw_feature_len);
				*/
				existing_features[existing_feature_count++] = hw_feature;
				sec_table_size += (sizeof(uint16_t) * (hw_feature->hw_feature_len - 1))
					+ sizeof(hw_feature_attrib_t);
				sec_table_entries++;
			}
			hw_feature = (char *)hw_feature + sizeof(uint16_t) *
				(hw_feature->hw_feature_len - 1) + sizeof(hw_feature_attrib_t);
		}
	}
	/*Add 1 for the last '\0' in the string*/
	sec_table_size += len + 1 + sizeof(hw_feature_attrib_t);
	sec_table_entries++;

	if (sec_table_size & 1)
		sec_table_size++;

	/*
	printk("security table size: %d\n", sec_table_size);
	printk("security table entries: %d\n", sec_table_entries);
	*/

	sec_table = vmalloc(sec_table_size);
	memset(sec_table, 0, sec_table_size);

	((hw_feature_table_header_t *)sec_table)->hw_feature_table_version =
		HW_SECURITY_TABLE_VERSION;
	((hw_feature_table_header_t *)sec_table)->hw_feature_table_len =
		sec_table_size / sizeof(uint16_t);
	((hw_feature_table_header_t *)sec_table)->hw_feature_attrib_count =
		sec_table_entries;

	hw_feature = (hw_feature_attrib_t *)(sec_table + sizeof(hw_feature_table_header_t));
	hw_feature->hw_feature_type = (int)data;
	hw_feature->hw_feature_len = (len + 1) / sizeof(uint16_t);
	if ((len + 1) & 1)
		hw_feature->hw_feature_len++;

	if (copy_from_user(&hw_feature->hw_feature_value, buffer, len)) {
		printk("security feature update: copy error\n");
		return -EFAULT;
	}

	/* printk("New Feature ( type: %d, len: %d )\n",
			hw_feature->hw_feature_type, hw_feature->hw_feature_len);
	*/

	for (i = 0; i < existing_feature_count; i++) {
		hw_feature = (hw_feature_attrib_t *)((char *)hw_feature +
			sizeof(uint16_t) * (hw_feature->hw_feature_len - 1) +
			sizeof(hw_feature_attrib_t));
		memcpy(hw_feature, existing_features[i],
			(existing_features[i]->hw_feature_len - 1) * sizeof(uint16_t) +
			sizeof(hw_feature_attrib_t));
	}

	memcpy(hw_table, sec_table, sec_table_size);

	fact_para->length[0] = factpara_len + sec_table_size;
	fact_para->checksum = fact_para_checksum(fact_para);

	/*
	printk("length: %d\n", fact_para->length[0]);
	printk("cksum: %08x\n", fact_para->checksum);
	*/

	mtd_read(mtd, 0, gs_factory_cfg_size_flash, &mtdread_len, factory_buf);
	if (mtdread_len != gs_factory_cfg_size_flash) {
		printk("sec_feature_proc_write: error reading factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	memcpy(factory_buf + GS_FACTORY_CFG_OFFSET,
		factory_para, GS_FACTORY_CFG_SIZE_FLASH);
	if (erase_helper(mtd, 0, gs_factory_cfg_size_flash) < 0) {
		printk("sec_feature_proc_write: error erasing factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	mtd_write(mtd, 0, gs_factory_cfg_size_flash, &mtdread_len, factory_buf);
	if (mtdread_len != gs_factory_cfg_size_flash) {
		printk("sec_feature_proc_write: error writing factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	set_factory_para_dirty();
	vfree(factory_buf);
	vfree(sec_table);
	put_mtd_device(mtd);

	return len;
}
#endif

static int sec_feature_proc_open(struct inode *inode, struct file *file)
{
	int res = 0;

	res = seq_open(file, &sec_feature_seq_ops);
	if (!res)
		((struct seq_file *)file->private_data)->private = PDE_DATA(inode);

	return res;
}

static const struct file_operations sec_feature_proc_fops = {
	.open = sec_feature_proc_open,
	.read = seq_read,
#ifndef DISABLE_SECURITY_PROCFS_WRITE
	.write = sec_feature_proc_write,
#endif
	.llseek = seq_lseek,
	.release = seq_release
};

/******************************************
* devinfo_proc
******************************************/
static int devinfo_proc_show(struct seq_file *m, void *v)
{
	void *data = (void *)m->private;
	int len = 0;
	int checksum_valid = 1;
	char *nickstr = HW_UNKNOWN;
	char *default_mac = "00:0B:82:FF:FF:FF";
	char *pn;
	char mac_str[64];
	unsigned short int required_bits[4];
	GS_FACTORYPARA *factory_para;
	hw_info_t *hw_info;

	factory_para = fact_para_get();

	if (factory_para->checksum != fact_para_checksum(factory_para)) {
		printk("INVALID CHECKSUM\n");
		checksum_valid = 0;
	}

	hw_info = hw_feature_get();

	switch ((int)data) {
	case imodel_key:
		seq_printf(m, "%04x", factory_para->hw_model);
		break;
	case idev_prefix:
		printk("WARNING -- DEPRECATED PROC ENTRY PREFIX ACCESSED\n");
		seq_printf(m, "%s", hw_info->modelstr);
		break;
	case irequired_bits:
		if (checksum_valid) {
			required_bits[0] = factory_para->RequiredBits[0];
			required_bits[1] = factory_para->RequiredBits[1];
			required_bits[2] = factory_para->RequiredBits[2];
			required_bits[3] = factory_para->RequiredBits[3];
			/* This gets byte swapped in flash by hexpnxbin, we want all instances
			* of required bits to look the same for convenience */
			required_bits[0] = ((required_bits[0] & 0x00FF) << 8) |
				((required_bits[0] & 0xFF00) >> 8);
			required_bits[1] = ((required_bits[1] & 0x00FF) << 8) |
				((required_bits[1] & 0xFF00) >> 8);
			required_bits[2] = ((required_bits[2] & 0x00FF) << 8) |
				((required_bits[2] & 0xFF00) >> 8);
			required_bits[3] = ((required_bits[3] & 0x00FF) << 8) |
				((required_bits[3] & 0xFF00) >> 8);
		} else
			memset(required_bits, 0, sizeof(required_bits));

		seq_printf(m, "%04x%04x%04x%04x",
			required_bits[0],
			required_bits[1],
			required_bits[2],
			required_bits[3]);
			break;
	case iPN:
		if (checksum_valid) {
			pn = (char *)factory_para->PartNumber;
			seq_printf(m, "%s", pn);
		} else
			seq_printf(m, HW_UNKNOWN);
		break;
	case idev_alias:
		nickstr = hw_info->modelalias;
		seq_printf(m, "%s", nickstr);
		break;
	case idev_id:
		nickstr = hw_info->modelstr;
		seq_printf(m, "%s", nickstr);
		break;
	case idev_data:
		len = ((factory_para->length[1] << 16) | factory_para->length[0]);
		if (len > sizeof(GS_FACTORYPARA))
		len = sizeof(GS_FACTORYPARA);
		/* How did we pass checksum if this is true? */
		for (; len > 0; len--) {
			if (checksum_valid)
				seq_printf(m, "%c", *(char *)factory_para);
			else
				seq_printf(m, "%c", 0);
		}
		break;
	case idev_mac:
		if (checksum_valid) {
			sprintf(mac_str, "%02x:%02x:%02x:%02x:%02x:%02x",
			factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
			factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
			factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF);
			seq_printf(m, "%s", mac_str);
		} else
			seq_printf(m, "%s", default_mac);
		break;
	case idev_rev:
		if (checksum_valid) {
			pn = (char *)factory_para->PartNumber;
			sprintf(mac_str, "V%c.%c%c", pn[8], pn[9], pn[10]);
			seq_printf(m, "%s", mac_str);
		} else
			seq_printf(m, HW_UNKNOWN);
		break;
	case ifact_ver:
		/* This is the default fact cfg values of the legacy fact sector.
		* RevID0:01 01 01 00#
		* RevID1:00 00 00 00#
		* RevID2:01 01 01 00#
		* RevID3:00 00 00 00#
		* The first 2 bytes will be repurposed for Fact cfg sector version control
		*/

		if (checksum_valid) {
			char * fact_ver = NULL;
			char default_rev_id[16] = {
				0x01, 0x01, 0x01, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x01, 0x01, 0x01, 0x00,
				0x00, 0x00, 0x00, 0x00};

			char fact_cfg_version_str[64] = {0};

			fact_ver = (char *)factory_para->RevID;

			/* If the RevID is same as the default one then do not print it. */
			if (memcmp(default_rev_id, fact_ver, 4))
				sprintf(fact_cfg_version_str, "%d", fact_ver[0]);

			seq_printf(m, "%s", fact_cfg_version_str);
		} else
			seq_printf(m, "%s", HW_UNKNOWN);

		break;
	}

	return 0;
}

static ssize_t devinfo_proc_write(struct file *file,
		const char __user *buffer,
		size_t len,
		loff_t *offset)
{
	void *data = PDE_DATA(file_inode(file));
	GS_FACTORYPARA *factory_para;
	unsigned short mac[3];
	struct mtd_info *mtd;
	int mtdread_len;
	uint8_t *factory_buf;
	unsigned int gs_factory_cfg_size_flash = GS_FACTORY_CFG_SIZE_FLASH;

	if ((int)data != idev_data)
		return -EINVAL;

	if (len != sizeof(mac)) {
		printk("devinfo_proc_write: invalid size %zu\n", len);
		return -EINVAL;
	}

	if (copy_from_user(mac, buffer, 6)) {
		printk("devinfo_proc_write: copy error\n");
		return -EFAULT;
	}

	factory_para = fact_para_get();
	if (factory_para->checksum !=
		fact_para_checksum(factory_para)) {
		printk("devinfo_proc_write: checksum fail\n");
		return -EINVAL;
	}

	if (!(factory_para->mac[0] == 0x0b00 &&
		factory_para->mac[1] == 0x0082 &&
		factory_para->mac[2] == 0x0000)) {
		printk("devinfo_proc_write: mac already set\n");
		return -EFAULT;
	}

	mtd = get_mtd_device_nm("bastic_fact");
	if (mtd <= 0) {
		printk("devinfo_proc_write: can't get mtd device\n");
		return -EFAULT;
	}

#if (defined(CONFIG_MTD_NAND) && defined(CONFIG_HT))
	gs_factory_cfg_size_flash = mtd->size;
#endif

	factory_buf = vmalloc(gs_factory_cfg_size_flash);
	if (factory_buf == NULL) {
		printk("devinfo_proc_write: can't allocate buffer\n");
		put_mtd_device(mtd);
		return -EFAULT;
	}

	printk("devinfo_proc_write: setting MAC %04x:%04x:%04x\n",
		mac[0], mac[1], mac[2]);

	factory_para->mac[0] = mac[0];
	factory_para->mac[1] = mac[1];
	factory_para->mac[2] = mac[2];

	factory_para->checksum = fact_para_checksum(factory_para);

	mtd_read(mtd, 0, gs_factory_cfg_size_flash, &mtdread_len, factory_buf);
	if (len != gs_factory_cfg_size_flash) {
		printk("devinfo_proc_write: error reading factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	memcpy(factory_buf + GS_FACTORY_CFG_OFFSET,
		factory_para, GS_FACTORY_CFG_SIZE_FLASH);
	if (erase_helper(mtd, 0, gs_factory_cfg_size_flash) < 0) {
		printk("devinfo_proc_write: error erasing factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	mtd_write(mtd, 0, gs_factory_cfg_size_flash, &mtdread_len, factory_buf);
	if (mtdread_len != gs_factory_cfg_size_flash) {
		printk("devinfo_proc_write: error writing factory part\n");
		put_mtd_device(mtd);
		vfree(factory_buf);
		return -EFAULT;
	}

	set_factory_para_dirty();
	vfree(factory_buf);
	put_mtd_device(mtd);

	return len;
}

static int devinfo_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, devinfo_proc_show, PDE_DATA(inode));
}

static const struct file_operations devinfo_proc_fops = {
	.open = devinfo_proc_open,
	.read = seq_read,
	.write = devinfo_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

/******************************************
* cfg_reload_proc
******************************************/
static int cfg_reload_proc_show(struct seq_file *m, void *v)
{
	seq_printf(m, "%u", cfg_reload_count);
	return 0;
}

static ssize_t cfg_reload_proc_write(struct file *file,
		const char __user *buffer,
		size_t len,
		loff_t *offset)
{
	/* use this to trigger factory reload after writing security table by factory tool */
	set_factory_para_dirty();
	cfg_reload_count++;

	return len;
}

static int cfg_reload_proc_open(struct inode *inode,
	struct file *file)
{
	return single_open(file, cfg_reload_proc_show, PDE_DATA(inode));
}

static const struct file_operations cfg_reload_proc_fops = {
	.open = cfg_reload_proc_open,
	.read = seq_read,
	.write = cfg_reload_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

/************************
 * ProcFS Entries
 ************************/
struct gs_procfs_entry gs_procfs_misc_entry[] = {
#define GS_PROCFS_ENTRY(name) \
	{#name, &name##_proc_fops, 0}

	GS_PROCFS_ENTRY(reset_button),
	GS_PROCFS_ENTRY(oem_override),
	GS_PROCFS_ENTRY(cfg_reload),

#undef GS_PROCFS_ENTRY
};
#define NUM_PROCFS_MISC_ENTRIES \
	(sizeof(gs_procfs_misc_entry) / sizeof(struct gs_procfs_entry))

/************************
 * Devinfo Entries
  ************************/
struct gs_procfs_entry gs_devinfo_entry[] = {
#define GS_DEVINFO_ENTRY(name, value) \
	{#name, &devinfo_proc_fops, value}
	GS_DEVINFO_ENTRY(PN, iPN),
	GS_DEVINFO_ENTRY(dev_alias, idev_alias),
	GS_DEVINFO_ENTRY(dev_data, idev_data),
	GS_DEVINFO_ENTRY(dev_id, idev_id),
	GS_DEVINFO_ENTRY(dev_mac, idev_mac),
	GS_DEVINFO_ENTRY(dev_rev, idev_rev),
	GS_DEVINFO_ENTRY(required_bits, irequired_bits),
	GS_DEVINFO_ENTRY(model_key, imodel_key),
	GS_DEVINFO_ENTRY(dev_prefix, idev_prefix),
	GS_DEVINFO_ENTRY(fact_ver, ifact_ver),
#undef GS_DEVINFO_ENTRY
	};
#define NUM_DEVINFO_ENTRIES \
	(sizeof(gs_devinfo_entry) / sizeof(struct gs_procfs_entry))

/************************
 * HW Feature Entries
  ************************/
struct gs_procfs_entry gs_hw_features_entry[] = {
#define HW_FEATURE_ENTRY(name, value) {#name, &hw_feature_proc_fops, value}
	HW_FEATURE_ENTRY(vendor, HW_FEATURE_VENDOR),
	HW_FEATURE_ENTRY(model_str, HW_FEATURE_MODELSTR),
	HW_FEATURE_ENTRY(model_alias, HW_FEATURE_MODELALIAS),
	HW_FEATURE_ENTRY(default_fwkey, HW_FEATURE_FWKEY),
	HW_FEATURE_ENTRY(oem_id, HW_FEATURE_OEM_ID),
	HW_FEATURE_ENTRY(vendor_fullname, HW_FEATURE_VENDOR_FULLNAME),
	HW_FEATURE_ENTRY(eth, HW_FEATURE_ETH),
	HW_FEATURE_ENTRY(eth_interface, HW_FEATURE_ETH_INTERFACE),
	HW_FEATURE_ENTRY(handset, HW_FEATURE_HANDSET),
	HW_FEATURE_ENTRY(nuv_dev, HW_FEATURE_NUV_DEV),
	HW_FEATURE_ENTRY(si3217x_dev, HW_FEATURE_SI3217X_DEV),
	HW_FEATURE_ENTRY(nuv_neon, HW_FEATURE_NUV_NEON),
	HW_FEATURE_ENTRY(num_accts, HW_FEATURE_NUM_ACCTS),
	HW_FEATURE_ENTRY(ext, HW_FEATURE_EXT),
	HW_FEATURE_ENTRY(num_mpk, HW_FEATURE_NUM_MPK),
	HW_FEATURE_ENTRY(num_softkeys, HW_FEATURE_NUM_SOFTKEYS),
	HW_FEATURE_ENTRY(gs_model, HW_FEATURE_MODEL_BASE),
	HW_FEATURE_ENTRY(cpu, HW_FEATURE_CPU),
	HW_FEATURE_ENTRY(flash, HW_FEATURE_FLASH),
	HW_FEATURE_ENTRY(flash_size, HW_FEATURE_FLASH_SIZE),
	HW_FEATURE_ENTRY(dect, HW_FEATURE_DECT),
	HW_FEATURE_ENTRY(dect_hs, HW_FEATURE_DECT_HS),
	HW_FEATURE_ENTRY(dect_region, HW_FEATURE_DECT_REGION),
	HW_FEATURE_ENTRY(bluetooth, HW_FEATURE_BLUETOOTH),
	HW_FEATURE_ENTRY(poe_class, HW_FEATURE_POE_CLASS),
	HW_FEATURE_ENTRY(rtl8363_smi, HW_FEATURE_RTL8363_SMI),
	HW_FEATURE_ENTRY(dap1912, HW_FEATURE_DAP1912),
	HW_FEATURE_ENTRY(num_linekeys, HW_FEATURE_NUM_LINEKEYS),
	HW_FEATURE_ENTRY(headset_mic_detect, HW_FEATURE_HEADSET_MIC_DETECT),
	HW_FEATURE_ENTRY(hw_manufact_id, HW_FEATURE_MANUFACT_ID),
	HW_FEATURE_ENTRY(hw_manufact_opt_1, HW_FEATURE_MANUFACT_OPT_1),
	HW_FEATURE_ENTRY(hw_manufact_opt_2, HW_FEATURE_MANUFACT_OPT_2),
	HW_FEATURE_ENTRY(slic_vendor, HW_FEATURE_SLIC_VENDOR),
	HW_FEATURE_ENTRY(slic_dcdc_type, HW_FEATURE_SLIC_DCDC_TYPE),
	HW_FEATURE_ENTRY(slic_dcdc_vin, HW_FEATURE_SLIC_DCDC_VIN),
	HW_FEATURE_ENTRY(slic_hw_prot, HW_FEATURE_SLIC_HW_PROT),
	HW_FEATURE_ENTRY(num_slic_dev, HW_FEATURE_NUM_SLIC_DEV),
	HW_FEATURE_ENTRY(num_ch_per_slic, HW_FEATURE_NUM_CH_PER_SLIC),
	HW_FEATURE_ENTRY(num_daa_dev, HW_FEATURE_NUM_DAA_DEV),
	HW_FEATURE_ENTRY(num_ch_per_daa, HW_FEATURE_NUM_CH_PER_DAA),
	HW_FEATURE_ENTRY(neon_support, HW_FEATURE_NEON_SUPPORT),
	HW_FEATURE_ENTRY(max_ren_per_ch, HW_FEATURE_MAX_REN_PER_CH),
	HW_FEATURE_ENTRY(lcd_backlight, HW_FEATURE_LCD_BACKLIGHT),
	HW_FEATURE_ENTRY(require_pem, HW_FEATURE_REQUIRE_PEM),
	HW_FEATURE_ENTRY(require_sn, HW_FEATURE_REQUIRE_SN),
	HW_FEATURE_ENTRY(require_admin_pwd, HW_FEATURE_REQUIRE_ADMIN_PWD),
	HW_FEATURE_ENTRY(require_xml_key, HW_FEATURE_REQUIRE_XML_KEY),
	HW_FEATURE_ENTRY(require_hash, HW_FEATURE_REQUIRE_HASH),
	HW_FEATURE_ENTRY(eth_clock_source, HW_FEATURE_ETH_CLOCK_SOURCE),
	HW_FEATURE_ENTRY(eth_reset, HW_FEATURE_ETH_RST),
};
#define NUM_HW_FEATURE_ENTRIES \
	(sizeof(gs_hw_features_entry) / sizeof(struct gs_procfs_entry))


/************************
 * Sec Feature Entries
  ************************/
struct gs_procfs_entry sec_feature_entry[] = {
#define SEC_FEATURE_ENTRY(name, value) {#name, &sec_feature_proc_fops, value}
	SEC_FEATURE_ENTRY(private_key, SEC_FEATURE_PRIVATE_KEY),
	SEC_FEATURE_ENTRY(certificate, SEC_FEATURE_CERTIFICATE),
	SEC_FEATURE_ENTRY(checksum, SEC_FEATURE_CHECKSUM),
	SEC_FEATURE_ENTRY(sn, SEC_FEATURE_SN),
	SEC_FEATURE_ENTRY(hash, SEC_FEATURE_VONAGE_HASH),
	SEC_FEATURE_ENTRY(password, SEC_FEATURE_VONAGE_PASSWORD),
	SEC_FEATURE_ENTRY(xml_key, SEC_FEATURE_VONAGE_XML_KEY),
};
#define NUM_SEC_FEATURE_ENTRIES \
	(sizeof(sec_feature_entry) / sizeof(struct gs_procfs_entry))

static int __init gs_procfs_init(void)
{
	int i;
	struct proc_dir_entry *gs_procfs_root;
	struct proc_dir_entry *gs_devinfo_root;
	struct proc_dir_entry *gs_hw_feature_root;
	struct proc_dir_entry *gs_sec_feature_root;
	struct proc_dir_entry *procfs_entry;

	memset(&oem_override, 0, sizeof(oem_override));

	gs_procfs_root = proc_mkdir("gxp", NULL);

	/* install misc entries */
	for (i = 0; i < NUM_PROCFS_MISC_ENTRIES; i++) {
		procfs_entry = proc_create(gs_procfs_misc_entry[i].name,
			S_IFREG | S_IRUGO, gs_procfs_root,
			gs_procfs_misc_entry[i].fops);
		if (!procfs_entry) {
			printk("gs_procfs_misc_entry creation failure\n");
			return -1;
		}
	}

	/* install devinfo entries */
	gs_devinfo_root = proc_mkdir("dev_info", gs_procfs_root);
	for (i = 0; i < NUM_DEVINFO_ENTRIES; i++) {
		procfs_entry = proc_create_data(gs_devinfo_entry[i].name,
			S_IFREG | S_IRUGO, gs_devinfo_root,
			gs_devinfo_entry[i].fops,
			(int *)gs_devinfo_entry[i].data);
		if (!procfs_entry) {
			printk("gs_devinfo_entry creation failure\n");
			return -1;
		}
	}

	/* install hw_features entries */
	gs_hw_feature_root = proc_mkdir("hw_features", gs_devinfo_root);
	for (i = 0; i < NUM_HW_FEATURE_ENTRIES; i++) {
		procfs_entry = proc_create_data(gs_hw_features_entry[i].name,
			S_IFREG | S_IRUGO, gs_hw_feature_root,
			gs_hw_features_entry[i].fops,
			(int *)gs_hw_features_entry[i].data);
		if (!procfs_entry) {
			printk("gs_hw_features_entry creation failure\n");
			return -1;
		}
	}

	/* install security entries */
	gs_sec_feature_root = proc_mkdir("security", gs_devinfo_root);
	for (i = 0; i < NUM_SEC_FEATURE_ENTRIES; i++) {
		procfs_entry = proc_create_data(sec_feature_entry[i].name,
			S_IFREG | S_IRUGO, gs_sec_feature_root,
			sec_feature_entry[i].fops,
			(int *)sec_feature_entry[i].data);
		if (!procfs_entry) {
			printk("gs procfs sec_feature creation failure\n");
			return -1;
		}
	}

	return 0;
}

static void __exit gs_procfs_exit(void)
{
	return;
}

char *get_syslog_id(void)
{
	static int init;

	if (!init) {
		GS_FACTORYPARA *factory_para;
		hw_info_t *hw_info;

		hw_info = hw_get_hw_feature();
		factory_para = fact_para_get();
		memset(syslog_id, 0, sizeof(syslog_id));
		sprintf(syslog_id, "%s_KERNEL [%02x:%02x:%02x:%02x:%02x:%02x]",
			hw_info->modelalias,
			factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
			factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
			factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF);
	}

	return syslog_id;
}

EXPORT_SYMBOL(get_syslog_id);
EXPORT_SYMBOL(hw_feature_get);
EXPORT_SYMBOL(fact_para_get);

subsys_initcall(gs_procfs_init);
module_exit(gs_procfs_exit);

#if defined(CONFIG_HT)
static int reset_button_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;

	reset_gpio = of_get_named_gpio(np, "RESET_BUTTOM", 0);
	printk("reset_button: %s\n", gpio_is_valid(reset_gpio) ? "valid" : "invalid");
	if (gpio_is_valid(reset_gpio)) {
		gpio_request(reset_gpio, "grandstream-reset");
		gpio_direction_input(reset_gpio);
	}

	return 0;
}

static int reset_button_remove(struct platform_device *pdev)
{
	gpio_free(reset_gpio);
	return 0;
}

static const struct of_device_id reset_button_of_match[] = {
	{ .compatible = "grandstream,reset", },
	{},
};
MODULE_DEVICE_TABLE(of, reset_button_of_match);

static struct platform_driver reset_button_driver = {
	.probe		= reset_button_probe,
	.remove		= reset_button_remove,
	.driver		= {
		.name		= "reset-button-driver",
		.owner		= THIS_MODULE,
		.of_match_table = reset_button_of_match,
	},
};

static int __init gs_reset_button_init(void)
{
	return platform_driver_register(&reset_button_driver);
}

static void __exit gs_reset_button_exit(void)
{
	platform_driver_unregister(&reset_button_driver);
}

module_init(gs_reset_button_init);
module_exit(gs_reset_button_exit);
#endif

