/****************************************************************************
 *
 * FILENAME:        $RCSfile: provision.c
 *
 * LAST MODIFIED:   Date: 2017/05/17
 *
 * DESCRIPTION:     provision support
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2017 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

#include <linux/types.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <mach/factory.h>
#include <linux/module.h>
#include <provision.h>

#include <linux/mtd/ubi.h>

/*
* must be updated for every new model inthe table
* reference from drivers/mtd/mtdpart.c
*/
struct mtd_part {
	struct mtd_info mtd;
	struct mtd_info *master;
	uint64_t offset;
	struct list_head list;
};
#define PART(x)  ((struct mtd_part *)(x))

typedef struct {
	char *name;
	int mtd_num;
	int mtd_type;
	char *ubi_volname;
} prov_parts_t;

static int prov_parts_init = 0;
prov_parts_t prov_parts[] = {
#if defined(CONFIG_HT818)
	{"uboot",  -1, MTD_NANDFLASH, NULL},
	{"uboot2", -1, MTD_NANDFLASH, NULL},
	{"core",   -1, MTD_UBIVOLUME, "core"},
	{"core2",  -1, MTD_UBIVOLUME, "core"},
	{"base",   -1, MTD_UBIVOLUME, "base"},
	{"base2",  -1, MTD_UBIVOLUME, "base"},
	{"prog",   -1, MTD_UBIVOLUME, "prog"},
	{"prog2",  -1, MTD_UBIVOLUME, "prog"},
#else
#error NO PRODUCT DEFINED
#endif
};

#define NUM_PROV_ENTRIES	(sizeof(prov_parts) / sizeof(prov_parts_t))

static int provision_parts_group;

static int __init prov_parts_group_setup(char *str)
{
	if (str[0] == '1')
		provision_parts_group = 0;
	else
		provision_parts_group = 1;

	return 1;
}
__setup("bootpartition=", prov_parts_group_setup);

extern int gluebi_get_mtdname(struct mtd_info *mtd, char*name, int name_len);
static int prov_parts_setup(void)
{
	int i, num = 0;
	struct mtd_info *mtd;
	char *match_name;

	if (prov_parts_init)
		return 0;

	do {
		mtd  = get_mtd_device(NULL, num);
		if (IS_ERR(mtd)) {
			break;
		}

		for (i = 0; i < NUM_PROV_ENTRIES; i++) {
			if (prov_parts[i].mtd_num >= 0)
				continue;

			/* MTD_NANDFLASH */
			if (MTD_UBIVOLUME != prov_parts[i].mtd_type)
				match_name = prov_parts[i].name;
			else
				match_name = prov_parts[i].ubi_volname;

			if (strcmp(mtd->name, match_name) == 0) {
				if (prov_parts[i].mtd_type != mtd->type) {
					printk("%s provision part type is incorrect\n",
						prov_parts[i].name);
					continue;
				}
				prov_parts[i].mtd_num = mtd->index;
				printk("Found provision part %s(mtd%d)\n",
					prov_parts[i].name, mtd->index);
				break;
			}
		}

		put_mtd_device(mtd);
		num++;
	} while(1);
	prov_parts_init = 1;

	return 0;
}

void gs_add_mtd_header_offset(struct mtd_info *mtd)
{
	int i;
	char *cmp_name;

	mtd->hdr_offset = 0;
	for (i = 0; i < NUM_PROV_ENTRIES; i++) {
		if (prov_parts[i].mtd_type != MTD_UBIVOLUME)
			cmp_name = prov_parts[i].name;
		else
			cmp_name = prov_parts[i].ubi_volname;

		if (strcmp(mtd->name, cmp_name) == 0) {
			mtd->hdr_offset = HEADER_BLOCK_SIZE;
			break;
		}
	}
}

static int prov_part_grp_proc_show(struct seq_file *m, void *v)
{
	seq_printf(m, "%d", provision_parts_group);
	return 0;
}

static int prov_part_grp_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, prov_part_grp_proc_show, PDE_DATA(inode));
}

static const struct file_operations prov_part_grp_proc_ops = {
	.open = prov_part_grp_proc_open,
	.read = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int prov_read_imghdr_proc_show(struct seq_file *m, void *v)
{
	int partition = (int)m->private;
	size_t len, retlen = 0;
	pImageHeader p;
	int j = 0;
	char *data = NULL;
	struct mtd_info *mtd;
	loff_t off = 0;

	p = (pImageHeader)vmalloc(sizeof(ImageHeader));
	if (!p)
		return -ENOMEM;

	len = sizeof(ImageHeader);
	memset(p, 0, sizeof(ImageHeader));

	if (prov_parts[partition].mtd_num < 0) {
		printk("mtd index for provision part is invalid\n");
		return 0;
	}

	mtd  = get_mtd_device(NULL, prov_parts[partition].mtd_num);
	if (!mtd) {
		printk("%s: unable to open mtd device\n", __func__);
		return -ENOENT;
	}

	if ( prov_parts[partition].mtd_type ==  MTD_NANDFLASH ) {
		while ( mtd->_block_isbad (mtd, off) ) {
			off += mtd->erasesize;
			if ( off >= mtd->size ) break;
		}
	}
	if ( off < mtd->size ) {
		mtd_read(mtd, off, len, &retlen, (char *)p);
		if (retlen != len) {
			printk("%s: read_imghdr size %zu != %zu \n",
				__func__, retlen, len);
		}
	} else {
		printk("Partition %s is all bad blocks\n", prov_parts[partition].name);
		len = 0;
	}

	put_mtd_device(mtd);

	data = (char *)p;
	for (j = 0; j < len; j++) {
		seq_printf(m, "%c", *data);
		data++;
	}

	vfree(p);

	return 0;
}

static int prov_read_imghdr_proc_open(struct inode *inode, struct file *file)
{
	prov_parts_setup();

	return single_open(file, prov_read_imghdr_proc_show, PDE_DATA(inode));
}

static const struct file_operations prov_read_imghdr_proc_ops = {
	.open  = prov_read_imghdr_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int __init gs_provfs_init(void)
{
	int i;
	struct proc_dir_entry *gs_provfs_root;
	struct proc_dir_entry *procfs_entry;

	gs_provfs_root = proc_mkdir("provision", NULL);

	procfs_entry = proc_create("provision_group",
		S_IFREG | S_IRUGO, gs_provfs_root, &prov_part_grp_proc_ops);

	gs_provfs_root = proc_mkdir("partition", gs_provfs_root);

	for (i = 0; i < NUM_PROV_ENTRIES; i++) {
		procfs_entry = proc_create_data(prov_parts[i].name,
			S_IFREG | S_IRUGO, gs_provfs_root,
			&prov_read_imghdr_proc_ops,
			(int *)(i)); /* Partition Number */
	}

	return 0;
}

static void __exit gs_provfs_exit(void)
{
	return;
}

subsys_initcall(gs_provfs_init);
module_exit(gs_provfs_exit);

/* EOF */
