/*
 *  Copyright (C) 2016 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Common Clock Framework support for DSPG platforms
 */
#include <linux/clk-provider.h>
#include <linux/delay.h>
#include <linux/of.h>

#include "clk-pll.h"

#define PLL_UNIT_CTRL					0x00
#define PLL_UNIT_LOCK_CNT				0x08
#define PLL_CFG1					0x0c
#define PLL_CFG2					0x10

static void dvf101_pll_read_configuration(struct dspg_pll *pll)
{
	pll->refdiv = dspg_pll_read_reg_field(pll, PLL_CFG2,  0, 0x3f);
	pll->fbdiv = dspg_pll_read_reg_field(pll, PLL_CFG2,  8,  0xfff);
	pll->postdiv1 = dspg_pll_read_reg_field(pll, PLL_CFG2,  24, 0x7);
	pll->postdiv2 = dspg_pll_read_reg_field(pll, PLL_CFG2, 28,  0x7);
}

static void dvf101_pll_write_configuration(struct dspg_pll *pll)
{
	dspg_pll_write_reg_field(pll, PLL_CFG1,  0,  0x3f, pll->refdiv);
	dspg_pll_write_reg_field(pll, PLL_CFG2,  8, 0xfff, pll->fbdiv);
	dspg_pll_write_reg_field(pll, PLL_CFG2, 24,   0x7, pll->postdiv1);
	dspg_pll_write_reg_field(pll, PLL_CFG2, 28,   0x7, pll->postdiv2);
}

static int dvf101_pll_in_automatic_mode(struct dspg_pll *pll)
{
	if ((dspg_pll_read_reg(pll, PLL_UNIT_CTRL) >> 5) & 1)
		return 0;
	return 1;
}

static int dvf101_pll_is_locked(struct dspg_pll *pll)
{
	if ((dspg_pll_read_reg(pll, PLL_CFG1) >> 24) & 1)
		return 1;
	return 0;
}

static int dvf101_pll_is_power_down(struct dspg_pll *pll)
{
	if ((dspg_pll_read_reg(pll, PLL_UNIT_CTRL) >> 2) & 1)
		return 1;
	return 0;
}

static int dvf101_pll_clk_in_use(struct dspg_pll *pll)
{
	return (dspg_pll_read_reg(pll, PLL_UNIT_CTRL) >> 7) & 1;
}

static void dvf101_pll_set_power_down(struct dspg_pll *pll, unsigned long value)
{
	dspg_pll_write_reg_field(pll, PLL_UNIT_CTRL,  2, 1, value);
}

static void dvf101_pll_set_bypass(struct dspg_pll *pll, unsigned long value)
{
	dspg_pll_write_reg_field(pll, PLL_UNIT_CTRL,  4, 1, value);
}

static void dvf101_pll_set_lock_cnt(struct dspg_pll *pll, unsigned long value)
{
	dspg_pll_write_reg_field(pll, PLL_UNIT_LOCK_CNT,  0, 0xfffff, value);
}

static void dvf101_init_pll(struct dspg_pll *pll, struct device_node *node,
			    struct clk_init_data *init)
{
	dvf101_pll_read_configuration(pll);

	if (dvf101_pll_in_automatic_mode(pll))
		init->flags |= CLK_IGNORE_UNUSED;

	init->flags |= CLK_SET_RATE_NO_REPARENT;

	pll->lock_cnt = 1499;
	of_property_read_u32(node, "lock-cnt", (u32 *)&pll->lock_cnt);
}

static int dvf101_pll_enable(struct clk_hw *hw)
{
	struct dspg_pll *pll = to_dspg_pll(hw);

	if (!dvf101_pll_is_power_down(pll))
		/* PLL already enabled */
		return 0;

	if (dvf101_pll_in_automatic_mode(pll)) {
		dvf101_pll_write_configuration(pll);

		/* set lock count */
		dvf101_pll_set_lock_cnt(pll, pll->lock_cnt);

		dvf101_pll_set_power_down(pll, 0);
		while (!dvf101_pll_clk_in_use(pll))
			;
		return 0;
	}
	dvf101_pll_write_configuration(pll);

	/* set lock count */
	dvf101_pll_set_lock_cnt(pll, pll->lock_cnt);

	dvf101_pll_set_power_down(pll, 0);
	while (!dvf101_pll_is_locked(pll))
			;
	dvf101_pll_set_bypass(pll, 1);
	while (!dvf101_pll_clk_in_use(pll))
		;

	return 0;
}

static void dvf101_pll_disable(struct clk_hw *hw)
{
	struct dspg_pll *pll = to_dspg_pll(hw);

	if (dvf101_pll_in_automatic_mode(pll)) {
		dvf101_pll_set_power_down(pll, 1);
		while (dvf101_pll_clk_in_use(pll))
			;
		return;
	}

	dvf101_pll_set_bypass(pll, 0);
	while (dvf101_pll_clk_in_use(pll))
		;
	dvf101_pll_set_power_down(pll, 1);
}

static int dvf101_pll_is_enabled(struct clk_hw *hw)
{
	struct dspg_pll *pll = to_dspg_pll(hw);

	if (dvf101_pll_in_automatic_mode(pll)) {
		/* pll is always on in automatic mode */
		return 1;
	}

	if (dvf101_pll_is_power_down(pll))
		return 0;
	if (!dvf101_pll_is_locked(pll))
		return 0;
	return 1;
}

static unsigned long dvf101_pll_recalc_rate(struct clk_hw *hw,
					    unsigned long parent_rate)
{
	struct dspg_pll *pll = to_dspg_pll(hw);

	if (!dvf101_pll_clk_in_use(pll))
		return parent_rate;

	dvf101_pll_read_configuration(pll);

	return ((parent_rate / pll->refdiv) * pll->fbdiv) /
		(pll->postdiv1 * pll->postdiv2);
}

static const struct clk_ops dspg_dvf101_pll_ops = {
	.enable = dvf101_pll_enable,
	.disable = dvf101_pll_disable,
	.is_enabled = dvf101_pll_is_enabled,
	.recalc_rate = dvf101_pll_recalc_rate,
};

const struct dspg_pll_data dspg_dvf101_pll_data = {
	.init = dvf101_init_pll,
	.ops = &dspg_dvf101_pll_ops,
};

static const struct clk_ops dspg_dvf101_pll_always_on_ops = {
	.is_enabled = dvf101_pll_is_enabled,
	.recalc_rate = dvf101_pll_recalc_rate,
};

const struct dspg_pll_data dspg_dvf101_pll_always_on_data = {
	.init = dvf101_init_pll,
	.ops = &dspg_dvf101_pll_always_on_ops,
};
