/*
 * drivers/gpio/dspg-gpio.c
 *
 *  Copyright (c) 2012 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <asm-generic/gpio.h>
#include <linux/errno.h>
#include <linux/gpio-dspg.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/platform_device.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <linux/types.h>

#define EXTINT_CFG          0x2A0
#define EXTINT_EDGE_CLEAR   0x2AC

#define EXTINT_FLAG_MASKED	(1 << 0)
#define EXTINT_FLAG_BOTH_EDGES	(1 << 1)

struct dspg_gpio;

struct extint {
	unsigned bit;
	u16 hwirq;
	u16 flags;
	struct dspg_gpio *dspg_gpio;
	struct irq_domain *irq_domain;
	int upstream_irq;
};

struct dspg_gpio {
	struct gpio_chip chip;
	void *regs;
	struct dspg_gpio_pdata pdata;
	struct irq_domain *extint_domain;
	int nr_extints;
	struct extint extints[16];
};

static DEFINE_SPINLOCK(extint_lock);

#define GET_REG(priv, reg, gpio) ( \
	__raw_readl((priv)->regs + (__BANK(gpio) * 0x60) + (reg)) & (1 << __PIN(gpio)) \
	)

#define SET_REG(priv, reg, gpio) \
	__raw_writel(1 << __PIN(gpio), (priv)->regs + (__BANK(gpio) * 0x60) + (reg))


static int gpio_get_pull_enable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_PULL_CTRL, gpio);
}

/*
 * Returns: 0: Output, 1: Input
 */
static int gpio_get_direction(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_DIR, gpio);
}

static void dspg_gpio_set_value(struct dspg_gpio *dspg_gpio, unsigned gpio, int value)
{
	if (value)
		SET_REG(dspg_gpio, XGP_DATA_SET, gpio);
	else
		SET_REG(dspg_gpio, XGP_DATA_CLR, gpio);
}

static int gpio_get_enable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_EN, gpio);
}

static int gpio_get_opendrain(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_OD, gpio);
}

static int gpio_get_keeper(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_KEEP, gpio);
}

/*
 * Returns: 0: pull down, 1: pull up
 */
static int gpio_get_pull_selection(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_PULL_TYPE, gpio);
}

/*
 * Tests whether the Pull-Up/Dn internal resistor can be disconnected, in order
 * to save power.
 */
static void tst_pull_disconnect(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	if (!gpio_get_pull_enable(dspg_gpio, gpio))
		return;

	if (gpio_get_direction(dspg_gpio, gpio) == 0) {
		/* Output gpio */

		/* Note:
		 * "Wired-OR" is achived by enabling "Open-Drain" and
		 * No Pull-Down resistor, but the user may connect
		 * a Pull-Up resistor, if no such External resistor
		 * exists. So "open-drain" must be excluded !
		 */
		if (!gpio_get_opendrain(dspg_gpio, gpio))
			SET_REG(dspg_gpio, XGP_PULL_DIS, gpio);
	} else {
		/* Input gpio */

		if (gpio_get_keeper(dspg_gpio, gpio))
			SET_REG(dspg_gpio, XGP_PULL_DIS, gpio);
	}
}

static int
dspg_gpiolib_request(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	if (dspg_gpio->pdata.nx_masks[__BANK(offset)] & (1 << __PIN(offset))) {
		dev_warn(chip->dev, "request for non-existent %cGPIO%-2d denied!\n",
			 'A' + __BANK(offset), __PIN(offset));
		return -ENODEV;
	}

	if (!gpio_get_enable(dspg_gpio, offset)) {
		dev_warn(chip->dev, "request for %cGPIO%-2d denied: "
			 "pin is not configured as GPIO!\n",
			 'A' + __BANK(offset), __PIN(offset));
		return -ENODEV;
	}

	return 0;
}

static int
dspg_gpiolib_direction_input(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	SET_REG(dspg_gpio, XGP_DIR_IN, offset);
	tst_pull_disconnect(dspg_gpio, offset);

	return 0;
}

static int
dspg_gpiolib_direction_output(struct gpio_chip *chip, unsigned offset, int value)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	dspg_gpio_set_value(dspg_gpio, offset, value);
	SET_REG(dspg_gpio, XGP_DIR_OUT, offset);
	tst_pull_disconnect(dspg_gpio, offset);

	return 0;
}

static int
dspg_gpiolib_get(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	return !!GET_REG(dspg_gpio, XGP_DATA, offset);
}

static void
dspg_gpiolib_set(struct gpio_chip *chip, unsigned offset, int value)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	dspg_gpio_set_value(dspg_gpio, offset, value);
}

static void
dspg_gpiolib_show(struct seq_file *s, struct gpio_chip *chip)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);
	const char *label;
	unsigned i;

	for (i = 0; i < chip->ngpio; i++) {
		if (!(label = gpiochip_is_requested(chip, i)))
			continue;

		seq_printf(s, " %cGPIO%-2d (%-20.20s) %s %s%s",
			'A' + __BANK(i), __PIN(i), label,
			gpio_get_direction(dspg_gpio, i) ? "in " : "out",
			dspg_gpiolib_get(chip, i) ? "hi" : "lo",
			gpio_get_pull_enable(dspg_gpio, i)
				? (gpio_get_pull_selection(dspg_gpio, i) ? " pu" : " pd")
				: ""
			);
		seq_printf(s, "\n");
	}
}

static int dspg_gpiolib_to_irq(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);
	int i;

	/* search for matching GPIO and then map through exting IRQ domain */
	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		if (dspg_gpio->pdata.extint_gpios[i] == offset)
			return irq_create_mapping(dspg_gpio->extint_domain, i);
	}

	return -ENODEV;
}

#ifdef CONFIG_OF
static int dspg_gpiolib_xlate(struct gpio_chip *gc,
			      const struct of_phandle_args *gpiospec, u32 *flags)
{
	struct dspg_gpio *dspg_gpio = container_of(gc, struct dspg_gpio, chip);
	unsigned pin;

	pin = gpiospec->args[0] * 32 + gpiospec->args[1];

	if (pin >= gc->ngpio || gpiospec->args[1] > 32)
		return -EINVAL;

	if (dspg_gpio->pdata.nx_masks[__BANK(pin)] & (1 << __PIN(pin)))
		return -ENODEV;

	if (flags)
		*flags = gpiospec->args[2];

	return pin;
}
#endif

static void extint_adjust_pol(struct extint *extint)
{
	unsigned long value;
	int pol, shift = extint->bit * 2;

	pol = !dspg_gpiolib_get(&extint->dspg_gpio->chip,
		extint->dspg_gpio->pdata.extint_gpios[extint->hwirq]);

	value = readl(extint->dspg_gpio->regs + EXTINT_CFG);
	value &= ~(0x1UL << shift);
	value |= pol << shift;
	writel(value, extint->dspg_gpio->regs + EXTINT_CFG);
}

static void extint_irq_ack(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	writel(1 << extint->bit, extint->dspg_gpio->regs + EXTINT_EDGE_CLEAR);

	if (extint->flags & EXTINT_FLAG_BOTH_EDGES) {
		spin_lock_irqsave(&extint_lock, flags);
		extint_adjust_pol(extint);
		spin_unlock_irqrestore(&extint_lock, flags);
	}
}

static void extint_irq_mask(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	spin_lock_irqsave(&extint_lock, flags);
	if (!(extint->flags & EXTINT_FLAG_MASKED)) {
		disable_irq_nosync(extint->upstream_irq);
		extint->flags |= EXTINT_FLAG_MASKED;
	}
	spin_unlock_irqrestore(&extint_lock, flags);
}

static void extint_irq_unmask(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	spin_lock_irqsave(&extint_lock, flags);
	if (extint->flags & EXTINT_FLAG_MASKED) {
		if (extint->flags & EXTINT_FLAG_BOTH_EDGES)
			extint_adjust_pol(extint);

		enable_irq(extint->upstream_irq);
		extint->flags &= ~EXTINT_FLAG_MASKED;
	}
	spin_unlock_irqrestore(&extint_lock, flags);
}

static int extint_irq_set_type(struct irq_data *data, unsigned int type)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags, value;
	int pol = 0; /* 0 = active low; 1 = active high */
	int mode = 0;     /* 0 = level; 1 = edge */
	int shift;

	type &= IRQ_TYPE_SENSE_MASK;

	spin_lock_irqsave(&extint_lock, flags);

	if ((type & IRQ_TYPE_EDGE_BOTH) == IRQ_TYPE_EDGE_BOTH) {
		/*
		 * emulate EDGE_BOTH by setting LEVEL_HIGH or LEVEL_LOW
		 * depending on current gpio value
		 */
		extint->flags |= EXTINT_FLAG_BOTH_EDGES;
		pol = !dspg_gpiolib_get(&extint->dspg_gpio->chip,
			extint->dspg_gpio->pdata.extint_gpios[extint->hwirq]);
	} else {
		extint->flags &= ~EXTINT_FLAG_BOTH_EDGES;

		if (type & IRQ_TYPE_EDGE_BOTH) {
			/* specific edge type irq */
			mode = 1;

			if (type & IRQ_TYPE_EDGE_RISING)
				pol = 1;
		} else if (type & IRQ_TYPE_LEVEL_HIGH)
			pol = 1;
	}

	shift = extint->bit << 1; /* extint * 2 */

	value = readl(extint->dspg_gpio->regs + EXTINT_CFG);
	value &= ~(0x3UL << shift);
	value |= ((mode <<1) | pol) << shift;
	writel(value, extint->dspg_gpio->regs + EXTINT_CFG);

	spin_unlock_irqrestore(&extint_lock, flags);

	if (type & (IRQ_TYPE_LEVEL_LOW | IRQ_TYPE_LEVEL_HIGH))
		irq_set_handler_locked(data, handle_level_irq);
	else if (type & (IRQ_TYPE_EDGE_FALLING | IRQ_TYPE_EDGE_RISING))
		irq_set_handler_locked(data, handle_edge_irq);

	return 0;
}

static int extint_irq_set_wake(struct irq_data *data, unsigned int enable)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	return irq_set_irq_wake(extint->upstream_irq, enable);
}

static struct irq_chip extint_irq_chip = {
	.name		= "gpio",
	.irq_ack	= extint_irq_ack,
	.irq_mask	= extint_irq_mask,
	.irq_unmask	= extint_irq_unmask,
	.irq_set_type	= extint_irq_set_type,
	.irq_set_wake	= extint_irq_set_wake,
};

static irqreturn_t extint_irq_handler(int irq, void *dev_id)
{
	unsigned long flags;
	struct extint *extint = dev_id;

	local_irq_save(flags);
	generic_handle_irq(irq_linear_revmap(extint->irq_domain, extint->hwirq));
	local_irq_restore(flags);

	return IRQ_HANDLED;
}

/*
 * This lock class tells lockdep that EXTINTs are in a different category than
 * their parents, so it won't report false recursion.
 */
static struct lock_class_key gpio_lock_class;

static int extint_irq_map(struct irq_domain *dom, unsigned int irq,
			  irq_hw_number_t hw_irq)
{
	struct dspg_gpio *dspg_gpio = dom->host_data;

	irq_set_lockdep_class(irq, &gpio_lock_class);
	irq_set_chip_data(irq, &(dspg_gpio->extints[hw_irq]));
	irq_set_chip_and_handler(irq, &extint_irq_chip, handle_level_irq);

	return 0;
}

static struct irq_domain_ops extint_irq_ops = {
	.map	= extint_irq_map,
	.xlate	= irq_domain_xlate_twocell,
};

static int __init dspg_gpio_probe(struct platform_device *pdev)
{
	struct dspg_gpio_pdata *pdata = pdev->dev.platform_data;
	struct dspg_gpio *dspg_gpio;
	struct resource *res;
	int i, ret = 0, extsize = 0;

	dspg_gpio = devm_kzalloc(&pdev->dev, sizeof(*dspg_gpio), GFP_KERNEL);
	if (!dspg_gpio)
		return -ENOMEM;

	if (!pdata) {
		struct device_node *np = pdev->dev.of_node;
		u32 extint_gpios[48];

		if (!np) {
			dev_err(&pdev->dev, "missing platform data!\n");
			return -EINVAL;
		}

		ret = of_property_read_u32(np, "dspg,banks",
			&dspg_gpio->pdata.num_banks);
		if (ret) {
			dev_err(&pdev->dev, "missing 'dspg,banks' property!\n");
			return ret;
		}
		if (dspg_gpio->pdata.num_banks > ARRAY_SIZE(dspg_gpio->pdata.nx_masks)) {
			dev_err(&pdev->dev, "too many banks!\n");
			return -EINVAL;
		}

		ret = of_property_read_u32_array(np, "dspg,nx_masks",
			dspg_gpio->pdata.nx_masks, dspg_gpio->pdata.num_banks);
		if (ret && ret != -EINVAL) {
			dev_err(&pdev->dev, "invalid 'dspg,nx_masks' property!\n");
			return ret;
		}

		dspg_gpio->nr_extints = 16;
		ret = (int)of_find_property(np, "dspg,extints", &extsize);
		if (ret)
			dspg_gpio->nr_extints = extsize / sizeof(u32) / 3;

		ret = of_property_read_u32_array(np, "dspg,extints",
			extint_gpios, dspg_gpio->nr_extints * 3);
		if (ret) {
			dev_err(&pdev->dev, "missing 'dspg,extints' property!\n");
			return ret;
		}

		for (i = 0; i < dspg_gpio->nr_extints; i++) {
			dspg_gpio->pdata.extint_gpios[i] =
				extint_gpios[i*3] * 32 + extint_gpios[i*3+1];
			dspg_gpio->extints[i].bit = extint_gpios[i*3+2];
		}
	} else
		dspg_gpio->pdata = *pdata;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -EINVAL;

	dspg_gpio->regs = devm_ioremap_resource(&pdev->dev, res);
	if (!dspg_gpio->regs) {
		dev_err(&pdev->dev, "failed to map registers\n");
		return -ENOMEM;
	}

	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		res = platform_get_resource(pdev, IORESOURCE_IRQ, i);
		if (!res) {
			dev_err(&pdev->dev, "Missing IRQ resource\n");
			return -EINVAL;
		}
		dspg_gpio->extints[i].upstream_irq = res->start;
	}

	dspg_gpio->chip.dev			= &pdev->dev;
	dspg_gpio->chip.label			= "gpio-dspg";
	dspg_gpio->chip.owner			= THIS_MODULE;
	dspg_gpio->chip.request			= dspg_gpiolib_request;
	dspg_gpio->chip.direction_input		= dspg_gpiolib_direction_input;
	dspg_gpio->chip.get			= dspg_gpiolib_get;
	dspg_gpio->chip.direction_output	= dspg_gpiolib_direction_output;
	dspg_gpio->chip.set			= dspg_gpiolib_set;
	dspg_gpio->chip.dbg_show		= dspg_gpiolib_show;
	dspg_gpio->chip.to_irq			= dspg_gpiolib_to_irq;
	dspg_gpio->chip.ngpio			= dspg_gpio->pdata.num_banks * 32;
#ifdef CONFIG_OF
	dspg_gpio->chip.of_node			= pdev->dev.of_node;
	dspg_gpio->chip.of_gpio_n_cells		= 3;
	dspg_gpio->chip.of_xlate		= dspg_gpiolib_xlate;
#endif

	dspg_gpio->extint_domain = irq_domain_add_linear(pdev->dev.of_node,
		dspg_gpio->nr_extints, &extint_irq_ops, dspg_gpio);
	if (!dspg_gpio->extint_domain) {
		dev_err(&pdev->dev, "cannot register IRQ domain!\n");
		return -ENOMEM;
	}

	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		int irq = dspg_gpio->extints[i].upstream_irq;

		dspg_gpio->extints[i].dspg_gpio = dspg_gpio;
		dspg_gpio->extints[i].irq_domain = dspg_gpio->extint_domain;
		dspg_gpio->extints[i].hwirq = i;
		dspg_gpio->extints[i].flags = EXTINT_FLAG_MASKED;

		irq_set_status_flags(irq, IRQ_NOAUTOEN);
		ret = devm_request_irq(&pdev->dev, irq, extint_irq_handler, 0,
				       "gpio-dspg", &(dspg_gpio->extints[i]));
		if (ret) {
			dev_err(&pdev->dev, "cannot request upstream irq %d!\n",
				irq);
			return ret;
		}
	}

	ret = gpiochip_add(&dspg_gpio->chip);
	if (ret)
		return ret;

	dev_info(&pdev->dev, "%d banks\n", dspg_gpio->pdata.num_banks);

	platform_set_drvdata(pdev, dspg_gpio);
	return 0;
}

static int dspg_gpio_remove(struct platform_device *pdev)
{
	struct dspg_gpio *dspg_gpio = platform_get_drvdata(pdev);

	gpiochip_remove(&dspg_gpio->chip);

	return 0;
}

static const struct of_device_id dspg_gpio_dt_match[] = {
	{ .compatible = "dspg,gpio", },
	{}
};

static struct platform_driver dspg_gpio_driver = {
	.driver		= {
		.name	= "gpio-dspg",
		.owner	= THIS_MODULE,
		.of_match_table = dspg_gpio_dt_match,
	},
	.remove		= dspg_gpio_remove,
};

static int __init dspg_gpio_init(void)
{
	return platform_driver_probe(&dspg_gpio_driver, dspg_gpio_probe);
}
subsys_initcall(dspg_gpio_init);
