/**
 *  Copyright (C) 2016 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/mmc/host.h>
#include <linux/mmc/dw_mmc.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>

#include <linux/sysfs.h>

#include "dw_mmc.h"
#include "dw_mmc-pltfm.h"

#define DW_MCI_DSPG_MAX_TUNE_VALS	20
/*#define DW_MCI_DSPG_DEBUG_TUNING*/

struct dw_mci_dspg_priv_data {
	struct clk *clk_d;
	struct clk *clk_s;
	struct clk *main_in;

	int shift_d;
	int shift_s;
	int tune_drive_clk;
};

static inline int
find_best_value(int results[], int values, int interval_start)
{
	int best_value;
	int i = interval_start;
	int longest = 0;
	int cur_length = 0;
	int cur_sum = 0;

	do {
		if (results[i % values] == 1) {
			cur_length++;
			cur_sum += i;
		} else if (results[i % values] == 0) {
			if (cur_length > longest) {
				longest = cur_length;
				best_value = (cur_sum / cur_length) % values;
			}
			cur_length = 0;
			cur_sum = 0;
		} else {
			return -EINVAL;
		}

		i++;
	} while ((i % values) != interval_start);

	return best_value;
}

static int
dw_mci_dspg_tune_clk_in(struct dw_mci_slot *slot, u32 opcode,
			int *best_phase)
{
	struct dw_mci *host = slot->host;
	struct dw_mci_dspg_priv_data *priv = host->priv;
	struct mmc_host *mmc = slot->mmc;
	int results[DW_MCI_DSPG_MAX_TUNE_VALS];
	unsigned int tried_phases;
	int i;
	int interval_start = -1;
	int good_phases = 0;

	for (tried_phases = 0; tried_phases < DW_MCI_DSPG_MAX_TUNE_VALS;
	     tried_phases++) {
		if (clk_set_phase(priv->clk_s, tried_phases))
			break;
		udelay(100);
#ifdef DW_MCI_DSPG_DEBUG_TUNING
		dev_info(host->dev, "    sample phase %d:", tried_phases);
#endif

		if (mmc_send_tuning(mmc, opcode, NULL)) {
#ifdef DW_MCI_DSPG_DEBUG_TUNING
			dev_info(host->dev, "    ERR");
#endif
			results[tried_phases] = 0;
		} else {
#ifdef DW_MCI_DSPG_DEBUG_TUNING
			dev_info(host->dev, "    OK");
#endif
			results[tried_phases] = 1;
			good_phases++;
		}
	}

	if (!tried_phases) {
		dev_err(host->dev, "could not set drive clock phase");
		return -EIO;
	}

	for (i = 0; i < tried_phases; i++)
		if (results[i] == 0 && results[(i + 1) % tried_phases] == 1)
			interval_start = (i + 1) % tried_phases;

	if (interval_start < 0) {
		/* all phases are valid or invalid */
		*best_phase = 0;
		return good_phases;
	}

	*best_phase = find_best_value(results, tried_phases, interval_start);

	return good_phases;
}

static int
dw_mci_dspg_execute_tuning(struct dw_mci_slot *slot, u32 opcode)
{
	struct dw_mci *host = slot->host;
	struct dw_mci_dspg_priv_data *priv = host->priv;
	int in_phase;

	if (priv->tune_drive_clk) {
		int best_out_phase = -1;
		int best_in_phase = -1;
		int best = 0;
		int out_phase, cur;

		for (out_phase = 0; out_phase < DW_MCI_DSPG_MAX_TUNE_VALS;
		     out_phase++) {
			if (clk_set_phase(priv->clk_d, out_phase))
				break;
			udelay(100);

			cur = dw_mci_dspg_tune_clk_in(slot, opcode, &in_phase);
#ifdef DW_MCI_DSPG_DEBUG_TUNING
			dev_info(host->dev, "drive phase %d score: %d",
				 out_phase, cur);
#endif
			if (cur > best) {
				best = cur;
				best_out_phase = out_phase;
				best_in_phase = in_phase;
			}
		}

		if (best_out_phase < 0 || best_in_phase < 0)
			return -EIO;

		clk_set_phase(priv->clk_d, best_out_phase);
		clk_set_phase(priv->clk_s, best_in_phase);
		udelay(100);

		priv->shift_d = best_out_phase;
		priv->shift_s = best_in_phase;
		priv->tune_drive_clk = 0;
	} else {
		dw_mci_dspg_tune_clk_in(slot, opcode, &in_phase);

		if (in_phase < 0)
			return -EIO;

		clk_set_phase(priv->clk_s, in_phase);
		udelay(100);

		priv->shift_s = in_phase;
	}

	return 0;
}

static int
dw_mci_dspg_parse_dt99(struct dw_mci *host)
{
	struct dw_mci_dspg_priv_data *priv;
	const struct dw_mci_drv_data *drv_data = host->drv_data;

	priv = devm_kzalloc(host->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	if (of_property_read_bool(host->dev->of_node, "dspg,non-removable"))
		*drv_data->caps |= MMC_CAP_NONREMOVABLE;

	host->priv = priv;
	dev_set_drvdata(host->dev, host);

	return 0;
}

static ssize_t
tune_store(struct device *dev, struct device_attribute *attr,
	   const char *buf, size_t size)
{
	struct dw_mci *host = dev_get_drvdata(dev);
	struct dw_mci_dspg_priv_data *priv = host->priv;

	priv->tune_drive_clk = 1;

	return size;
}

static ssize_t
drive_shift_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct dw_mci *host = dev_get_drvdata(dev);
	struct dw_mci_dspg_priv_data *data = host->priv;

	return sprintf(buf, "%d\n", data->shift_d);
}

static ssize_t
sample_shift_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct dw_mci *host = dev_get_drvdata(dev);
	struct dw_mci_dspg_priv_data *data = host->priv;

	return sprintf(buf, "%d\n", data->shift_s);
}

static DEVICE_ATTR_WO(tune);
static DEVICE_ATTR_RO(drive_shift);
static DEVICE_ATTR_RO(sample_shift);

static struct attribute *dw_mci_dspg_attrs[] = {
	&dev_attr_tune.attr,
	&dev_attr_drive_shift.attr,
	&dev_attr_sample_shift.attr,
	NULL
};
ATTRIBUTE_GROUPS(dw_mci_dspg);

static int
dw_mci_dspg_parse_dt(struct dw_mci *host)
{
	struct dw_mci_dspg_priv_data *priv;
	const struct dw_mci_drv_data *drv_data = host->drv_data;

	priv = devm_kzalloc(host->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	priv->clk_d = devm_clk_get(host->dev, "clk-d");
	if (IS_ERR(priv->clk_d)) {
		dev_err(host->dev,
			"Could not parse 'clk-d' property.  Tuning will be disabled.\n");
		return -EIO;
	}
	priv->clk_s = devm_clk_get(host->dev, "clk-s");
	if (IS_ERR(priv->clk_s)) {
		dev_err(host->dev,
			"Could not parse 'clk-s' property.  Tuning will be disabled.\n");
		return -EIO;
	}
	priv->main_in = devm_clk_get(host->dev, "main-in");
	if (IS_ERR(priv->main_in))
		dev_info(host->dev,
			"Could not parse 'main-in' property.\n");

	if (of_property_read_u32(host->dev->of_node, "dspg,shift-d",
				 &priv->shift_d))
		priv->shift_d = -1;
	if (of_property_read_u32(host->dev->of_node, "dspg,shift-s",
				 &priv->shift_s))
		priv->shift_s = -1;

	if (of_property_read_bool(host->dev->of_node, "dspg,non-removable"))
		*drv_data->caps |= MMC_CAP_NONREMOVABLE;

	host->priv = priv;
	dev_set_drvdata(host->dev, host);

	if (sysfs_create_group(&host->dev->kobj, dw_mci_dspg_groups[0]))
		dev_info(host->dev,
			"Could not create sysfs group.  Drive clock tuning will not be accessible.");

	return 0;
}

static int
dw_mci_dspg_setup_clocks(struct dw_mci *host)
{
	struct dw_mci_dspg_priv_data *priv = host->priv;

	clk_prepare_enable(priv->clk_d);
	clk_prepare_enable(priv->clk_s);
	if (!IS_ERR_OR_NULL(priv->main_in))
		clk_prepare_enable(priv->main_in);

	if (priv->shift_d >= 0)
		clk_set_phase(priv->clk_d, priv->shift_d);
	if (priv->shift_s >= 0)
		clk_set_phase(priv->clk_s, priv->shift_s);

	return 0;
}

static void
dw_mci_dspg_set_ios(struct dw_mci *host, struct mmc_ios *ios)
{
	struct clk *main_in =
		((struct dw_mci_dspg_priv_data *)host->priv)->main_in;

	if (ios->clock == 0)
		return;

	if (!IS_ERR_OR_NULL(main_in)) {
		if (ios->clock <= 400000)
			clk_set_rate(main_in,    4000000);
		else if (ios->clock <=  25000000)
			clk_set_rate(main_in,  250000000);
		else if (ios->clock <=  52000000)
			clk_set_rate(main_in,  500000000);
		else if (ios->clock <= 100000000)
			clk_set_rate(main_in, 1000000000);
		else if (ios->clock <= 152000000)
			clk_set_rate(main_in, 1400000000);
		else
			dev_err(host->dev,
				"%u exceeds maximum clock rate 152 MHz",
				ios->clock);
	}

	/* the block internal divider should always be set to 1 (bypass) */
	clk_set_rate(host->ciu_clk, ios->clock);
	host->bus_hz = clk_get_rate(host->ciu_clk);
}

static unsigned long dw_mci_dspg_caps_mmc_evb99[1] = {
	MMC_CAP_SD_HIGHSPEED | MMC_CAP_MMC_HIGHSPEED,
};

static unsigned long dw_mci_dspg_caps_mmc_evb101[1] = {
	/* EVB101 does not support automatic I/O voltage switching yet */
	/*MMC_CAP_1_8V_DDR | */MMC_CAP_SD_HIGHSPEED | MMC_CAP_MMC_HIGHSPEED /*|
	MMC_CAP_UHS_SDR12 | MMC_CAP_UHS_SDR25 | MMC_CAP_UHS_SDR50 |
	MMC_CAP_UHS_SDR104 | MMC_CAP_UHS_DDR50*/,
};

static unsigned long dw_mci_dspg_caps_emmc_evb101[1] = {
	/* UHS is not supported as we cannot switch to 1.8V; DDR52 mode at 3.3V
	 * and 4Bit mode are implicitly assumed to be supported
	 */
	MMC_CAP_SD_HIGHSPEED | MMC_CAP_MMC_HIGHSPEED,
};

static void
dw_mci_dspg_set_ios99(struct dw_mci *host, struct mmc_ios *ios)
{
	int i;

	for (i = 0; i < host->num_slots; i++) {
		struct dw_mci_slot *slot = host->slot[i];

		if (!slot)
			continue;

		set_bit(DW_MMC_CARD_NO_USE_HOLD, &slot->flags);
	}
}

static const struct dw_mci_drv_data dspg_drv_data_mmc_evb99 = {
	.set_ios		= dw_mci_dspg_set_ios99,
	.parse_dt		= dw_mci_dspg_parse_dt99,
	.caps			= dw_mci_dspg_caps_mmc_evb99,
};

static const struct dw_mci_drv_data dspg_drv_data_mmc_evb101 = {
	.execute_tuning		= dw_mci_dspg_execute_tuning,
	.set_ios		= dw_mci_dspg_set_ios,
	.parse_dt		= dw_mci_dspg_parse_dt,
	.setup_clock		= dw_mci_dspg_setup_clocks,
	.caps			= dw_mci_dspg_caps_mmc_evb101,
};

static const struct dw_mci_drv_data dspg_drv_data_emmc_evb101 = {
	.execute_tuning		= dw_mci_dspg_execute_tuning,
	.set_ios		= dw_mci_dspg_set_ios,
	.parse_dt		= dw_mci_dspg_parse_dt,
	.setup_clock		= dw_mci_dspg_setup_clocks,
	.caps			= dw_mci_dspg_caps_emmc_evb101,
};

static const struct of_device_id dw_mci_dspg_match[] = {
	{
		.compatible = "dspg,dw-mmc-evb99",
		.data = &dspg_drv_data_mmc_evb99
	},
	{
		.compatible = "dspg,dw-mmc-evb101",
		.data = &dspg_drv_data_mmc_evb101
	},
	{
		.compatible = "dspg,dw-emmc-evb101",
		.data = &dspg_drv_data_emmc_evb101
	},
	{ /* sentinel */ },
};

static int
dw_mci_dspg_probe(struct platform_device *pdev)
{
	const struct dw_mci_drv_data *drv_data;
	const struct of_device_id *match;

	if (!pdev->dev.of_node)
		return -ENODEV;

	match = of_match_node(dw_mci_dspg_match, pdev->dev.of_node);
	drv_data = match->data;

	return dw_mci_pltfm_register(pdev, drv_data);
}

static int
dw_mci_dspg_remove(struct platform_device *pdev)
{
	struct dw_mci *host = dev_get_drvdata(&pdev->dev);
	struct dw_mci_dspg_priv_data *priv = host->priv;

	if (!IS_ERR_OR_NULL(priv->main_in))
		clk_disable_unprepare(priv->main_in);
	clk_disable_unprepare(priv->clk_s);
	clk_disable_unprepare(priv->clk_d);

	return dw_mci_pltfm_remove(pdev);
}

#ifdef CONFIG_PM_SLEEP
static int
dw_mci_dspg_suspend(struct device *dev)
{
	struct dw_mci *host = dev_get_drvdata(dev);

	return dw_mci_suspend(host);
}

static int
dw_mci_dspg_resume(struct device *dev)
{
	struct dw_mci *host = dev_get_drvdata(dev);

	return dw_mci_resume(host);
}
#else
static int dw_mci_dspg_suspend(struct device *dev) { return 0; }
static int dw_mci_dspg_resume(struct device *dev) { return 0; }
#endif

static SIMPLE_DEV_PM_OPS(dw_mci_dspg_pmops,
			 dw_mci_dspg_suspend,
			 dw_mci_dspg_resume);

static struct platform_driver dw_mci_dspg_pltfm_driver = {
	.probe	= dw_mci_dspg_probe,
	.remove	= dw_mci_dspg_remove,
	.driver	= {
		.name		= "dwmmc_dspg",
		.of_match_table	= dw_mci_dspg_match,
		.pm		= &dw_mci_dspg_pmops,
	},
};

module_platform_driver(dw_mci_dspg_pltfm_driver);

MODULE_AUTHOR("DSPG Technologies GmbH");
MODULE_AUTHOR("Christoph Kloecker <christoph.kloecker@dspg.com>");
MODULE_DESCRIPTION("DSPG Specific DW-MSHC Driver Extension");
MODULE_LICENSE("GPL v2");
