/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright © 2010 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * Modified by Grandstream for dvf99
 *
 */
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/phy.h>
#include "athrs17_phy.h"

/* for PHY related debugging messages */
#undef S17_PHY_DEBUG
#ifdef S17_PHY_DEBUG
#define DPRINTF_PHY(_fmt,...) do { \
	printk(MODULE_NAME":"_fmt, __VA_ARGS__);\
} while (0)
#define DPRINTF_PHY2(_fmt,...)
#else
#define DPRINTF_PHY(_fmt,...)
#define DPRINTF_PHY2(_fmt,...)
#endif

#define S17_WAN_PORT    5
#define S17_WAN_VID 2
#define S17_LAN_VID 1

#define DRV_LOG(DBG_SW, X0, X1, X2, X3, X4, X5, X6)
#define DRV_MSG(x,a,b,c,d,e,f)
#define DRV_PRINT(DBG_SW,X)

#define S17_LAN_PORT_VLAN          S17_LAN_VID
#define S17_WAN_PORT_VLAN         S17_WAN_VID

#define ENET_UNIT_GE1 1
#define ENET_UNIT_GE0 0 /* Connected to the switch */

#define TRUE    1
#define FALSE   0

#define S17_PHY0_ADDR   0x0
#define S17_PHY1_ADDR   0x1
#define S17_PHY2_ADDR   0x2
#define S17_PHY3_ADDR   0x3
#define S17_PHY4_ADDR   0x4
#define S17_IND_PHY 4

#define MODULE_NAME "ATHRS17"

#define S17_PHY_MAX 5

/* Range of valid PHY IDs is [MIN..MAX] */
#define S17_ID_MIN 0
#define S17_ID_MAX (S17_PHY_MAX-1)

/* Convenience macros to access myPhyInfo */
#define S17_IS_ENET_PORT(phyUnit) (athrPhyInfo[phyUnit].isEnetPort)
#define S17_IS_PHY_ALIVE(phyUnit) (athrPhyInfo[phyUnit].isPhyAlive)
#define S17_ETHUNIT(phyUnit) (athrPhyInfo[phyUnit].ethUnit)
#define S17_PHYBASE(phyUnit) (athrPhyInfo[phyUnit].phyBase)
#define S17_PHYADDR(phyUnit) (athrPhyInfo[phyUnit].phyAddr)


//=================
// Typedefs
//=================
/* PHY selections and access functions */
typedef enum
{
	PHY_SRCPORT_INFO,
	PHY_PORTINFO_SIZE,
} PHY_CAP_TYPE;

typedef enum
{
	PHY_SRCPORT_NONE,
	PHY_SRCPORT_VLANTAG,
	PHY_SRCPORT_TRAILER,
} PHY_SRCPORT_TYPE;

/*
 * Track per-PHY port information.
 */
typedef struct
{
	BOOL   isEnetPort;       /* normal enet port */
	BOOL   isPhyAlive;       /* last known state of link */
	int    ethUnit;          /* MAC associated with this phy port */
	uint32_t phyBase;
	uint32_t phyAddr;          /* PHY registers associated with this phy port */
	uint32_t VLANTableSetting; /* Value to be written to VLAN table */
#ifdef S17_8023AZ_FEATURE
	int azFeature;
#endif
} athrPhyInfo_t;

static struct net_device* g_netdev;
/*
 * Per-PHY information, indexed by PHY unit number
 * MAC port 0 - CPU port 0x100
 * All ports are connected to GE0 of the SoC MAC
 * LAN/WAN seperation by VLAN tags (port 0-3 with VLANID 1; port 4 with VLANID 2
 */
static athrPhyInfo_t athrPhyInfo[] =
{
	{
		TRUE,   /* phy port 0 -- MAC port 1 0x200 */
		FALSE,
		ENET_UNIT_GE1,
		0,
		S17_PHY0_ADDR,
		S17_LAN_PORT_VLAN
	},

	{
		TRUE,   /* phy port 1 -- MAC port 2 0x300 */
		FALSE,
		ENET_UNIT_GE1,
		0,
		S17_PHY1_ADDR,
		S17_LAN_PORT_VLAN
	},

	{
		TRUE,   /* phy port 2 -- MAC port 3 0x400 */
		FALSE,
		ENET_UNIT_GE1,
		0,
		S17_PHY2_ADDR,
		S17_LAN_PORT_VLAN
	},

	{
		TRUE,   /* phy port 3 -- MAC port 4 0x500 */
		FALSE,
		ENET_UNIT_GE1,
		0,
		S17_PHY3_ADDR,
		S17_LAN_PORT_VLAN
	},

	{
		TRUE,   /* phy port 4 -- WAN port or MAC port 5 0x600 */
		FALSE,
		ENET_UNIT_GE0,
		0,
		S17_PHY4_ADDR,
		S17_WAN_PORT_VLAN   /* set as WAN port */
	},

	{
		FALSE,  /* phy port 5 -- CPU port (no RJ45 connector) */
		TRUE,
		ENET_UNIT_GE0,
		0,
		0x00,
		S17_LAN_PORT_VLAN    /* Send to all ports */
	},
};

static const athr_switch_intf_t default_rgmii_only =
{
	.mac0_type      = RGMII,
	.mac6_type      = NOT_CONNECT,
	.mac5_type      = DONT_CARE,
	.wan_port       = S17_WAN_PORT,
	.vlan_id        = 0,
	.wan2_port      = 0xff,
	.use_vlan_table = 0,
};

static const athr_switch_intf_t default_rgmii_sgmii =
{
	.mac0_type      = SGMII,
	.mac6_type      = RGMII,
	.mac5_type      = DONT_CARE,
	.wan_port       = S17_WAN_PORT,
	.vlan_id        = 1,
	.wan2_port      = 0xff,
	.use_vlan_table = 0,
};

static const athr_switch_intf_t default_sgmii_only =
{
	.mac0_type      = SGMII,
	.mac6_type      = NOT_CONNECT,
	.mac5_type      = DONT_CARE,
	.wan_port       = S17_WAN_PORT,
	.vlan_id        = 1,
	.wan2_port      = 0xff,
	.use_vlan_table = 0,
};

static const athr_switch_intf_t default_rgmii_rgmii =
{
	.mac0_type      = RGMII,
	.mac6_type      = RGMII,
	.mac5_type      = DONT_CARE,
	.wan_port       = S17_WAN_PORT,
	.vlan_id        = 1,
	.wan2_port      = 0xff,
	.use_vlan_table = 0,
};

static uint16_t phy_reg_read(int dev_id, uint32_t phy_addr, uint8_t reg);
static void phy_reg_write(int dev_id, uint32_t phy_addr, uint8_t reg, uint16_t data);
static void athrs17_reg_rmw(unsigned int reg_addr, unsigned int reg_val);
static int athrs17_is_az_fixed(void);

void athrs17_phy_init(struct net_device *ndev,
	athr_switch_intf_type_t interface, int master)
{
	int phy_addr;
	unsigned int value;

	g_netdev = ndev;
	if (!g_netdev) {
		printk("athrs17_phy_init: netdev is null!!!\n");
		return;
	}

	if (interface == RGMII) {
		printk("Athers switch interface mode: RGMII\n");
		/* configure for RGMII */

		athrs17_reg_write(S17_GLOFW_CTRL1_REG,
			S17_BROAD_DPALL | S17_MULTI_FLOOD_DPALL |
			S17_UNI_FLOOD_DPALL);
		athrs17_reg_write(S17_PWS_REG, S17_PWS_CHIP_AR8327);

#ifdef CONFIG_HT818
		/* QCA8334 MAC0 RXCLK_DELAY configure by S17 P5PAD_MODE_REG */
		athrs17_reg_write(S17_P5PAD_MODE_REG, S17_MAC5_RGMII_RXCLK_DELAY);
		athrs17_reg_write(S17_P0PAD_MODE_REG, S17_MAC0_RGMII_EN |
			(1 << S17_MAC0_RGMII_RXCLK_SHIFT));
#else
		athrs17_reg_write(S17_P0PAD_MODE_REG, S17_MAC0_RGMII_EN |
			S17_MAC0_RGMII_TXCLK_DELAY |
			S17_MAC0_RGMII_RXCLK_DELAY |
			(1 << S17_MAC0_RGMII_TXCLK_SHIFT) |
			(1 << S17_MAC0_RGMII_RXCLK_SHIFT));
#endif
		athrs17_reg_write(S17_P0STATUS_REG, S17_PORT_STATUS_DEFAULT_1000M);

#ifdef HYBRID_SWITCH_PORT6_USED
		/* configure AR8327 mac6 as RGMII mode */
		athrs17_reg_write(S17_PWS_REG, S17_PWS_CHIP_AR8327);
		athrs17_reg_write(S17_P6PAD_MODE_REG, S17_MAC6_RGMII_EN |
			S17_MAC6_RGMII_TXCLK_DELAY |
			S17_MAC6_RGMII_RXCLK_DELAY |
			(2 << S17_MAC6_RGMII_RXCLK_SHIFT) |
			(1 << S17_MAC6_RGMII_TXCLK_SHIFT));
		athrs17_reg_write(S17_P6STATUS_REG, S17_PORT6_STATUS_DEFAULT);
#else
		athrs17_reg_write(S17_P6PAD_MODE_REG, S17_MAC6_RGMII_RXCLK_DELAY |
			(0 << S17_MAC6_RGMII_RXCLK_SHIFT));
#endif

		/* AR8327/AR8328 v1.0 fixup */
		if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff)== S17_CHIPID_V1_0) {
			for (phy_addr = 0x0; phy_addr <= S17_PHY_MAX; phy_addr++) {
				/* For 100M waveform */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x0);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x02ea);
				/* Turn On Gigabit Clock */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x3d);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x68a0);
			}
		}

		/* AR8337 v1.0 fixup */
		if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff) == S17C_CHIPID_V1_0) {
			for (phy_addr = 0x0; phy_addr <= S17_PHY_MAX; phy_addr++) {
				/* Turn On Gigabit Clock */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x3d);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x6820);
			}
		}
	} else if (interface == RMII) {
		printk("Athers switch interface mode: RMII(%s)n",
			master ? "master" : "slave");
		/* configure for RMII */
		value = athrs17_reg_read(0x10);
		value |= 0x10000000;
		athrs17_reg_write(0x10, value);

		value = athrs17_reg_read(0x30);
		value |= 0x00000001;

		athrs17_reg_write(0x30, value);

		if (master) 	/* Qualccomm switch master mode*/
			athrs17_reg_write(0x4, 0x40000000);
		else	/* Qualccomm switch slave mode*/
			athrs17_reg_write(0x4, 0x20000000);

		athrs17_reg_write(0x0624, 0x007f7f7f);
		/* RMII 100MB/s */
		athrs17_reg_write(0x007c, 0x007d);

		/* RMII 10MB/s
		athrs17_reg_write(0x007c, 0x007c);
		*/
	}

	/* enable HOL by default */
	for (phy_addr = 0; phy_addr < S17_MAC_MAX; phy_addr++) {
		switch (phy_addr) {
		case 0:
		case 5:
		case 6:
			athrs17_reg_write(S17_PORT0_HOL_CTRL0 + phy_addr * 8,
				S17_HOL_CTRL0_WAN);
			break;

		default:
			athrs17_reg_write(S17_PORT0_HOL_CTRL0 + phy_addr * 8,
				S17_HOL_CTRL0_LAN);
			break;
		}

		athrs17_reg_write(S17_PORT0_HOL_CTRL1 + phy_addr * 8,
			S17_HOL_CTRL1_DEFAULT);
	}

	/* CPU port enable, 2byte Atheros Header,
	* Enable Port0 Atheros TX/RX Header */
	athrs17_reg_write(S17_GLOFW_CTRL0_REG, 0x001004f0);
	athrs17_reg_write(S17_HDRCTRL_REG, 0x0);
	athrs17_reg_write(S17_P0HDRCTRL_REG, 0x0000000a);

#ifdef S17_8023AZ_FEATURE
	if (!athrs17_is_az_fixed()) {
		printk("S17 LPI disable All\n");
		athrs17_reg_write(S17_EEE_CTRL_REG, S17_LPI_DISABLE_ALL);
	}
#endif

	/* Turn all LEDs off .*/
	/* LED control */
	athrs17_reg_write(S17_LED_CTRL0_REG, 0);
	athrs17_reg_write(S17_LED_CTRL1_REG, 0);
	athrs17_reg_write(S17_LED_CTRL2_REG, 0);
}

void athrs17_port_isolation(int isolat)
{
	/*For *8334 */
	u32 p0reg, p2reg, p3reg, p6reg;

	p0reg = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	p2reg = athrs17_reg_read(S17_P2LOOKUP_CTRL_REG);
	p3reg = athrs17_reg_read(S17_P3LOOKUP_CTRL_REG);
	p6reg = athrs17_reg_read(S17_P6LOOKUP_CTRL_REG);

	if (isolat) {
		p0reg = (p0reg & 0xffffff80) | PORT2_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p2reg = (p2reg & 0xffffff80) | PORT0_BITMAP;
		p3reg = (p3reg & 0xffffff80) | PORT0_BITMAP;
		p6reg = (p6reg & 0xffffff80) | PORT0_BITMAP;
	} else {
		p0reg = (p0reg & 0xffffff80) | PORT2_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p2reg = (p2reg & 0xffffff80) | PORT0_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p3reg = (p3reg & 0xffffff80) | PORT0_BITMAP |
			PORT2_BITMAP | PORT6_BITMAP;
		p6reg = (p6reg & 0xffffff80) | PORT0_BITMAP |
			PORT2_BITMAP | PORT3_BITMAP;
	}

	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG, p0reg);
	athrs17_reg_write(S17_P2LOOKUP_CTRL_REG, p2reg);
	athrs17_reg_write(S17_P3LOOKUP_CTRL_REG, p3reg);
	athrs17_reg_write(S17_P6LOOKUP_CTRL_REG, p6reg);
}

static uint16_t phy_reg_read(int dev_id, uint32_t phy_addr, uint8_t reg)
{
	uint16_t data = 0;
	struct net_local *priv = netdev_priv(g_netdev);
	struct mii_bus *miibus = priv->mii_bus;

	data = miibus->read(miibus, phy_addr, reg);
	return data;
}

static void phy_reg_write(int dev_id, uint32_t phy_addr,
	uint8_t reg, uint16_t data)
{
	struct net_local *priv = netdev_priv(g_netdev);
	struct mii_bus *miibus = priv->mii_bus;
	miibus->write(miibus, phy_addr, reg, data);
}

void athrs17_lan_port_enable(void)
{
	unsigned int value;

	value = athrs17_reg_read(S17_P2STATUS_REG);
	value |= S17_TXMAC_EN;
	value |= S17_LINK_EN;
	value |= S17_FLOW_LINK_EN;
	athrs17_reg_write(S17_P2STATUS_REG, value);
	printk("lan port enabled.\n");
}

void athrs17_lan_port_disable(void)
{
	unsigned int value;

	value = athrs17_reg_read(S17_P2STATUS_REG);
	value &= ~(S17_TXMAC_EN);
	value &= ~(S17_LINK_EN);
	value &= ~(S17_FLOW_LINK_EN);
	athrs17_reg_write(S17_P2STATUS_REG, value);
	printk("lan port disabled.\n");
}

void athrs17_pc_port_enable(void)
{
	athrs17_reg_write(S17_P2STATUS_REG, S17_PORT_STATUS_DEFAULT);
	printk("PC port enabled.\n");
}

void athrs17_pc_port_disable(void)
{
	athrs17_reg_write(S17_P2STATUS_REG, 0);
	printk("PC port disabled.\n");
}

void athrs17_pc_port_mirror_enable(void)
{
	unsigned int value;

	// Set mirror port number 2.
	value = athrs17_reg_read(S17_GLOFW_CTRL0_REG);
	value &= ~(0x0F << S17_MIRROR_PORT_SHIFT);
	athrs17_reg_write(S17_GLOFW_CTRL0_REG, value | (2 << S17_MIRROR_PORT_SHIFT));

	// Enable mirror for incoming packets to port 0.
	value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG, value | (1 << 25));

	// Enable mirror for outgoing packets from port 0.
	value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
	athrs17_reg_write(S17_PORT0_HOL_CTRL1, value | (1 << 16));

	printk("PC port mirror enabled.\n");
}

void athrs17_pc_port_mirror_disable(void)
{
	unsigned int value;

	// Disable mirror for incoming packets to port 0.
	value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG, value & ~(1 << 25));

	// Disable mirror for outgoing packets from port 0.
	value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
	athrs17_reg_write(S17_PORT0_HOL_CTRL1, value & ~(1 << 16));

	printk("PC port mirror disabled.\n");
}

static int athrs17_port_status(int port)
{
	static int last_status = 0;
	uint16_t phyHwStatus, phyHwControl;
	uint32_t phyBase;
	uint32_t phyAddr;

	phyBase = S17_PHYBASE( port );
	phyAddr = S17_PHYADDR( port );

	if ( last_status ) {
		/* link was previously up */
		phyHwStatus = phy_reg_read(phyBase, phyAddr, S17_PHY_SPEC_STATUS);

		/* See if we've lost link */
		if (!(phyHwStatus & S17_STATUS_LINK_PASS)) {
			/* Flush the port from ARL entry */
			athrs17_reg_rmw(S17_ATU_FUNC_REG, ( port << 8 ));

			last_status = 0;
		}
	} else {
		/* last known link status was DEAD */
		/* Check for reset complete */
		phyHwStatus = phy_reg_read(phyBase, phyAddr, S17_PHY_STATUS);
		if (!S17_RESET_DONE(phyHwStatus))
			return 0;

		phyHwControl = phy_reg_read(phyBase, phyAddr, S17_PHY_CONTROL);
		/* Check for AutoNegotiation complete */
		if ((!(phyHwControl & S17_CTRL_AUTONEGOTIATION_ENABLE)) ||
			S17_AUTONEG_DONE( phyHwStatus )) {
			phyHwStatus = phy_reg_read(phyBase, phyAddr, S17_PHY_SPEC_STATUS);

			if (phyHwStatus & S17_STATUS_LINK_PASS)
				last_status = 1;
		}
	}

	return last_status;
}

int athrs17_lan_status(void)
{
	return athrs17_port_status(2);
}

int athrs17_pc_status(void)
{
	return athrs17_port_status(1);
}

unsigned int athrs17_reg_read(unsigned int reg_addr)
{
	uint32_t reg_word_addr;
	uint32_t phy_addr, tmp_val, reg_val;
	uint16_t phy_val;
	uint8_t phy_reg;

	/* change reg_addr to 16-bit word address, 32-bit aligned */
	reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

	/* configure register high address */
	phy_addr = 0x18;
	phy_reg = 0x0;
	phy_val = (uint16_t) ((reg_word_addr >> 8) & 0x1ff);  /* bit16-8 of reg address */
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* For some registers such as MIBs, since it is read/clear, we should */
	/* read the lower 16-bit register then the higher one */

	/* read register in lower address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
	phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
	reg_val = (uint32_t) phy_reg_read(0, phy_addr, phy_reg);

	/* read register in higher address */
	reg_word_addr++;
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
	phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
	tmp_val = (uint32_t) phy_reg_read(0, phy_addr, phy_reg);
	reg_val |= (tmp_val << 16);

	return reg_val;
}

void athrs17_reg_write(unsigned int reg_addr,unsigned int reg_val)
{
	uint32_t reg_word_addr;
	uint32_t phy_addr;
	uint16_t phy_val;
	uint8_t phy_reg;

	/* change reg_addr to 16-bit word address, 32-bit aligned */
	reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

	/* configure register high address */
	phy_addr = 0x18;
	phy_reg = 0x0;
	phy_val = (uint16_t) ((reg_word_addr >> 8) & 0x1ff);  /* bit16-8 of reg address */
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* For S17 registers such as ARL and VLAN, since they include BUSY bit */
	/* in higher address, we should write the lower 16-bit register then the */
	/* higher one */

	/* write register in lower address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
	phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
	phy_val = (uint16_t) (reg_val & 0xffff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* read register in higher address */
	reg_word_addr++;
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
	phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
	phy_val = (uint16_t) ((reg_val >> 16) & 0xffff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);
}

static void athrs17_reg_rmw(unsigned int reg_addr, unsigned int reg_val)
{
	reg_val |= athrs17_reg_read(reg_addr);
	athrs17_reg_write(reg_addr,reg_val);
}

#ifdef S17_8023AZ_FEATURE
static int athrs17_is_az_fixed(void)
{
	if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff) > S17_CHIPID_V1_1)
		return 1;
	return 0;
}
#endif

#if defined(S17_DUMP_MIB)
unsigned long long athrs17_mib_txbytes(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXBYTE1);
	pktcnt += (athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXBYTE2) * (1LL<<32));

	return pktcnt;
}

unsigned long long athrs17_mib_rxbytes(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXBYTE1);
	pktcnt += (athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXBYTE2) *
		(1LL<< 32));

	return pktcnt;
}

unsigned long long athrs17_mib_txok(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX64B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX128B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX256B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX512B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX1024B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TX1518B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXMAXB);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXLATECOL);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXABORTCOL);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXUNDERRUN);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXEXCDEFER);

	return pktcnt;
}

unsigned long long athrs17_mib_rxok(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX64B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX128B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX256B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX512B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX1024B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RX1518B);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXMAXB);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXFCSERR);
	pktcnt -= athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXALIGNERR);

	return pktcnt;
}

unsigned long long athrs17_mib_txcol(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXCOL);

	return pktcnt;
}

unsigned long long athrs17_mib_rxerr(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_RXFCSERR);

	return pktcnt;
}

unsigned long long athrs17_mib_txerr(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXOVERSIZE);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXUNDERRUN);
	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXLATECOL);

	return pktcnt;
}

unsigned long long athrs17_mib_txabortcol(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXABORTCOL);

	return pktcnt;
}

unsigned long long athrs17_mib_txexcdefer(int port)
{
	unsigned long long pktcnt = 0;

	pktcnt += athrs17_reg_read(
		S17_MIB_PORT0 + 0x100 * port + S17_MIB_TXEXCDEFER);

	return pktcnt;
}

int athrs17_mib_set(int enable)
{
	if (0 == enable) {
		athrs17_reg_write(S17_MODULE_EN_REG,
			athrs17_reg_read(S17_MODULE_EN_REG) & (~S17_MODULE_MIB_EN));
		athrs17_reg_write(S17_MIB_REG, 0);
	} else {
		athrs17_reg_write(S17_MODULE_EN_REG,
			athrs17_reg_read(S17_MODULE_EN_REG) | S17_MODULE_MIB_EN);
		athrs17_reg_write(S17_MIB_REG,
			athrs17_reg_read(S17_MIB_REG) | S17_MIB_CPU_KEEP);
	}

	printk("athrs17_mib_set: %s\n", enable ? "enabled" : "disabled");

	return 0;
}
#endif

/* EOF */
