/*
 * Cause: GRP260x requires both rtl8304 and rtl8363
 * Solution: Avoid redefinition errors with other Realtek switch APIs
 *
 * */
#include <linux/kernel.h>
#include <linux/string.h>
#include <rtk_switch.h>
#include <port.h>
#include <cpu.h>
#include <rtl8367c_asicdrv_port.h>
#include "rtl8363_func.h"
#include "rtk_switch.h"
#include "rtk_error.h"
#include "rtk_types.h"
#include "mirror.h"
#include "vlan.h"
#include "eee.h"
#include "smi.h"

/* Initial settings */
static int lan_port = UTP_PORT1;
static int wan_port = UTP_PORT3;
static int cpu_port = EXT_PORT0;

void rtl8363_init_port(int lan, int wan)
{
	wan_port = wan;
	lan_port = lan;
	printk("rtl8363 use wan_port %d, lan_port %d\n", wan_port, lan_port);
}

int rtl8363_switch_phy_link_state(int port)
{
	rtk_port_linkStatus_t linkStatus = PORT_LINKDOWN;
	rtk_port_speed_t speed;
	rtk_port_duplex_t duplex;
	rtk_api_ret_t ret;

	if (port == wan_port) {
		ret = rtk_port_phyStatus_get(wan_port, &linkStatus, &speed, &duplex);
		if (ret != RT_ERR_OK)
			printk("%s:%d failed to get link status %d\n", __func__, __LINE__, ret);
	} else {
		ret = rtk_port_phyStatus_get(lan_port, &linkStatus, &speed, &duplex);
		if (ret != RT_ERR_OK)
			printk("%s:%d failed to get link status %d\n", __func__, __LINE__, ret);
	}

	return linkStatus;
}

int rtl8363_port_isolation(int isolat)
{
	rtk_portmask_t CpuMask, LanMask, WanMask;

	RTK_PORTMASK_CLEAR(WanMask);
	RTK_PORTMASK_CLEAR(LanMask);
	RTK_PORTMASK_CLEAR(CpuMask);
	if (isolat) {
		RTK_PORTMASK_PORT_SET(WanMask, cpu_port);
		RTK_PORTMASK_PORT_SET(LanMask, cpu_port);
		RTK_PORTMASK_PORT_SET(CpuMask, lan_port);
		RTK_PORTMASK_PORT_SET(CpuMask, wan_port);
	} else {
		RTK_PORTMASK_PORT_SET(WanMask, cpu_port);
		RTK_PORTMASK_PORT_SET(WanMask, lan_port);
		RTK_PORTMASK_PORT_SET(LanMask, cpu_port);
		RTK_PORTMASK_PORT_SET(LanMask, wan_port);
		RTK_PORTMASK_PORT_SET(CpuMask, lan_port);
		RTK_PORTMASK_PORT_SET(CpuMask, wan_port);
	}
	rtk_port_isolation_set(lan_port, &LanMask);
	rtk_port_isolation_set(wan_port, &WanMask);
	rtk_port_isolation_set(cpu_port, &CpuMask);

	return 0;
}

void rtl8363_lan_port_enable(int enable)
{
	rtk_api_ret_t ret = RT_ERR_OK;

	if (enable) {
		rtk_port_adminEnable_set(lan_port, ENABLED);
		if (ret != RT_ERR_OK)
			printk("%s:%d failed set lan port enable %d\n", __func__, __LINE__, ret);
	}
	else {
		rtk_port_adminEnable_set(lan_port, DISABLED);
		if (ret != RT_ERR_OK)
			printk("%s:%d failed set lan port disable %d\n", __func__, __LINE__, ret);
	}
}

rtk_api_ret_t rtl8363_set_eee(rtk_enable_t enable)
{
	rtk_api_ret_t ret;

	ret = rtk_eee_init();
	if (ret != RT_ERR_OK) {
		printk("%s:%d failed set eee init %d\n", __func__, __LINE__, ret);
	}

	ret = rtk8363_rtk_eee_portEnable_set(wan_port, enable);
	ret = rtk8363_rtk_eee_portEnable_set(lan_port, enable);

	return ret;
}

rtk_api_ret_t rtl8363_switch_init(struct mii_bus *bus, smi_mdio_read read, smi_mdio_write write)
{
	rtk_port_mac_ability_t mac_cfg;
	rtk_mode_ext_t mode;
	rtk_data_t txDelay, rxDelay;
	rtk_api_ret_t ret;

	rtl_init_mii_bus(bus, read, write);

	ret = rtk_switch_init();

	/* Set extension port 0 to RGMII with Force mode, 1000M, Full-duplex, enable TX&RX pause*/
	mode = MODE_EXT_RGMII;
	mac_cfg.forcemode = MAC_FORCE;
	mac_cfg.speed = PORT_SPEED_1000M;
	mac_cfg.duplex = PORT_FULL_DUPLEX;
	mac_cfg.link = PORT_LINKUP;
	mac_cfg.nway = DISABLED;
	mac_cfg.txpause = ENABLED;
	mac_cfg.rxpause = ENABLED;
	ret = rtk_port_macForceLinkExt_set(EXT_PORT0, mode, &mac_cfg);

	txDelay = 1;
	rxDelay = 4;
	ret = rtk_port_rgmiiDelayExt_set( EXT_PORT0, txDelay, rxDelay );
	ret = rtk_port_rgmiiDelayExt_get(EXT_PORT0, &txDelay, &rxDelay);

	ret = rtl8363_set_eee(DISABLED);
	ret = rtk_port_phyEnableAll_set(ENABLED);
	ret = rtk_cpu_enable_set(ENABLED);
	ret = rtk_cpu_tagPort_set(cpu_port, CPU_INSERT_TO_ALL);

	return ret;
}

int rtl8363_switch_mdio_read(struct mii_bus *bus, int phyaddr, int phyreg)
{
	unsigned int RegData;
	int ret;

	if (RTL_PHY_MASK & (1 << phyaddr))
		return 0;

	ret = rtl8367c_getAsicPHYReg(phyaddr, phyreg, &RegData);
	if (ret != 0)
		printk("%s %d return err\n", __func__, __LINE__);

	return RegData;
}

int rtl8363_switch_mdio_write(struct mii_bus *bus, int phyaddr, int phyreg, u16 phydata)
{
	int ret;

	if (RTL_PHY_MASK & (1 << phyaddr))
		return 0;

	ret = rtl8367c_setAsicPHYReg(phyaddr, phyreg, phydata);
	if (ret != 0)
		printk("%s %d return err\n", __func__, __LINE__);

	return 0;
}

int rtl8363_switch_mdio_reset(struct mii_bus *bus)
{
	return 0;
}

