/*
 * Core driver for the DVF99 pin controller
 *
 * Copyright (C) 2015 DSPG
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_gpio.h>
#include <linux/of_irq.h>
#include <linux/of_address.h>
#include <linux/regmap.h>
#include <linux/mfd/syscon.h>
#include <linux/pinctrl/machine.h>
#include <linux/pinctrl/pinconf.h>
#include <linux/pinctrl/pinctrl.h>
#include <linux/pinctrl/pinmux.h>
#include <linux/platform_device.h>
#include <linux/gpio-dspg.h>
#include <linux/seq_file.h>
#include <linux/gpio/driver.h>
#include "core.h"

#define DVF_GPIO_PINS_PER_BANK 32

struct dvf_pio_control {
	u32 rt_pin_mask;
	struct regmap_field *alt;
};

struct dvf_pctl_data {
	/* register offset information */
	int alt, oe, pu, od;
};

struct dvf_pinconf {
	int		pin;
	const char	*name;
	unsigned long	config;
	int		altfunc;
};

struct dvf_padconf {
	int		pad;
	int		value;
};

struct dvf_pmx_func {
	const char	*name;
	const char	**groups;
	unsigned	ngroups;
};

struct dvf_pctl_group {
	const char		*name;
	unsigned int		*pins;
	unsigned		npins;
	unsigned		npads;
	struct dvf_pinconf	*pin_conf;
	struct dvf_padconf	*pad_conf;
};

struct save_state {
	int function;
	unsigned long config;
	unsigned int direction;
	unsigned int out_val;
};

struct dvf_gpio_bank {
	u32 offset;
	struct gpio_chip		gpio_chip;
	struct pinctrl_gpio_range	range;
	void __iomem			*base;
	struct dvf_pio_control		pc;
	struct irq_domain		*domain;
	unsigned long			irq_edge_conf;
	struct save_state		save[32];
	spinlock_t			lock;
};

struct dvf_pinctrl {
	struct device			*dev;
	struct pinctrl_dev		*pctl;
	struct dvf_gpio_bank		*banks;
	int				nbanks;
	struct dvf_pmx_func		*functions;
	int				nfunctions;
	struct dvf_pctl_group		*groups;
	int				ngroups;
	struct regmap			*regmap;
	const struct dvf_pctl_data	*data;
	void __iomem			*irqmux_base;
	struct dspg_gpio		*gpio;
};

#define EXTINT_CFG          0x2A0
#define EXTINT_EDGE_CLEAR   0x2AC

#define EXTINT_FLAG_MASKED	(1 << 0)
#define EXTINT_FLAG_BOTH_EDGES	(1 << 1)

struct dspg_gpio;

struct extint {
	unsigned bit;
	u16 hwirq;
	u16 flags;
	struct dspg_gpio *dspg_gpio;
	struct irq_domain *irq_domain;
	int upstream_irq;
};

struct dspg_gpio {
	struct gpio_chip chip;
	void *regs;
	struct dspg_gpio_pdata pdata;
	struct irq_domain *extint_domain;
	int nr_extints;
	struct extint extints[16];
};

static DEFINE_SPINLOCK(extint_lock);

#define GET_REG(priv, reg, gpio) ( \
	__raw_readl((priv)->regs + (__BANK(gpio) * 0x60) + (reg)) & (1 << __PIN(gpio)) \
	)

#define SET_REG(priv, reg, gpio) \
	__raw_writel(1 << __PIN(gpio), (priv)->regs + (__BANK(gpio) * 0x60) + (reg))


static int gpio_get_pull_enable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_PULL_CTRL, gpio);
}

/*
 * Returns: 0: Output, 1: Input
 */
static int gpio_get_direction(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_DIR, gpio);
}

static void dspg_gpio_set_value(struct dspg_gpio *dspg_gpio, unsigned gpio, int value)
{
	if (value)
		SET_REG(dspg_gpio, XGP_DATA_SET, gpio);
	else
		SET_REG(dspg_gpio, XGP_DATA_CLR, gpio);
}

static int gpio_get_enable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_EN, gpio);
}

static void gpio_disable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	SET_REG(dspg_gpio, XGP_EN_CLR, gpio);
}

static void gpio_enable(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	SET_REG(dspg_gpio, XGP_EN_SET, gpio);
}

static int gpio_get_opendrain(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_OD, gpio);
}

static int gpio_get_keeper(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_KEEP, gpio);
}

/*
 * Returns: 0: pull down, 1: pull up
 */
static int gpio_get_pull_selection(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	return !!GET_REG(dspg_gpio, XGP_PULL_TYPE, gpio);
}

/*
 * Tests whether the Pull-Up/Dn internal resistor can be disconnected, in order
 * to save power.
 */
static void tst_pull_disconnect(struct dspg_gpio *dspg_gpio, unsigned gpio)
{
	if (!gpio_get_pull_enable(dspg_gpio, gpio))
		return;

	if (gpio_get_direction(dspg_gpio, gpio) == 0) {
		/* Output gpio */

		/* Note:
		 * "Wired-OR" is achived by enabling "Open-Drain" and
		 * No Pull-Down resistor, but the user may connect
		 * a Pull-Up resistor, if no such External resistor
		 * exists. So "open-drain" must be excluded !
		 */
		if (!gpio_get_opendrain(dspg_gpio, gpio))
			SET_REG(dspg_gpio, XGP_PULL_DIS, gpio);
	} else {
		/* Input gpio */

		if (gpio_get_keeper(dspg_gpio, gpio))
			SET_REG(dspg_gpio, XGP_PULL_DIS, gpio);
	}
}

static int
dspg_gpiolib_request(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	if (dspg_gpio->pdata.nx_masks[__BANK(offset)] & (1 << __PIN(offset))) {
		dev_warn(chip->parent,
			 "request for non-existent %cGPIO%-2d denied!\n",
			 'A' + __BANK(offset), __PIN(offset));
		return -ENODEV;
	}

	if (!gpio_get_enable(dspg_gpio, offset)) {
		dev_warn(chip->parent, "enabling %cGPIO%-2d\n",
			 'A' + __BANK(offset), __PIN(offset));
		gpio_enable(dspg_gpio, offset);
		//return -ENODEV;
	}

	return 0;
}

static int
dspg_gpiolib_direction_input(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	SET_REG(dspg_gpio, XGP_DIR_IN, offset);
	tst_pull_disconnect(dspg_gpio, offset);

	return 0;
}

static int
dspg_gpiolib_direction_output(struct gpio_chip *chip, unsigned offset, int value)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	dspg_gpio_set_value(dspg_gpio, offset, value);
	SET_REG(dspg_gpio, XGP_DIR_OUT, offset);
	tst_pull_disconnect(dspg_gpio, offset);

	return 0;
}

static int
dspg_gpiolib_get(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	return !!GET_REG(dspg_gpio, XGP_DATA, offset);
}

static void
dspg_gpiolib_set(struct gpio_chip *chip, unsigned offset, int value)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);

	dspg_gpio_set_value(dspg_gpio, offset, value);
}

static void
dspg_gpiolib_show(struct seq_file *s, struct gpio_chip *chip)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);
	const char *label;
	unsigned i;

	for (i = 0; i < chip->ngpio; i++) {
		if (!(label = gpiochip_is_requested(chip, i)))
			continue;

		seq_printf(s, " %cGPIO%-2d (%-20.20s) %s %s%s",
			'A' + __BANK(i), __PIN(i), label,
			gpio_get_direction(dspg_gpio, i) ? "in " : "out",
			dspg_gpiolib_get(chip, i) ? "hi" : "lo",
			gpio_get_pull_enable(dspg_gpio, i)
				? (gpio_get_pull_selection(dspg_gpio, i) ? " pu" : " pd")
				: ""
			);
		seq_printf(s, "\n");
	}
}

static int dspg_gpiolib_to_irq(struct gpio_chip *chip, unsigned offset)
{
	struct dspg_gpio *dspg_gpio = container_of(chip, struct dspg_gpio, chip);
	int i;

	/* search for matching GPIO and then map through exting IRQ domain */
	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		if (dspg_gpio->pdata.extint_gpios[i] == offset)
			return irq_create_mapping(dspg_gpio->extint_domain, i);
	}

	return -ENODEV;
}

#ifdef CONFIG_OF
static int dspg_gpiolib_xlate(struct gpio_chip *gc,
			      const struct of_phandle_args *gpiospec, u32 *flags)
{
	struct dspg_gpio *dspg_gpio = container_of(gc, struct dspg_gpio, chip);
	unsigned pin;

	pin = gpiospec->args[0] * 32 + gpiospec->args[1];

	if (pin >= gc->ngpio || gpiospec->args[1] > 32)
		return -EINVAL;

	if (dspg_gpio->pdata.nx_masks[__BANK(pin)] & (1 << __PIN(pin)))
		return -ENODEV;

	if (flags)
		*flags = gpiospec->args[2];

	return pin;
}
#endif

static void extint_adjust_pol(struct extint *extint)
{
	unsigned long value;
	int pol, shift = extint->bit * 2;

	pol = !dspg_gpiolib_get(&extint->dspg_gpio->chip,
		extint->dspg_gpio->pdata.extint_gpios[extint->hwirq]);

	value = readl(extint->dspg_gpio->regs + EXTINT_CFG);
	value &= ~(0x1UL << shift);
	value |= pol << shift;
	writel(value, extint->dspg_gpio->regs + EXTINT_CFG);
}

static void extint_irq_ack(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	writel(1 << extint->bit, extint->dspg_gpio->regs + EXTINT_EDGE_CLEAR);

	if (extint->flags & EXTINT_FLAG_BOTH_EDGES) {
		spin_lock_irqsave(&extint_lock, flags);
		extint_adjust_pol(extint);
		spin_unlock_irqrestore(&extint_lock, flags);
	}
}

static void extint_irq_mask(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	spin_lock_irqsave(&extint_lock, flags);
	if (!(extint->flags & EXTINT_FLAG_MASKED)) {
		disable_irq_nosync(extint->upstream_irq);
		extint->flags |= EXTINT_FLAG_MASKED;
	}
	spin_unlock_irqrestore(&extint_lock, flags);
}

static void extint_irq_unmask(struct irq_data *data)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags;

	spin_lock_irqsave(&extint_lock, flags);
	if (extint->flags & EXTINT_FLAG_MASKED) {
		if (extint->flags & EXTINT_FLAG_BOTH_EDGES)
			extint_adjust_pol(extint);

		enable_irq(extint->upstream_irq);
		extint->flags &= ~EXTINT_FLAG_MASKED;
	}
	spin_unlock_irqrestore(&extint_lock, flags);
}

static int extint_irq_set_type(struct irq_data *data, unsigned int type)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	unsigned long flags, value;
	int pol = 0; /* 0 = active low; 1 = active high */
	int mode = 0;     /* 0 = level; 1 = edge */
	int shift;

	type &= IRQ_TYPE_SENSE_MASK;

	spin_lock_irqsave(&extint_lock, flags);

	if ((type & IRQ_TYPE_EDGE_BOTH) == IRQ_TYPE_EDGE_BOTH) {
		/*
		 * emulate EDGE_BOTH by setting LEVEL_HIGH or LEVEL_LOW
		 * depending on current gpio value
		 */
		extint->flags |= EXTINT_FLAG_BOTH_EDGES;
		pol = !dspg_gpiolib_get(&extint->dspg_gpio->chip,
			extint->dspg_gpio->pdata.extint_gpios[extint->hwirq]);
	} else {
		extint->flags &= ~EXTINT_FLAG_BOTH_EDGES;

		if (type & IRQ_TYPE_EDGE_BOTH) {
			/* specific edge type irq */
			mode = 1;

			if (type & IRQ_TYPE_EDGE_RISING)
				pol = 1;
		} else if (type & IRQ_TYPE_LEVEL_HIGH)
			pol = 1;
	}

	shift = extint->bit << 1; /* extint * 2 */

	value = readl(extint->dspg_gpio->regs + EXTINT_CFG);
	value &= ~(0x3UL << shift);
	value |= ((mode <<1) | pol) << shift;
	writel(value, extint->dspg_gpio->regs + EXTINT_CFG);

	spin_unlock_irqrestore(&extint_lock, flags);

	if (type & (IRQ_TYPE_LEVEL_LOW | IRQ_TYPE_LEVEL_HIGH))
		irq_set_handler_locked(data, handle_level_irq);
	else if (type & (IRQ_TYPE_EDGE_FALLING | IRQ_TYPE_EDGE_RISING))
		irq_set_handler_locked(data, handle_edge_irq);

	return 0;
}

static int extint_irq_set_wake(struct irq_data *data, unsigned int enable)
{
	struct extint *extint = irq_data_get_irq_chip_data(data);
	return irq_set_irq_wake(extint->upstream_irq, enable);
}

static struct irq_chip extint_irq_chip = {
	.name		= "gpio",
	.irq_ack	= extint_irq_ack,
	.irq_mask	= extint_irq_mask,
	.irq_unmask	= extint_irq_unmask,
	.irq_set_type	= extint_irq_set_type,
	.irq_set_wake	= extint_irq_set_wake,
};

static irqreturn_t extint_irq_handler(int irq, void *dev_id)
{
	unsigned long flags;
	struct extint *extint = dev_id;

	local_irq_save(flags);
	generic_handle_irq(irq_linear_revmap(extint->irq_domain, extint->hwirq));
	local_irq_restore(flags);

	return IRQ_HANDLED;
}

/*
 * This lock class tells lockdep that EXTINTs are in a different category than
 * their parents, so it won't report false recursion.
 */
static struct lock_class_key gpio_lock_class;

static int extint_irq_map(struct irq_domain *dom, unsigned int irq,
			  irq_hw_number_t hw_irq)
{
	struct dspg_gpio *dspg_gpio = dom->host_data;

	irq_set_lockdep_class(irq, &gpio_lock_class);
	irq_set_chip_data(irq, &(dspg_gpio->extints[hw_irq]));
	irq_set_chip_and_handler(irq, &extint_irq_chip, handle_level_irq);

	return 0;
}

static struct irq_domain_ops extint_irq_ops = {
	.map	= extint_irq_map,
	.xlate	= irq_domain_xlate_twocell,
};

/* Pinctrl Groups */
static int
dvf_pctl_get_groups_count(struct pinctrl_dev *pctldev)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);

	return info->ngroups;
}

static const char *
dvf_pctl_get_group_name(struct pinctrl_dev *pctldev, unsigned selector)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);

	return info->groups[selector].name;
}

static int
dvf_pctl_get_group_pins(struct pinctrl_dev *pctldev, unsigned selector,
			const unsigned **pins, unsigned *npins)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);

	if (selector >= info->ngroups)
		return -EINVAL;

	*pins = info->groups[selector].pins;
	*npins = info->groups[selector].npins;

	return 0;
}

static const inline struct dvf_pctl_group *
dvf_pctl_find_group_by_name(const struct dvf_pinctrl *info, const char *name)
{
	int i;

	for (i = 0; i < info->ngroups; i++) {
		if (!strcmp(info->groups[i].name, name))
			return &info->groups[i];
	}

	return NULL;
}

static int
dvf_pctl_dt_node_to_map(struct pinctrl_dev *pctldev, struct device_node *np,
			struct pinctrl_map **map, unsigned *num_maps)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);
	const struct dvf_pctl_group *grp;
	struct pinctrl_map *new_map;
	struct device_node *parent;
	int map_num, i;

	grp = dvf_pctl_find_group_by_name(info, np->name);
	if (!grp) {
		dev_err(info->dev, "unable to find group for node %s\n",
			np->name);
		return -EINVAL;
	}

	map_num = grp->npins + 1;
	new_map = devm_kzalloc(pctldev->dev,
			       sizeof(*new_map) * map_num, GFP_KERNEL);
	if (!new_map)
		return -ENOMEM;

	parent = of_get_parent(np);
	if (!parent) {
		devm_kfree(pctldev->dev, new_map);
		return -EINVAL;
	}

	*map = new_map;
	*num_maps = map_num;
	new_map[0].type = PIN_MAP_TYPE_MUX_GROUP;
	new_map[0].data.mux.function = parent->name;
	new_map[0].data.mux.group = np->name;
	of_node_put(parent);

	/* create config map per pin */
	new_map++;
	for (i = 0; i < grp->npins; i++) {
		new_map[i].type = PIN_MAP_TYPE_CONFIGS_PIN;
		new_map[i].data.configs.group_or_pin =
				pin_get_name(pctldev, grp->pins[i]);
		new_map[i].data.configs.configs = &grp->pin_conf[i].config;
		new_map[i].data.configs.num_configs = 1;
	}
	dev_info(pctldev->dev, "maps: function %s group %s num %d\n",
		(*map)->data.mux.function, grp->name, map_num);

	return 0;
}

static void
dvf_pctl_dt_free_map(struct pinctrl_dev *pctldev, struct pinctrl_map *map,
		     unsigned num_maps)
{
}

static struct pinctrl_ops dvf_pctlops = {
	.get_groups_count	= dvf_pctl_get_groups_count,
	.get_group_pins		= dvf_pctl_get_group_pins,
	.get_group_name		= dvf_pctl_get_group_name,
	.dt_node_to_map		= dvf_pctl_dt_node_to_map,
	.dt_free_map		= dvf_pctl_dt_free_map,
};

/* Pinmux */
static int
dvf_pmx_get_funcs_count(struct pinctrl_dev *pctldev)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);

	return info->nfunctions;
}

static const char *
dvf_pmx_get_fname(struct pinctrl_dev *pctldev, unsigned selector)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);

	return info->functions[selector].name;
}

static int
dvf_pmx_get_groups(struct pinctrl_dev *pctldev, unsigned selector,
		   const char * const **grps, unsigned * const ngrps)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);
	*grps = info->functions[selector].groups;
	*ngrps = info->functions[selector].ngroups;

	return 0;
}

#define SPCR1 0x54

static void
pad_configure(struct dvf_pinctrl *info, int pad, int value)
{
	regmap_update_bits(info->regmap, SPCR1 + 4*(pad/32), 1 << (pad % 32),
			   (!!value) << (pad % 32));
}

static int
dvf_pmx_set_mux(struct pinctrl_dev *pctldev, unsigned fselector,
		unsigned group_selector)
{
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);
	struct dvf_pinconf *conf = info->groups[group_selector].pin_conf;
	struct dvf_padconf *pad = info->groups[group_selector].pad_conf;
	int i;

	for (i = 0; i < info->groups[group_selector].npins; i++) {
		dev_dbg(pctldev->dev,
			"GPIO disable %d: pin %d (%cGPIO%d), func %d\n",
			i, conf[i].pin, (conf[i].pin/32) + 'A',
			conf[i].pin % 32, conf[i].altfunc);
		gpio_disable(info->gpio, conf[i].pin);
	}
	for (i = 0; i < info->groups[group_selector].npads; i++) {
		dev_dbg(pctldev->dev, "PAD configure %d: pad %d, value %d\n",
			i, pad[i].pad, pad[i].value);
		pad_configure(info, pad[i].pad, pad[i].value);
	}

	return 0;
}

static struct pinmux_ops dvf_pmxops = {
	.get_functions_count	= dvf_pmx_get_funcs_count,
	.get_function_name	= dvf_pmx_get_fname,
	.get_function_groups	= dvf_pmx_get_groups,
	.set_mux		= dvf_pmx_set_mux,
};

/* Pinconf  */
static int
dvf_pinconf_set(struct pinctrl_dev *pctldev, unsigned pin_id,
		unsigned long *config, unsigned num_configs)
{
	/*int pin = pin_id; //dvf_gpio_pin(pin_id);
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);
	struct dvf_pio_control *pc = dvf_get_pio_control(pctldev, pin_id);*/

	return 0;
}

static int
dvf_pinconf_get(struct pinctrl_dev *pctldev, unsigned pin_id,
		unsigned long *config)
{
	/* int pin = pin_id; //dvf_gpio_pin(pin_id);
	struct dvf_pinctrl *info = pinctrl_dev_get_drvdata(pctldev);
	struct dvf_pio_control *pc = dvf_get_pio_control(pctldev, pin_id);

	*config = 0; */
	return 0;
}

static void
dvf_pinconf_dbg_show(struct pinctrl_dev *pctldev,  struct seq_file *s,
		     unsigned pin_id)
{
	unsigned long config;

	mutex_unlock(&pctldev->mutex);
	dvf_pinconf_get(pctldev, pin_id, &config);
	mutex_lock(&pctldev->mutex);
	seq_printf(s, "xxx");
}

static struct pinconf_ops dvf_confops = {
	.pin_config_get		= dvf_pinconf_get,
	.pin_config_set		= dvf_pinconf_set,
	.pin_config_dbg_show	= dvf_pinconf_dbg_show,
};

/*
 * Each pin is represented in of the below forms.
 * <gpio>
 */
static int
dvf_pinctrl_dt_parse_groups(struct device_node *np,
			    struct dvf_pctl_group *grp,
			    struct dvf_pinctrl *info, int idx)
{
	/* bank pad direction val altfunction */
	const __be32 *list;
	struct property *pp;
	struct dvf_pinconf *conf;
	struct dvf_padconf *pad;
	struct device_node *pins;
	int i = 0, j = 0, k, npins = 0, npads = 0, nr_props;

	pins = of_get_child_by_name(np, "dvf,pins");
	if (!pins)
		return -ENODATA;

	for_each_property_of_node(pins, pp) {
		/* Skip those we do not want to proceed */
		if (!strcmp(pp->name, "name"))
			continue;
		if (!strcmp(pp->name, "pad")) {
			npads += (pp->length/sizeof(__be32)) / 2;
			/* dev_dbg(pctldev->dev,
				"found pad with %d entries (%d)\n",
				npads, pp->length); */
			continue;
		}

		if (pp && (pp->length/sizeof(__be32)) == 4) {
			npins++;
		} else {
			pr_warn("Invalid dvf,pins in %s node 0x%p %d\n", np->name, pp, pp->length/sizeof(__be32));
			return -EINVAL;
		}
	}

	grp->npins = npins;
	grp->npads = npads;
	grp->name = np->name;
	grp->pins = devm_kzalloc(info->dev, npins * sizeof(u32), GFP_KERNEL);
	grp->pin_conf = devm_kzalloc(info->dev,
					npins * sizeof(*conf), GFP_KERNEL);
	grp->pad_conf = devm_kzalloc(info->dev,
					npads * sizeof(*pad), GFP_KERNEL);

	if (!grp->pins || !grp->pin_conf || !grp->pad_conf)
		return -ENOMEM;

	/* <gpio> */
	for_each_property_of_node(pins, pp) {
		if (!strcmp(pp->name, "name"))
			continue;
		if (!strcmp(pp->name, "pad")) {
			nr_props = pp->length/sizeof(u32);
			list = pp->value;
			for (k = 0; k < nr_props/2; k++) {
				pad = &grp->pad_conf[j];
				pad->pad = be32_to_cpup(list++);
				pad->value = be32_to_cpup(list++);
				/* dev_dbg(pctldev->dev, "found pad %d -> %d\n",
					pad->pad, pad->value); */
				j++;
			}
			continue;
		}
		nr_props = pp->length/sizeof(u32);
		list = pp->value;
		conf = &grp->pin_conf[i];

		conf->pin = of_get_named_gpio(pins, pp->name, 0);
		conf->name = pp->name;
		grp->pins[i] = conf->pin;
		i++;
		/* dev_dbg(pctldev->dev, "found gpio %d (%s)\n", conf->pin,
			conf->name); */
	}
	of_node_put(pins);

	return 0;
}

static int
dvf_pctl_parse_functions(struct device_node *np, struct dvf_pinctrl *info,
			 u32 index, int *grp_index)
{
	struct device_node *child;
	struct dvf_pmx_func *func;
	struct dvf_pctl_group *grp;
	int ret, i;

	func = &info->functions[index];
	func->name = np->name;
	//dev_dbg(pctldev->dev, "function %d has name %s\n", index, func->name);
	func->ngroups = of_get_child_count(np);
	if (func->ngroups <= 0) {
		dev_err(info->dev, "No groups defined\n");
		return -EINVAL;
	}
	func->groups = devm_kzalloc(info->dev,
			func->ngroups * sizeof(char *), GFP_KERNEL);
	if (!func->groups)
		return -ENOMEM;

	i = 0;
	for_each_child_of_node(np, child) {
		func->groups[i] = child->name;
		grp = &info->groups[*grp_index];
		*grp_index += 1;
		ret = dvf_pinctrl_dt_parse_groups(child, grp, info, i++);
		if (ret)
			return ret;
	}
	dev_info(info->dev, "Function[%d\t name:%s,\tgroups:%d]\n",
				index, func->name, func->ngroups);

	return 0;
}

static void
dvf_pctl_dt_child_count(struct dvf_pinctrl *info, struct device_node *np)
{
	struct device_node *child;

	for_each_child_of_node(np, child) {
		if (of_property_read_bool(child, "gpio-controller")) {
			info->nbanks++;
		} else {
			info->nfunctions++;
			info->ngroups += of_get_child_count(child);
		}
	}
}

static struct of_device_id dvf_pinctrl_of_match[] = {
	{ .compatible = "dspg,pinctrl-dvf" },
	{ /* sentinel */ }
};

static int
dvf_pctl_probe_dt(struct platform_device *pdev,
		  struct pinctrl_desc *pctl_desc, struct dvf_pinctrl *info)
{
	struct dspg_gpio_pdata *pdata = pdev->dev.platform_data;
	struct dspg_gpio *dspg_gpio;
	struct resource *res;
	int extsize = 0;
	int ret = 0;
	int i = 0, j = 0, k = 0;
	struct pinctrl_pin_desc *pdesc;
	struct device_node *np = pdev->dev.of_node;
	struct device_node *child;
	int grp_index = 0;

	dspg_gpio = kzalloc(sizeof(*dspg_gpio), GFP_KERNEL);
	if (!dspg_gpio)
		return -ENOMEM;

	info->gpio = dspg_gpio;

	if (!pdata) {
		struct device_node *np = pdev->dev.of_node;
		u32 extint_gpios[48];

		if (!np) {
			dev_err(&pdev->dev, "missing platform data!\n");
			return -EINVAL;
		}

		ret = of_property_read_u32(np, "dspg,banks",
			&dspg_gpio->pdata.num_banks);
		if (ret) {
			dev_err(&pdev->dev, "missing 'dspg,banks' property!\n");
			return ret;
		}
		if (dspg_gpio->pdata.num_banks > ARRAY_SIZE(dspg_gpio->pdata.nx_masks)) {
			dev_err(&pdev->dev, "too many banks!\n");
			return -EINVAL;
		}

		ret = of_property_read_u32_array(np, "dspg,nx_masks",
			dspg_gpio->pdata.nx_masks, dspg_gpio->pdata.num_banks);
		if (ret && ret != -EINVAL) {
			dev_err(&pdev->dev, "invalid 'dspg,nx_masks' property!\n");
			return ret;
		}

		dspg_gpio->nr_extints = 16;
		ret = (int)of_find_property(np, "dspg,extints", &extsize);
		if (ret)
			dspg_gpio->nr_extints = extsize / sizeof(u32) / 3;

		ret = of_property_read_u32_array(np, "dspg,extints",
			extint_gpios, dspg_gpio->nr_extints * 3);
		if (ret) {
			dev_err(&pdev->dev, "missing 'dspg,extints' property!\n");
			return ret;
		}

		for (i = 0; i < dspg_gpio->nr_extints; i++) {
			dspg_gpio->pdata.extint_gpios[i] =
				extint_gpios[i*3] * 32 + extint_gpios[i*3+1];
			dspg_gpio->extints[i].bit = extint_gpios[i*3+2];
		}
	} else
		dspg_gpio->pdata = *pdata;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -EINVAL;

	dspg_gpio->regs = devm_ioremap_resource(&pdev->dev, res);
	if (!dspg_gpio->regs) {
		dev_err(&pdev->dev, "failed to map registers\n");
		return -ENOMEM;
	}

	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		res = platform_get_resource(pdev, IORESOURCE_IRQ, i);
		if (!res) {
			dev_err(&pdev->dev, "Missing IRQ resource\n");
			return -EINVAL;
		}
		dspg_gpio->extints[i].upstream_irq = res->start;
	}

	dspg_gpio->chip.parent			= &pdev->dev;
	dspg_gpio->chip.label			= "gpio-dspg";
	dspg_gpio->chip.owner			= THIS_MODULE;
	dspg_gpio->chip.request			= dspg_gpiolib_request;
	dspg_gpio->chip.direction_input		= dspg_gpiolib_direction_input;
	dspg_gpio->chip.get			= dspg_gpiolib_get;
	dspg_gpio->chip.direction_output	= dspg_gpiolib_direction_output;
	dspg_gpio->chip.set			= dspg_gpiolib_set;
	dspg_gpio->chip.dbg_show		= dspg_gpiolib_show;
	dspg_gpio->chip.to_irq			= dspg_gpiolib_to_irq;
	dspg_gpio->chip.ngpio			= dspg_gpio->pdata.num_banks * 32;
#ifdef CONFIG_OF
	dspg_gpio->chip.of_node			= pdev->dev.of_node;
	dspg_gpio->chip.of_gpio_n_cells		= 3;
	dspg_gpio->chip.of_xlate		= dspg_gpiolib_xlate;
#endif

	dspg_gpio->extint_domain = irq_domain_add_linear(pdev->dev.of_node,
		dspg_gpio->nr_extints, &extint_irq_ops, dspg_gpio);
	if (!dspg_gpio->extint_domain) {
		dev_err(&pdev->dev, "cannot register IRQ domain!\n");
		return -ENOMEM;
	}

	for (i = 0; i < dspg_gpio->nr_extints; i++) {
		int irq = dspg_gpio->extints[i].upstream_irq;

		dspg_gpio->extints[i].dspg_gpio = dspg_gpio;
		dspg_gpio->extints[i].irq_domain = dspg_gpio->extint_domain;
		dspg_gpio->extints[i].hwirq = i;
		dspg_gpio->extints[i].flags = EXTINT_FLAG_MASKED;

		irq_set_status_flags(irq, IRQ_NOAUTOEN);
		ret = devm_request_irq(&pdev->dev, irq, extint_irq_handler, 0,
				       "gpio-dspg", &(dspg_gpio->extints[i]));
		if (ret) {
			dev_err(&pdev->dev, "cannot request upstream irq %d!\n",
				irq);
			return ret;
		}
	}

	ret = gpiochip_add(&dspg_gpio->chip);
	if (ret)
		return ret;

	dev_info(&pdev->dev, "%d banks\n", dspg_gpio->pdata.num_banks);

	platform_set_drvdata(pdev, dspg_gpio);

	dvf_pctl_dt_child_count(info, np);
	info->nbanks = dspg_gpio->pdata.num_banks;
	if (!info->nbanks) {
		dev_err(&pdev->dev, "you need atleast one gpio bank\n");
		return-EINVAL;
	}

	dev_info(&pdev->dev, "nbanks = %d\n", info->nbanks);
	dev_info(&pdev->dev, "nfunctions = %d\n", info->nfunctions);
	dev_info(&pdev->dev, "ngroups = %d\n", info->ngroups);

	info->functions = devm_kzalloc(&pdev->dev,
		info->nfunctions * sizeof(*info->functions), GFP_KERNEL);

	info->groups = devm_kzalloc(&pdev->dev,
			info->ngroups * sizeof(*info->groups), GFP_KERNEL);

	info->banks = devm_kzalloc(&pdev->dev,
			info->nbanks * sizeof(*info->banks), GFP_KERNEL);

	if (!info->functions || !info->groups || !info->banks)
		return -ENOMEM;

	info->regmap = syscon_regmap_lookup_by_phandle(np, "dvf,syscfg");
	if (IS_ERR(info->regmap)) {
		dev_err(info->dev, "No syscfg phandle specified\n");
		return PTR_ERR(info->regmap);
	}

	pctl_desc->npins = info->nbanks * DVF_GPIO_PINS_PER_BANK;
	pdesc = devm_kzalloc(&pdev->dev,
			sizeof(*pdesc) * pctl_desc->npins, GFP_KERNEL);
	if (!pdesc)
		return -ENOMEM;

	pctl_desc->pins = pdesc;

	i = 0;
	for_each_child_of_node(np, child) {
		ret = dvf_pctl_parse_functions(child, info,
					i++, &grp_index);
		if (ret) {
			dev_err(&pdev->dev, "No functions found.\n");
			return ret;
		}
	}
	for (i = 0; i < info->nbanks; i++) {
		for (j = 0; j < DVF_GPIO_PINS_PER_BANK; j++, k++) {
			pdesc->number = (i * DVF_GPIO_PINS_PER_BANK) + j;
			pdesc->name = kasprintf(GFP_KERNEL, "%c[%d]",
						'A' + i, j);
			pdesc++;
		}
	}

	return 0;
}

static int
dvf_pinctrl_probe(struct platform_device *pdev)
{
	struct dvf_pinctrl *info;
	struct pinctrl_desc *pctl_desc;
	int ret;

	if (!pdev->dev.of_node) {
		dev_err(&pdev->dev, "device node not found.\n");
		return -EINVAL;
	}

	pctl_desc = devm_kzalloc(&pdev->dev, sizeof(*pctl_desc), GFP_KERNEL);
	if (!pctl_desc)
		return -ENOMEM;

	info = devm_kzalloc(&pdev->dev, sizeof(*info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	info->dev = &pdev->dev;
	platform_set_drvdata(pdev, info);
	ret = dvf_pctl_probe_dt(pdev, pctl_desc, info);
	if (ret)
		return ret;

	pctl_desc->owner	= THIS_MODULE;
	pctl_desc->pctlops	= &dvf_pctlops;
	pctl_desc->pmxops	= &dvf_pmxops;
	pctl_desc->confops	= &dvf_confops;
	pctl_desc->name		= dev_name(&pdev->dev);

	info->pctl = pinctrl_register(pctl_desc, &pdev->dev, info);
	if (!info->pctl) {
		dev_err(&pdev->dev, "Failed pinctrl registration\n");
		return -EINVAL;
	}

	printk("successfully registered dvf pinctrl\n");

	return 0;
}

static struct platform_driver dvf_pinctrl_driver = {
	.driver = {
		.name = "dvf-pinctrl",
		.owner = THIS_MODULE,
		.of_match_table = dvf_pinctrl_of_match,
	},
	.probe = dvf_pinctrl_probe,
};


static int __init dvf_pinctrl_init(void)
{
	return platform_driver_register(&dvf_pinctrl_driver);
}
arch_initcall(dvf_pinctrl_init);
