/*
 * linux/drivers/staging/dspg/usb-phy-mux.c
 *
 *  Copyright (C) 2016 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */
#include <linux/err.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/regmap.h>
#include <linux/of_device.h>
#include <linux/mfd/syscon.h>
#include <linux/sysfs.h>

#define USB_CSR                 0x58

struct usb_phy_mux {
	struct device *dev;
	struct regmap_field *f;
	unsigned int route_to_host;
};

static const struct reg_field usb_phy_mux_regfield =
	REG_FIELD(USB_CSR, 0,  BIT(0));

static int set_route(struct usb_phy_mux *pdata)
{
	int ret;

	ret = regmap_field_write(pdata->f, pdata->route_to_host);
	if (ret)
		dev_err(pdata->dev, "failed to set USB PHY muxer");

	return ret;
}

static const char *xlate_mux_setting[] = {
	"device", "host",
};

static ssize_t mux_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct usb_phy_mux *pdata = dev_get_drvdata(dev);

	return sprintf(buf, "%s\n", xlate_mux_setting[pdata->route_to_host]);
}

static DEVICE_ATTR_RO(mux);

static const struct attribute *dvf101_usb_phy_mux_sysfs[] = {
	&dev_attr_mux.attr,
	NULL,
};

static int usb_phy_mux_probe(struct platform_device *pdev)
{
	int ret;
	struct device_node *np = pdev->dev.of_node;
	struct regmap *map;
	struct usb_phy_mux *pdata;

	pdata = devm_kzalloc(&pdev->dev, sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return -ENOMEM;
	pdata->dev = &pdev->dev;
	dev_set_drvdata(&pdev->dev, pdata);

	map = syscon_regmap_lookup_by_phandle(np, "dvf,syscfg");
	if (IS_ERR(map))
		return PTR_ERR(map);

	pdata->f = devm_regmap_field_alloc(&pdev->dev, map,
					   usb_phy_mux_regfield);
	if (IS_ERR(pdata->f))
		return PTR_ERR(pdata->f);

	pdata->route_to_host = !!of_property_read_bool(np, "port2-to-host");

	ret = set_route(pdata);
	if (ret)
		return ret;

	ret = sysfs_create_files(&pdev->dev.kobj, dvf101_usb_phy_mux_sysfs);
	if (ret)
		return ret;

	dev_info(&pdev->dev, "PHY2 routed to %s",
		 xlate_mux_setting[pdata->route_to_host]);
	dev_info(&pdev->dev, "successfully registered USB PHY muxer");

	return 0;
}

static struct of_device_id usb_phy_mux_of_device_ids[] = {
	{ .compatible = "dspg,dvf101-usb-phy-mux" },
	{ },
};

static struct platform_driver usb_phy_mux_driver = {
	.probe = usb_phy_mux_probe,
	.driver = {
		.name = "usb-phy-mux",
		.owner = THIS_MODULE,
		.of_match_table = usb_phy_mux_of_device_ids,
	},
};

static int __init usb_phy_mux_init(void)
{
	return platform_driver_register(&usb_phy_mux_driver);
}
arch_initcall(usb_phy_mux_init);

MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("USB PHY muxer");
MODULE_AUTHOR("DSP Group");
