/*
 * Copyright 2013 Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <mach/led_ht.h>
#include <linux/init.h>

/************Defines************/
/* blink interval in ms */

#define INITIAL_LED_TIMER_BLINK_PATTERN	(3 * 1000)

/************Typedefs************/

/************Locals************/

/*Private Functions*/
static void blink_timer_expired(unsigned long arg);

/* Private Data */

static struct led_t led[] = {
#if defined(CONFIG_HT)
	{"LED_PHONE1", 0},
	{"LED_PHONE2", 0},
#if defined(CONFIG_HT818)
	{"LED_PHONE3", 0},
	{"LED_PHONE4", 0},
	{"LED_PHONE5", 0},
	{"LED_PHONE6", 0},
	{"LED_PHONE7", 0},
	{"LED_PHONE8", 0},
	{"LED_WAN", 0},
	{"LED_POWER", 0},
	{"LED_RJ45",   0}
#else
	{"LED_WAN", 0},
	{"LED_POWER", 0},
	{"LED_RJ45", 0}
#endif
#else
#error "No product defined"
#endif
};

#define NUM_LEDS (sizeof(led) / sizeof(struct led_t))

static int led_pattern_num;
static struct led_t led_pattern[NUM_LEDS];
static struct led_t led_rj45;

static struct timer_list blink_timer;
static int initial_led_state = 1;
static int initial_blink_count = 1;
/* updated by of_property_read_bool(np, "enable-active-high")) */
static int active_high = 1;

#define LED_ON		(active_high == 1 ? 1 : 0)
#define LED_OFF		(active_high == 1 ? 0 : 1)

/******************************
	Functions
********************************/
struct led_t *gs_get_led_pattern_ptr(int *led_num)
{
	*led_num = led_pattern_num;
	return &led_pattern[0];
}
EXPORT_SYMBOL(gs_get_led_pattern_ptr);

struct led_t *gs_get_led_rj45_ptr(void)
{
	return &led_rj45;
}
EXPORT_SYMBOL(gs_get_led_rj45_ptr);

int gs_led_set(const char *led_name, const int on)
{
	int i;

	for (i = 0; i < NUM_LEDS; i++) {
		if (strcmp(led[i].name, led_name) == 0) {
			if (on) {
					gpio_set_value(led[i].gpio, LED_ON);
			} else {
					gpio_set_value(led[i].gpio, LED_OFF);
			}
			return 0;
		}
	}
	return -1;
}
EXPORT_SYMBOL(gs_led_set);

static void blink_timer_expired(unsigned long arg)
{
	int i;
	int on;

	/* Toggle all the LEDs to OFF */
	initial_led_state ^= 1;
	on = initial_led_state;

	printk("LED_HT: Turning all(%d) LEDs ON/OFF\n", led_pattern_num + 1);

	for (i = 0; (i < led_pattern_num && i < NUM_LEDS); i++) {
		if (on)
			gpio_set_value(led_pattern[i].gpio, LED_ON);
		else
			gpio_set_value(led_pattern[i].gpio, LED_OFF);
	}

	if (on)
		gpio_set_value(led_rj45.gpio, LED_ON);
	else
		gpio_set_value(led_rj45.gpio, LED_OFF);

	initial_blink_count--;
	if (initial_blink_count) {
		blink_timer.expires = jiffies + INITIAL_LED_TIMER_BLINK_PATTERN / 10;
		add_timer(&blink_timer);
		printk("restarting led timer (%d left)\n", initial_blink_count);
	}
}

static int led_probe(struct platform_device *pdev)
{
	struct led_t *drvdata;
	struct device_node *np = pdev->dev.of_node;
	int i;

	drvdata = &led[0];
	platform_set_drvdata(pdev, drvdata);

	memset(led_pattern, 0, sizeof(led_pattern));
	led_pattern_num = 0;

	for (i = 0; i < NUM_LEDS; i++) {
		drvdata[i].gpio = of_get_named_gpio(np, drvdata[i].name, 0);
		if (gpio_is_valid(drvdata[i].gpio)) {
			printk("%s: valid\n", drvdata[i].name);
			gpio_request(drvdata[i].gpio, "grandstream-led");
			gpio_direction_output(drvdata[i].gpio, 1);
#ifdef CONFIG_HT818
			if (strcmp(led[i].name, "LED_RJ45") == 0)
				memcpy(&led_rj45, &led[i], sizeof(struct led_t));
			memcpy(&led_pattern[led_pattern_num++], &led[i], sizeof(struct led_t));
#else
			if (strcmp(led[i].name, "LED_RJ45") == 0)
				memcpy(&led_rj45, &led[i], sizeof(struct led_t));
			else
				memcpy(&led_pattern[led_pattern_num++], &led[i], sizeof(struct led_t));
#endif
			gpio_set_value(led[i].gpio, 0);
		} else
			led[i].name[0] = 0;
	}

	if (of_property_read_bool(np, "enable-active-high"))
		active_high = 1;
	else
		active_high = 0;

	/* Bug 33891 - HT802 [Vonage] P1: Bootloader should momentarily flash Internet\Phone1\Phone2
		as soon as the core processor initializes to indicate the core is functional.
		Solution: Start a timer to turn all the leds off after initialization.
	*/
	init_timer(&blink_timer);
	blink_timer.data = 0;
	blink_timer.function = blink_timer_expired;
	blink_timer.expires = jiffies + INITIAL_LED_TIMER_BLINK_PATTERN / 10;
	add_timer(&blink_timer);
	/* this flag gets toggled to blink the LEDs */
	initial_led_state = 1;
	/* This count get decremented on every timer expire */
	initial_blink_count = 1;
	return 0;
}

static int  led_remove(struct platform_device *pdev)
{
	int i;
	struct led_t *drvdata = platform_get_drvdata(pdev);

	for (i = 0; i < NUM_LEDS; i++) {
		if (gpio_is_valid(drvdata[i].gpio))
		gpio_free(drvdata[i].gpio);
	}
	return 0;
}

static const struct of_device_id led_of_match[]  = {
	{ .compatible = "grandstream,led", },
	{},
};
MODULE_DEVICE_TABLE(of, led_of_match);

static struct platform_driver gs_led_driver = {
	.probe		= led_probe,
	.remove		= led_remove,
	.driver		= {
		.name		= "led-driver",
		.owner		= THIS_MODULE,
		.of_match_table = led_of_match,	/* for historical reasons, this will remain as 'led' in the device tree */
	},
};
static int __init gs_led_init(void)
{
	return platform_driver_register(&gs_led_driver);
}
module_init(gs_led_init);

static void __exit gs_led_exit(void)
{
	platform_driver_unregister(&gs_led_driver);
}
module_exit(gs_led_exit);

MODULE_AUTHOR("Steven Frick (sfrick@grandstream.com)");
MODULE_DESCRIPTION("HRT Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:led-driver");
