/*
 * gs_slic.c  --  Silicon Labs SPI driver
 *
 * Copyright (C) 2017 Grandstream,Inc
 *
 * 20170608  jwli@grandstream.cn
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

//#define V_DEBUG
//#define DEBUG

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/spi/spi.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/tlv.h>
#include <asm/div64.h>
#include <linux/gpio.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/of_gpio.h>

#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>
#include <asm/uaccess.h>
#include "gs-slic.h"


/* Must keep this private pointer in the local scope of the file so 
   get_gs_slic_priv() can correctly return it to the bsp driver of the slic  
*/
static struct gs_slic_priv * ptr_gs_slic[NUM_SPI_BANKS] = { 0, 0, 0, 0, 0, 0, 0, 0 };

static int gs_slic_spi_probe(struct spi_device *spi)
{
	struct gs_slic_priv *gslic = NULL;
	struct device_node *np;
	int reset_gpio = -1, cs_gpio, ret;
	unsigned int id;

	np = spi->dev.of_node;
	if (np == NULL)
		dev_info(&spi->dev, "of node is null\n");

	/* Get the reset line associated wtih the devices on this SPI bank */
	reset_gpio = of_get_named_gpio(np, "reset-gpio", 0);

	if ((cs_gpio = of_get_named_gpio(np, "cs-gpio", 0)) < 0)
		cs_gpio = -1;

	/* Read the SPI id and verify it before moving on */
	ret = of_property_read_u32(np, "reg", &id);
	if (ret && ret != -EINVAL) {
		dev_err(&spi->dev, "Error parsing id value in devtree\n");
		return -ENODEV;
	}

	if (id >= NUM_SPI_BANKS) {
		dev_info(&spi->dev,
			"gs-slic:Driver support only %d SPI banks. Must increase NUM_SPI_BANKS macro\n",
			NUM_SPI_BANKS);
		return -ENODEV;
	}

	gslic = kzalloc(sizeof(struct gs_slic_priv), GFP_KERNEL);
	if (gslic == NULL)
		return -ENOMEM;

	gslic->reset = reset_gpio;
	gslic->id = id;
	gslic->cs = cs_gpio;

	spi_set_drvdata(spi, gslic);
	printk("gs-slic: spi_dev[%d] = 0x%8p\n", id, spi);
	ptr_gs_slic[id] = gslic;
	printk("gs-slic: ptr_gs_slic[%d] = 0x%8p\n", id, ptr_gs_slic[id]);

	/*
	* setup spi parameters; this makes sure that parameters we request
	* are acceptable by the spi driver
	*/
	spi->mode = SPI_MODE_3; /* clk active high */
	spi->bits_per_word = 8;

	ret = spi_setup(spi);
	if (ret < 0) {
		dev_err(&spi->dev, "spi_setup() failed\n");
		goto err_gslic;
	}

	gslic->spi = spi;

	ret = gpio_request(gslic->reset, "gs-slic reset");
	if (ret < 0) {
		dev_err(&spi->dev, "failed requesting reset gpio\n");
		goto err_gslic;
	}

	/* Resetting the SPI devices will be done by the low level drivers  */
	/* 	gs_slic_reset(gslic->reset); */
	dev_set_name(&spi->dev, "%s.%u", "gs-slic", gslic->id);

	dev_info(&spi->dev, "successfully registered\n");

	return ret;

err_gslic:
	kfree(gslic);
	return ret;
}

/* Helper function to allow low level BSP driver finish all the setups */
struct gs_slic_priv * get_gs_slic_priv(int spi_dev_id)
{
	printk("calling get_gs_slic return spi_dev[%d] = 0x%8p\n", spi_dev_id, ptr_gs_slic[spi_dev_id]);

	return ptr_gs_slic[spi_dev_id];
}
EXPORT_SYMBOL(get_gs_slic_priv);

static int gs_slic_spi_remove(struct spi_device *spi)
{
	struct gs_slic_priv *gslic = spi_get_drvdata(spi);

	kfree(gslic);
	return 0;
}

static const struct of_device_id gs_slic_dt_match[] = {
	{ .compatible = "grandstream,gs_slic_spi", },
	{}
};

static struct spi_driver gs_slic_spi_driver = {
	.driver = {
		.name	= "gs-slic",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
		.of_match_table = gs_slic_dt_match,
	},
	.probe	= gs_slic_spi_probe,
	.remove	= gs_slic_spi_remove,
};

static int __init gs_slic_module_init(void)
{
	return spi_register_driver(&gs_slic_spi_driver);
}

static void gs_slic_module_exit(void)
{
	spi_unregister_driver(&gs_slic_spi_driver);
}

module_init(gs_slic_module_init);
module_exit(gs_slic_module_exit);

MODULE_DESCRIPTION("Silicon Labs SPI driver driver");
MODULE_AUTHOR("GRANDSTREAM");
MODULE_LICENSE("GPL");
