/*
 * gs-tdm.c -- Grandstream TDM driver
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 * 2nd Jan, 2012:
 * Adapted to kernel 3.4 by Sujan.S <sujan.somisetty@dspg.com>
 *
 * Copyright (C) 2017 Grandstream, Inc.
 *  19 jun, 2019
 *  Ported gs-tdm base on "DSPG ASoC CPU DAI driver" by <jwli@grandstream.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/reset.h>
#include <mach/platform.h>
#include <mach/coma-tdm.h>
#include <mach/hardware.h>
#include <linux/proc_fs.h>

#include <linux/clk-provider.h>

#ifdef V_DEBUG
#ifndef DEBUG
#define DEBUG
#endif
#define vdbg_prt(fmt, ...) \
		printk("GS-HT Pltfrm TDM: %s(): " fmt, __func__, ##__VA_ARGS__)
#else
#define vdbg_prt(fmt, ...)
#endif

#ifdef DEBUG
#define dbg_prt(fmt, ...) \
		printk("GS-HT Pltfrm TDM: %s(): " fmt, __func__, ##__VA_ARGS__)

#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
#define dbg_prt(fmt, ...)
#define DIR_SRT(dir)
#endif

/*****************************************************************************/

/*
 * The structure is per TDM and shared between the playback and capture
 * substreams. Fortunately there's no need for a dedicated mutex because the
 * access to the struct is implicitly guarded by the 'pcm_mutex' of the ASoC
 * core.
 */
struct dvf_tdm_priv {
	uint dai_id;
	void *reg_base;
	resource_size_t iores_mem;
	int iores_irq;
	struct clk *clk;
	uint tdm_clk_active;
	uint tdm_clk_suspended;
	spinlock_t tdm_clk_lock;

	int opencnt         : 2;
	int fs_master       : 1;
	int clk_master      : 1;
	int play_configured : 1;
	int capt_configured : 1;
	int play_active     : 1;
	int capt_active     : 1;

	int css_to_own_tdm  : 1; /* Pending request to TDM-ownership switch-over:
			          * '0' - to APP; '1' - to CSS
			          */
	int css_owns_tdm    : 1; /* Current state of TDM ownership:
				  * '0' - by APP; '1' - by CSS;
				  */

	int tdm_config_only  :1; /* This flag indicates to configure TDM only,
				    not to send grant to CSS */
	int tdm_sample_bits;	/* num SCLKs per pcm sample, over TDM bus */
	int alsa_sample_bits;	/* num bits per alsa pcm sample */
	int rate;		/* sample-rate (Fs), in Hz */
	int alsa_channels;	/* num channels (TDM time-slots) */
	uint dai_format;
	uint num_tdm_slots; /* TDM slots configurable via param */

	struct reset_control *rc;
};

static struct dvf_tdm_priv *gs_tdm[CONFIG_DVF_NR_TDMS];
static int css_own_tdm[CONFIG_DVF_NR_TDMS];

/* Note: There are 2 TDM machines in the dvf99 chip, each one having its
 *       own TDM-clock. However, there is a Single (global) DP-clock, driving
 *       the codec chip(or codec chips !) mounted on the board.
 */

#define DEVICE_NAME "dvf-tdm"

#define DRV_NAME "dvf-tdm"

enum eSYS_CLK {
	/* Add for other system-clock freq */
	kSYS_CLK_DEFAULT,
	kSYS_CLK_MAX
};

enum eFS_HZ {
	kFS_HZ_8000,
	kFS_HZ_11025,
	kFS_HZ_16000,
	kFS_HZ_22050,
	kFS_HZ_32000,
	kFS_HZ_44100,
	kFS_HZ_48000,
	kFS_HZ_MAX
};

/*****************************************************************************/

/* TDM registers: */
#define TDM_ENABLE                   0x0000
#define TDM_CONFIG_1                 0x0004
#define TDM_CONFIG_1_AUTO_SYNC       (1 << 13)
#define TDM_CONFIG_1_CLK_MSTR_SLV    (1 << 12)
#define TDM_CONFIG_1_FSYNC_MSTR_SLV  (1 << 11)
#define TDM_CONFIG_1_SYNC_EDGE       (1 << 10)
#define TDM_CONFIG_1_RX_EDGE         (1 <<  9)
#define TDM_CONFIG_1_TX_EDGE         (1 <<  8)
#define TDM_CONFIG_1_TIME_SLOTS      0x1F
#define TDM_CONFIG_2                 0x0008
#define TDM_CONFIG_2_SHIFT           (1 << 9)
#define TDM_CONFIG_2_TRUNCATE        (1 << 8)
#define TDM_CONFIG_2_BANK_SIZE       0x07
#define TDM_FSYNC                    0x0010
#define TDM_FSYNC_DUR                0x0014
#define TDM_FIFO_WM                  0x0018
#define TDM_INT_EN                   0x0020
#define TDM_INT_EN_RX_FULL           (1 << 3)
#define TDM_INT_EN_TX_EMPTY          (1 << 1)
#define TDM_INT_EN_TX_WM             (1 << 0)
#define TDM_INT_EN_RX_WM             (1 << 2)
#define TDM_INT_CLR                  0x0028
#define TDM_INT_CLR_TX_FLAGS         (1 << 3)
#define TDM_INT_CLR_FSYNC            (1 << 2)
#define TDM_INT_CLR_RX_OVRN          (1 << 1)
#define TDM_INT_CLR_TX_UNDRN         (1 << 0)
#define TDM_INT_STATUS               0x002C
#define TDM_INT_STATUS_RX_OVRN       (1 << 3)
#define TDM_INT_STATUS_TX_UNDRN      (1 << 1)
#define TDM_STATUS                   0x0030
#define TDM_TX_FIFO                  0x0038
#define TDM_RX_FIFO                  0x003C
#define TDM_TX_REG_0                 0x0A00
#define TDM_RX_REG_0                 0x1200
#define TDM_FDA_REG_00               0x1800
#define TDM_FDA_REG_01               0x1804
#define TDM_FDA_REG_02               0x1808
#define TDM_FDA_REG_03               0x180C

#define TDM_REGLEN 0x2000

/* The TDM FIFOs registers are 20 bits wide.
 * In the dvf99 TDM machine, the 20 MSbits are taken
 * when writing/reading the Tx/Rx fifo with a u32 value.
 * Note that this is opposite to the dw-52/74 TDM machine - 20 LSbits.
 */

/* The TDM Tx & Rx FIFOs have room for total of 128 samples -
 * e.g. for 64 TDM frames, where a frame consists of 2 time-slots
 * (Left+Right channels, interleaved).
 */
/* #define TDM_FIFO_DEPTH	64 */

/* For HT818's 16 TDM frames, where a frame consists of 8 time-slots */
#define TDM_FIFO_DEPTH   16

/* When using DMA to write/read the TDM FIFOs,
 * we set the fifo water-mark level to trigger a dma
 * transaction. This is done by hardware. In this case,
 * the WM interrupts are not used, thus disabled.
 *
 * NOTE:
 * ----
 * TDM_FIFO_LVL_FOR_DMA Must(!) be >= N, where N is defined in
 * arch/arm/mach-dmw/devices.c, by the value of "transfer_length" fields
 * of the DMA channels used by the 3 TDM machines, as:
 *
 *	.transfer_length = (2 * N) - 1,	// N audio frames (stereo samples)
 *
 * "transfer_length" is the DMA Burst length, in u32 units.
 *
 * TDM_FIFO_LVL_FOR_DMA Should be equal to N.
 * It can be larger then N, but it CANNOT be Smaller !!!
 * If it is smaller, pcm data in the TDM TX Fifo might be over-written!
 */
#define	TDM_FIFO_LVL_FOR_DMA	(TDM_FIFO_DEPTH / 4)

#define TDM_TX_FIFO_LEVEL  TDM_FIFO_LVL_FOR_DMA  /* emptiness level, in frames */
#define TDM_RX_FIFO_LEVEL  TDM_FIFO_LVL_FOR_DMA  /* fulness   level, in frames */

#define TDM_SLOT(_index, _cd, _width) \
	(((_index) << 8) | (0x03 << 6) | (!!(_cd) << 5) | (_width & 0x1F))

/* CSS code works vs. the TDM FIFOs using IRQ.
 * It passes 4 audio frames (stereo-samples) per
 * Water-Mark interrupt.
 */
//#define TDM_FIFO_LVL_FOR_CSS (TDM_FIFO_DEPTH - 4)
#define TDM_FIFO_LVL_FOR_CSS (TDM_FIFO_DEPTH - 2)

/*****************************************************************************/

static inline unsigned int
round_up_power_2(unsigned int x)
{
	x--;
	x |= x >> 1;  // handle  2 bit numbers
	x |= x >> 2;  // handle  4 bit numbers
	x |= x >> 4;  // handle  8 bit numbers
	x |= x >> 8;  // handle 16 bit numbers
	x |= x >> 16; // handle 32 bit numbers
	x++;

	return x;
}

static inline unsigned int
log_base_2(unsigned int x)
{
	unsigned int ans = 0;

	while (x >>= 1)
		ans++;

	return ans;
}

static unsigned long
tdm_read(struct dvf_tdm_priv *tdm, int reg)
{
	return readl(tdm->reg_base + reg);
}

static void
tdm_write(struct dvf_tdm_priv *tdm, int reg, unsigned long value)
{
	writel(value, tdm->reg_base + reg);
}

/* Note:
 * Can't(!) use reference-counting on TDM clock enable/disable,
 * since ALSA APIs _hw_params() and _hw_free() are Not symetric:
 * _hw_free() may, and is, called 2 times or more for a single
 * _hw_params().
 */
static int
tdm_clk_enable(struct dvf_tdm_priv *tdm)
{
	unsigned long flags;
	int ret = 0;

	dbg_prt();

	spin_lock_irqsave(&tdm->tdm_clk_lock, flags);
	if (tdm->tdm_clk_active == 0) {
		tdm->tdm_clk_active = 1;
		clk_prepare_enable(tdm->clk);
		ret = 0;
	} else {
		ret = 1;
	}
	spin_unlock_irqrestore(&tdm->tdm_clk_lock, flags);

	vdbg_prt("%s %s\n", "TDM clock:", ((ret == 0) ? "Enabled now" : "Already Enabled"));

	return ret;
}

static int
tdm_clk_disable(struct dvf_tdm_priv *tdm)
{
	unsigned long flags;
	int ret = 0;

	dbg_prt();

	spin_lock_irqsave(&tdm->tdm_clk_lock, flags);
	if (tdm->tdm_clk_active != 0) {
		tdm->tdm_clk_active = 0;
		clk_disable_unprepare(tdm->clk);
		ret = 0;
	} else {
		ret = 1;
	}
	spin_unlock_irqrestore(&tdm->tdm_clk_lock, flags);

	vdbg_prt("%s %s\n", "TDM clock:", ((ret == 0) ? "Disabled now" : "Already Disabled"));

	return ret;
}

static void
dvf_tdm_reset(struct dvf_tdm_priv *tdm)
{
	int ret;

	dbg_prt();

	/* Disable the TDM machine */
	tdm_write(tdm, TDM_ENABLE, 0);

	/* TDM clock must be Disabled when doing the TDM Reset,
	 * Must wait(!) for clock to Cease: the worst case is
	 * with the slowest clock.
	 */
	tdm_clk_disable(tdm);

	/* Delay >= 20 uSec, according to latest spec recommendation.
	 * A.M. April 3, 2012:
	 * However, according to slowest fs=8KHz --> sclck = 8*32=256KHz,
	 * you need to wait only 1/256KHz = ~4 uSec
	 */
	udelay(20);

	/* Reset TDM block - the reset suppose to be Asynchronous */
	ret = reset_control_assert(tdm->rc);
	if (ret)
		vdbg_prt("reset set failed");

	udelay(2);

	ret = reset_control_deassert(tdm->rc);
	if (ret)
		vdbg_prt("reset release failed");

	/* A.M. April 3, 2012:
	 * Note: Keep the clock to the TDM block Disabled !
	 */
}

static int
dvf_tdm_configure(struct dvf_tdm_priv *tdm)
{
	unsigned int slots;
	unsigned int alsa_bps, tdm_bps;
	unsigned int tdm_reg;
	unsigned int tx_fifo_wm_lvl, rx_fifo_wm_lvl;
	ulong master_sclk_rate = 0;
	unsigned int i;

	dbg_prt();

#ifdef CONFIG_DSPG_CSS_TDM
	if (tdm->css_owns_tdm) {
		/* Resume APP ownership over the TDM machine,
		 * Before(!) touching the TDM machine.
		 */
		if (coma_tdm_revoke(tdm->dai_id) == 0) {
			tdm->css_owns_tdm = 0;
			vdbg_prt("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoked OK");
		} else {
			vdbg_prt("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoke Failed");
		}
	}
#endif /* CONFIG_DSPG_CSS_TDM */

	/* Note:
	 * "bps" here means Bits-per-Sample.
	 *
	 * When the DVF99 TDM machine is the Master, i.e sourcing out the
	 * TDM SCLK & FSync clocks, there might be cases in which we would like
	 * to produce More SCLK clocks than the number of bits in the pcm sample.
	 * This is to achive better SCLK(BCLK) frequency accuracy.
	 * The SCLK frequency depends on Sample-Rate;
	 * its accuracy depends on the driving PLL frequency setting,
	 * the clock-tree further dividers resolution and the Sample-Rate Family
	 * (Fs = 8/12/11.025KHz multiplications).
	 * Example: to Tx a Fs=44.1KHz 16-bit pcm sample, where PLL freq=147.456MHz,
	 * we'll use 19 sckls (instaed of 16).
	 *
	 * Since the FIFO regs width is 20 bits, we are limitted to Tx/Rx
	 * 20 bits (20 sclks) in a tdm time-slot.
	 * Another HW restriction: minimal number of bits in a time-slot is 8 (8 sclks).
	 * (8 <= nun sclks per time-slot <= 20).
	 * However, the (ALSA) PCM sample may contain 16 (commonly), 20, 24, etc. bits.
	 *
	 * When there is a need to Tx/Rx more than 20 bits per time-slot,
	 * we can use Two (!) time-slots per pcm sample:
	 *  - the first  time-slot is of Audio/Voice type, having upto 20 bits.
	 *  - the second time-slot is of Control/Status type, having the rest of the bits,
	 *    but not less then 8.
	 *
	 * Currently, we Do Not support more than 20 bits per pcm sample.
	 */

	/* Note: the term "bps" means here "Bits per Sample" ! */

	alsa_bps = tdm->alsa_sample_bits;

	if ((tdm->clk_master) && (tdm->fs_master)) {

		/* dvf99 is TDM Master */
		/* ------------------- */

		/* vs. other codecs */
		tdm_bps = alsa_bps;
		master_sclk_rate = tdm->rate * tdm_bps * tdm->num_tdm_slots;

		/* Set Actual SCLK (BCLK) frequency over TDM bus,
		 * generated by the dvf99, as TDM Master !
		 */
		master_sclk_rate = clk_round_rate(tdm->clk, master_sclk_rate);

		clk_set_rate(tdm->clk, master_sclk_rate);
	} else {
		/* dvf99 is TDM Slave */
		/* ------------------ */

		tdm_bps = alsa_bps;
	}

	printk("GS-HT Pltfrm TDM: %s %u %s %lu\n",
		 "Final tdm_bps:", tdm_bps,
		 "Final sclk rate:", master_sclk_rate);

	/* A.M. April 3, 2012:
	* Problematic: looks like you should Not enable the Clock to the
	* TDM block before programming its registers. Though the TDM block itself
	* is disbled here (TDM_ENABLE = 0), if the clock is enabled before setting
	* the registers, the machine won't always sync correctly on FSync
	* as Slave or produce badeFSync timing as Master !
	* The FSync then deviates from the pcm frame real starting point,
	* causing the "metallic sound" phenomenon.
	* There is No hw explanation provided by the chip-design team.
	*
	* Workaround: do most of the TDM machine registers configuration,
	* including its reset-sequence, while the TDM clock is disabled.
	* This is CONTRARY to the chip-design directives since dw-52 !
	*/

	/*
	 * We should sw-reset the DVF99 TDM block perior to any TDM configuration
	 * change according to the data sheet.
	 * Note: this function now Leaves the TDM clock Disabled!
	 */
	dvf_tdm_reset(tdm);
	
	/* Re-enable the clock to the TDM block */
	tdm_clk_enable(tdm);

	/* Configure Frame Descriptor Array. Might insert some NULL samples if
	 * required.
	 */
	if (tdm->num_tdm_slots == 0) {
		vdbg_prt("Resetting num of slots to 2\n");
		tdm->num_tdm_slots = 2;
	}
	slots = tdm->num_tdm_slots;
	for (i=0; i<tdm->num_tdm_slots; i++)
		tdm_write(tdm, TDM_FDA_REG_00 + i*4, TDM_SLOT(i, 0, tdm_bps));

	/* Note:
	 * - Here we set the TDM bus configuration:
	 *	DVF99 TDM Slave/Master;
	 *	I2S format Only(!) (for now).
	 *
	 * Remember the DVF99 TDM machine limitation:
	 *
	 *   - as Master:
	 *     ----------
	 *     it can produce only Rising FSync pulse signal,
	 *     (at rising or falling SCLK edge - ok), but Cannot(!) produce
	 *     falling FSync signal. Thus, for true I2S format, we need to Shift
	 *     the Rising of the FSync (which is LRC signal in I2S) in such way that the
	 *     Falling edge would take place one bit before the start of Frame,
	 *     which is start of the Left channel. This means:
	 *
	 *     - for 2 time-slots (up to 20 bits per sample), with N bits per time-slot :
	 *       Trigger the FSync on N-1 bit of the First time slot.
	 *       e.g. on bit 15 of the first time-slot, for 16-bits time-slots.
	 *
	 *   - as Slave:
	 *     ---------
	 *     it syncs only(!) on Rising edge of FSync.
	 *     So we need to Invert the FSync polarity, generated at the Wolfson.
	 *     However, we Cannot do this, since the Wolfson will put the ADC data
	 *     on the second tdm-slot (Right channel) - we can't live with that.
	 *     So we have Left/Right channel swap!
	 *
	 *
	 *     Slave or Master:
	 *     ----------------
	 *     Tx & Rx edges can, and shoud, be configured, for I2S, as follows:
	 *     - Tx data at Falling edge of SCLK
	 *     - Rx data at Rising edge of SCLK
	 */

	/* Configure TDM as Master/Slave SYNC/TX/RX Edges, 2 time-slots */
	tdm_reg = 0x0000;
	tdm_reg |= slots - 1;			/* Number of time-slots in frame (2 or 4) */
	tdm_reg |= TDM_CONFIG_1_TX_EDGE;	/* Tx data on SCLK Falling edge */
	tdm_reg &= ~TDM_CONFIG_1_RX_EDGE;	/* Rx data on SCLK Rising edge */

	/* FSync rise at SCLK Falling edge:
	 *   Matters only if Master; Irrelevant for Slave.
	 */
	switch (tdm->dai_format & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_IB_NF:
		tdm_reg &= ~(TDM_CONFIG_1_TX_EDGE);	/* Tx data on SCLK Rising edge */
		tdm_reg &= ~(TDM_CONFIG_1_SYNC_EDGE);	/* SCLK and FSYNC should be in phase */
		break;
	default:
		tdm_reg |= TDM_CONFIG_1_SYNC_EDGE;
		break;
	}

	if (unlikely(tdm->fs_master)) {
		/* DVF99 TDM is FSync Master */
		tdm_reg |=  TDM_CONFIG_1_FSYNC_MSTR_SLV;/* Source-out FSync clock */
	} else {
		/* DVF99 TDM is FSync Slave */
		tdm_reg &= ~TDM_CONFIG_1_FSYNC_MSTR_SLV;/* Sink-in FSync clock */
		tdm_reg |=  TDM_CONFIG_1_AUTO_SYNC;	/* Resync on FSync, after FSync loss/jitter */
	}
	if (unlikely(tdm->clk_master))
		tdm_reg |=  TDM_CONFIG_1_CLK_MSTR_SLV;	/* Source-out SCLK clock */
	else
		tdm_reg &= ~TDM_CONFIG_1_CLK_MSTR_SLV;	/* Sink-in SCLK clock */

	tdm_write(tdm, TDM_CONFIG_1, tdm_reg);

	/* Configure MSB fisrt, 64(32) banks (in a fifo line), 2(4) fifo lines */
	tdm_reg = 0x0000;
	tdm_reg |= TDM_CONFIG_2_SHIFT;    /* Shift Out/In MSB first */
	tdm_reg |= TDM_CONFIG_2_TRUNCATE; /* On Tx/Rx - Allign data bits if less then 20 */
	/* bank size computation */
	tdm_reg |= log_base_2(round_up_power_2(slots)) - 1;

	vdbg_prt("Number of slots = %d, Config-2 register set to 0x%x\n",
		 slots, tdm_reg);

	tdm_write(tdm, TDM_CONFIG_2, tdm_reg);

	/* Configure LRC Rise on last bit of Left channel -
	 * to produce LRC Fall at 1 bit before data frame starts (standard I2S)
	 */
	switch (tdm->dai_format & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_IB_NF:
		tdm_reg = 0x0000; /* For FXO DAI, FSYNC should rise at first slot and first bit */
		break;
	default:
		tdm_reg = 0x0000;
		if (likely(slots == 2)) {
			tdm_reg |= ((tdm_bps - 1) << 8);/* Last bit of */
			tdm_reg |= (0 & 0x001F);	/* First time slot */
		}
		break;
	}
	tdm_write(tdm, TDM_FSYNC, tdm_reg);

	/* Configure LRC length, in sclks */
	tdm_write(tdm, TDM_FSYNC_DUR, tdm_bps - 1);

	/* Configure TX & RX FIFOs Water-Mark */
	tdm_reg = 0x0000;

#if 1 /* USE FIFO LVL FOR CSS default */
	tx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS;
	rx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS;
#else
	if (likely(!tdm->css_to_own_tdm && !tdm->tdm_config_only)) {
		tx_fifo_wm_lvl = TDM_TX_FIFO_LEVEL;
		rx_fifo_wm_lvl = TDM_RX_FIFO_LEVEL;
	} else {
		tx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS; //TDM_FIFO_DEPTH/slots -2;
		rx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS; //TDM_FIFO_DEPTH/slots -2;
	}
#endif
	/* Future: for more then 20 bits in per TDM channels,
	 * using two time-slots per L/R channel
	 */
	if (unlikely(slots == 4)) {
		tx_fifo_wm_lvl >>= 1; /* :2 */
		rx_fifo_wm_lvl >>= 1; /* :2 */
	}

	/* Latest documented changes in the dvf99 TDM machine:
	 * - RX_LVL field is bits 13-8.
	 * - TX_LVL and RX_LVL actually trigger DMA or WM-Interrupt
	 *   at field value + 1.
	 */
	tdm_reg |= (tx_fifo_wm_lvl & 0x003F) << 0;
	tdm_reg |= (rx_fifo_wm_lvl & 0x003F) << 8; /* bits 13:8 */
	tdm_write(tdm, TDM_FIFO_WM, tdm_reg);

	/* "Clear" TDM FIFOs (!):
	 * Note: this is needed in order to Deassert the TDM Water-Mark
	 * interrupt indications, at the TDM machine, in case they are
	 * asserted.
	 * Otherwise, an interrupt will be asserted immediately after
	 * unmasking these interrupts, Even Though the machine is
	 * Disabled (TDM_ENABLE has 0)!!!
	 * This is because the TDM WM interrupts are of "Level" type
	 * (Not "Edge-trigger").
	 */
	/* Fill-in Tx fifo */
	i = tx_fifo_wm_lvl + 1;
	i *= slots / 2;
	while (i--) {
		tdm_write(tdm, TDM_TX_FIFO, 0x0000); /* Left */
		tdm_write(tdm, TDM_TX_FIFO, 0x0000); /* Right */
	}
	/* Empty Rx fifo */
	i = rx_fifo_wm_lvl + 1;
	i *= slots / 2;
	while (i--) {
		tdm_read(tdm, TDM_RX_FIFO); /* Left */
		tdm_read(tdm, TDM_RX_FIFO); /* Right */
	}

	/* TODO - add handling for grant timeout (CSS is in panic) */

#ifdef CONFIG_DSPG_CSS_TDM
	if (unlikely(tdm->css_to_own_tdm)) {
		if (!tdm->css_owns_tdm) {
			/* TDM ownership should be given to CSS.
			 *
			 * The TDM machine will now be handled by the CSS'es TDM driver.
			 * The APP now finishes the TDM machine configuration to meet
			 * the CSS needs and Ceases to handle it !
			 */

			/* Disable TDM IRQ (at the APP-ICU!) */
			disable_irq(tdm->iores_irq);

			/* Enable Tx/Rx Water-Mark and Underrun/Overrun interrupts
			 * (to be handled by the CSS!)
			 */
			tdm_write(tdm, TDM_INT_EN,
				  TDM_INT_EN_TX_EMPTY | TDM_INT_EN_RX_FULL | /* For normal operation */
				  TDM_INT_EN_TX_WM | TDM_INT_EN_RX_WM);    /* To catch errors */

			/* Handover TDM ownership to CSS: */
			if (coma_tdm_grant(tdm->dai_id, tdm->rate,
					   slots, tdm_bps) == 0) {
				tdm->css_owns_tdm = 1;
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Granted to CSS");
			} else {
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Grant Failed");
			}
		}
	} else {
		/* TDM ownership should stay owned by the APP */
		tdm->css_owns_tdm = 0;
		vdbg_prt("%s %u\n", "APP owns tdm", tdm->dai_id);
	}
#endif /* CONFIG_DSPG_CSS_TDM */

	tdm_write(tdm, TDM_ENABLE, 0x1);

	return 0;
}

static int dvf_tdm_revoke(struct dvf_tdm_priv *tdm)
{
#ifdef CONFIG_DSPG_CSS_TDM
	if (tdm->css_owns_tdm) {
		/* Resume APP ownership over the TDM machine,
		 * Before(!) touching the TDM machine.
		 */
		if (coma_tdm_revoke(tdm->dai_id) == 0) {
			tdm->css_owns_tdm = 0;
			printk("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoked OK");
		} else {
			printk("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoke Failed");
		}
	}

#endif /* CONFIG_DSPG_CSS_TDM */
	tdm_write(tdm, TDM_ENABLE, 0x1);

	return 0;
}

static int dvf_tdm_grant(struct dvf_tdm_priv *tdm)
{
	tdm_write(tdm, TDM_ENABLE, 0x0);

#ifdef CONFIG_DSPG_CSS_TDM
	if (unlikely(tdm->css_to_own_tdm)) {
		if (!tdm->css_owns_tdm) {
			/* TDM ownership should be given to CSS.
			 *
			 * The TDM machine will now be handled by the CSS'es TDM driver.
			 * The APP now finishes the TDM machine configuration to meet
			 * the CSS needs and Ceases to handle it !
			 */

			/* Disable TDM IRQ (at the APP-ICU!) */
			disable_irq(tdm->iores_irq);

			/* Enable Tx/Rx Water-Mark and Underrun/Overrun interrupts
			 * (to be handled by the CSS!)
			 */
			tdm_write(tdm, TDM_INT_EN,
				TDM_INT_EN_TX_EMPTY | TDM_INT_EN_RX_FULL | /* For normal operation */
				TDM_INT_EN_TX_WM | TDM_INT_EN_RX_WM);	   /* To catch errors */

			/* Handover TDM ownership to CSS: */
			if (coma_tdm_grant(tdm->dai_id, tdm->rate,
					tdm->num_tdm_slots, tdm->alsa_sample_bits) == 0) {
				tdm->css_owns_tdm = 1;
				printk("%s %u %s\n",
					"tdm", tdm->dai_id, "Granted to CSS");
			} else {
				printk("%s %u %s\n",
					"tdm", tdm->dai_id, "Grant Failed");
			}
		}
	} else {
		/* TDM ownership should stay owned by the APP */
		tdm->css_owns_tdm = 0;
		printk("%s %u\n", "APP owns tdm", tdm->dai_id);
	}
#endif /* CONFIG_DSPG_CSS_TDM */

	return 0;
}

static irqreturn_t
dvf_tdm_irq(int irq, void *dev_id)
{
	struct dvf_tdm_priv *tdm = dev_id;
	irqreturn_t ret = IRQ_NONE;
	uint pending;

	pending = tdm_read(tdm, TDM_INT_STATUS) & tdm_read(tdm, TDM_INT_EN);

	if (pending) {
		if (pending & (TDM_INT_STATUS_RX_OVRN | TDM_INT_STATUS_TX_UNDRN)) {
			if (pending & TDM_INT_STATUS_TX_UNDRN) {
				printk(KERN_WARNING "TDM %u Underrun !!!\n",
				       tdm->dai_id);

			}
			if (pending & TDM_INT_STATUS_RX_OVRN) {
				printk(KERN_WARNING "TDM %u Overrun !!!\n",
				       tdm->dai_id);
			}

			/* Clear X-RUN interrupts (in TDM machine) */
			tdm_write(tdm, TDM_INT_CLR, TDM_INT_CLR_RX_OVRN | TDM_INT_CLR_TX_UNDRN);
		} else {
			/* Other interrupts should not happen */
			// ... Unless machine is TDM Slave and FSYNC_INT enabled
			// and FSync-jitter happens.

			/* Clear All TDM interrupts (in TDM machine) */
			tdm_write(tdm, TDM_INT_CLR, 0xFFFFF);

			printk(KERN_WARNING "Unexpected TDM interrupt !!! pending = 0x%X\n", pending);
		}
		ret = IRQ_HANDLED;
	}

	return ret;
}

/************************
* css_own_tdm_proc
*************************/
static int dvf_css_own_tdm_proc_show(struct seq_file *m, void *v)
{
	void *data = (void *)m->private;
	uint tdm_id = *(uint *)data;

	seq_printf(m, "%d", css_own_tdm[tdm_id % CONFIG_DVF_NR_TDMS]);
	return 0;
}

static ssize_t dvf_css_own_tdm_proc_write(struct file *file,
		const char __user *buffer,
		size_t len,
		loff_t *offset)
{
	void *data = PDE_DATA(file_inode(file));
	uint tdm_id = *(uint *)data;
	static struct dvf_tdm_priv *tdm;

	tdm = gs_tdm[tdm_id];

	if ( buffer[0] == '1' ) {
		tdm->css_to_own_tdm = 1;
		dvf_tdm_grant(tdm);
		css_own_tdm[tdm_id] = 1;
	} else {
		tdm->css_owns_tdm = 0;
		dvf_tdm_revoke(tdm);
		css_own_tdm[tdm_id] = 0;
	}

	return len;
}

static int dvf_css_own_tdm_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, dvf_css_own_tdm_proc_show, PDE_DATA(inode));
}

static const struct file_operations dvf_css_own_tdm_proc_fops = {
	.open = dvf_css_own_tdm_proc_open,
	.read = seq_read,
	.write = dvf_css_own_tdm_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

static struct proc_dir_entry *gs_procfs_root;

static int
dspg_tdm_dev_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	struct dvf_tdm_priv *tdm;
	int ret;

	struct proc_dir_entry * procfs_entry;
	struct resource *iores;
	char own_tdm_name[20];

	dbg_prt();

	tdm = devm_kzalloc(&pdev->dev,sizeof(struct dvf_tdm_priv), GFP_KERNEL);
	if (!tdm) {
		dev_err(&pdev->dev, "Can't allocate dvf_tdm_priv");
		return -ENOMEM;
	}
	dev_set_drvdata(&pdev->dev, tdm);

	ret = of_property_read_u32(np, "id", &tdm->dai_id);
	if (ret) {
		dev_err(&pdev->dev, "missing 'id' property!\n");
		return -EINVAL;
	}

	if (tdm->dai_id >= CONFIG_DVF_NR_TDMS) {
		dev_err(&pdev->dev, "invalid 'id' property!\n");
		return -EINVAL;
	}

	/* get dev clock && resources for gs tdm */
	tdm->rc = devm_reset_control_get(&pdev->dev, NULL);
	if (IS_ERR(tdm->rc))
		return PTR_ERR(tdm->rc);

	tdm->clk = devm_clk_get(&pdev->dev, NULL);
	if (IS_ERR(tdm->clk)) {
		ret = PTR_ERR(tdm->clk);
		dev_err(&pdev->dev, "No clock!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		"gs (tdm) id", tdm->dai_id, "clk_get OK");

	tdm->iores_irq = platform_get_irq(pdev, 0);
	if (tdm->iores_irq < 0) {
		ret = tdm->iores_irq;
		dev_err(&pdev->dev, "No IRQ!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		"gs (tdm) id", tdm->dai_id, "platform_get_irq OK");

	iores = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!iores) {
		dev_err(&pdev->dev, "Mem iores missing!\n");
		return -EINVAL;
	}
	vdbg_prt("%s %u %s\n",
		"gs (tdm) id", tdm->dai_id, "platform_get_resource OK");

	tdm->iores_mem = iores->start;

	tdm->reg_base = devm_ioremap(&pdev->dev, iores->start, TDM_REGLEN);
	if (!tdm->reg_base) {
		dev_err(&pdev->dev, "Cannot remap registers!\n");
		return -ENOMEM;
	}
	vdbg_prt("%s %u %s\n",
		 "gs (tdm) id", tdm->dai_id, "ioremap_nocache OK");

	ret = devm_request_irq(&pdev->dev, tdm->iores_irq, dvf_tdm_irq, 0,
			"dvf_tdm_irq", tdm);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request IRQ!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		 "gs (tdm) id", tdm->dai_id, "request_irq OK");

	spin_lock_init(&tdm->tdm_clk_lock);

	/* TDM configuration for HT */
	tdm->tdm_clk_active = 0;
	tdm->tdm_clk_suspended = 0;

	/* For HT818, we will configure tdm ownership in libcordless,
	 * and here, we configure the ownership to APP as default.
	 */
	tdm->css_to_own_tdm = 0;
	tdm->css_owns_tdm = 0;
	tdm->tdm_config_only = 0;

	tdm->alsa_sample_bits = 16;	/* Defualt PCM sample width, in bits */
	tdm->tdm_sample_bits = 16;	/* Default TDM time-slot width, in SCLK cycles:
					 * (1 cycle per bit) */

	tdm->dai_format = SND_SOC_DAIFMT_IB_NF | SND_SOC_DAIFMT_CBS_CFS;

	tdm->num_tdm_slots = 8;
	tdm->rate = 8000;
	tdm->clk_master = 1;
	tdm->fs_master = 1;

	dvf_tdm_configure(tdm);

	vdbg_prt("gs (tdm %p)id %u probed ok\n", tdm, tdm->dai_id);

	gs_tdm[tdm->dai_id] = tdm;

	sprintf(own_tdm_name, "css_own_tdm%d", tdm->dai_id);
	procfs_entry = proc_create_data(own_tdm_name,
		S_IFREG | S_IRUGO, gs_procfs_root,
		&dvf_css_own_tdm_proc_fops,
		&tdm->dai_id);

	if (!procfs_entry) {
		printk("dspg_tdm_dev_probe: css_own_tdm entry creation failure\n");
		ret = ENOENT;
	}

	return ret;

}

static int
dspg_tdm_dev_remove(struct platform_device *pdev)
{
	struct dvf_tdm_priv *tdm = dev_get_drvdata(&pdev->dev);
	char own_tdm_name[20];

	sprintf(own_tdm_name, "css_own_tdm%d", tdm->dai_id);
	remove_proc_entry(own_tdm_name, gs_procfs_root);
	
	free_irq(tdm->iores_irq, tdm);
	iounmap(tdm->reg_base);

	release_mem_region(tdm->iores_mem, TDM_REGLEN);

	if (!IS_ERR(tdm->rc))
		reset_control_assert(tdm->rc);

	if (!IS_ERR(tdm->clk))
		clk_disable_unprepare(tdm->clk);

	return 0;
}

static int
dspg_tdm_dev_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct dvf_tdm_priv *tdm = dev_get_drvdata(&pdev->dev);

	dbg_prt();

	if (tdm) {
		if (!tdm->css_owns_tdm) {
			/* Disable dai's TDM machine clock,
			 * If (!) Is active upon 'suspend'
			 */
			if (tdm->tdm_clk_active) {
				vdbg_prt("%s %u\n",
					 "Disabling TDM clock of TDM id",
					 tdm->dai_id);
				tdm->tdm_clk_suspended = 1;
				tdm_clk_disable(tdm);
			} else {
				tdm->tdm_clk_suspended = 0;
			}
		} else {
			vdbg_prt("%s %u %s\n", "TDM id", tdm->dai_id,
				 "owned by CSS. Ignoring!");
		}

	} else {
		vdbg_prt("%s\n", "No driver_data. Ignoring!");
	}

	return 0;
}

static int
dspg_tdm_dev_resume(struct platform_device * pdev)
{
	struct dvf_tdm_priv *tdm = dev_get_drvdata(&pdev->dev);

	dbg_prt();

	if (tdm) {
		if (!tdm->css_owns_tdm) {
			/* Re-Enable dai's TDM machine clock,
			 * If (!) Was active upon 'suspend'
			 */
			if (tdm->tdm_clk_suspended) {
				vdbg_prt("%s %u\n",
					 "Re-Enabling TDM clock of TDM id",
					 tdm->dai_id);
				tdm->tdm_clk_suspended = 0;
				tdm_clk_enable(tdm);
			}
		} else {
			vdbg_prt("%s %u %s\n",
				 "TDM id",
				 tdm->dai_id,
				 "owned by CSS. Ignoring!");
		}
	} else {
		vdbg_prt("%s\n", "No driver_data. Ignoring!");
	}

	return 0;
}

static const struct of_device_id dspg_tdm_dt_match[] = {
	{ .compatible = "dspg,tdm", },
	{}
};

static struct platform_driver dspg_tdm_driver = {
	.probe		= dspg_tdm_dev_probe,
	.remove		= dspg_tdm_dev_remove,
	.suspend	= dspg_tdm_dev_suspend,
	.resume		= dspg_tdm_dev_resume,
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = dspg_tdm_dt_match,
	},
};

static int __init
dspg_tdm_init(void)
{
	int ret;
	gs_procfs_root = proc_mkdir("gs", NULL);
	if (!gs_procfs_root) {
		printk("dspg_tdm_init: proc mkdir gs failed\n");
		return -1;
	}

	ret = platform_driver_register(&dspg_tdm_driver);
	return ret;
}
module_init(dspg_tdm_init);

static void __exit
dspg_tdm_exit(void)
{
	platform_driver_unregister(&dspg_tdm_driver);

	if (gs_procfs_root) {
		remove_proc_entry("gs", NULL);
		gs_procfs_root = NULL;
	}
}
module_exit(dspg_tdm_exit);

MODULE_DESCRIPTION("DVF Grandstream HT platform TDM driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:" DRV_NAME);
