/*
 * kernel/time/timer_gs.c
 *
 * List timers for grandstream
 *
 * Copyright(C) 2018, Grandstream Networks, Inc., Carl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/seq_file.h>
#include <linux/kallsyms.h>

#include <asm/uaccess.h>

#include "tick-internal.h"


#define MAX_GS_TIMER_SIZE (1024)
#define GS_TIMER_INVALID_ID (-1)
#define IS_VALID_TIMER(x) (((x) >= 0) && ((x) < MAX_GS_TIMER_SIZE))
#define GS_TIMER_ABS(x) ((x) >= 0 ? (x) : -(x))
#define GS_TIMER_JIFFIES
#if !defined(GS_TIMER_JIFFIES)
#error ("timer with gptc interrupt count has not been supported, please define GS_TIMER_JIFFIES")
#endif


struct _gs_timer {
	int id;
	bool enable;
	bool periodical;
	u64 base;
	u64 expire;
	void *p;
	void (*call_back)(int id, void *p);
};

static struct _gs_timer gstimer[MAX_GS_TIMER_SIZE];
static u64 gs_timer_int_count;

static int __init gs_timer_init(void)
{
	int id;

	gs_timer_int_count = 0;
	for ( id = 0; id < MAX_GS_TIMER_SIZE; ++id ) {
		gstimer[id].id = GS_TIMER_INVALID_ID;
		gstimer[id].enable = false;
		gstimer[id].periodical = false;
		gstimer[id].base = 0; 
		gstimer[id].expire = 0;
		gstimer[id].p = NULL;
		gstimer[id].call_back = NULL;
	}
	return 0;
}

int gs_timer_create(void (*call_back)(int id, void *p), 
	void* p)
{
	int id;

	for ( id = 0; id < MAX_GS_TIMER_SIZE; ++id ) {
		if (!IS_VALID_TIMER(gstimer[id].id)) {
			gstimer[id].id = id;
			gstimer[id].enable = false;
			gstimer[id].periodical = false;
			gstimer[id].base = 0; 
			gstimer[id].expire = 0;
			gstimer[id].p = p;
			gstimer[id].call_back = call_back;

			return gstimer[id].id;
		}
	}

	return GS_TIMER_INVALID_ID;
}
EXPORT_SYMBOL_GPL(gs_timer_create);

bool gs_timer_destory(int id)
{
	if (IS_VALID_TIMER(id)) {
		if (id == gstimer[id].id) {
			gstimer[id].id = GS_TIMER_INVALID_ID;
			gstimer[id].enable = false;
			return true;
		}
	}
	return false;
}
EXPORT_SYMBOL_GPL(gs_timer_destory);

bool gs_timer_set(int id, u64 expire, bool periodical, bool enable)
{
	if (IS_VALID_TIMER(id)) {
		if ((id == gstimer[id].id) 
			&& (enable != gstimer[id].enable)
		)
		{
			if (enable) {
				gstimer[id].periodical = periodical;
#if defined(GS_TIMER_JIFFIES)
				gstimer[id].base = get_jiffies_64() + (u64)msecs_to_jiffies(expire);
#else
				gstimer[id].base = gs_timer_int_count;
#endif 
				gstimer[id].expire = expire;
			} else {
			}
			gstimer[id].enable = enable;
		}
		return true;
	}

	return false;
}
EXPORT_SYMBOL_GPL(gs_timer_set);

/* This needs to be called in a periodical interrupt */
void gs_timer_running(unsigned long gptc_period_rate)
{
	int id;

	++gs_timer_int_count;
	for ( id = 0; id < MAX_GS_TIMER_SIZE; ++id ) {

		if (gstimer[id].enable) {

#if defined(GS_TIMER_JIFFIES)
			if (time_after_eq64(get_jiffies_64(), gstimer[id].base))
#else
			/* I dont know how to convert period_rate to msec */
			if ((GS_TIMER_ABS(gs_timer_int_count - gstimer[id].base) * gptc_period_rate) >= gstimer[id].expire)
#endif
			{
				if (gstimer[id].call_back) gstimer[id].call_back(gstimer[id].id, gstimer[id].p);
				if (gstimer[id].periodical) {
#if defined(GS_TIMER_JIFFIES)
					gstimer[id].base = get_jiffies_64();
#else
					gstimer[id].base = gs_timer_int_count;
#endif 
				} else {
					gstimer[id].enable = false;
				}
			}
		}
	}
}
EXPORT_SYMBOL_GPL(gs_timer_running);


__initcall(gs_timer_init);
