/*
 * csr8811.c -- CSR8811 BlueCore: Bluetooth ALSA SoC Audio Codec driver
 *
 * Copyright (C) 2016 DSP Group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <sound/soc.h>
#include <sound/pcm_params.h>

#define CSR8811_FORMATS (SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_S16_LE)

#define CSR8811_RATES	(SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_48000)

static int
csr8811_dai_set_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt)
{
	struct snd_soc_codec *codec = codec_dai->codec;

	/* Set Master/Slave */
	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		dev_dbg(codec->dev, "TDM Master");
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		dev_dbg(codec->dev, "TDM Slave");
		break;
	default:
		return -EINVAL;
	}

	/* Set Format */
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		dev_dbg(codec->dev, "TDM format is I2S");
		break;
	case SND_SOC_DAIFMT_DSP_B:
		dev_dbg(codec->dev, "TDM format is Right-Justified");
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		dev_dbg(codec->dev, "TDM format is Left-Justified");
		break;
	default:
		return -EINVAL;
	}

	/* Set TDM clocks Polarity */
	switch (fmt & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_NB_NF:
		dev_dbg(codec->dev, "normal BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_IF:
		dev_dbg(codec->dev, "inverted BCLK(sclk); inverted LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_NF:
		dev_dbg(codec->dev, "inverted BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_NB_IF:
		dev_dbg(codec->dev, "normal BCLK(sclk); inverted LRC(FSync)");
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int
csr8811_dai_hw_params(struct snd_pcm_substream *substream,
		      struct snd_pcm_hw_params *params,
		      struct snd_soc_dai *dai)
{
	struct snd_soc_codec *codec = dai->codec;

	/* Bit width */
	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S8:
		dev_dbg(codec->dev, "PCM sample-length 8 bits");
		break;
	case SNDRV_PCM_FORMAT_S16_LE:
		dev_dbg(codec->dev, "PCM sample-length 16 bits; LE");
		break;
	default:
		return -EINVAL;
	}

	/* Filter coefficient (per sample-rate) */
	switch (params_rate(params)) {
	case 8000:
		dev_dbg(codec->dev, "fs = 8KHz");
		break;
	case 48000:
		dev_dbg(codec->dev, "fs = 48KHz");
		break;
	case 11025:
	case 12000:
	case 16000:
	case 22050:
	case 24000:
	case 32000:
	case 44100:
	default:
		dev_dbg(codec->dev, "fs of %uHz is not supported",
			params_rate(params));
		return -EINVAL;
	}

	return 0;
}

static struct snd_soc_dai_ops csr8811_dai_ops = {
	.set_fmt	= csr8811_dai_set_fmt,
	.hw_params	= csr8811_dai_hw_params,
};

struct snd_soc_dai_driver csr8811_dai = {
	.name = "csr8811-hifi",
	.playback = {
		.stream_name	= "Playback",
		.channels_min	= 1,
		.channels_max	= 2,
		.rates		= CSR8811_RATES,
		.formats	= CSR8811_FORMATS,
	},
	.capture = {
		.stream_name	= "Capture",
		.channels_min	= 1,
		.channels_max	= 2,
		.rates		= CSR8811_RATES,
		.formats	= CSR8811_FORMATS,
	},
	.ops = &csr8811_dai_ops,
};

static struct snd_soc_codec_driver soc_codec_dev_csr8811;

static int
csr8811_platform_probe(struct platform_device *pdev)
{
	int ret;

	ret = snd_soc_register_codec(&pdev->dev, &soc_codec_dev_csr8811,
				     &csr8811_dai, 1);
	if (ret < 0 && ret != -EPROBE_DEFER) {
		dev_err(&pdev->dev, "failed to register codec: %d\n", ret);
		return ret;
	}

	if (ret == 0)
		dev_info(&pdev->dev, "successfully registered\n");

	return ret;
}

static int
csr8811_platform_remove(struct platform_device *pdev)
{
	snd_soc_unregister_codec(&pdev->dev);
	return 0;
}

static const struct of_device_id dvf_evb_dt_match[] = {
	{ .compatible = "dspg,csr8811", },
	{ }
};

static struct platform_driver csr8811_platform_driver = {
	.driver = {
		.name	= "csr8811",
		.owner	= THIS_MODULE,
		.of_match_table = dvf_evb_dt_match,
	},
	.probe  = csr8811_platform_probe,
	.remove = csr8811_platform_remove,
};

static int __init csr8811_module_init(void)
{
	return platform_driver_register(&csr8811_platform_driver);
}

static void __exit csr8811_module_exit(void)
{
	platform_driver_unregister(&csr8811_platform_driver);
}
module_init(csr8811_module_init);
module_exit(csr8811_module_exit);

MODULE_DESCRIPTION("ASoC CSR8811 codec driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL v2");
