/*
 * dcx81.c -- DCX81 CMBS: ALSA SoC Audio Codec driver
 *
 * Copyright (C) 2011 DSPG (DSP Group)
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Note:
 * This is a "stub" asoc codec-driver, actually having no interaction with the
 * DCX81 codec chip at all! It only allows registration of the codec within
 * the ALSA SoC framework.
 * The actual chip configuration is done by the BlueZ - Linux Bluetooth protocol stack -
 * at the user-space, via HCI commands, sent to the chip over UART.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/initval.h>
#include <sound/tlv.h>
#include <asm/div64.h>

#ifdef DEBUG
	#define dbg_prt(fmt, ...) \
		pr_debug("DCX81 asoc Cdec-drv: %s(): " fmt, __func__, ##__VA_ARGS__)

	#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
	#define dbg_prt(fmt, ...)
	#define DIR_SRT(dir)
#endif

#define DCX81_FORMATS (SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_S16_LE)

#define DCX81_RATES	(SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_48000)

static struct snd_soc_codec *g_dcx81_codec;

/* codec private data */
struct dcx81_priv {
	struct snd_soc_codec codec;
};

/* ASoC Controls */
static const struct snd_kcontrol_new dcx81_snd_controls[] = {
};

/* ASoC DAPM (Dynamic Audio Power Management) Widgets */
static const struct snd_soc_dapm_widget dcx81_dapm_widgets[] = {
};

/* ASoC audio-route Map */
static const struct snd_soc_dapm_route audio_route[] = {
};

/* DCX81 Codec-DAI Operations: */

static int
dcx81_dai_set_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt)
{	

	dbg_prt();

	/* Set Master/Slave */
	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		dbg_prt("%s\n", "DCX81 is TDM Master");
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		dbg_prt("%s\n", "DCX81 is TDM Slave");
		break;
	default:
		return -EINVAL;
	}

	/* Set Format */
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		dbg_prt("%s\n", "TDM format is I2S");
		break;
	case SND_SOC_DAIFMT_RIGHT_J:
		dbg_prt("%s\n", "TDM format is Right-Justified");
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		dbg_prt("%s\n", "TDM format is Left-Justified");
		break;
	default:
		return -EINVAL;
	}

	/* Set TDM clocks Polarity */
	switch (fmt & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_NB_NF:
		dbg_prt("%s\n", "TDM clocks polarity: normal BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_IF:
		dbg_prt("%s\n", "TDM clocks polarity: inverted BCLK(sclk); inverted LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_NF:
		dbg_prt("%s\n", "TDM clocks polarity: inverted BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_NB_IF:
		dbg_prt("%s\n", "TDM clocks polarity: normal BCLK(sclk); inverted LRC(FSync)");
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int
dcx81_dai_hw_params(struct snd_pcm_substream *substream,
		      struct snd_pcm_hw_params *params,
		      struct snd_soc_dai *dai)
{

	dbg_prt("%s\n", DIR_STR(substream->stream));
	
	/* Bit width */
	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S8:
		dbg_prt("%s\n", "PCM sample-length (vs. ALSA) = 8 bits");
		break;
	case SNDRV_PCM_FORMAT_S16_LE:
		dbg_prt("%s\n", "PCM sample-length (vs. ALSA) = 16 bits; LE");
		break;
	default:
		return -EINVAL;
	}

	/* Filter coefficient (per sample-rate) */
	switch (params_rate(params)) {
	case 8000:
		dbg_prt("%s\n", "fs = 8KHz");
		break;
	case 48000:
		dbg_prt("%s\n", "fs = 48KHz");
		break;

	case 11025:
	case 12000:
	case 16000:
	case 22050:
	case 24000:
	case 32000:
	case 44100:
	default:
		dbg_prt("%s %u %s\n",
			 "Unsupported fs:", (params_rate(params)) / 1000, "KHz");
		return -EINVAL;
	}

	return 0;
}

static struct snd_soc_dai_ops dcx81_dai_ops = {
	.set_fmt	= dcx81_dai_set_fmt,
	.hw_params	= dcx81_dai_hw_params,
};

/* DCX81 Codec DAI: */

struct snd_soc_dai_driver dcx81_dai = {
	.name = "dcx81-cmbs",
	.playback = {
		.stream_name	= "Playback",
		.channels_min	= 1,
		.channels_max	= 2,
		.rates		= DCX81_RATES,
		.formats	= DCX81_FORMATS,
	},
	.capture = {
		.stream_name	= "Capture",
		.channels_min	= 1,
		.channels_max	= 2,
		.rates		= DCX81_RATES,
		.formats	= DCX81_FORMATS,
	},
	.ops = &dcx81_dai_ops,
};

/* DCX81 Codec device: */

static int
dcx81_set_bias_level(struct snd_soc_codec *codec,
		       enum snd_soc_bias_level level)
{
	dbg_prt();

	switch (level) {
	case SND_SOC_BIAS_ON:
		dbg_prt("%s\n", "Bias ON");
		break;
	case SND_SOC_BIAS_PREPARE:
		dbg_prt("%s\n", "Bias PREPARE");
		break;
	case SND_SOC_BIAS_STANDBY:
		dbg_prt("%s\n", "Bias STANDBY");
		break;
	case SND_SOC_BIAS_OFF:
		dbg_prt("%s\n", "Bias OFF");
		break;
	default:
		dbg_prt("%s %u\n", "Unsupported Bias level:", level);
		return -EINVAL;
	}

	/* Change to new state */
	snd_soc_codec_get_dapm(codec)->bias_level = level;

	return 0;
}

static int
dcx81_dev_probe(struct snd_soc_codec *codec)
{
	struct snd_soc_dapm_context *dapm = snd_soc_codec_get_dapm(codec);
	int ret = 0;

	dbg_prt();

	codec->control_data = NULL;

	dcx81_set_bias_level(codec, SND_SOC_BIAS_STANDBY);

	/* Register controls */
	ret = snd_soc_add_codec_controls(codec, dcx81_snd_controls,
					 ARRAY_SIZE(dcx81_snd_controls));
	if (ret < 0) {
		dev_err(codec->dev, "failed to add Controls: %d\n", ret);
		return ret;
	}

	/* Register the DAPM Widgets */
	ret = snd_soc_dapm_new_controls(dapm, dcx81_dapm_widgets,
					ARRAY_SIZE(dcx81_dapm_widgets));
	if (ret < 0) {
		dev_err(codec->dev, "failed to add Widgets: %d\n", ret);
		return ret;
	}

	/* Set up the DCX81 audio map */
	ret = snd_soc_dapm_add_routes(dapm, audio_route,
				      ARRAY_SIZE(audio_route));
	if (ret < 0) {
		dev_err(codec->dev, "failed to add Routes: %d\n", ret);
		return ret;
	}

	return 0;
}

static int
dcx81_dev_remove(struct snd_soc_codec *codec)
{
	dbg_prt();

	dcx81_set_bias_level(codec, SND_SOC_BIAS_OFF);

	return 0;
}

static int
dcx81_dev_suspend(struct snd_soc_codec *codec)
{
	dbg_prt();

	dcx81_set_bias_level(codec, SND_SOC_BIAS_OFF);

	return 0;
}

static int
dcx81_dev_resume(struct snd_soc_codec *codec)
{
	dbg_prt();

	dcx81_set_bias_level(codec, SND_SOC_BIAS_STANDBY);

	return 0;
}

static struct snd_soc_codec_driver soc_codec_dev_dcx81 = {
	.probe   = dcx81_dev_probe,
	.remove  = dcx81_dev_remove,
	.suspend = dcx81_dev_suspend,
	.resume  = dcx81_dev_resume,
	.set_bias_level = dcx81_set_bias_level,

	/* No codec registers ! */
	.reg_cache_size	= 0,
	.reg_word_size = 0,
	.reg_cache_default = NULL,
};

/* DCX81 Codec Driver:
 * --------------------
 */
static int
dcx81_register(struct dcx81_priv *dcx81)
{
	int ret;
	struct snd_soc_codec *codec = &(dcx81->codec);

	dbg_prt();

	if (g_dcx81_codec != NULL) {
		dev_err(codec->dev, "Another DCX81 is registered. Aborting!\n");
		return -EINVAL;
	}

	codec->component.name = "dcx81";
	snd_soc_codec_set_drvdata(codec, dcx81);

	/* Register the codec */
	ret = snd_soc_register_codec(codec->dev, &soc_codec_dev_dcx81,
				     &dcx81_dai, 1);
	if (ret != 0) {
		dev_err(codec->dev, "Failed to register Codec and its Dai: %d\n", ret);
		return ret;
	}

	/* Keep global pointer to codec structure (!) */
	g_dcx81_codec = codec;

	return 0;
}

static void
dcx81_unregister(struct dcx81_priv *dcx81)
{
	struct snd_soc_codec *codec = &(dcx81->codec);;

	dbg_prt();

	dcx81_set_bias_level(codec, SND_SOC_BIAS_OFF);

	/* Deregister the codec and its dai */
	snd_soc_unregister_codec(codec->dev);

	g_dcx81_codec = NULL;
}

static int
dcx81_platform_probe(struct platform_device *pdev)
{
	struct dcx81_priv *dcx81 = NULL;
	struct snd_soc_codec *codec = NULL;

	dbg_prt();

	/* Allocate private data */
	dcx81 = kzalloc(sizeof(struct dcx81_priv), GFP_KERNEL);
	if (dcx81 == NULL) {
		return -ENOMEM;
	}

	/* Save pointer to private data on the device */
	pdev->dev.platform_data = dcx81;

	/* Save pointer to device on the priv data */
	codec = &(dcx81->codec);
	codec->dev = &(pdev->dev);

	/* Register the codec and its dai */
	return dcx81_register(dcx81);
}

static int
dcx81_platform_remove(struct platform_device *pdev)
{
	struct dcx81_priv *dcx81 = pdev->dev.platform_data;

	dbg_prt();

	/* Deregister the codec and its dai */
	dcx81_unregister(dcx81);

	kfree(dcx81);

	return 0;
}

static const struct of_device_id dvf_evb_dt_match[] = {
	{ .compatible = "dvf,cmbs-dcx81-codec", },
	{ }
};

static struct platform_driver dcx81_platform_driver = {
	.driver = {
		.name	= "cmbs-dcx81-codec",
		.owner	= THIS_MODULE,
		.of_match_table = dvf_evb_dt_match,
	},
	.probe  = dcx81_platform_probe,
	.remove = dcx81_platform_remove,
};

static int __init
dcx81_mod_init(void)
{
	return platform_driver_register(&dcx81_platform_driver);
}
module_init(dcx81_mod_init);

static void __exit
dcx81_mod_exit(void)
{
	platform_driver_unregister(&dcx81_platform_driver);
}
module_exit(dcx81_mod_exit);

MODULE_DESCRIPTION("ASoC DCX81 codec driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");

