/*
 * dvf-evb-dcx81.c -- ASoC Machine Driver of DVF99 EVB board for DCX81 connectivity,
 *
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <mach/platform.h>
#include <mach/gpio.h>
#include <asm/atomic.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <sound/dvf-tdm.h>
#include <sound/snd-dvf.h>
#include <linux/of.h>

#define DVF_DCX81_NUM_TDM_SLOTS 16
#define DCX81_BITS_PER_SLOT     16
#define DVF_DCX81_FSYNC_RATE    8000
#define DVF_DCX81_NUM_CHANNEL   1
#define CODEC_NAME "cmbs-dcx81-codec"

#define DRIVER_NAME "snd-dvf-evb-cmbs"

#ifdef V_DEBUG
	#ifndef DEBUG
		#define DEBUG
	#endif
	#define vdbg_prt(fmt, ...) \
		pr_debug("DVF99-CMBS asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)
#else
	#define vdbg_prt(fmt, ...)
#endif


#ifdef DEBUG
	#define dbg_prt(fmt, ...) \
		pr_debug("DVF99-CMBS asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)

	#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
	#define dbg_prt(fmt, ...)
	#define DIR_STR(stream_dir)
#endif

#define DMW_MAX_NUM_DAIs 1

/* asoc Cpu-Dai (TDM) device name */
const char* dcx81_cpu_dai_name[] = {"dvf-tdm.0", "dvf-tdm.1"};
static int g_dvf_dcx81_configure = 0;
static int g_dvf_dcx81_grant     = 0;
static const char *dcx81_configure_function[] = {"Off", "On"};
static const char *dcx81_grant_function[]     = {"Revoke", "Grant"};
static const struct soc_enum dcx81_configure_enum = SOC_ENUM_SINGLE_EXT(2, dcx81_configure_function);
static const struct soc_enum dcx81_grant_enum     = SOC_ENUM_SINGLE_EXT(2, dcx81_grant_function);

static struct snd_soc_dai_link dcx81_dai_link[DMW_MAX_NUM_DAIs];

static int brd_dcx81_dai_init(struct snd_soc_pcm_runtime *rtd);

/* DCX81 section: */
/* ---------------- */

void
dvf_tdm_set_telephony(struct snd_soc_dai *cpu_dai, int tel_rqst);
void
dvf_tdm_set_slots(struct snd_soc_dai *cpu_dai, int slots);

static struct snd_soc_dai_link brd_dcx81_dai_link = {
	.name = "dvf99-dcx81_Dai_Link",
	.stream_name = "dvf99-dcx81_PCM",

	/* asoc Codec-Dai device name */
	.codec_dai_name = "dcx81-cmbs",

	.init = brd_dcx81_dai_init,
};

static int
dcx81_get_configure(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = g_dvf_dcx81_configure;

	vdbg_prt("%s %s\n", "TDM Configure is:",
		(g_dvf_dcx81_configure == 0) ? "Off" : "On");

	return 0;
}

static int
dcx81_set_configure(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_card *card = snd_kcontrol_chip(kcontrol);
	struct snd_soc_pcm_runtime *rtd = list_entry(&card->rtd_list,
			struct snd_soc_pcm_runtime, list);
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;
	unsigned int dai_format;
	int ret;

	if (g_dvf_dcx81_configure == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
		  "TDM Configure is Unchanged:",
		  (g_dvf_dcx81_configure == 0) ? "Off" : "On");
		return 0;
	}

	if(0 == ucontrol->value.integer.value[0]){
		g_dvf_dcx81_configure = 0;
		if(g_dvf_dcx81_grant == 1){
			g_dvf_dcx81_grant = 0;
			dvf_tdm_configure_reg(cpu_dai,codec_dai,DCX81_BITS_PER_SLOT,
					DVF_DCX81_FSYNC_RATE,DVF_DCX81_NUM_CHANNEL,1);
		}
		vdbg_prt("%s\n","Reset TDM Configure");
		return 0;
	}
	/* Set DAI's TDM configuration: */
	dai_format = 0;
	dai_format = SND_SOC_DAIFMT_DSP_B;
	dai_format |= SND_SOC_DAIFMT_IB_NF; /* Normal SCLK(BCLK) & FSync(LRC) polarity */
	dai_format |= SND_SOC_DAIFMT_CBS_CFS; /* Codec(!) point-of-view - si3050 is Slave */

	vdbg_prt("%s\n", "DVF99 TDM is Master, SI32178 is Slave");

	/* Set cpu DAI configuration */
	vdbg_prt("%s", "Calling CPU DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0)
		return ret;

	/* Set the TDM DAI's slots using proprietary function */
        dvf_tdm_set_slots(cpu_dai, DVF_DCX81_NUM_TDM_SLOTS);

	ret = dvf_tdm_configure_reg(cpu_dai,codec_dai,DCX81_BITS_PER_SLOT,
					DVF_DCX81_FSYNC_RATE,DVF_DCX81_NUM_CHANNEL,0);
	if (ret < 0)
		return ret;

	g_dvf_dcx81_configure = ucontrol->value.integer.value[0];

	vdbg_prt("%s %s\n", "Updating TDM Configure to:",
	  (g_dvf_dcx81_configure == 0) ? "Off" : "On");

	return 1;
}

static int
dcx81_get_grant(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = g_dvf_dcx81_grant;

	vdbg_prt("%s %s\n", "TDM Grant is:",
		(g_dvf_dcx81_grant == 0) ? "Revoke" : "Grant");

	return 0;
}

static int
dcx81_set_grant(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_card *card = snd_kcontrol_chip(kcontrol);
	struct snd_soc_pcm_runtime *rtd =
		list_entry(&card->rtd_list, struct snd_soc_pcm_runtime, list);
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;

	if (g_dvf_dcx81_grant == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
			"TDM Grant value is  Unchanged:",
			(g_dvf_dcx81_grant == 0) ? "Revoke" : "Grant");
		return 0;
	}
	g_dvf_dcx81_grant = ucontrol->value.integer.value[0];

	dvf_tdm_grant(cpu_dai,g_dvf_dcx81_grant);

	vdbg_prt("%s %s\n", "Updating TDM Grant :",
		(g_dvf_dcx81_grant == 0) ? "Revoke" : "Grant");

	return 1;
}

static int
dcx81_dai_link_init(struct snd_soc_card *card)
{
	int nCnt;
	int num_dai_links = 0;
	struct snd_soc_dai_link *dai_link;
	struct device_node *codec_node;
	const int* prop = NULL;
	unsigned int dai_id = 1; //default TDM device;

	dbg_prt();

	codec_node = of_find_node_by_name(0, CODEC_NAME);
	if (!codec_node) {
		vdbg_prt(" Codec Node not found \n");
		return -1;
	}

        prop = of_get_property(codec_node, "dai_id", NULL);

	if(prop){
		dai_id = be32_to_cpup(prop);
		if(dai_id >1){
			vdbg_prt(" dai_id is not configured correctly \n");
			return -1;
		}
	}else{
		vdbg_prt(" dai_id is not configuration is missing \n");
		return -1;
	}

	brd_dcx81_dai_link.cpu_dai_name =  dcx81_cpu_dai_name[dai_id];

	memcpy( &dcx81_dai_link[num_dai_links++],
		&brd_dcx81_dai_link,
		sizeof(struct snd_soc_dai_link) );

	for (nCnt = 0; nCnt < num_dai_links; nCnt++) {
		dai_link = &card->dai_link[nCnt];
		dai_link->codec_of_node = codec_node;
	}

	return num_dai_links;
}

static const struct snd_kcontrol_new dcx81_controls[] = {
	SOC_ENUM_EXT("TDM_Configure", dcx81_configure_enum, dcx81_get_configure, dcx81_set_configure),
	SOC_ENUM_EXT("TDM_Grant",     dcx81_grant_enum,     dcx81_get_grant,     dcx81_set_grant),
};

static struct snd_soc_card g_dspg_dvf99_evb_cmbs_card;

static int
brd_dcx81_dai_init(struct snd_soc_pcm_runtime *rtd)
{
	int ret;
	struct snd_soc_codec *codec = rtd->codec;
	struct snd_soc_dapm_context *dapm = snd_soc_codec_get_dapm(codec);
	struct snd_soc_card *card = rtd->card;

	dbg_prt();

	ret = snd_soc_add_card_controls(card, dcx81_controls,
					 ARRAY_SIZE(dcx81_controls));
	if (ret) {
		vdbg_prt("%s\n", "Telephony control adding Failed");
		return ret;
	}

	g_dvf_dcx81_configure = 0;
	g_dvf_dcx81_grant = 0;

	vdbg_prt("%s\n", "Syncing dapm");
	ret = snd_soc_dapm_sync(dapm);
	if (ret) {
		vdbg_prt("%s\n", "Dapm syncing Failed");
		return ret;
	}

	return 0;
}

static struct snd_soc_card g_dspg_dvf99_evb_cmbs_card = {
	.name		= "dspg-dvf99-evb-cmbs-card",
	.dai_link	= dcx81_dai_link,  /* DAI-link Array(!) */
	.num_links	= 0, /* overridden by dcx81_snd_probe() */

	/* Note: new to ASoC:
	 * Card's (not Codec's !) set_bias_level().
	 * May be used for certain activities, e.g. diasbling
	 * card's power amplifiers or enabling them just upon BIAS_ON;
	 * this is if they are Not declared as card's DAPM Widgests,
	 * as part of the audio Routes.
	 */
	.set_bias_level		= NULL,
	.set_bias_level_post	= NULL,
};

static int
dcx81_snd_probe(struct platform_device *pdev)
{
	int ret;
	struct snd_soc_card *card = &g_dspg_dvf99_evb_cmbs_card;

	dbg_prt();

	/* Initialize DAI-link array and num of DAIs that will be registered */
	card->num_links = dcx81_dai_link_init(card);

	/* Book keeping : */

	/* note: platform_set_drvdata() here saves pointer to the card's data
	 * on the device's 'struct device_private *p'->driver_data
	 */
	card->dev = &pdev->dev;

	ret = snd_soc_register_card(card);
	if (ret) {
		dev_err(&pdev->dev,"snd_soc_register_card Failed\n");
		goto ERR_CLEAR;
	}
	dev_info(&pdev->dev,"DVF99 CMBS ASoc Card registred\n");

	return 0;

ERR_CLEAR:

	return ret;
}

static int
dcx81_snd_remove(struct platform_device *pdev)
{
	struct snd_soc_card *card = platform_get_drvdata(pdev);

	dbg_prt();

	snd_soc_unregister_card(card);

	return 0;
}

static struct of_device_id dspg_dvf_cmbs_match[] = {
	{ .compatible = "dspg,dvf99-cmbs-mach-drv" },
	{ },
};

static struct platform_driver g_dcx81_dvf99_evb_snd_drv = {
	.driver = {
		.name = DRIVER_NAME,
		.owner = THIS_MODULE,
		.of_match_table = dspg_dvf_cmbs_match,
		.pm = &snd_soc_pm_ops,
	},
	.probe = dcx81_snd_probe,
	.remove = dcx81_snd_remove,
};

int
dvf_snd_cmbs_init(void)
{
	return platform_driver_register(&g_dcx81_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_cmbs_init);

void
dvf_snd_cmbs_exit(void)
{
	platform_driver_unregister(&g_dcx81_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_cmbs_exit);

static int __init
brd_dvf_evb_mod_init(void)
{
	return 0;
}

static void __exit
brd_dvf_evb_mod_exit(void)
{
}

module_init(brd_dvf_evb_mod_init);
module_exit(brd_dvf_evb_mod_exit);

MODULE_DESCRIPTION("ASoC machine driver for DSPG DVF99 EVB board");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
