/*
 * dspg-tdm.c -- DSPG ASoC CPU DAI driver
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 * 2nd Jan, 2012:
 * Adapted to kernel 3.4 by Sujan.S <sujan.somisetty@dspg.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/reset.h>
#include <mach/platform.h>
#include <mach/coma-tdm.h>
#include <mach/hardware.h>

#include <sound/dvf-tdm.h>

#ifdef V_DEBUG
#ifndef DEBUG
#define DEBUG
#endif
#define vdbg_prt(fmt, ...) \
		pr_debug("asoc Pltfrm TDM: %s(): " fmt, __func__, ##__VA_ARGS__)
#else
#define vdbg_prt(fmt, ...)
#endif

#ifdef DEBUG
#define dbg_prt(fmt, ...) \
		pr_debug("asoc Pltfrm TDM: %s(): " fmt, __func__, ##__VA_ARGS__)

#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
#define dbg_prt(fmt, ...)
#define DIR_SRT(dir)
#endif

/*****************************************************************************/

/*
 * The structure is per TDM and shared between the playback and capture
 * substreams. Fortunately there's no need for a dedicated mutex because the
 * access to the struct is implicitly guarded by the 'pcm_mutex' of the ASoC
 * core.
 */
struct dvf_tdm_priv {
	uint dai_id;
	void *reg_base;
	int iores_irq;
	struct clk *clk;
	uint tdm_clk_active;
	uint tdm_clk_suspended;
	spinlock_t tdm_clk_lock;

	int opencnt         : 2;
	int fs_master       : 1;
	int clk_master      : 1;
	int play_configured : 1;
	int capt_configured : 1;
	int play_active     : 1;
	int capt_active     : 1;

	int css_to_own_tdm  : 1;	/* Pending request to TDM-ownership switch-over:
					* '0' - to APP; '1' - to CSS
					*/
	int css_owns_tdm    : 1; /* Current state of TDM ownership:
				  * '0' - by APP; '1' - by CSS;
				  */

	int tdm_config_only  :1; /* This flag indicates to configure TDM only,
				    not to send grant to CSS */
	int tdm_sample_bits;	/* num SCLKs per pcm sample, over TDM bus */
	int alsa_sample_bits;	/* num bits per alsa pcm sample */
	int rate;		/* sample-rate (Fs), in Hz */
	int alsa_channels;	/* num channels (TDM time-slots) */
	uint dai_format;
	uint num_tdm_slots; /* TDM slots configurable via param */

	struct snd_soc_dai *linked_codec_dai;

	struct snd_pcm_substream *play_substream;
	struct snd_pcm_substream *capt_substream;
	struct snd_soc_dai_driver *dai;

	struct reset_control *rc;
};

/* Note: There are 2 TDM machines in the dvf99 chip, each one having its
 *       own TDM-clock. However, there is a Single (global) DP-clock, driving
 *       the codec chip(or codec chips !) mounted on the board.
 */

#define DEVICE_NAME "dvf-tdm"

#define DRV_NAME "dvf-tdm"

#define NUM_DVF101_TDMS		3

enum eSYS_CLK {
	/* Add for other system-clock freq */
	kSYS_CLK_DEFAULT,
	kSYS_CLK_MAX
};

enum eFS_HZ {
	kFS_HZ_8000,
	kFS_HZ_11025,
	kFS_HZ_16000,
	kFS_HZ_22050,
	kFS_HZ_32000,
	kFS_HZ_44100,
	kFS_HZ_48000,
	kFS_HZ_MAX
};

static const struct snd_soc_component_driver dvf_tdm_component = {
	.name = "dvf_tdm",
};

/*****************************************************************************/

static int g_tel_rqst[CONFIG_DVF_NR_TDMS];

/*****************************************************************************/

/* TDM registers: */
#define TDM_ENABLE                   0x0000
#define TDM_CONFIG_1                 0x0004
#define TDM_CONFIG_1_AUTO_SYNC       (1 << 13)
#define TDM_CONFIG_1_CLK_MSTR_SLV    (1 << 12)
#define TDM_CONFIG_1_FSYNC_MSTR_SLV  (1 << 11)
#define TDM_CONFIG_1_SYNC_EDGE       (1 << 10)
#define TDM_CONFIG_1_RX_EDGE         (1 <<  9)
#define TDM_CONFIG_1_TX_EDGE         (1 <<  8)
#define TDM_CONFIG_1_TIME_SLOTS      0x1F
#define TDM_CONFIG_2                 0x0008
#define TDM_CONFIG_2_SHIFT           (1 << 9)
#define TDM_CONFIG_2_TRUNCATE        (1 << 8)
#define TDM_CONFIG_2_BANK_SIZE       0x07
#define TDM_FSYNC                    0x0010
#define TDM_FSYNC_DUR                0x0014
#define TDM_FIFO_WM                  0x0018
#define TDM_INT_EN                   0x0020
#define TDM_INT_EN_RX_FULL           (1 << 3)
#define TDM_INT_EN_TX_EMPTY          (1 << 1)
#define TDM_INT_EN_TX_WM             (1 << 0)
#define TDM_INT_EN_RX_WM             (1 << 2)
#define TDM_INT_CLR                  0x0028
#define TDM_INT_CLR_TX_FLAGS         (1 << 3)
#define TDM_INT_CLR_FSYNC            (1 << 2)
#define TDM_INT_CLR_RX_OVRN          (1 << 1)
#define TDM_INT_CLR_TX_UNDRN         (1 << 0)
#define TDM_INT_STATUS               0x002C
#define TDM_INT_STATUS_RX_OVRN       (1 << 3)
#define TDM_INT_STATUS_TX_UNDRN      (1 << 1)
#define TDM_STATUS                   0x0030
#define TDM_TX_FIFO                  0x0038
#define TDM_RX_FIFO                  0x003C
#define TDM_TX_REG_0                 0x0A00
#define TDM_RX_REG_0                 0x1200
#define TDM_FDA_REG_00               0x1800
#define TDM_FDA_REG_01               0x1804
#define TDM_FDA_REG_02               0x1808
#define TDM_FDA_REG_03               0x180C

#define TDM_REGLEN 0x2000

/* The TDM FIFOs registers are 20 bits wide.
 * In the dvf99 TDM machine, the 20 MSbits are taken
 * when writing/reading the Tx/Rx fifo with a u32 value.
 * Note that this is opposite to the dw-52/74 TDM machine - 20 LSbits.
 */

/* The TDM Tx & Rx FIFOs have room for total of 128 samples -
 * e.g. for 64 TDM frames, where a frame consists of 2 time-slots
 * (Left+Right channels, interleaved).
 */
#define TDM_FIFO_DEPTH	64

/* When using DMA to write/read the TDM FIFOs,
 * we set the fifo water-mark level to trigger a dma
 * transaction. This is done by hardware. In this case,
 * the WM interrupts are not used, thus disabled.
 *
 * NOTE:
 * ----
 * TDM_FIFO_LVL_FOR_DMA Must(!) be >= N, where N is defined in
 * arch/arm/mach-dmw/devices.c, by the value of "transfer_length" fields
 * of the DMA channels used by the 3 TDM machines, as:
 *
 *	.transfer_length = (2 * N) - 1,	// N audio frames (stereo samples)
 *
 * "transfer_length" is the DMA Burst length, in u32 units.
 *
 * TDM_FIFO_LVL_FOR_DMA Should be equal to N.
 * It can be larger then N, but it CANNOT be Smaller !!!
 * If it is smaller, pcm data in the TDM TX Fifo might be over-written!
 */
#define	TDM_FIFO_LVL_FOR_DMA	(TDM_FIFO_DEPTH / 4)

#define TDM_TX_FIFO_LEVEL  TDM_FIFO_LVL_FOR_DMA  /* emptiness level, in frames */
#define TDM_RX_FIFO_LEVEL  TDM_FIFO_LVL_FOR_DMA  /* fulness   level, in frames */

#define TDM_SLOT(_index, _cd, _width) \
	(((_index) << 8) | (0x03 << 6) | (!!(_cd) << 5) | (_width & 0x1F))

/* CSS code works vs. the TDM FIFOs using IRQ.
 * It passes 4 audio frames (stereo-samples) per
 * Water-Mark interrupt.
 */
//#define TDM_FIFO_LVL_FOR_CSS (TDM_FIFO_DEPTH - 4)
#define TDM_FIFO_LVL_FOR_CSS (TDM_FIFO_DEPTH - 2)

/*****************************************************************************/

static inline unsigned int
round_up_power_2(unsigned int x)
{
	x--;
	x |= x >> 1;  // handle  2 bit numbers
	x |= x >> 2;  // handle  4 bit numbers
	x |= x >> 4;  // handle  8 bit numbers
	x |= x >> 8;  // handle 16 bit numbers
	x |= x >> 16; // handle 32 bit numbers
	x++;

	return x;
}

static inline unsigned int
log_base_2(unsigned int x)
{
	unsigned int ans = 0;

	while (x >>= 1)
		ans++;

	return ans;
}

static unsigned long
tdm_read(struct dvf_tdm_priv *tdm, int reg)
{
	return readl(tdm->reg_base + reg);
}

static void
tdm_write(struct dvf_tdm_priv *tdm, int reg, unsigned long value)
{
	writel(value, tdm->reg_base + reg);
}

/* Note:
 * Can't(!) use reference-counting on TDM clock enable/disable,
 * since ALSA APIs _hw_params() and _hw_free() are Not symetric:
 * _hw_free() may, and is, called 2 times or more for a single
 * _hw_params().
 */
static int
tdm_clk_enable(struct dvf_tdm_priv *tdm)
{
	unsigned long flags;
	int ret = 0;

	dbg_prt();

	spin_lock_irqsave(&tdm->tdm_clk_lock, flags);
	if (tdm->tdm_clk_active == 0) {
		tdm->tdm_clk_active = 1;
		clk_enable(tdm->clk);
		ret = 0;
	} else {
		ret = 1;
	}
	spin_unlock_irqrestore(&tdm->tdm_clk_lock, flags);

	vdbg_prt("%s %s\n", "TDM clock:", ((ret == 0) ? "Enabled now" : "Already Enabled"));

	return ret;
}

static int
tdm_clk_disable(struct dvf_tdm_priv *tdm)
{
	unsigned long flags;
	int ret = 0;

	dbg_prt();

	spin_lock_irqsave(&tdm->tdm_clk_lock, flags);
	if (tdm->tdm_clk_active != 0) {
		tdm->tdm_clk_active = 0;
		clk_disable(tdm->clk);
		ret = 0;
	} else {
		ret = 1;
	}
	spin_unlock_irqrestore(&tdm->tdm_clk_lock, flags);

	vdbg_prt("%s %s\n", "TDM clock:", ((ret == 0) ? "Disabled now" : "Already Disabled"));

	return ret;
}

static void
dvf_tdm_reset(struct dvf_tdm_priv *tdm)
{
	int ret;

	dbg_prt();

	/* Disable the TDM machine */
	tdm_write(tdm, TDM_ENABLE, 0);

	/* TDM clock must be Disabled when doing the TDM Reset,
	 * Must wait(!) for clock to Cease: the worst case is
	 * with the slowest clock.
	 */
	tdm_clk_disable(tdm);

	/* Delay >= 20 uSec, according to latest spec recommendation.
	 * A.M. April 3, 2012:
	 * However, according to slowest fs=8KHz --> sclck = 8*32=256KHz,
	 * you need to wait only 1/256KHz = ~4 uSec
	 */
	udelay(20);

	/* Reset TDM block - the reset suppose to be Asynchronous */
	ret = reset_control_assert(tdm->rc);
	if (ret)
		vdbg_prt("reset set failed");

	udelay(2);

	ret = reset_control_deassert(tdm->rc);
	if (ret)
		vdbg_prt("reset release failed");

	/* A.M. April 3, 2012:
	 * Note: Keep the clock to the TDM block Disabled !
	 */
}

static int
dvf_tdm_configure(struct dvf_tdm_priv *tdm)
{
	unsigned int slots;
	unsigned int alsa_bps, tdm_bps;
	unsigned int tdm_reg;
	unsigned int tx_fifo_wm_lvl, rx_fifo_wm_lvl;
	ulong master_sclk_rate = 0;
	unsigned int i;

	dbg_prt();

#ifdef CONFIG_DSPG_CSS_TDM
	if (tdm->css_owns_tdm) {
		/* Resume APP ownership over the TDM machine,
		 * Before(!) touching the TDM machine.
		 */
		if (coma_tdm_revoke(tdm->dai_id) == 0) {
			tdm->css_owns_tdm = 0;
			vdbg_prt("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoked OK");
		} else {
			vdbg_prt("%s %u %s\n",
				 "tdm", tdm->dai_id, "Revoke Failed");
		}
	}

#endif /* CONFIG_DSPG_CSS_TDM */
	/* Note:
	 * "bps" here means Bits-per-Sample.
	 *
	 * When the DVF99 TDM machine is the Master, i.e sourcing out the
	 * TDM SCLK & FSync clocks, there might be cases in which we would like
	 * to produce More SCLK clocks than the number of bits in the pcm sample.
	 * This is to achive better SCLK(BCLK) frequency accuracy.
	 * The SCLK frequency depends on Sample-Rate;
	 * its accuracy depends on the driving PLL frequency setting,
	 * the clock-tree further dividers resolution and the Sample-Rate Family
	 * (Fs = 8/12/11.025KHz multiplications).
	 * Example: to Tx a Fs=44.1KHz 16-bit pcm sample, where PLL freq=147.456MHz,
	 * we'll use 19 sckls (instaed of 16).
	 *
	 * Since the FIFO regs width is 20 bits, we are limitted to Tx/Rx
	 * 20 bits (20 sclks) in a tdm time-slot.
	 * Another HW restriction: minimal number of bits in a time-slot is 8 (8 sclks).
	 * (8 <= nun sclks per time-slot <= 20).
	 * However, the (ALSA) PCM sample may contain 16 (commonly), 20, 24, etc. bits.
	 *
	 * When there is a need to Tx/Rx more than 20 bits per time-slot,
	 * we can use Two (!) time-slots per pcm sample:
	 *  - the first  time-slot is of Audio/Voice type, having upto 20 bits.
	 *  - the second time-slot is of Control/Status type, having the rest of the bits,
	 *    but not less then 8.
	 *
	 * Currently, we Do Not support more than 20 bits per pcm sample.
	 */

	/* Note: the term "bps" means here "Bits per Sample" ! */

	alsa_bps = tdm->alsa_sample_bits;

	if ((tdm->clk_master) && (tdm->fs_master)) {

		/* dvf99 is TDM Master */
		/* ------------------- */

		/* vs. other codecs */
		tdm_bps = alsa_bps;
		master_sclk_rate = tdm->rate * tdm_bps * tdm->num_tdm_slots;

		/* Set Actual SCLK (BCLK) frequency over TDM bus,
		 * generated by the dvf99, as TDM Master !
		 */
		master_sclk_rate = clk_round_rate(tdm->clk, master_sclk_rate);
		clk_set_rate(tdm->clk, master_sclk_rate);
	} else {
		/* dvf99 is TDM Slave */
		/* ------------------ */

		tdm_bps = alsa_bps;
	}

	vdbg_prt("%s %u %s %lu\n",
		 "Final tdm_bps:", tdm_bps,
		 "Final sclk rate:", master_sclk_rate);

	/* A.M. April 3, 2012:
	* Problematic: looks like you should Not enable the Clock to the
	* TDM block before programming its registers. Though the TDM block itself
	* is disbled here (TDM_ENABLE = 0), if the clock is enabled before setting
	* the registers, the machine won't always sync correctly on FSync
	* as Slave or produce badeFSync timing as Master !
	* The FSync then deviates from the pcm frame real starting point,
	* causing the "metallic sound" phenomenon.
	* There is No hw explanation provided by the chip-design team.
	*
	* Workaround: do most of the TDM machine registers configuration,
	* including its reset-sequence, while the TDM clock is disabled.
	* This is CONTRARY to the chip-design directives since dw-52 !
	*/

	/*
	 * We should sw-reset the DVF99 TDM block perior to any TDM configuration
	 * change according to the data sheet.
	 * Note: this function now Leaves the TDM clock Disabled!
	 */
	dvf_tdm_reset(tdm);

	/* Re-enable the clock to the TDM block */
	tdm_clk_enable(tdm);

	/* Configure Frame Descriptor Array. Might insert some NULL samples if
	 * required.
	 */
	if (tdm->num_tdm_slots == 0) {
		vdbg_prt("Resetting num of slots to 2\n");
		tdm->num_tdm_slots = 2;
	}
	slots = tdm->num_tdm_slots;
	for (i = 0; i < tdm->num_tdm_slots; i++)
		tdm_write(tdm, TDM_FDA_REG_00 + i*4, TDM_SLOT(i, 0, tdm_bps));

	/* Note:
	 * - Here we set the TDM bus configuration:
	 *	DVF99 TDM Slave/Master;
	 *	I2S format Only(!) (for now).
	 *
	 * Remember the DVF99 TDM machine limitation:
	 *
	 *   - as Master:
	 *     ----------
	 *     it can produce only Rising FSync pulse signal,
	 *     (at rising or falling SCLK edge - ok), but Cannot(!) produce
	 *     falling FSync signal. Thus, for true I2S format, we need to Shift
	 *     the Rising of the FSync (which is LRC signal in I2S) in such way that the
	 *     Falling edge would take place one bit before the start of Frame,
	 *     which is start of the Left channel. This means:
	 *
	 *     - for 2 time-slots (up to 20 bits per sample), with N bits per time-slot :
	 *       Trigger the FSync on N-1 bit of the First time slot.
	 *       e.g. on bit 15 of the first time-slot, for 16-bits time-slots.
	 *
	 *   - as Slave:
	 *     ---------
	 *     it syncs only(!) on Rising edge of FSync.
	 *     So we need to Invert the FSync polarity, generated at the Wolfson.
	 *     However, we Cannot do this, since the Wolfson will put the ADC data
	 *     on the second tdm-slot (Right channel) - we can't live with that.
	 *     So we have Left/Right channel swap!
	 *
	 *
	 *     Slave or Master:
	 *     ----------------
	 *     Tx & Rx edges can, and shoud, be configured, for I2S, as follows:
	 *     - Tx data at Falling edge of SCLK
	 *     - Rx data at Rising edge of SCLK
	 */

	/* Configure TDM as Master/Slave SYNC/TX/RX Edges, 2 time-slots */
	tdm_reg = 0x0000;
	tdm_reg |= slots - 1;			/* Number of time-slots in frame (2 or 4) */
	tdm_reg |= TDM_CONFIG_1_TX_EDGE;	/* Tx data on SCLK Falling edge */
	tdm_reg &= ~TDM_CONFIG_1_RX_EDGE;	/* Rx data on SCLK Rising edge */

	/* FSync rise at SCLK Falling edge:
	 *   Matters only if Master; Irrelevant for Slave.
	 */
	switch (tdm->dai_format & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_IB_NF:
		tdm_reg &= ~(TDM_CONFIG_1_TX_EDGE);	/* Tx data on SCLK Rising edge */
		tdm_reg &= ~(TDM_CONFIG_1_SYNC_EDGE);	/* SCLK and FSYNC should be in phase */
		break;
	default:
		tdm_reg |= TDM_CONFIG_1_SYNC_EDGE;
		break;
	}

	if (unlikely(tdm->fs_master)) {
		/* DVF99 TDM is FSync Master */
		tdm_reg |=  TDM_CONFIG_1_FSYNC_MSTR_SLV;/* Source-out FSync clock */
	} else {
		/* DVF99 TDM is FSync Slave */
		tdm_reg &= ~TDM_CONFIG_1_FSYNC_MSTR_SLV;/* Sink-in FSync clock */
		tdm_reg |=  TDM_CONFIG_1_AUTO_SYNC;	/* Resync on FSync, after FSync loss/jitter */
	}
	if (unlikely(tdm->clk_master))
		tdm_reg |=  TDM_CONFIG_1_CLK_MSTR_SLV;	/* Source-out SCLK clock */
	else
		tdm_reg &= ~TDM_CONFIG_1_CLK_MSTR_SLV;	/* Sink-in SCLK clock */

	tdm_write(tdm, TDM_CONFIG_1, tdm_reg);

	/* Configure MSB fisrt, 64(32) banks (in a fifo line), 2(4) fifo lines */
	tdm_reg = 0x0000;
	tdm_reg |= TDM_CONFIG_2_SHIFT;    /* Shift Out/In MSB first */
	tdm_reg |= TDM_CONFIG_2_TRUNCATE; /* On Tx/Rx - Allign data bits if less then 20 */
	/* bank size computation */
	tdm_reg |= log_base_2(round_up_power_2(slots)) - 1;

	vdbg_prt("Number of slots = %d, Config-2 register set to 0x%x\n",
		 slots, tdm_reg);

	tdm_write(tdm, TDM_CONFIG_2, tdm_reg);

	/* Configure LRC Rise on last bit of Left channel -
	 * to produce LRC Fall at 1 bit before data frame starts (standard I2S)
	 */
	switch (tdm->dai_format & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_IB_NF:
		tdm_reg = 0x0000; /* For FXO DAI, FSYNC should rise at first slot and first bit */
		break;
	default:
		tdm_reg = 0x0000;
		if (likely(slots == 2)) {
			tdm_reg |= ((tdm_bps - 1) << 8);/* Last bit of */
			tdm_reg |= (0 & 0x001F);	/* First time slot */
		}
		break;
	}
	tdm_write(tdm, TDM_FSYNC, tdm_reg);

	/* Configure LRC length, in sclks */
	/*tdm_write(tdm, TDM_FSYNC_DUR, tdm_bps - 1);*/
	tdm_write(tdm, TDM_FSYNC_DUR, 2);

	/* Configure TX & RX FIFOs Water-Mark */
	tdm_reg = 0x0000;
	if (likely(!tdm->css_to_own_tdm && !tdm->tdm_config_only)) {
		tx_fifo_wm_lvl = TDM_TX_FIFO_LEVEL;
		rx_fifo_wm_lvl = TDM_RX_FIFO_LEVEL;
	} else {
		tx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS; //TDM_FIFO_DEPTH/slots -2;
		rx_fifo_wm_lvl = TDM_FIFO_LVL_FOR_CSS; //TDM_FIFO_DEPTH/slots -2;
	}

	/* Future: for more then 20 bits in per TDM channels,
	 * using two time-slots per L/R channel
	 */
	if (unlikely(slots == 4)) {
		tx_fifo_wm_lvl >>= 1; /* :2 */
		rx_fifo_wm_lvl >>= 1; /* :2 */
	}

	/* Latest documented changes in the dvf99 TDM machine:
	 * - RX_LVL field is bits 13-8.
	 * - TX_LVL and RX_LVL actually trigger DMA or WM-Interrupt
	 *   at field value + 1.
	 */
	tdm_reg |= (tx_fifo_wm_lvl & 0x003F) << 0;
	tdm_reg |= (rx_fifo_wm_lvl & 0x003F) << 8; /* bits 13:8 */
	tdm_write(tdm, TDM_FIFO_WM, tdm_reg);


	/* "Clear" TDM FIFOs (!):
	 * Note: this is needed in order to Deassert the TDM Water-Mark
	 * interrupt indications, at the TDM machine, in case they are
	 * asserted.
	 * Otherwise, an interrupt will be asserted immediately after
	 * unmasking these interrupts, Even Though the machine is
	 * Disabled (TDM_ENABLE has 0)!!!
	 * This is because the TDM WM interrupts are of "Level" type
	 * (Not "Edge-trigger").
	 */
	/* Fill-in Tx fifo */
	i = tx_fifo_wm_lvl + 1;
	i *= slots / 2;
	while (i--) {
		tdm_write(tdm, TDM_TX_FIFO, 0x0000); /* Left */
		tdm_write(tdm, TDM_TX_FIFO, 0x0000); /* Right */
	}
	/* Empty Rx fifo */
	i = rx_fifo_wm_lvl + 1;
	i *= slots / 2;
	while (i--) {
		tdm_read(tdm, TDM_RX_FIFO); /* Left */
		tdm_read(tdm, TDM_RX_FIFO); /* Right */
	}

	/* TODO - add handling for grant timeout (CSS is in panic) */

#ifdef CONFIG_DSPG_CSS_TDM
	if (unlikely(tdm->css_to_own_tdm)) {
		if (!tdm->css_owns_tdm) {
			/* TDM ownership should be given to CSS.
			 *
			 * The TDM machine will now be handled by the CSS'es TDM driver.
			 * The APP now finishes the TDM machine configuration to meet
			 * the CSS needs and Ceases to handle it !
			 */

			/* Disable TDM IRQ (at the APP-ICU!) */
			disable_irq(tdm->iores_irq);

			/* Enable Tx/Rx Water-Mark and Underrun/Overrun interrupts
			 * (to be handled by the CSS!)
			 */
			tdm_write(tdm, TDM_INT_EN,
				  TDM_INT_EN_TX_EMPTY | TDM_INT_EN_RX_FULL | /* For normal operation */
				  TDM_INT_EN_TX_WM | TDM_INT_EN_RX_WM);	   /* To catch errors */

			/* Handover TDM ownership to CSS: */
			if (coma_tdm_grant(tdm->dai_id, tdm->rate,
					   slots, tdm_bps) == 0) {
				tdm->css_owns_tdm = 1;
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Granted to CSS");
			} else {
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Grant Failed");
			}
		}
	} else {
		/* TDM ownership should stay owned by the APP */
		tdm->css_owns_tdm = 0;
		vdbg_prt("%s %u\n", "APP owns tdm", tdm->dai_id);
	}
#endif /* CONFIG_DSPG_CSS_TDM */

	return 0;
}

/* New functions are added to configure the TDM Register separately
 * and sending grant to CSS, these two can be handled by control
 * command sends from application via codec driver */
int
dvf_tdm_configure_reg(struct snd_soc_dai *cpu_dai,
		      struct snd_soc_dai *codec_dai, int alsa_sample_bits,
		      int rate, int channels, int reset_tdm)
{

	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt();

	/* The explicit setting of codec dai with TDM Platform driver */
	tdm->linked_codec_dai = codec_dai;

	if (!reset_tdm) {
		tdm->alsa_sample_bits = alsa_sample_bits;
		tdm->rate = rate;
		tdm->alsa_channels = channels;
		/* Initialized to 2 by default for I2S */
		if (tdm->num_tdm_slots == 0)
			tdm->num_tdm_slots = 2;
		/* The variable tdm_config_only is indicate only
		 * configuration, but grant to CSS will follow */
		tdm->tdm_config_only = 1;
		vdbg_prt("%s\n\t\t%s %u  %s %u  %s %u %s %u\n",
			 "ALSA PCM params:",
			 "bits-per-sample:", tdm->alsa_sample_bits,
			 "sample-rate:", tdm->rate,
			 "num channels:", tdm->alsa_channels,
			 "num-slots", tdm->num_tdm_slots);
		dvf_tdm_configure(tdm);
	} else {
		if (tdm->css_owns_tdm) {
			/* Resume APP ownership over the TDM machine */
			if (coma_tdm_revoke(tdm->dai_id) == 0) {
				tdm->css_owns_tdm = 0;
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Revoke OK");
			} else {
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Revoke failed");
			}

			/* Entirely Reset the TDM machine:
			 * the TDM machine gets Disabled, its interrupts cleared,
			 * and all of its registers get their default values.
			 */
			dvf_tdm_reset(tdm);

			/* Enable TDM interrupt (at the APP ICU !) */
			enable_irq(tdm->iores_irq);
		}
	}

	return 0;
}

/* The function dvf_tdm_grant will send grant/revoke to CSS.
 * This is used to separate out TDM reg configuration and grant sending.
 * Now in run time application can call grant provided TDM register is set.
 */

void
dvf_tdm_grant(struct snd_soc_dai *cpu_dai, int grant_val)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt();

	if (grant_val) {
		tdm_clk_enable(tdm);
		/* Disable TDM IRQ (at the APP-ICU!) */
		disable_irq(tdm->iores_irq);

		/* Enable Tx/Rx Water-Mark and Underrun/Overrun interrupts
		 * (to be handled by the CSS!)
		 */
		tdm_write(tdm, TDM_INT_EN,
			  TDM_INT_EN_TX_EMPTY | TDM_INT_EN_RX_FULL | /* For normal operation */
			  TDM_INT_EN_TX_WM | TDM_INT_EN_RX_WM);      /* To catch errors */

		if (coma_tdm_grant(tdm->dai_id, tdm->rate, tdm->num_tdm_slots,
				   tdm->alsa_sample_bits) == 0) {
			tdm->css_owns_tdm = 1;
			vdbg_prt("tdm %d granted to CSS\n", tdm->dai_id);
		}
	} else {
		if (coma_tdm_revoke(tdm->dai_id) == 0) {
			tdm->css_owns_tdm = 0;
			vdbg_prt("tdm %d revoked to CSS\n", tdm->dai_id);
		}
	}
}

static void
tdm_apply_constraints(struct dvf_tdm_priv *tdm,
		      struct snd_pcm_substream *substream)
{
	dbg_prt("%s\n", DIR_STR(substream->stream));

	/* Set the constraints according to the already configured stream */
	snd_pcm_hw_constraint_minmax(substream->runtime,
				     SNDRV_PCM_HW_PARAM_RATE,
				     tdm->rate,
				     tdm->rate);

	snd_pcm_hw_constraint_minmax(substream->runtime,
				     SNDRV_PCM_HW_PARAM_SAMPLE_BITS,
				     tdm->alsa_sample_bits,
				     tdm->alsa_sample_bits);
}

static int
tdm_check_constraints(struct dvf_tdm_priv *tdm,
		      struct snd_pcm_hw_params *params)
{
	dbg_prt();

	/* Check the constraints according to the already configured stream */
	if ((params_rate(params) != tdm->rate) ||
	    (snd_pcm_format_width(params_format(params)) != tdm->alsa_sample_bits)) {
		vdbg_prt("%s\n", "Constraints matching Failed!");
		return -EINVAL;
	}

	return 0;
}

static int
dvf_tdm_startup(struct snd_pcm_substream *substream,
		struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt("%s %s\n", cpu_dai->name, DIR_STR(substream->stream));

	if (tdm->opencnt++ == 0) {
		vdbg_prt("%s %u\n",
			 "Enabling TDM clock of tdm id", cpu_dai->id);
		tdm_clk_enable(tdm);
	}

	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		vdbg_prt("%s\n", "Playback stream startup");
		tdm->play_substream = substream;
		if (tdm->capt_configured) {
			tdm_apply_constraints(tdm, substream);
			vdbg_prt("%s\n", "While Capture already configured");
		}
		break;

	case SNDRV_PCM_STREAM_CAPTURE:
		vdbg_prt("%s\n", "Capture stream startup");
		tdm->capt_substream = substream;
		if (tdm->play_configured) {
			tdm_apply_constraints(tdm, substream);
			vdbg_prt("%s\n", "While Playback already configured");
		}
		break;
	}

	return 0;
}

static void
dvf_tdm_shutdown(struct snd_pcm_substream *substream,
		 struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt("%s %s\n", cpu_dai->name, DIR_STR(substream->stream));

	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		vdbg_prt("%s\n", "Playback shutdown");
		tdm->play_substream = NULL;
		break;

	case SNDRV_PCM_STREAM_CAPTURE:
		vdbg_prt("%s\n", "Capture shutdown");
		tdm->capt_substream = NULL;
		break;
	}

	if (--tdm->opencnt == 0) {
		vdbg_prt("%s %u\n",
			 "Disabling TDM clock of tdm id", cpu_dai->id);
		tdm_clk_disable(tdm);
	}
}

static int
dvf_tdm_prepare(struct snd_pcm_substream *substream,
		struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;

	dbg_prt("%s\n", DIR_STR(substream->stream));

	tdm->css_to_own_tdm = g_tel_rqst[tdm->dai_id];

	/* Keep the codec-dai spouse.
	 * Note that the same cpu-dai (tdm machine) can be "linked"
	 * to different codecs over different dai-links. So this changes
	 * at run-time, according to the dai-link used!
	 */
	tdm->linked_codec_dai = codec_dai;

	/*
	 * Configure TDM machine (including its Reset) just before the first
	 * direction is started.
	 * This is especially needed after an underrun/overrun has occured
	 * And
	 * when Dect Telephony is involved (ownership of the TDM machine by the CSS),
	 * since the state of the relevant asoc control, in asound.conf, is actually
	 * set After dvf_tdm_hw_params() is called and Before dvf_tdm_prepare() gets
	 * called.
	 */
	if ((!tdm->play_active) &&
	     (!tdm->capt_active) &&
	     (!tdm->css_owns_tdm)) {
		return dvf_tdm_configure(tdm);
	} else {
		vdbg_prt("%s\n\t\t%s %u %s %u %s %u\n",
			 "dvf_tdm_configure()Skipped !!!",
			 "play_active:",  (uint)(tdm->play_active),
			 "capt_active:",  (uint)(tdm->capt_active),
			 "css_owns_tdm:", (uint)(tdm->css_owns_tdm));
	}

	return 0;
}

static int
dvf_tdm_trigger(struct snd_pcm_substream *substream, int cmd,
		struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt("%s\n", DIR_STR(substream->stream));

	/* Note:
	 * Under DECT phone-call, where the TDM ownership has been granted to CSS,
	 * No triggers are expected. Will be handled anyway, just in case they come.
	 */

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
	case SNDRV_PCM_TRIGGER_RESUME:
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:

		if (tdm->css_owns_tdm) {
			vdbg_prt("%s\n", "START trigger: TDM Not owned by APP. Ignoring!");
			return -EPERM;
		}

		if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
			vdbg_prt("%s\n", "Playback Start/Resume/unPuase");
			tdm->play_active = 1;
			if (tdm->capt_active) {
				break;
			}
		} else {
			vdbg_prt("%s\n", "Capture Start/Resume/unPuase");
			tdm->capt_active = 1;
			if (tdm->play_active) {
				break;
			}
		}

		/* Clear interrupt status (before enabling interrupts) */
		tdm_write(tdm, TDM_INT_CLR, 0xFFFF);

		/* Enable Tx & Rx Underrun/Overrun error interrupts */
		tdm_write(tdm, TDM_INT_EN,
			  TDM_INT_EN_TX_EMPTY | TDM_INT_EN_RX_FULL);

		/* Enable TDM machine */
		tdm_write(tdm, TDM_ENABLE, 1);
		break;

	case SNDRV_PCM_TRIGGER_STOP:
	case SNDRV_PCM_TRIGGER_SUSPEND:
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
			vdbg_prt("%s\n", "Playback Stop/Suspend/Puase");
			tdm->play_active = 0;
			if (tdm->capt_active)
				break;
		} else {
			vdbg_prt("%s\n", "Capture Stop/Suspend/Puase");
			tdm->capt_active = 0;
			if (tdm->play_active)
				break;
		}

		if (!tdm->css_owns_tdm) {
			/* Disable all TDM interrupts */
			tdm_write(tdm, TDM_INT_EN, 0);

			/* Disable TDM machine */
			tdm_write(tdm, TDM_ENABLE, 0);
		} else {
			vdbg_prt("%s\n", "STOP trigger: TDM Not owned by APP. Not disabling TDM!");
		}
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

/* Note: this fuction may be called more then once for the same session ! */
static int
dvf_tdm_hw_params(struct snd_pcm_substream *substream,
		  struct snd_pcm_hw_params *params,
		  struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);
	int alsa_bps;
	int ret;

	dbg_prt("%s\n", DIR_STR(substream->stream));

	/*
	 * The other direction might already be configured. All we can do in
	 * this case is to verify that the settings match as the TDM allows no
	 * dynamic reconfiguration.
	 */
	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		if (tdm->capt_configured) {
			ret = tdm_check_constraints(tdm, params);
			if (!ret) {
				tdm->play_configured = 1;
				vdbg_prt("%s\n", "Playback stream implicitly configured");
			} else {
				vdbg_prt("%s\n",
					 "Playback stream doesn't match Capture stream. Denied!");
			}
			return ret;
		}
		break;

	case SNDRV_PCM_STREAM_CAPTURE:
		if (tdm->play_configured) {
			ret = tdm_check_constraints(tdm, params);
			if (!ret) {
				tdm->capt_configured = 1;
				vdbg_prt("%s\n", "Capture stream implicitly configured");
			} else {
				vdbg_prt("%s\n",
					 "Capture stream doesn't match Playback stream. Denied!");
			}
			return ret;
		}
		break;
	}

	/* Save PCM parameters and Configure accordingly: */
	tdm->alsa_sample_bits = snd_pcm_format_width(params_format(params));
	tdm->rate = params_rate(params);
	tdm->alsa_channels = params_channels(params);
	if (tdm->num_tdm_slots == 0)
		tdm->num_tdm_slots = 2; // Initialized to 2 by default for I2S

	vdbg_prt("%s\n\t\t%s %u  %s %u  %s %u %s %u\n",
		 "ALSA PCM params:",
		 "bits-per-sample:", tdm->alsa_sample_bits,
		 "sample-rate:", tdm->rate,
		 "num channels:", tdm->alsa_channels,
		 "num-slots", tdm->num_tdm_slots);

	alsa_bps = tdm->alsa_sample_bits; /* Num bits in ALSA PCM sample (per channel) */
	if (unlikely((alsa_bps != 8) && (alsa_bps != 16))) {
		vdbg_prt("%s\n", "PCM width <> 8 or 16 bits - Not supported!");
		return -EINVAL;
	}

	/* If both playback and capture streams are open, and one of them
	 * is setting the hw parameters right now (since we are here), set
	 * constraints to the other stream to match the current one.
	 */
	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		if (tdm->capt_substream) {
			vdbg_prt("%s\n", "Apply constraints to Capture stream");
			tdm_apply_constraints(tdm, tdm->capt_substream);
		}
		tdm->play_configured = 1;
		vdbg_prt("%s\n", "Playback stream explicitly configured");
		break;

	case SNDRV_PCM_STREAM_CAPTURE:
		if (tdm->play_substream) {
			vdbg_prt("%s\n", "Apply constraints to Playback stream");
			tdm_apply_constraints(tdm, tdm->play_substream);
		}
		tdm->capt_configured = 1;
		vdbg_prt("%s\n", "Capture stream explicitly configured");
		break;
	}

	return 0;
}

/* Note: this fuction may be called more then once for the same session ! */
static int
dvf_tdm_hw_free(struct snd_pcm_substream *substream,
		struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt("%s\n", DIR_STR(substream->stream));

	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		tdm->play_configured = 0;
		vdbg_prt("%s\n", "Playback stream unconfigured");
		break;

	case SNDRV_PCM_STREAM_CAPTURE:
		tdm->capt_configured = 0;
		vdbg_prt("%s\n", "Capture stream unconfigured");
		break;
	}

#ifdef CONFIG_DSPG_CSS_TDM
	if ((!tdm->play_configured) && (!tdm->capt_configured)) {
		if (tdm->css_owns_tdm) {

			/* TODO - add handing for Revoke Timeout (CSS has curshed ) */

			/* Resume APP ownership over the TDM machine */
			if (coma_tdm_revoke(tdm->dai_id) == 0) {
				tdm->css_owns_tdm = 0;
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Revoked OK");
			} else {
				vdbg_prt("%s %u %s\n",
					 "tdm", tdm->dai_id, "Revoke Failed");
			}

			/* Entirely Reset the TDM machine:
			 * the TDM machine gets Disabled, its interrupts cleared,
			 * and all of its registers get their default values.
			 */
			dvf_tdm_reset(tdm);

			/* Enable TDM interrupt (at the APP ICU !) */
			enable_irq(tdm->iores_irq);
		}
	}
#endif /* CONFIG_DSPG_CSS_TDM */

	return 0;
}

static int
dvf_tdm_set_dai_fmt(struct snd_soc_dai *cpu_dai, unsigned int fmt)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	vdbg_prt("format requested 0x%x\n", fmt);

	tdm->dai_format = fmt;
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		break;
	case SND_SOC_DAIFMT_DSP_B:
		break;
	default:
		return -EINVAL;
	}

	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	/* Note: M/S setting is from CODEC point-of-view.
	 *	 Here we set the Opposite !
	 */

	case SND_SOC_DAIFMT_CBM_CFM:
		tdm->clk_master = 0;
		tdm->fs_master = 0;
		vdbg_prt("%s\n", "DVF99 is TDM Slave");
		break;
	case SND_SOC_DAIFMT_CBS_CFM:
		tdm->clk_master = 1;
		tdm->fs_master = 0;
		break;
	case SND_SOC_DAIFMT_CBM_CFS:
		tdm->clk_master = 0;
		tdm->fs_master = 1;
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		tdm->clk_master = 1;
		tdm->fs_master = 1;
		vdbg_prt("%s\n", "DVF99 is TDM Master");
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static irqreturn_t
dvf_tdm_irq(int irq, void *dev_id)
{
	struct dvf_tdm_priv *tdm = dev_id;
	irqreturn_t ret = IRQ_NONE;
	uint pending;
	unsigned long flags;

	pending = tdm_read(tdm, TDM_INT_STATUS) & tdm_read(tdm, TDM_INT_EN);

	if (pending) {
		if (pending & (TDM_INT_STATUS_RX_OVRN | TDM_INT_STATUS_TX_UNDRN)) {
			if (pending & TDM_INT_STATUS_TX_UNDRN) {
				printk(KERN_WARNING "TDM %u Underrun !!!\n",
				       tdm->dai_id);

			}
			if (pending & TDM_INT_STATUS_RX_OVRN) {
				printk(KERN_WARNING "TDM %u Overrun !!!\n",
				       tdm->dai_id);
			}

			/*
			 * DMA was too slow. We have to cancel both directions because
			 * the TDM has to be reset to recover.
			 */
			if (tdm->play_active) {
				vdbg_prt("%s\n", "TDM IRQ: Stopping Playback pcm stream !");
				snd_pcm_stream_lock_irqsave(tdm->play_substream, flags);
				snd_pcm_stop(tdm->play_substream, SNDRV_PCM_STATE_XRUN);
				snd_pcm_stream_unlock_irqrestore(tdm->play_substream, flags);
			}
			if (tdm->capt_active) {
				vdbg_prt("%s\n", "TDM IRQ: Stopping Capture pcm stream !");
				snd_pcm_stream_lock_irqsave(tdm->capt_substream, flags);
				snd_pcm_stop(tdm->capt_substream, SNDRV_PCM_STATE_XRUN);
				snd_pcm_stream_unlock_irqrestore(tdm->capt_substream, flags);
			}

			/* Clear X-RUN interrupts (in TDM machine) */
			tdm_write(tdm, TDM_INT_CLR, TDM_INT_CLR_RX_OVRN | TDM_INT_CLR_TX_UNDRN);
		} else {
			/* Other interrupts should not happen */
			// ... Unless machine is TDM Slave and FSYNC_INT enabled
			// and FSync-jitter happens.

			/* Clear All TDM interrupts (in TDM machine) */
			tdm_write(tdm, TDM_INT_CLR, 0xFFFFF);

			printk(KERN_WARNING "Unexpected TDM interrupt !!! pending = 0x%X\n", pending);
		}
		ret = IRQ_HANDLED;
	}

	return ret;
}

static int
dspg_tdm_dai_remove(struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt();

	reset_control_put(tdm->rc);
	devm_free_irq(cpu_dai->dev, tdm->iores_irq, tdm);
	devm_iounmap(cpu_dai->dev, tdm->reg_base);
	devm_clk_put(cpu_dai->dev, tdm->clk);

	return 0;
}

static int
dspg_tdm_dai_probe(struct snd_soc_dai *cpu_dai)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);
	struct platform_device *pdev = to_platform_device(cpu_dai->dev);
	struct resource *iores;
	int ret = 0;

	dbg_prt();

	tdm->rc = devm_reset_control_get(cpu_dai->dev, NULL);
	if (IS_ERR(tdm->rc))
		return PTR_ERR(tdm->rc);

	tdm->clk = devm_clk_get(cpu_dai->dev, NULL);
	if (IS_ERR(tdm->clk)) {
		ret = PTR_ERR(tdm->clk);
		dev_err(cpu_dai->dev, "No clock!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "clk_get OK");

	clk_prepare(tdm->clk);

	tdm->iores_irq = platform_get_irq(pdev, 0);
	if (tdm->iores_irq < 0) {
		ret = tdm->iores_irq;
		dev_err(cpu_dai->dev, "No IRQ!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "platform_get_irq OK");

	iores = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!iores) {
		dev_err(cpu_dai->dev, "Mem iores missing!\n");
		return -EINVAL;
	}
	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "platform_get_resource OK");

	tdm->reg_base = devm_ioremap(cpu_dai->dev, iores->start, TDM_REGLEN);
	if (!tdm->reg_base) {
		dev_err(cpu_dai->dev, "Cannot remap registers!\n");
		return -ENOMEM;
	}
	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "ioremap_nocache OK");

	ret = devm_request_irq(cpu_dai->dev, tdm->iores_irq, dvf_tdm_irq, 0,
			       "dvf_tdm_irq", tdm);
	if (ret) {
		dev_err(cpu_dai->dev, "Cannot request IRQ!\n");
		return ret;
	}
	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "request_irq OK");

	spin_lock_init(&tdm->tdm_clk_lock);
	tdm->tdm_clk_active = 0;
	tdm->tdm_clk_suspended = 0;

	tdm->css_to_own_tdm = 0;
	tdm->css_owns_tdm = 0;
	tdm->tdm_config_only = 0;

	tdm->alsa_sample_bits = 16;	/* Defualt PCM sample width, in bits */
	tdm->tdm_sample_bits = 16;	/* Default TDM time-slot width, in SCLK cycles:
					 * (1 cycle per bit) */
	tdm->dai_id = cpu_dai->id;

	vdbg_prt("%s %u %s\n",
		 "cpu-dai(tdm) id", cpu_dai->id, "probed ok");

	return 0;
}

static snd_pcm_sframes_t
dvf_tdm_delay_report(struct snd_pcm_substream *substream,
		     struct snd_soc_dai *cpu_dai)
{
	/* Note:
	 * The DMA buffers, in dvf-pcm.c, are defined to hold
	 * one single max period (number of audio-frames), and dynamically hold
	 * excactly one Period; thus the delay (in number of frames) is 1 period
	 * to start with.
	 * In addition to that, there are audio-frames waiting to be TX in the
	 * TX-FIFO (after being DMA'ed) or accumulated in the RX_FIFO (before
	 * being DMA'ed).
	 */

	switch (substream->stream) {

	case SNDRV_PCM_STREAM_PLAYBACK:
		return substream->runtime->period_size +
		       (TDM_FIFO_DEPTH - TDM_TX_FIFO_LEVEL);	/* 48 */

	case SNDRV_PCM_STREAM_CAPTURE:
		return substream->runtime->period_size +
		       TDM_RX_FIFO_LEVEL;			/* 16 */
	}

	return 0;
}

void
dvf_tdm_set_slots(struct snd_soc_dai *cpu_dai, int slots)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	vdbg_prt("%s %u setting slots to %u\n",
		 "cpu-dai(tdm) id", cpu_dai->id, slots);

	tdm->num_tdm_slots = slots;
}

static struct snd_soc_dai_ops dvf_tdm_dai_ops = {
	.set_fmt	= dvf_tdm_set_dai_fmt,
	.startup	= dvf_tdm_startup,
	.shutdown	= dvf_tdm_shutdown,
	.hw_params	= dvf_tdm_hw_params,
	.hw_free	= dvf_tdm_hw_free,
	.prepare	= dvf_tdm_prepare,
	.trigger	= dvf_tdm_trigger,
	.delay		= dvf_tdm_delay_report,

	/// A.M. Is symmetric_rates fixed in
	/// kernel 2.6.39 or still 'broken' ?
	//.symmetric_rates = 1,
};

void
dvf_tdm_set_telephony(struct snd_soc_dai *cpu_dai, int tel_rqst)
{
	struct dvf_tdm_priv *tdm = snd_soc_dai_get_drvdata(cpu_dai);

	dbg_prt();
	vdbg_prt("%s %s %s.  %s %s\n",
		 "Pending TEL Request for cpu-dai",
		 tdm->dai->name,
		 ((tel_rqst) ? ": Tel to ON" : ": Tel to OFF"),
		 "Curr TDM Owner:", ((g_tel_rqst[tdm->dai_id]) ? "CSS" : "APP"));

	if (tel_rqst) {
		if (!g_tel_rqst[tdm->dai_id])
			vdbg_prt("%s\n", "Pending APP --> CSS switch");
	} else {
		if (g_tel_rqst[tdm->dai_id])
			vdbg_prt("%s\n", "Pending CSS --> APP switch");
	}

	g_tel_rqst[tdm->dai_id] = tel_rqst;
}
EXPORT_SYMBOL(dvf_tdm_set_telephony);

#define DVF_TDM_DAI(_id)						 \
{									 \
	.name = DEVICE_NAME"."__stringify(_id),				 \
	.id = (_id),							 \
	.probe = dspg_tdm_dai_probe,					 \
	.remove = dspg_tdm_dai_remove,					 \
	.ops = &dvf_tdm_dai_ops,					 \
	.playback = {							 \
		.channels_min = 1,					 \
		.channels_max = 2,					 \
		.rates = SNDRV_PCM_RATE_8000_48000,			 \
		.formats = SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_S16_LE,\
	},								 \
	.capture = {							 \
		.channels_min = 1,					 \
		.channels_max = 2,					 \
		.rates = SNDRV_PCM_RATE_8000_48000,			 \
		.formats = SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_S16_LE,\
	},								 \
}

static struct snd_soc_dai_driver dvf_tdm_dai[NUM_DVF101_TDMS] = {
	DVF_TDM_DAI(0),
	DVF_TDM_DAI(1),
	DVF_TDM_DAI(2),
};

static int
dspg_tdm_dev_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	struct dvf_tdm_priv *tdm;
	int ret;

	dbg_prt();

	tdm = devm_kzalloc(&pdev->dev, sizeof(struct dvf_tdm_priv), GFP_KERNEL);
	if (!tdm) {
		dev_err(&pdev->dev, "Can't allocate dvf_tdm_priv");
		return -ENOMEM;
	}
	dev_set_drvdata(&pdev->dev, tdm);

	ret = of_property_read_u32(np, "id", &tdm->dai_id);
	if (ret) {
		dev_err(&pdev->dev, "missing 'id' property!\n");
		return -EINVAL;
	}

	if (tdm->dai_id >= CONFIG_DVF_NR_TDMS) {
		dev_err(&pdev->dev, "invalid 'id' property!\n");
		return -EINVAL;
	}

	tdm->dai = &dvf_tdm_dai[tdm->dai_id];

	ret = snd_soc_register_component(&pdev->dev, &dvf_tdm_component, tdm->dai, 1);
	if (ret)
		vdbg_prt("%s %u %s\n", "Registering cpu-dai", tdm->dai->id, "Failed !!!");

	return ret;

}

static int
dspg_tdm_dev_remove(struct platform_device *pdev)
{
	snd_soc_unregister_component(&pdev->dev);

	return 0;
}

static int
dspg_tdm_dev_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct dvf_tdm_priv *tdm = dev_get_drvdata(&pdev->dev);

	dbg_prt();

	if (tdm) {
		if (!tdm->css_owns_tdm) {
			/* Disable dai's TDM machine clock,
			 * If (!) Is active upon 'suspend'
			 */
			if (tdm->tdm_clk_active) {
				vdbg_prt("%s %u\n",
					 "Disabling TDM clock of TDM id",
					 tdm->dai_id);
				tdm->tdm_clk_suspended = 1;
				tdm_clk_disable(tdm);
			} else {
				tdm->tdm_clk_suspended = 0;
			}
		} else {
			vdbg_prt("%s %u %s\n", "TDM id", tdm->dai_id,
				 "owned by CSS. Ignoring!");
		}

	} else {
		vdbg_prt("%s\n", "No driver_data. Ignoring!");
	}

	return 0;
}

static int
dspg_tdm_dev_resume(struct platform_device *pdev)
{
	struct dvf_tdm_priv *tdm = dev_get_drvdata(&pdev->dev);

	dbg_prt();

	if (tdm) {
		if (!tdm->css_owns_tdm) {
			/* Re-Enable dai's TDM machine clock,
			 * If (!) Was active upon 'suspend'
			 */
			if (tdm->tdm_clk_suspended) {
				vdbg_prt("%s %u\n",
					 "Re-Enabling TDM clock of TDM id",
					 tdm->dai_id);
				tdm->tdm_clk_suspended = 0;
				tdm_clk_enable(tdm);
			}
		} else {
			vdbg_prt("%s %u %s\n",
				 "TDM id",
				 tdm->dai_id,
				 "owned by CSS. Ignoring!");
		}
	} else {
		vdbg_prt("%s\n", "No driver_data. Ignoring!");
	}

	return 0;
}

static const struct of_device_id dspg_tdm_dt_match[] = {
	{ .compatible = "dspg,tdm", },
	{}
};

static struct platform_driver dspg_tdm_driver = {
	.probe		= dspg_tdm_dev_probe,
	.remove		= dspg_tdm_dev_remove,
	.suspend	= dspg_tdm_dev_suspend,
	.resume		= dspg_tdm_dev_resume,
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = dspg_tdm_dt_match,
	},
};

static int __init
dspg_tdm_init(void)
{
	return platform_driver_register(&dspg_tdm_driver);
}
module_init(dspg_tdm_init);

static void __exit
dspg_tdm_exit(void)
{
	platform_driver_unregister(&dspg_tdm_driver);
}
module_exit(dspg_tdm_exit);

MODULE_DESCRIPTION("DVF ASoC platform CPU DAI driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:" DRV_NAME);
