/*
 * dvf-evb.c -- ASoC Machine Driver for DVF99 EVB board,
 *
 *  Copyright (C) 2015 DSP Group Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <mach/platform.h>
#include <mach/gpio.h>
#include <asm/atomic.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/dvf-tdm.h>
#include <sound/soc-dapm.h>
#include <linux/of.h>

#define DRIVER_NAME "snd-dvf-evb-bt"
#define DVF_BT_NUM_TDM_SLOTS 2


#ifdef V_DEBUG
	#ifndef DEBUG
		#define DEBUG
	#endif
	#define vdbg_prt(fmt, ...) \
		pr_debug("DVF99-BT asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)
#else
	#define vdbg_prt(fmt, ...)
#endif


#ifdef DEBUG
	#define dbg_prt(fmt, ...) \
		pr_debug("DVF99-BT asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)

	#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
	#define dbg_prt(fmt, ...)
	#define DIR_STR(stream_dir)
#endif

static int g_dvf_tel_state = 0;

#define CSR8811_BITS_PER_SLOT     16
#define DVF_CSR8811_FSYNC_RATE_NB    8000
#define DVF_CSR8811_FSYNC_RATE_WB    16000
#define DVF_CSR8811_NUM_CHANNEL   0
static int g_dvf_csr8811_configure;
static int g_dvf_csr8811_grant;
static const char * const csr8811_configure_function[] = {"Off", "NB", "WB"};
static const char * const csr8811_grant_function[]     = {"Revoke", "Grant"};
static const struct soc_enum tdm_configure_enum
	= SOC_ENUM_SINGLE_EXT(3, csr8811_configure_function);
static const struct soc_enum tdm_grant_enum
	= SOC_ENUM_SINGLE_EXT(2, csr8811_grant_function);


static const char *tel_function[] = {"Off", "On"};
static const struct soc_enum tel_enum = SOC_ENUM_SINGLE_EXT(2, tel_function);

#define DMW_MAX_NUM_DAIs 1
static struct snd_soc_dai_link brd_dvf_dai_link[DMW_MAX_NUM_DAIs];

void dvf_tdm_set_telephony(struct snd_soc_dai *cpu_dai, int tel_rqst);

/* CSR csr8811 Bluetooth - optional */
static int brd_csr8811_dai_init(struct snd_soc_pcm_runtime *rtd);

/* ASoC Controls, DAPM Widgets and Routes: */
static void dvf_set_tel_state(struct snd_pcm_substream *substream);

/* CSR8811 section: */
/* ---------------- */

static int
brd_csr8811_hw_params(struct snd_pcm_substream *substream,
		      struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;

	unsigned int dai_format;
	int ret;

	dbg_prt("%s\n", DIR_STR(substream->stream));

	switch(params_rate(params)) {
	case 8000:
	case 16000:
		/* slot 0: 16bits, rx/tx, fsync start at bit 0, 1 bit wide */
		snd_soc_dai_set_tdm_slot(cpu_dai,
					 (3 << 6),
					 1 | (0 << 1) | (1 << 6),
					 0,
					 16);
		/* slot 1: 16bits, rx/tx */
		snd_soc_dai_set_tdm_slot(cpu_dai,
					 (3 << 6),
					 0,
					 1,
					 16);
		break;
	default:
		break;
	}

	/* STEP 1: */

	/* Set DAI's TDM configuration: */
	dai_format = 0;
	dai_format |= SND_SOC_DAIFMT_I2S; /* TODO changes for CSR8811 !!! */

#ifdef CONFIG_SOC_CODEC_TDM_MASTER
	/* DVF is TDM Slave */
	dai_format |= SND_SOC_DAIFMT_IB_NF;	/* invert BCLK + nor FRM */
	dai_format |= SND_SOC_DAIFMT_CBM_CFM;	/* Codec(!) point-of-view - Csr is Master */
	vdbg_prt("%s\n", "DVF99 TDM is Slave, Csr is Master");
#else
	/* DVF is TDM Master */
	dai_format |= SND_SOC_DAIFMT_IB_NF;	/* invert BCLK + nor FRM */
	dai_format |= SND_SOC_DAIFMT_CBS_CFS;	/* Codec(!) point-of-view - Csr is Slave */
	vdbg_prt("%s\n", "DVF99 TDM is Master, Csr is Slave");
#endif /* !CONFIG_SOC_CODEC_TDM_MASTER */

	/* Set cpu DAI configuration */
	vdbg_prt("%s", "Calling CPU DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0) {
		return ret;
	}

	/* Set codec DAI configuration */
	vdbg_prt("%s\n", "Calling CODEC DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(codec_dai, dai_format);
	if (ret < 0) {
		return ret;
	}

	return 0;
}

static int brd_csr8811_prepare(struct snd_pcm_substream *substream)
{
	dbg_prt("%s\n", DIR_STR(substream->stream));

	vdbg_prt("%s\n", "Setting TEL state");

	dvf_set_tel_state(substream);

	return 0;
}

static struct snd_soc_ops brd_csr8811_ops = {
	.hw_params	= brd_csr8811_hw_params,
	.prepare	= brd_csr8811_prepare,
};

static struct snd_soc_dai_link brd_csr8811_dai_link = {
	.name = "dvf99-csr8811_Dai_Link",
	.stream_name = "dvf99-csr8811_I2S",

	/* asoc Cpu-Dai (TDM) device name */
	.cpu_dai_name = "dvf-tdm.1",

	/* asoc Codec-Dai device name */
	.codec_dai_name = "csr8811-hifi",

	.init = brd_csr8811_dai_init,
	.ops = &brd_csr8811_ops,
};

#define CODEC_NAME "bt-csr8811-codec"
static int
brd_dvf_dai_link_init(struct snd_soc_card *card)
{
	int nCnt;
	int num_dai_links = 0;
	struct snd_soc_dai_link *dai_link;
	struct device_node *codec_node;

	codec_node = of_find_node_by_name(0, CODEC_NAME);
	if (!codec_node) {
		printk(" Codec Node not found \n");
		return -1;
	}

	memcpy( &brd_dvf_dai_link[num_dai_links++],
		&brd_csr8811_dai_link,
		sizeof(struct snd_soc_dai_link) );
	for (nCnt = 0; nCnt < num_dai_links; nCnt++) {
		dai_link = &card->dai_link[nCnt];
		dai_link->codec_of_node = codec_node;
	}

	return num_dai_links;
}

static void
dvf_set_tel_state(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;

	dbg_prt();

	vdbg_prt("%s %s\n", "Requested TEL state is",
		(g_dvf_tel_state) ? "ON" : "OFF");

	vdbg_prt("%s %s\n",
		 "Calling dvf_tdm_set_telephony() for cpu-dai",
		 cpu_dai_name);

	dvf_tdm_set_telephony(cpu_dai, g_dvf_tel_state);

	vdbg_prt("%s\n", "Calling dvf_pcm_set_telephony()");
}

static int
dvf_get_tel(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = g_dvf_tel_state;

	vdbg_prt("%s %s\n", "TEL state is:",
		(g_dvf_tel_state == 0) ? "Off" : "On");

	return 0;
}

static int
dvf_set_tel(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	if (g_dvf_tel_state == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
			"TEL state Unchanged:",
			(g_dvf_tel_state == 0) ? "Off" : "On");
		return 0;
	}

	g_dvf_tel_state = ucontrol->value.integer.value[0];

	vdbg_prt("%s %s\n", "Updating TEL state to:",
		(g_dvf_tel_state == 0) ? "Off" : "On");

	return 1;
}

static const struct snd_kcontrol_new tel_controls[] = {
	SOC_ENUM_EXT("Telephony", tel_enum, dvf_get_tel, dvf_set_tel),
};

static int
dvf_tdm_get_configure(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol){
	ucontrol->value.integer.value[0] = g_dvf_csr8811_configure;
	vdbg_prt("%s %s\n", "TDM Configure is:",
		(g_dvf_csr8811_configure == 0) ? "Off" : "WB/NB");
	return 0;
}

static int
dvf_tdm_set_configure(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol){
	struct snd_soc_dai *cpu_dai = snd_kcontrol_chip(kcontrol);
	struct snd_soc_pcm_runtime *rtd = container_of(&cpu_dai,
		struct snd_soc_pcm_runtime, cpu_dai);
	struct snd_soc_dai *codec_dai = rtd->codec_dai;	unsigned int dai_format;
	int ret;

	if (g_dvf_csr8811_configure == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
			"TDM Configure is Unchanged:",
			(g_dvf_csr8811_configure == 0) ? "Off" : "WB/NB");
		return 0;
	}
	vdbg_prt("rtd '%p' codec_dai '%p' cpu_dai '%p'\n",
		rtd, codec_dai, cpu_dai);
	if (ucontrol->value.integer.value[0] == 0) {
		g_dvf_csr8811_configure = 0;
		if (g_dvf_csr8811_grant == 1)
			g_dvf_csr8811_grant = 0;
		dvf_tdm_configure_reg(cpu_dai, codec_dai, CSR8811_BITS_PER_SLOT,
			DVF_CSR8811_FSYNC_RATE_NB, DVF_CSR8811_NUM_CHANNEL, 1);
		vdbg_prt("%s\n", "Reset TDM Configure");
		return 0;
	}
#if 1
	/* For BT we don't need to do the DAI setting,
	 * let's use the kernel default
	 */
	/* Set DAI's TDM configuration: */
	dai_format = 0x0000;
	dai_format = SND_SOC_DAIFMT_I2S;
	/* Normal SCLK(BCLK) & FSync(LRC) polarity */
	dai_format |= SND_SOC_DAIFMT_IB_NF;
#ifdef CONFIG_SOC_CODEC_TDM_MASTER
	/* Codec(!) point-of-view - CSR8811 is Master */
	/* DVF101 TDM is Slave */
	dai_format |= SND_SOC_DAIFMT_CBM_CFM;
	vdbg_prt("%s\n", "DVF99 TDM is Slave, CSR8811 is Master");
#else
	/* Codec(!) point-of-view - CSR8811 is Slave */
	/* DVF101 TDM is Master */
	dai_format |= SND_SOC_DAIFMT_CBS_CFS;
	vdbg_prt("%s\n", "DVF99 TDM is Master, CSR8811 is Slave");
#endif /* !CONFIG_SOC_CODEC_TDM_MASTER */
	/* Set cpu DAI configuration */
	vdbg_prt("%s", "Calling CPU DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0) {
		vdbg_prt("Failed to set the dai format '%d'\n", ret);
		return ret;
		}
#endif
	/* Set the TDM DAI's slots using proprietary function */
	dvf_tdm_set_slots(cpu_dai, DVF_BT_NUM_TDM_SLOTS);
	ret = dvf_tdm_configure_reg(cpu_dai, codec_dai, CSR8811_BITS_PER_SLOT,
		(ucontrol->value.integer.value[0] == 1) ?
		DVF_CSR8811_FSYNC_RATE_NB : DVF_CSR8811_FSYNC_RATE_WB,
		DVF_CSR8811_NUM_CHANNEL, 0);
	if (ret < 0) {
		vdbg_prt("Failed to do the TDM configuration '%d'\n", ret);
		return ret;
	}
	g_dvf_csr8811_configure = ucontrol->value.integer.value[0];
	vdbg_prt("%s %s\n", "Updating TDM Configure to:",
		(g_dvf_csr8811_configure == 0) ? "Off"
		: (g_dvf_csr8811_configure == 1) ? "NB" : "WB");
	return 1;
}

static int
dvf_tdm_get_grant(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = g_dvf_csr8811_grant;

	vdbg_prt("%s %s\n", "TDM Grant is:",
	(g_dvf_csr8811_grant == 0) ? "Revoke" : "Grant");

	return 0;
}

static int
dvf_tdm_set_grant(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_dai *cpu_dai = snd_kcontrol_chip(kcontrol);

	if (g_dvf_csr8811_grant == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
			"TDM Grant value is  Unchanged:",
			(g_dvf_csr8811_grant == 0) ? "Revoke" : "Grant");
	return 0;
	}

	g_dvf_csr8811_grant = ucontrol->value.integer.value[0];

	dvf_tdm_grant(cpu_dai, g_dvf_csr8811_grant);

	vdbg_prt("%s %s\n", "Updating TDM Grant :",
		(g_dvf_csr8811_grant == 0) ? "Revoke" : "Grant");

	return 1;
}

static const struct snd_kcontrol_new dvf_tdm_controls[] = {
	SOC_ENUM_EXT("TDM_Configure", tdm_configure_enum,
		dvf_tdm_get_configure, dvf_tdm_set_configure),
	SOC_ENUM_EXT("TDM_Grant", tdm_grant_enum,
		dvf_tdm_get_grant, dvf_tdm_set_grant),
};

static int
brd_csr8811_dai_init(struct snd_soc_pcm_runtime *rtd)
{
	int ret;
	struct snd_soc_codec *codec = rtd->codec;
	struct snd_soc_dapm_context *dapm = snd_soc_codec_get_dapm(codec);

	dbg_prt();

	vdbg_prt("%s\n", "Adding Telephony control");
	ret = snd_soc_add_dai_controls(rtd->cpu_dai, dvf_tdm_controls,
		ARRAY_SIZE(dvf_tdm_controls));
	if (ret) {
		vdbg_prt("%s\n", "Telephony control adding Failed");
		return ret;
	}

	vdbg_prt("%s\n", "Syncing dapm");
	ret = snd_soc_dapm_sync(dapm);
	if (ret) {
		vdbg_prt("%s\n", "Dapm syncing Failed");
		return ret;
	}

	return 0;
}

static struct snd_soc_card g_dspg_dvf99_evb_bt_card = {
	.name		= "dspg-dvf99-evb-bt-card",
	.dai_link	= brd_dvf_dai_link,  /* DAI-link Array(!) */
	.num_links	= 0, /* overridden by dspg_snd_probe() */

	/* Note: new to ASoC:
	 * Card's (not Codec's !) set_bias_level().
	 * May be used for certain activities, e.g. diasbling
	 * card's power amplifiers or enabling them just upon BIAS_ON;
	 * this is if they are Not declared as card's DAPM Widgests,
	 * as part of the audio Routes.
	 */
	.set_bias_level		= NULL,
	.set_bias_level_post	= NULL,
};

static int
dspg_snd_probe(struct platform_device *pdev)
{
	int ret;
	struct snd_soc_card *card = &g_dspg_dvf99_evb_bt_card;

	dbg_prt();

	/* Initialize DAI-link array and num of DAIs that will be registered */
	card->num_links = brd_dvf_dai_link_init(card);

	/* Book keeping : */

	/* note: platform_set_drvdata() here saves pointer to the card's data
	 * on the device's 'struct device_private *p'->driver_data
	 */
	card->dev = &pdev->dev;

	ret = snd_soc_register_card(card);
	if (ret) {
		dev_err(&pdev->dev,"snd_soc_register_card Failed\n");
		goto ERR_CLEAR;
	}
	dev_info(&pdev->dev,"DVF99 BT ASoc Card registred\n");

	return 0;

ERR_CLEAR:

	return ret;
}

static int
dspg_snd_remove(struct platform_device *pdev)
{
	struct snd_soc_card *card = platform_get_drvdata(pdev);

	dbg_prt();

	snd_soc_unregister_card(card);

	return 0;
}

static struct of_device_id dspg_dvf_bt_match[] = {
	{ .compatible = "dspg,dvf99-bt-mach-drv" },
	{ },
};

static struct platform_driver g_brd_dvf99_evb_snd_drv = {
	.driver = {
		.name = DRIVER_NAME,
		.owner = THIS_MODULE,
		.of_match_table = dspg_dvf_bt_match,
		.pm = &snd_soc_pm_ops,
	},
	.probe = dspg_snd_probe,
	.remove = dspg_snd_remove,
};

int
dvf_snd_bt_init(void)
{
	return platform_driver_register(&g_brd_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_bt_init);

void
dvf_snd_bt_exit(void)
{
	platform_driver_unregister(&g_brd_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_bt_exit);

static int __init
brd_dvf_evb_mod_init(void)
{
	return 0;
}

static void __exit
brd_dvf_evb_mod_exit(void)
{
}

module_init(brd_dvf_evb_mod_init);
module_exit(brd_dvf_evb_mod_exit);

MODULE_DESCRIPTION("ASoC machine driver for DSPG DVF99 EVB board");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
