/*
 * dvf-evb-si32178.c -- ASoC Machine Driver for si32178 on DVF99 EVB board,
 *
 * 10 Jan, 2013:
 * Adapted to kernel 3.4 by Sujan.S <sujan.somisetty@dspg.com>
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <mach/platform.h>
#include <mach/gpio.h>
#include <asm/atomic.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <sound/dvf-tdm.h>
#include <linux/of.h>

#define DRIVER_NAME "snd-dvf-evb-fxs"
#ifdef V_DEBUG
	#ifndef DEBUG
		#define DEBUG
	#endif
	#define vdbg_prt(fmt, ...) \
		pr_debug( "Evb-si32178 asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)
#else
	#define vdbg_prt(fmt, ...)
#endif


#ifdef DEBUG
	#define dbg_prt(fmt, ...) \
		pr_debug( "Evb-si32178 asoc Mach-drv: %s(): " fmt, __func__, ##__VA_ARGS__)

	#define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
	#define dbg_prt(fmt, ...)
	#define DIR_STR(stream_dir)
#endif

static int g_dvf_tel_state = 0;
#define DVF_FXS_NUM_TDM_SLOTS 4

static const char *tel_function[] = {"Off", "On"};
static const struct soc_enum tel_enum = SOC_ENUM_SINGLE_EXT(2, tel_function);

#define DMW_MAX_NUM_DAIs 1
static struct snd_soc_dai_link brd_dvf_dai_link[DMW_MAX_NUM_DAIs];

/* CSR si32178 Bluetooth - optional */
static int brd_si32178_dai_init(struct snd_soc_pcm_runtime *rtd);

void dvf_tdm_set_telephony(struct snd_soc_dai *cpu_dai, int tel_rqst);

/* ASoC Controls, DAPM Widgets and Routes: */
static void dvf_set_tel_state(struct snd_pcm_substream *substream);

/* SI32178 section: */
/* ---------------- */

static int
brd_si32178_hw_params(struct snd_pcm_substream *substream,
		      struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;
	unsigned int bits;
	unsigned int dai_format;
	int ret;

	dbg_prt("%s\n", DIR_STR(substream->stream));

	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S8:
		bits = 8;
		break;
	case SNDRV_PCM_FORMAT_S16_LE:
		bits = 16;
		break;
	default:
		return -EINVAL;
	}

	/* slot 0: rx/tx, fsync start at bit 0, 1 bit wide */
	snd_soc_dai_set_tdm_slot(cpu_dai,
				 (3 << 6),
				 1 | (0 << 1) | (1 << 6),
				 0,
				 bits);
	/* slot 1: rx/tx, no fsync */
	snd_soc_dai_set_tdm_slot(cpu_dai,
				 (3 << 6),
				 0,
				 1,
				 bits);
	/* slot 2: rx/tx, no fsync */
	snd_soc_dai_set_tdm_slot(cpu_dai,
				 (3 << 6),
				 0,
				 2,
				 bits);
	/* slot 3: rx/tx, no fsync */
	snd_soc_dai_set_tdm_slot(cpu_dai,
				 (3 << 6),
				 0,
				 3,
				 bits);

	/* Set DAI's TDM configuration: */
	dai_format = 0;
	dai_format = SND_SOC_DAIFMT_DSP_B;
	dai_format |= SND_SOC_DAIFMT_IB_NF; /* Normal SCLK(BCLK) & FSync(LRC) polarity */
	dai_format |= SND_SOC_DAIFMT_CBS_CFS; /* Codec(!) point-of-view - si3050 is Slave */
	vdbg_prt("%s\n", "DMW96 TDM is Master, SI32178 is Slave");

	/* Set cpu DAI configuration */
	vdbg_prt("%s", "Calling CPU DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0)
		return ret;

	/* Set codec DAI configuration */
	vdbg_prt("%s\n", "Calling CODEC DAI set_fmt(): ");
	ret = snd_soc_dai_set_fmt(codec_dai, dai_format);
	if (ret < 0)
		return ret;

	return 0;
}

static int brd_si32178_prepare(struct snd_pcm_substream *substream)
{
	dbg_prt("%s\n", DIR_STR(substream->stream));

	vdbg_prt("%s\n", "Setting TEL state");

	dvf_set_tel_state(substream);

	return 0;
}

static struct snd_soc_ops brd_si32178_ops = {
	.hw_params	= brd_si32178_hw_params,
	.prepare	= brd_si32178_prepare,
};

static struct snd_soc_dai_link brd_si32178_dai_link = {
	.name = "dvf99-si32178_Dai_Link",
	.stream_name = "dvf99-si32178_spi",

	/* asoc Cpu-Dai (TDM) device name */
	.cpu_dai_name = "5f00000.tdm",

	/* asoc Codec-Dai device name */
	.codec_dai_name = "si32178",

	.init = brd_si32178_dai_init,
	.ops = &brd_si32178_ops,
};

#define CODEC_NAME "dvf_si32178_spi"
static int
brd_dvf_dai_link_init(struct snd_soc_card *card)
{
	int nCnt;
	int num_dai_links = 0;
	struct snd_soc_dai_link *dai_link;
	struct device_node *codec_node;

	codec_node = of_find_node_by_name(0, CODEC_NAME);
	if (!codec_node) {
		printk(" Codec Node not found \n");
		return -1;
	}

	memcpy( &brd_dvf_dai_link[num_dai_links++],
		&brd_si32178_dai_link,
		sizeof(struct snd_soc_dai_link) );
	for (nCnt = 0; nCnt < num_dai_links; nCnt++) {
		dai_link = &card->dai_link[nCnt];
		dai_link->codec_of_node = codec_node;
	}

	return num_dai_links;
}

static void
dvf_set_tel_state(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;

	dbg_prt();

	vdbg_prt("%s %s\n", "Requested TEL state is",
		(g_dvf_tel_state) ? "ON" : "OFF");

	vdbg_prt("%s %s\n",
		 "Calling dvf_tdm_set_telephony() for cpu-dai",
		 cpu_dai->name);

	dvf_tdm_set_telephony(cpu_dai, g_dvf_tel_state);
}

static int
dvf_get_tel(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = g_dvf_tel_state;

	vdbg_prt("%s %s\n", "TEL state is:",
		(g_dvf_tel_state == 0) ? "Off" : "On");

	return 0;
}

static int
dvf_set_tel(struct snd_kcontrol *kcontrol,
	    struct snd_ctl_elem_value *ucontrol)
{
	if (g_dvf_tel_state == ucontrol->value.integer.value[0]) {
		vdbg_prt("%s %s\n",
			"TEL state Unchanged:",
			(g_dvf_tel_state == 0) ? "Off" : "On");
		return 0;
	}

	g_dvf_tel_state = ucontrol->value.integer.value[0];

	vdbg_prt("%s %s\n", "Updating TEL state to:",
		(g_dvf_tel_state == 0) ? "Off" : "On");

	return 1;
}

static const struct snd_kcontrol_new tel_controls[] = {
	SOC_ENUM_EXT("Telephony", tel_enum, dvf_get_tel, dvf_set_tel),
};

static int
brd_si32178_dai_init(struct snd_soc_pcm_runtime *rtd)
{
	int ret = 0;
	struct snd_soc_codec *codec = rtd->codec;
	struct snd_soc_dapm_context *dapm = snd_soc_codec_get_dapm(codec);

	dbg_prt();

	vdbg_prt("%s\n", "Adding Telephony control");
	ret = snd_soc_add_codec_controls(codec, tel_controls,
					 ARRAY_SIZE(tel_controls));
	if (ret) {
		vdbg_prt("%s\n", "Telephony control adding Failed");
		return ret;
	}

	vdbg_prt("%s\n", "Syncing dapm");
	ret = snd_soc_dapm_sync(dapm);
	if (ret)
		vdbg_prt("%s\n", "Dapm syncing Failed");

	return 0;
}

static struct snd_soc_card g_dspg_dvf99_evb_fxs_card = {
	.name		= "dspg-dvf99-evb-fxs-card",
	.dai_link	= brd_dvf_dai_link,  /* DAI-link Array(!) */
	.num_links	= 0, /* overridden by dspg_snd_probe() */

	/* Note: new to ASoC:
	 * Card's (not Codec's !) set_bias_level().
	 * May be used for certain activities, e.g. disabling
	 * card's power amplifiers or enabling them just upon BIAS_ON;
	 * this is if they are not declared as card's DAPM widgets,
	 * as part of the audio routes.
	 */
	.set_bias_level		= NULL,
	.set_bias_level_post	= NULL,
};

static int
dspg_snd_probe(struct platform_device *pdev)
{
	int ret;
	struct snd_soc_card *card = &g_dspg_dvf99_evb_fxs_card;

	dbg_prt();

	/* Initialize DAI-link array and num of DAIs that will be registered */
	card->num_links = brd_dvf_dai_link_init(card);

	/* Book keeping : */

	/* note: platform_set_drvdata() here saves pointer to the card's data
	 * on the device's 'struct device_private *p'->driver_data
	 */
	card->dev = &pdev->dev;
	platform_set_drvdata(pdev, card);

	/* Register ASoC sound Card */

	ret = snd_soc_register_card(card);
	if (ret) {
		dev_err(&pdev->dev, "snd_soc_register_card Failed\n");
		goto ERR_CLEAR;
	}

	dev_info(&pdev->dev,"DVF99 FXS ASoc Card registered\n");

	return 0;

ERR_CLEAR:
	snd_soc_card_set_drvdata(card, NULL);
	platform_set_drvdata(pdev, NULL);
	card->dev = NULL;

	return ret;
}

static int
dspg_snd_remove(struct platform_device *pdev)
{
	struct snd_soc_card *card = platform_get_drvdata(pdev);
	struct wm897X_setup_data* priv_data = snd_soc_card_get_drvdata(card);

	dbg_prt();

	kfree(priv_data);
	snd_soc_unregister_card(card);

	return 0;
}

static struct of_device_id dspg_dvf_fxs_match[] = {
	{ .compatible = "dspg,dvf99-fxs-mach-drv" },
	{ },
};

static struct platform_driver g_brd_dvf99_evb_snd_drv = {
	.driver = {
		.name = DRIVER_NAME,
		.owner = THIS_MODULE,
		.of_match_table = dspg_dvf_fxs_match,
		.pm = &snd_soc_pm_ops,
	},
	.probe = dspg_snd_probe,
	.remove = dspg_snd_remove,
};

int
dvf_snd_fxs_init(void)
{
	return platform_driver_register(&g_brd_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_fxs_init);

void
dvf_snd_fxs_exit(void)
{
	platform_driver_unregister(&g_brd_dvf99_evb_snd_drv);
}
EXPORT_SYMBOL(dvf_snd_fxs_exit);

static int __init
brd_dvf_evb_mod_init(void)
{
	return 0;
}

static void __exit
brd_dvf_evb_mod_exit(void)
{
}

module_init(brd_dvf_evb_mod_init);
module_exit(brd_dvf_evb_mod_exit);
MODULE_DESCRIPTION("ASoC machine driver for FXS on DSPG DVF99 EVB board");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
