/*
 * (C) Copyright 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/clock.h>
#include <asm/arch/pads.h>
#include <asm/arch/hardware.h>
#include "crypto_block.h"

unsigned long securedev_pubkey[8] = {
	0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff
};

int do_secretkey(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	unsigned long val;

	/* update OTP registers values */
	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("securedev: error updating otp regs!\n");
		return 1;
	}

	/*
	 * Check OTP state
	 */

	/* abort if secret key bit is already set */
	val = crypto_otp_get_state();
	if (val & CRYPTO_STATE_SECRET_BURNED) {
		printf("securedev: secret key already burned! (state=0x%lx).\n", val);
		return 1;
	}

	/* 1. generate and burn a device specific secret key */
	ret = crypto_otp_gen_seckey();
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error generating secret key!\n");
		return 1;
	}

	/* 2. update OTP state */
	ret = crypto_otp_set_state(CRYPTO_STATE_SECRET_BURNED);
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error updating OTP state!\n");
		return 1;
	}

	printf("ok\n");

	return 0;
}

U_BOOT_CMD(secretkey, 1, 0, do_secretkey,
		"Program the following data to the OTP (eFuse):\n"
		"\t- device specific secret key + SECRET_KEY_BURNED bit.\n",
		""
);

int do_securedev(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	int i;
	int ok;
	unsigned long pubkey[8];
	unsigned long val;

	/* update OTP registers values */
	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("securedev: error updating otp regs!\n");
		return 1;
	}

	/*
	 * Check OTP state
	 */

	/* abort if public, private or enforce_sb bits are already set */
	val = crypto_otp_get_state();
	if (val & CRYPTO_STATE_PUBLIC_KEY_BURNED ||
	    val & CRYPTO_STATE_SECRET_BURNED ||
	    val & CRYPTO_STATE_ENFORCE_SB) {
		printf("securedev: device seems to be partially/secured! (state=0x%lx).\n", val);
		return 1;
	}

	/* abort if public key or secret key alreay contains data */
	val = crypto_otp_get_debug();
	if (val & CRYPTO_OTP_DEBUG_BANK1_BURNED ||
	    val & CRYPTO_OTP_DEBUG_SECKEY_BURNED) {
		printf("securedev: public/secret key already burned!\n");
		return 1;
	}

	/* abort if the public key digest to be burned has the default value 0xffffffff */
	ok = 0;
	for (i = 0; i < 8; i++) {
		if (securedev_pubkey[i] != 0xffffffff) {
			ok = 1;
			break;
		}
	}
	if (!ok) {
		printf("securedev: cannot secure the device because this u-boot image is not signed properly!\n");
		return 1;
	}

	/*
	 * Program OTP
	 */

	printf("securedev: program OTP values... ");

	/* 1. program public key digest */
	ret = crypto_otp_set_pubkey_digest(securedev_pubkey);
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error programming public key digest!\n");
		return 1;
	}

	/* 2. generate and burn a device specific secret key */
	ret = crypto_otp_gen_seckey();
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error generating secret key!\n");
		return 1;
	}

	/* 3. update OTP state */
	ret = crypto_otp_set_state(CRYPTO_STATE_SECRET_BURNED |
				   CRYPTO_STATE_PUBLIC_KEY_BURNED |
				   CRYPTO_STATE_ENFORCE_SB);
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error updating OTP state!\n");
		return 1;
	}

	printf("ok\n");

	/*
	 * Verify OTP
	 */

	printf("securedev: verify OTP values... ");

	/* update OTP registers values */
	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("error\n");
		printf("securedev: error updating otp regs!\n");
		return 1;
	}

	val = crypto_otp_get_state();
	if (!(val & CRYPTO_STATE_ENFORCE_SB) ||
	    !(val & CRYPTO_STATE_PUBLIC_KEY_BURNED) ||
	    !(val & CRYPTO_STATE_SECRET_BURNED)) {
		printf("error\n");
		printf("securedev: error in state register!\n");
		return 1;
	}

	val = crypto_otp_get_debug();
	if (!(val & CRYPTO_OTP_DEBUG_SECKEY_BURNED)) {
		printf("error\n");
		printf("securedev: secret key was not programmed!\n");
		return 1;
	}

	crypto_otp_get_pubkey_digest(pubkey);
	for (i=0; i<8; i++) {
		if (pubkey[i] != securedev_pubkey[i]) {
			printf("error\n");
			printf("securedev: error in public key digest\n");
			return 1;
		}
	}

	printf("ok\n");

	return 0;
}

U_BOOT_CMD(securedev, 1, 0, do_securedev,
		"Secure this device and enforce secure boot from now on.\n"
		"Program the following data to the OTP (eFuse):\n"
		"\t- public key sha256 + PUBLIC_KEY_DIGEST_BURNED bit.\n"
		"\t- device specific secret key + SECRET_KEY_BURNED bit.\n"
		"\t- ENFORCE_SB bit.\n",
		""
);

int do_otpdump(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int i;
	int ret;
	unsigned long digest[8];

	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("otpdump: error updating otp regs!\n");
		return 1;
	}

	crypto_otp_get_pubkey_digest(digest);

	printf("OTP state:                      ");
	printf("0x%08lx\n", crypto_otp_get_state());

	printf("OTP device-id:                  ");
	printf("0x%08lx\n", crypto_otp_get_devid());

	printf("OTP minimum software version:   ");
	printf("0x%08lx\n", crypto_otp_get_minver());

	printf("OTP debug register:             ");
	printf("0x%08lx\n", crypto_otp_get_debug());

	printf("OTP public key digest(sha256):  ");
	for (i = 7; i >= 0; i--) {
		printf("%08lx", digest[i]);
	}
	printf("\n");

	return 0;
}

U_BOOT_CMD(otpdump, 1, 0, do_otpdump,
		"Dump OTP (eFuse) values.\n",
		""
);

int do_otpsetdevid(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	unsigned long newval, oldval, program_bits;

	if (argc<2)
		return -1;

	newval = simple_strtoul(argv[1], NULL, 16);

	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("error: updating otp regs!\n");
		return 1;
	}

	oldval = crypto_otp_get_devid();
	printf("previous devid = 0x%lx\n", oldval);
	if (oldval == 0xffffffff) {
		printf("error: all bits are already programmed!\n");
		return 1;
	}

	// cannot reprogram eFuse bits so remove them
	program_bits = newval & (~oldval);

	printf("programming new device id (0x%lx)... ", newval);

	ret = crypto_otp_set_devid(program_bits);
	if (ret < 0) {
		printf("error\n");
		printf("error: device id program failed!\n");
		return 1;
	}
	printf("ok\n");

	return 0;
}

U_BOOT_CMD(otpsetdevid, 2, 0, do_otpsetdevid,
		"program device id OTP bits.\n",
		"value\n"
		"     - value to program"
);

int do_otpsetminswver(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	unsigned long newval, oldval, program_bits;

	if (argc < 2)
		return -1;

	newval = simple_strtoul(argv[1], NULL, 16);

	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("error: updating otp regs!\n");
		return 1;
	}

	oldval = crypto_otp_get_minver();
	printf("previous min software version = 0x%lx\n", oldval);
	if (oldval == 0xffffffff) {
		printf("error: all bits are already programmed!\n");
		return 1;
	}

	// cannot reprogram eFuse bits so remove them
	program_bits = newval & (~oldval);

	printf("programming new minimum software version (0x%lx)... ", newval);

	ret = crypto_otp_set_minver(program_bits);
	if (ret < 0) {
		printf("error\n");
		printf("error: min software version program failed!\n");
		return 1;
	}
	printf("ok\n");

	return 0;
}

U_BOOT_CMD(otpsetminswver, 2, 0, do_otpsetminswver,
		"program minimum software version OTP bits.\n",
		"value\n"
		"     - value to program"
);

int do_otpdisablejtag(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	unsigned long val;

	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("error: updating otp regs!\n");
		return 1;
	}

	val = crypto_otp_get_state();

	ret = crypto_otp_set_state(val | CRYPTO_STATE_DISABLE_JTAG_IN_SB);
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error updating OTP state!\n");
		return 1;
	}

	printf("ok\n");

	return 0;
}

U_BOOT_CMD(otpdisablejtag, 1, 0, do_otpdisablejtag,
		"Disable JTAG in OTP bits.\n",
		""
);

int do_otpjtagreenable(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	unsigned long val;

	ret = crypto_otp_update_regs();
	if (ret < 0) {
		printf("error: updating otp regs!\n");
		return 1;
	}

	val = crypto_otp_get_state();

	ret = crypto_otp_set_state(val | CRYPTO_STATE_RE_ENABLE);
	if (ret != 0) {
		printf("error\n");
		printf("securedev: error updating OTP state!\n");
		return 1;
	}

	printf("ok\n");

	return 0;
}

U_BOOT_CMD(otpjtagreenable, 1, 0, do_otpjtagreenable,
		"Re-enable JTAG in OTP bits.\n",
		""
);

int do_pubkeydump(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int i;

	for (i = 0; i < 8; i++) {
		printf("pubkey[%d] = 0x%08lx\n", i, securedev_pubkey[i]);
	}

	return 0;
}

U_BOOT_CMD(pubkeydump, 1, 0, do_pubkeydump,
		"Show u-boot internal public key digest.\n",
		""
);

