/*
 * DSPG crypto block driver
 *
 * (C) Copyright 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <malloc.h>
#include <asm/arch/hardware.h>
#include <asm/arch/clock.h>

#include "crypto_block.h"

static void crypto_writel(unsigned long reg, unsigned long val)
{
	writel(val, DVF99_CRYPTO_BASE + reg);
}

static unsigned long crypto_readl(unsigned long reg)
{
	return readl(DVF99_CRYPTO_BASE + reg);
}

/* crypto_init:
 *
 * enable crypto and PMU clocks
 **/
static void crypto_init(void)
{
	reset_release("pmu");
	clk_enable("pmu");
	reset_release("secure");
	clk_enable("secure");
	writel(0x5300, PMU_DAIF_EFUSE_CTRL); /* disable Vddq */
}

static int crypto_otp_do_task(unsigned long task)
{
	int num_retry;
	unsigned long clk;

	/* set clock source */
	clk = dvf99_get_main_clk();
	if (clk == 13824000)
		task |= CRYPTO_OTP_REQ_CLOCK_13_824;
	else if (clk == 25000000)
		task |= CRYPTO_OTP_REQ_CLOCK_25;
	else if (clk == 50000000)
		task |= CRYPTO_OTP_REQ_CLOCK_50;

	/* wait for controller ready */
	num_retry = 50;
	while (num_retry && (crypto_readl(CRYPTO_OTP_DEBUG) & CRYPTO_OTP_DEBUG_CNTL_READY) == 0) {
		num_retry--;
		mdelay(10);
	}
	if (!num_retry)
		return -1; // timeout

	// request opration
	crypto_writel(CRYPTO_OTP_PROGREQ, task);

	// kick the hardware
	crypto_writel(CRYPTO_OTP_PROGKICK, CRYPTO_OTP_KICK_MAGIC);

	// wait for hardware to finish
	num_retry = 50;
	while (num_retry && (crypto_readl(CRYPTO_OTP_PROGSTAT) & CRYPTO_OTP_STAT_COMPLETE) == 0) {
		num_retry--;
		mdelay(10);
	}
	if (!num_retry)
		return -1; // timeout

	if ((crypto_readl(CRYPTO_OTP_PROGSTAT) & CRYPTO_OTP_STAT_SUCCESS) == 0)
		return -1; // fail

	return 0;
}

int crypto_sha256(const unsigned char *message, unsigned long len,
		unsigned char *digest)
{
	struct cryptohw_desc *hw_desc;

	if (!len)
		return -1;

	hw_desc = (struct cryptohw_desc*)malloc(sizeof(struct cryptohw_desc));
	memset(hw_desc, 0, sizeof(struct cryptohw_desc));

	hw_desc->desc_ctl = CRYPTO_DESC_CTL_DESC_OWNER_HW | CRYPTO_DESC_CTL_MODE_SHA2_256;

	hw_desc->data_in_phys = (uint32_t)message;
	hw_desc->data_in_len = len;
	hw_desc->data_out_phys = (uint32_t)digest;
	hw_desc->data_out_len = 32;
	hw_desc->link_ptr = 0;

	crypto_writel(CRYPTO_REG_CHANNEL_HEADPTR, (unsigned long)hw_desc);
	crypto_writel(CRYPTO_REG_HEADPTR_SET, 1);

	/* wait for hw to finish */
	while (hw_desc->desc_ctl & CRYPTO_DESC_CTL_DESC_OWNER_HW)
		udelay(1);

	free(hw_desc);

	return 0;
}

int crypto_is_dev_secured()
{
	unsigned long reg;

	reg = crypto_readl(CRYPTO_OTP_STATE);
	if (reg & CRYPTO_STATE_ENFORCE_SB)
		return 1;

#ifdef CONFIG_DVF99_SECURE_BOOT_DEBUG
	/* For secure boot debugging, always say yes */
	return 1;
#endif

	return 0;
}

int crypto_otp_update_regs()
{
	int ret;

	crypto_init();

	/* update OTP state register */
	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_STATE |
				 CRYPTO_OTP_REQ_UPDATE |
				 CRYPTO_OTP_REQ_BANKSEL_0);
	if (ret != 0)
		return ret;

	/* update OTP device-id register */
	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_DEVICE_ID |
				 CRYPTO_OTP_REQ_UPDATE |
				 CRYPTO_OTP_REQ_BANKSEL_0);
	if (ret != 0)
		return ret;

	/* update OTP min sw version register */
	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_SW_VER |
				 CRYPTO_OTP_REQ_UPDATE |
				 CRYPTO_OTP_REQ_BANKSEL_0);
	if (ret != 0)
		return ret;

	/* update OTP pubkey registers */
	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_UPDATE |
				 CRYPTO_OTP_REQ_BANKSEL_1);
	if (ret != 0)
		return ret;

	return 0;
}

unsigned long crypto_otp_get_state()
{
	return crypto_readl(CRYPTO_OTP_STATE);
}

int crypto_otp_set_state(unsigned long val)
{
	int ret;

	crypto_init();

	// write the new state value
	crypto_writel(CRYPTO_KEYSTORE_79, val);

	/* enable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// enable the eFuse
	mdelay(20);
	writel(0x5303, PMU_DAIF_EFUSE_CTRL);	// enable the high power eFuse regulator
	mdelay(20);

	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_STATE |
				 CRYPTO_OTP_REQ_PROGRAM |
				 CRYPTO_OTP_REQ_BANKSEL_0);

	/* disable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// disable the high power eFuse regulator
	mdelay(20);
	writel(0x5300, PMU_DAIF_EFUSE_CTRL);	// disable the eFuse
	mdelay(20);

	if (ret != 0)
		return ret;

	return 0;
}

void crypto_otp_get_pubkey_digest(unsigned long pubkey_dgst[8])
{
	int i;

	/* read public key digest registers */
	for (i = 0; i < 8; i++)
		pubkey_dgst[i] = crypto_readl(CRYPTO_OTP_PUBKEY_DIGEST_0 + i*4);

#ifdef CONFIG_DVF99_SECURE_BOOT_DEBUG
  #define FAKE_PUBKEY 1
#endif
#if (FAKE_PUBKEY==1)
	/* test/keys/key1.pem */
	pubkey_dgst[0] = 0x6e794849;
	pubkey_dgst[1] = 0x71cb2df5;
	pubkey_dgst[2] = 0xc306844a;
	pubkey_dgst[3] = 0xd2012627;
	pubkey_dgst[4] = 0xf489b25b;
	pubkey_dgst[5] = 0x6555787a;
	pubkey_dgst[6] = 0x6fe622ea;
	pubkey_dgst[7] = 0xbb1745f6;
#elif (FAKE_PUBKEY==2)
	/* test/keys/key2.pem */
	pubkey_dgst[0] = 0xb248f411;
	pubkey_dgst[1] = 0xfad2f3bc;
	pubkey_dgst[2] = 0xafac3fe1;
	pubkey_dgst[3] = 0xe276423b;
	pubkey_dgst[4] = 0xcd36f02f;
	pubkey_dgst[5] = 0x093f0e53;
	pubkey_dgst[6] = 0xc3a3e943;
	pubkey_dgst[7] = 0x6e4dbbf7;
#elif (FAKE_PUBKEY==3)
	/* test/keys/key3.pem */
	pubkey_dgst[0] = 0xda11f4ca;
	pubkey_dgst[1] = 0x74b7d7a4;
	pubkey_dgst[2] = 0xf47e6440;
	pubkey_dgst[3] = 0xe9702922;
	pubkey_dgst[4] = 0x1bd6be73;
	pubkey_dgst[5] = 0xa5e1d7d2;
	pubkey_dgst[6] = 0x721cd5e3;
	pubkey_dgst[7] = 0x949b8471;
#elif (FAKE_PUBKEY==4)
	/* test/keys/key4.pem */
	pubkey_dgst[0] = 0x509fd4c2;
	pubkey_dgst[1] = 0xc9f5efc4;
	pubkey_dgst[2] = 0x64b0f29c;
	pubkey_dgst[3] = 0x54d8bd0a;
	pubkey_dgst[4] = 0xdbaeae19;
	pubkey_dgst[5] = 0x5076f408;
	pubkey_dgst[6] = 0x1f293e88;
	pubkey_dgst[7] = 0xee8f8a5c;
#elif (FAKE_PUBKEY==5)
	/* test/keys/key5.pem */
	pubkey_dgst[0] = 0xbfa8ce28;
	pubkey_dgst[1] = 0xfc5bf1e8;
	pubkey_dgst[2] = 0x691c3f31;
	pubkey_dgst[3] = 0x2bab5941;
	pubkey_dgst[4] = 0xec729189;
	pubkey_dgst[5] = 0x6c73be8b;
	pubkey_dgst[6] = 0x62f5592f;
	pubkey_dgst[7] = 0x3ba318cc;
#endif
}

int crypto_otp_set_pubkey_digest(unsigned long pubkey_dgst[8])
{
	int i;
	int ret;

	crypto_init();

	// abort if bank1(pubkey bank) already contains at least one bit burned
	if (crypto_readl(CRYPTO_OTP_DEBUG) & CRYPTO_OTP_DEBUG_BANK1_BURNED)
		return -1;

	// write the pubkey value
	for (i = 0; i < 8; i++)
		crypto_writel(CRYPTO_KEYSTORE_79 + i*4, pubkey_dgst[i]);

	/* enable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// enable the eFuse
	mdelay(20);
	writel(0x5303, PMU_DAIF_EFUSE_CTRL);	// enable the high power eFuse regulator
	mdelay(20);

	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_PROGRAM |
				 CRYPTO_OTP_REQ_BANKSEL_1);

	/* disable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// disable the high power eFuse regulator
	mdelay(20);
	writel(0x5300, PMU_DAIF_EFUSE_CTRL);	// disable the eFuse
	mdelay(20);

	if (ret != 0)
		return ret;

	return 0;
}

unsigned long crypto_otp_get_minver(void)
{
	return crypto_readl(CRYPTO_OTP_MINSWVER);
}

int crypto_otp_set_minver(unsigned long val)
{
	int ret;

	crypto_init();

	// write the new version value
	crypto_writel(CRYPTO_KEYSTORE_79, val);

	/* enable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// enable the eFuse
	mdelay(20);
	writel(0x5303, PMU_DAIF_EFUSE_CTRL);	// enable the high power eFuse regulator
	mdelay(20);

	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_SW_VER |
				 CRYPTO_OTP_REQ_PROGRAM |
				 CRYPTO_OTP_REQ_BANKSEL_0);

	/* disable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// disable the high power eFuse regulator
	mdelay(20);
	writel(0x5300, PMU_DAIF_EFUSE_CTRL);	// disable the eFuse
	mdelay(20);

	if (ret != 0)
		return ret;

	return 0;
}

unsigned long crypto_otp_get_devid(void)
{
	return crypto_readl(CRYPTO_OTP_DEVID);
}

int crypto_otp_set_devid(unsigned long val)
{
	int ret;

	crypto_init();

	// write the new version value
	crypto_writel(CRYPTO_KEYSTORE_79, val);

	/* enable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// enable the eFuse
	mdelay(20);
	writel(0x5303, PMU_DAIF_EFUSE_CTRL);	// enable the high power eFuse regulator
	mdelay(20);

	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_B0_OFFSET_DEVICE_ID |
				 CRYPTO_OTP_REQ_PROGRAM |
				 CRYPTO_OTP_REQ_BANKSEL_0);

	/* disable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// disable the high power eFuse regulator
	mdelay(20);
	writel(0x5300, PMU_DAIF_EFUSE_CTRL);	// disable the eFuse
	mdelay(20);

	if (ret != 0)
		return ret;

	return 0;
}

unsigned long crypto_otp_get_debug(void)
{
	return crypto_readl(CRYPTO_OTP_DEBUG);
}

int crypto_otp_gen_seckey(void)
{
	int ret;

	// abort if secret key was already burned
	if (crypto_readl(CRYPTO_OTP_DEBUG) & CRYPTO_OTP_DEBUG_SECKEY_BURNED)
		return -1;

	crypto_init();

	/* enable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// enable the eFuse
	mdelay(20);
	writel(0x5303, PMU_DAIF_EFUSE_CTRL);	// enable the high power eFuse regulator
	mdelay(20);

	ret = crypto_otp_do_task(CRYPTO_OTP_REQ_GEN_SECKEY |
				 CRYPTO_OTP_REQ_PROGRAM);

	/* disable programming voltage */
	writel(0x5301, PMU_DAIF_EFUSE_CTRL);	// disable the high power eFuse regulator
	mdelay(20);
	writel(0x5300, PMU_DAIF_EFUSE_CTRL);	// disable the eFuse
	mdelay(20);

	if (ret != 0)
		return ret;

	return 0;
}
