/*
 * (C) Copyright 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include "dspg_image.h"
#include "sboot.h"
#include "crypto_block.h"
#include "sha256.h"
#include "crc32.h"
#include <tommath.h>
#include <malloc.h>

int sboot_is_dev_secured(void)
{
	return crypto_is_dev_secured();
}

int sboot_verify_kernel(unsigned long *addr)
{
	struct dspg_image_header* hdr = (struct dspg_image_header*)(*addr);
	unsigned long kernel_addr;
	uint32_t crc;
	unsigned char sha256[32];
	unsigned long ret;

	/* check image header */
	crc = sec_crc32_le(~0u, (const unsigned char*)hdr, sizeof(struct dspg_image_header) - sizeof(uint32_t));
	if (crc != hdr->crc32) {
		// no signature header (probably a regular uImage)
		if (crypto_is_dev_secured()) {
			// never allow
			printf("sboot: device is secured, kernel image unsigned! (%d)\n", ERR_SBOOT_NOSIG);
			return ERR_SBOOT_NOSIG;
		}

		kernel_addr = *addr;
	} else {
		// sets address of real kernel image
		kernel_addr = *addr + hdr->load_offset;
	}

	if (!crypto_is_dev_secured()) {
		*addr = kernel_addr;
		return 0;
	}

	/* verify header signature */
	if (!(hdr->flags & HDRFLAG_SECURE)) {
		printf("sboot: no kernel signature (%d)\n", ERR_SBOOT_NOSIG);
		return ERR_SBOOT_NOSIG;
	}

	ret = sboot_check_hdr((void*)(*addr));
	if (ret != ERR_NONE) {
		printf("sboot: invalid kernel signature (%ld)\n", ret);
		return ret;
	}

	/* check kernel code digest */
	ret = crypto_sha256((const unsigned char*)kernel_addr, hdr->load_size, sha256);
	if (ret != ERR_NONE) {
		printf("sboot: invalid kernel code digest (%ld)\n", ret);
		return ret;
	}

	if (memcmp(sha256, hdr->sb_hdr.load_sha256, 32) != 0) {
		printf("sboot: invalid kernel code digest (%d)\n", ERR_SBOOT_INVIMG);
		return ERR_SBOOT_INVIMG;
	}

	*addr = kernel_addr;

	return 0;
}

/* (cb) = (mb)^(el) mod (nb) */
int expmod(unsigned char cb[256], uint8_t *mb, unsigned int mlen,
		unsigned long el, uint8_t *nb, unsigned int nlen)
{
	mp_int c, m, e, n;
	int size;
	int ret;

	memset(cb, 0, 256);

	mp_init(&c);
	mp_init(&m);
	mp_init(&e);
	mp_init(&n);

	/* set the numbers */
	if (mp_read_unsigned_bin(&m, mb, mlen) != 0)
		return -1;

	if (mp_set_int(&e, el) != 0)
		return -2;

	if (mp_read_unsigned_bin(&n, nb, nlen) != 0)
		return -3;

	/* c = m^e mod n */
	ret = s_mp_exptmod(&m, &e, &n, &c, 0);
	if (ret != 0)
		return ret;

	/* do the following to pad the binary
	   outcome with leading 0s */
	size = mp_unsigned_bin_size(&c);
	if (size > 256)
		return -5;
	cb += 256-size;

	if (mp_to_unsigned_bin(&c, cb) != 0)
		return -6;

	mp_clear(&c);
	mp_clear(&m);
	mp_clear(&e);
	mp_clear(&n);

	return 0;
}

/* rsa_verify
 *
 * return:
 * 0 - verification succeeded
 * !0 - verification failed
 * */
int rsa_verify(unsigned char *msg, unsigned int msglen,
	       unsigned char *sig, unsigned int siglen,
	       struct pubkey_t *pubkey)
{
	uint8_t sha256_digest[32];
	uint8_t sigref[256];
	uint8_t sigdec[256];
	int ret;

	/* decrypt signature */
	ret = expmod(sigdec, sig, siglen, pubkey->e, pubkey->n, 256);
	if (ret != 0)
		return ret;

	/* compute reference message signature */
	sha256(msg, msglen, sha256_digest);
	ret = EMSA_encode_sha256(sigref, sha256_digest);
	if (ret<0)
		return -2;

	/* compare decrypted and reference signatures */
	return memcmp(sigdec, sigref, 256);
}

unsigned long sboot_check_hdr(void *_hdr)
{
	struct dspg_image_header* hdr = _hdr;
	unsigned char hdr_pubkey_sha[32];
	unsigned long otp_pubkey_sha[8];
	int ret;
	size_t size;

	if (!(hdr->flags & HDRFLAG_SECURE))
		return ERR_SBOOT_NOSIG;

	/* check pulic key in header (digest of n+e) match value in OTP */
	sha256((unsigned char*)&(hdr->sb_hdr.pubkey),
			sizeof(hdr->sb_hdr.pubkey), hdr_pubkey_sha);
	crypto_otp_get_pubkey_digest(otp_pubkey_sha);

	if (memcmp(hdr_pubkey_sha, otp_pubkey_sha, 32) != 0)
		return ERR_SBOOT_INVPUBKEY;

	/* check header signature */
	size = (unsigned int)hdr->sb_hdr.signature - (unsigned int)hdr;
	ret = rsa_verify((uint8_t*)hdr, size, hdr->sb_hdr.signature,
			 sizeof(hdr->sb_hdr.signature),
			 &(hdr->sb_hdr.pubkey));
	if (ret != 0)
		return ERR_SBOOT_INVSIG;

	return ERR_NONE;
}

unsigned long sboot_check_img(unsigned char *img, size_t imglen,
					unsigned char expected_sha[32])
{
	unsigned char sha[32];

	sha256(img, imglen, sha);

	if (memcmp(expected_sha, sha, 32) != 0)
		return ERR_SBOOT_INVIMG;

	return ERR_NONE;
}
