/*
 * (C) Copyright 2015 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/serial.h>

#include "clock_internal.h"
#include "clock_regs.h"
#include "pll.h"

#define IF_CLK_RATE		13824000ul
#define REF_CLK_RATE		25000000ul
#define TDM1_CLK_RATE		25000000ul
#define TDM2_CLK_RATE		25000000ul
#define TDM3_CLK_RATE		25000000ul

static struct clk *
get_no_parent(const struct clk *clk)
{
	return NULL;
}

static unsigned long
get_parent_rate(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);

	if (!parent)
		return 0;
	return parent->get_rate(parent);
}

static int
clk_en(const struct clk *clk, int en);

static int
clk_is_enabled(const struct clk *clk);

static inline uint32_t
limit_div(const uint32_t div, const struct clk *clk)
{
	uint32_t max_div = ~(0xFFFFFFFF << clk->div.length);

	return div > max_div ? max_div : div;
}

static inline void
write_div(uint32_t div, const struct clk *clk)
{
	uint32_t reg = readl_cmu(clk->div.reg);

	reg &= ~(~(0xFFFFFFFF << clk->div.length) << clk->div.shift);
	reg |= div << clk->div.shift;

	writel_cmu(reg, clk->div.reg);
}

static int
set_rate(const struct clk *clk, unsigned long rate)
{
	const struct clk *parent = clk->get_parent(clk);
	uint32_t divider;

	if (clk->flags & OSC_IN || !parent)
		return -1;

	if (!rate)
		rate = 1;

	if (clk->flags & FIXED_DIV) {
		unsigned long cur_rate;

		/* clk could be a fixed divider */
		cur_rate = clk->get_rate(clk);
		divider = cur_rate ? (parent->get_rate(parent) / cur_rate) : 1;

		return parent->set_rate(parent, rate * divider);
	}

	divider = floor_div(parent->get_rate(parent), rate);
	if (!divider)
		divider = 1;
	if (clk->flags & EVEN_DIV && divider % 2)
		divider++;
	divider--;

	write_div(limit_div(divider, clk), clk);

	return 0;
}

#define CLK(_name,  _flags)						\
	struct clk _name = {						\
		.name = __stringify(_name),				\
		.flags = _flags,					\
		.is_enabled = clk_is_enabled,				\
		.enable = clk_en,					\
		.get_parent = get_no_parent,				\
		.get_rate = get_parent_rate,				\
		.set_rate = set_rate,					\
	}

#define GET_BIT(_reg, _shift)						\
	GET_BITS(_reg, _shift, 0x1)
#define GET_BITS(_reg, _shift, _mask)					\
	(((readl_cmu(_reg)) >> (_shift)) & (_mask))
#define GET_BITS2(PAIR, _mask)						\
	GET_BITS(PAIR, (_mask))

#define SET_BIT(_reg, _shift)						\
	writel_cmu(readl_cmu(_reg) | (1 << (_shift)), (_reg))
#define CLEAR_BIT(_reg, _shift)						\
	writel_cmu(readl_cmu(_reg) & ~(1 << (_shift)), (_reg))
#define INSERT_BIT(_reg, _shift, _val)					\
	((_val) ? SET_BIT(_reg, _shift) : CLEAR_BIT(_reg, _shift))

CLK(ref_clk,			OSC_IN | FIXED_DIV);
CLK(if_clk,			OSC_IN | FIXED_DIV);
CLK(tdm1_ext_in,		OSC_IN | FIXED_DIV);
CLK(tdm2_ext_in,		OSC_IN | FIXED_DIV);
CLK(tdm3_ext_in,		OSC_IN | FIXED_DIV);
CLK(pll1_bypass_sel,		FIXED_DIV);
CLK(pll2_bypass_sel,		FIXED_DIV);
CLK(pll3_bypass_sel,		FIXED_DIV);
CLK(pll3_ext_input_sel,		FIXED_DIV);
CLK(pll1_pre,			HAS_EN);
CLK(pll2_pre,			HAS_EN);
CLK(pll3_pre,			HAS_EN);
CLK(pll1,			FIXED_DIV);
CLK(pll2,			FIXED_DIV);
CLK(pll3,			FIXED_DIV);
CLK(pll4,			FIXED_DIV);
CLK(pll1_post,			HAS_EN);
CLK(pll2_post,			HAS_EN);
CLK(pll3_post,			HAS_EN);
CLK(sys_ahb_clk,		FIXED_DIV | IS_LEAF);
CLK(sys_apb_clk,		FIXED_DIV | IS_LEAF);
CLK(sys_rom_clk,		IS_LEAF);
CLK(css_pre,			HAS_EN);
CLK(css_cpu_clk,		HAS_EN | HAS_AHB | FIXED_DIV | IS_LEAF);
CLK(css_ahb_div,		0);
CLK(css_ahb_clk,		FIXED_DIV | IS_LEAF);
CLK(css_apb_clk,		FIXED_DIV | IS_LEAF);
CLK(css_adpcm,			HAS_AHB | IS_LEAF); /* TODO check CSS AHB/APB code */
CLK(css_gdmac_clk,		HAS_APB | HAS_AHB | IS_LEAF);
CLK(css_timer12_apb,		HAS_APB | IS_LEAF);
CLK(css_timer1,			IS_LEAF);
CLK(css_timer2,			IS_LEAF);
CLK(rtc_clk,			HAS_APB | IS_LEAF);
CLK(wd_clk,			HAS_APB | IS_LEAF);
CLK(cs_clk,			HAS_APB | IS_LEAF);
CLK(mcu_div,			0);
CLK(axi_sys_div,		EVEN_DIV);
CLK(axi_ddr_div,		EVEN_DIV);
CLK(axi_ddr_clk,		FIXED_DIV | IS_LEAF);
CLK(apu_clk_sel,		FIXED_DIV);
CLK(apu_ripple_div,		0);
CLK(rom_ahb_fixed,		FIXED_DIV);
CLK(axi_sys_clk,		FIXED_DIV | IS_LEAF);
CLK(dsi_clk,			HAS_APB | IS_LEAF);
CLK(apu_pre,			0);
CLK(apu_clk,			FIXED_DIV | IS_LEAF);
CLK(uart1_clk,			HAS_APB | IS_LEAF);
CLK(uart2_clk,			HAS_APB | IS_LEAF);
CLK(uart3_clk,			HAS_APB | IS_LEAF);
CLK(uart4_clk,			HAS_APB | IS_LEAF);
CLK(nfc_clk,			HAS_APB | HAS_AHB | IS_LEAF);
CLK(daif_clk,			HAS_APB | IS_LEAF);
CLK(pwm_gen_clk,		HAS_AHB | IS_LEAF);
CLK(crypto_clk,			HAS_APB | HAS_AHB | IS_LEAF);
CLK(mcu_gdmac_clk,		HAS_APB | HAS_AHB | IS_LEAF);
CLK(sec_clk,			HAS_AHB | IS_LEAF);
CLK(i2c1_clk,			HAS_APB | IS_LEAF);
CLK(i2c2_clk,			HAS_APB | IS_LEAF);
CLK(gpio_clk,			HAS_APB | IS_LEAF);
CLK(timer12_apb,		HAS_APB | IS_LEAF);
CLK(timer1,			IS_LEAF);
CLK(timer2,			IS_LEAF);
CLK(timer34_apb,		HAS_APB | IS_LEAF);
CLK(timer3,			IS_LEAF);
CLK(timer4,			IS_LEAF);
CLK(ccu_clk,			HAS_APB | HAS_AHB | IS_LEAF);
CLK(dcls_clk,			HAS_AHB | IS_LEAF);
CLK(bmp_clk,			HAS_AHB | FIXED_DIV | IS_LEAF);
CLK(drt_clk,			HAS_AHB | FIXED_DIV | IS_LEAF);
CLK(bmp_pre,			0);
CLK(clk_out_clk,		0); /* TODO */
CLK(emmc_pre,			0);
CLK(emmc_int,			FIXED_DIV);
CLK(emmc_clk,			HAS_AHB | FIXED_DIV | IS_LEAF);
CLK(sdio_pre,			0);
CLK(sdio_int,			FIXED_DIV);
CLK(sdio_clk,			HAS_AHB | FIXED_DIV | IS_LEAF);
CLK(refdiv_clk,			IS_LEAF);
CLK(eth_pre,			EVEN_DIV);
CLK(eth_chip_clk,		FIXED_DIV | IS_LEAF);
CLK(eth_mac_clk,		OSC_IN | FIXED_DIV | HAS_AXI); /* for enabling only */
CLK(gpu_clk,			IS_LEAF);
CLK(lcdc_clk,			HAS_AXI | IS_LEAF);
CLK(spi1_clk,			IS_LEAF);
CLK(spi2_clk,			FIXED_DIV | IS_LEAF);
CLK(spi3_clk,			IS_LEAF);
CLK(tdm1_clk,			HAS_APB | IS_LEAF);
CLK(tdm2_clk,			HAS_APB | IS_LEAF);
CLK(tdm3_clk,			HAS_APB | IS_LEAF);

/* TODO USB clocks */

struct clk *clocks[] = {
	&css_pre,
	&css_cpu_clk,
	&css_ahb_div,
	&css_ahb_clk,
	&css_apb_clk,
	&css_adpcm,
	&css_gdmac_clk,
	&css_timer12_apb,
	&css_timer1,
	&css_timer2,
	&uart1_clk,
	&uart2_clk,
	&uart3_clk,
	&uart4_clk,
	&nfc_clk,
	&daif_clk,
	&pwm_gen_clk,
	&crypto_clk,
	&mcu_gdmac_clk,
	&sec_clk,
	&i2c1_clk,
	&i2c2_clk,
	&gpio_clk,
	&timer12_apb,
	&timer1,
	&timer2,
	&timer34_apb,
	&timer3,
	&timer4,
	&ccu_clk,
	&rtc_clk,
	&wd_clk,
	&bmp_clk,
	&drt_clk,
	&bmp_pre,
	&sys_ahb_clk,
	&sys_apb_clk,
	&sys_rom_clk,
	&axi_sys_clk,
	&emmc_pre,
	&emmc_int,
	&emmc_clk,
	&sdio_pre,
	&sdio_int,
	&sdio_clk,
	&apu_clk,
	&dcls_clk,
	&tdm1_clk,
	&tdm2_clk,
	&tdm3_clk,
	&cs_clk,
	&refdiv_clk,
	&eth_pre,
	&eth_chip_clk,
	&eth_mac_clk,
	&gpu_clk,
	&lcdc_clk,
	&dsi_clk,
	&spi1_clk,
	&spi2_clk,
	&spi3_clk,
	&clk_out_clk, /* TODO */
	&pll1,
	&pll2,
	&pll3,
	&pll1_pre,
	&pll2_pre,
	&pll3_pre,
	&pll1_post,
	&pll2_post,
	&pll3_post,
	&apu_clk_sel,
	&apu_pre,
	&mcu_div,
	&if_clk,
	&axi_sys_div,
	&axi_ddr_div,
	&axi_ddr_clk,
	&ref_clk,
	&pll3_ext_input_sel,
	&pll1_bypass_sel,
	&pll2_bypass_sel,
	&pll3_bypass_sel,
	&apu_ripple_div,
	&tdm1_ext_in,
	&tdm2_ext_in,
	&tdm3_ext_in,
	NULL,
};

static struct clk *clk_out_parents[] = {
	NULL,			/* 0 */
	NULL,			/* 1 */
	NULL,			/* 2 */
	NULL,			/* 3 */
	NULL,			/* 4 */
	NULL,			/* 5 */
	NULL,			/* 6 */
	NULL,			/* 7 */
	NULL,			/* 8 */
	NULL,			/* 9 */
	NULL,			/* 10 */
	NULL,			/* 11 */
	NULL,			/* 12 */
	NULL,			/* 13 */
	NULL,			/* 14 */
	NULL,			/* 15 */
	NULL,			/* 16 */
	NULL,			/* 17 */
	NULL,			/* 18 */
	NULL,			/* 19 */
	&pll1_post,		/* 20 */
	&pll2_post,		/* 21 */
	&pll3_post,		/* 22 */
	NULL,			/* 23 */
	NULL,			/* 24 */
	&axi_ddr_clk,		/* 25 */
	&axi_sys_clk,		/* 26 */
	&sys_ahb_clk,		/* 27 */
	&sys_apb_clk,		/* 28 */
	&rom_ahb_fixed,		/* 29 */
	NULL,			/* 30 */
	NULL,			/* 31 */
	&css_cpu_clk,		/* 32 */
	&css_ahb_clk,		/* 33 */
	&css_apb_clk,		/* 34 */
	&bmp_clk,		/* 35 */
	NULL,			/* 36 */
	&tdm1_ext_in,		/* 37 TODO confirm */
	&tdm2_ext_in,		/* 38 TODO confirm */
	&tdm3_ext_in,		/* 39 TODO confirm */
	&rtc_clk,		/* 40 */
	&wd_clk,		/* 41 */
	&css_adpcm,		/* 42 */
	/* TODO */
};

static struct clk *
clk_out_get_parent(const struct clk *clk)
{
	return clk_out_parents[GET_BITS(clk->mux.reg, clk->mux.shift, 0x3F)];
}

static struct clk *
get_single_parent(const struct clk *clk)
{
	return clk->parents[0];
}

static struct clk *
get_mux_parent(const struct clk *clk) {
	return clk->parents[GET_BITS(clk->mux.reg, clk->mux.shift, 0x3)];
}

struct clk_parents {
	struct clk *clk;

	struct clk* (*get_parent)(const struct clk *clk);
	struct clk_bit mux;
	struct clk *parents[4];
};

static struct clk_parents relations[] = {
	{ &pll1_bypass_sel, get_mux_parent, { PLL1_BYPASS_SEL },
	  { &ref_clk, NULL, &if_clk } },
	{ &pll2_bypass_sel, get_mux_parent, { PLL2_BYPASS_SEL },
	  { &ref_clk, NULL, &if_clk } },
	{ &pll3_bypass_sel, get_mux_parent, { PLL3_BYPASS_SEL },
	  { &ref_clk, NULL, &if_clk } },
	{ &pll1_pre, get_mux_parent, { PLL1_INPUT_SEL },
	  { &ref_clk, NULL, &if_clk } },
	{ &pll2_pre, get_mux_parent, { PLL2_INPUT_SEL },
	  { &ref_clk, NULL, &if_clk } },
	{ &pll3_ext_input_sel, get_mux_parent, { PLL3_EXT_INPUT_SEL },
	  { &tdm1_ext_in, &tdm2_ext_in, &tdm3_ext_in } },
	{ &pll3_pre, get_mux_parent, { PLL3_INPUT_SEL },
	  { &ref_clk, NULL, &if_clk, &pll3_ext_input_sel } },

	{ &pll1, get_single_parent, { }, { &pll1_pre } },
	{ &pll2, get_single_parent, { }, { &pll2_pre } },
	{ &pll3, get_single_parent, { }, { &pll3_pre } },
	{ &pll4, get_single_parent, { }, { &ref_clk } },

	{ &pll1_post, get_mux_parent, { PLL1_CLK_IN_USE },
	  { &pll1_bypass_sel, &pll1, &pll1_bypass_sel, &pll1 } },
	{ &pll2_post, get_mux_parent, { PLL2_CLK_IN_USE },
	  { &pll2_bypass_sel, &pll2, &pll2_bypass_sel, &pll2 } },
	{ &pll3_post, get_mux_parent, { PLL3_CLK_IN_USE },
	  { &pll3_bypass_sel, &pll3, &pll3_bypass_sel, &pll3 } },

	{ &rtc_clk, get_single_parent, { }, { &ref_clk } },
	{ &wd_clk, get_single_parent, { }, { &ref_clk } },

	{ &cs_clk, get_single_parent, { }, { &pll1_post } },
	{ &mcu_div, get_single_parent, { }, { &pll1_post } },

	{ &uart1_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &uart2_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &uart3_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &uart4_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &nfc_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &daif_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &pwm_gen_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &crypto_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &mcu_gdmac_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &sec_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &i2c1_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &i2c2_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &gpio_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &timer12_apb, get_single_parent, { }, { &axi_sys_div } },
	{ &timer1, get_single_parent, { }, { &timer12_apb } },
	{ &timer2, get_single_parent, { }, { &timer12_apb } },
	{ &timer34_apb, get_single_parent, { }, { &axi_sys_div } },
	{ &timer3, get_single_parent, { }, { &timer34_apb } },
	{ &timer4, get_single_parent, { }, { &timer34_apb } },
	{ &ccu_clk, get_single_parent, { }, { &axi_sys_div } },

	{ &axi_sys_div, get_single_parent, { }, { &mcu_div } },
	{ &axi_ddr_div, get_single_parent, { }, { &mcu_div } },
	{ &axi_ddr_clk, get_single_parent, { }, { &axi_ddr_div } },
	{ &axi_sys_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &sys_ahb_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &sys_apb_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &sys_rom_clk, get_single_parent, { }, { &axi_sys_div } },

	{ &apu_clk_sel, get_mux_parent, { APU_CLK_SEL },
	  { &apu_ripple_div, &apu_pre, &apu_ripple_div, &apu_pre } },
	{ &apu_ripple_div, get_single_parent, { }, { &ref_clk } },
	{ &rom_ahb_fixed, get_single_parent, { }, { &mcu_div } },

	{ &clk_out_clk, clk_out_get_parent, { CLK_OUT_INPUT_SEL }, { } },
	{ &bmp_clk, get_single_parent, { }, { &bmp_pre } },
	{ &drt_clk, get_single_parent, { }, { &bmp_pre } },
	{ &bmp_pre, get_mux_parent, { BMP_INPUT_SEL },
	  { &if_clk, &apu_clk_sel, /* reserved */ } },
	{ &apu_pre, get_mux_parent, { APU_DIV_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &apu_clk, get_single_parent, { }, { &apu_clk_sel } },

	{ &css_pre, get_mux_parent, { CSS_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post, &ref_clk } },
	{ &css_cpu_clk, get_single_parent, { }, { &css_pre } },
	{ &css_ahb_div, get_single_parent, { }, { &css_pre } },
	{ &css_ahb_clk, get_single_parent, { }, { &css_ahb_div } },
	{ &css_apb_clk, get_single_parent, { }, { &css_ahb_div } },
	{ &css_adpcm, get_single_parent, { }, { &css_ahb_div } },
	{ &css_gdmac_clk, get_single_parent, { }, { &css_ahb_div } },
	{ &css_timer12_apb, get_single_parent, { }, { &css_ahb_div } },
	{ &css_timer1, get_single_parent, { }, { &css_timer12_apb } },
	{ &css_timer2, get_single_parent, { }, { &css_timer12_apb } },
	/************************* 'regular' blocks **************************/
	{ &dcls_clk, get_mux_parent, { DCLS_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &tdm1_clk, get_mux_parent, { TDM1_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &tdm2_clk, get_mux_parent, { TDM2_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &tdm3_clk, get_mux_parent, { TDM3_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &sdio_pre, get_mux_parent, { SDIO_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &sdio_int, get_single_parent, { }, { &sdio_pre } },
	{ &sdio_clk, get_single_parent, { }, { &sdio_int } },
	{ &emmc_pre, get_mux_parent, { EMMC_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &emmc_int, get_single_parent, { }, { &emmc_pre } },
	{ &emmc_clk, get_single_parent, { }, { &emmc_int } },
	{ &refdiv_clk, get_mux_parent, { REFDIV_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &eth_pre, get_mux_parent, { ETH_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &eth_chip_clk, get_single_parent, { }, { &eth_pre } },
	{ &gpu_clk, get_mux_parent, { GPU_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &lcdc_clk, get_mux_parent, { LCDC_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &spi1_clk, get_mux_parent, { SPI1_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	{ &spi2_clk, get_single_parent, { }, { &axi_sys_div } },
	{ &spi3_clk, get_mux_parent, { SPI3_INPUT_SEL },
	  { &pll1_post, &pll2_post, &pll3_post } },
	/*********************************************************************/
	{ &dsi_clk, get_single_parent, { }, { &ref_clk } },

	{ /* sentinel */ },
};

static void
clk_init_parent_funcs(void)
{
	unsigned j, i = 0;
	struct clk_parents *funcs = &relations[i];
	struct clk *clk = funcs->clk;

	while (clk) {
		clk->get_parent = funcs->get_parent;
		clk->mux = funcs->mux;
		for (j = 0; j < 4; j++)
			clk->parents[j] = funcs->parents[j];

		funcs = &relations[++i];
		clk = funcs->clk;
	}
}

static unsigned long
fixed_divider_16(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);

	if (!parent)
		return 0;
	return parent->get_rate(parent) / 16;
}

static unsigned long
fixed_divider_2(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);

	if (!parent)
		return 0;
	return parent->get_rate(parent) / 2;
}

static unsigned long
div_get_rate(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);
	uint32_t divider;

	if (!parent)
		return 0;

	divider = GET_BITS(clk->div.reg, clk->div.shift,
			   ~(0xFFFFFFFF << clk->div.length)) + 1;

	return parent->get_rate(parent) / divider;
}

static unsigned long
bmp_pre_get_rate(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);
	uint32_t divider;

	if (!parent)
		return 0;

	divider = GET_BITS(clk->div.reg, clk->div.shift,
			   ~(0xFFFFFFFF << clk->div.length)) + 1;
	if (divider != 1 && divider != 12)
		return 0;

	return parent->get_rate(parent) / divider;
}

#define ABS(a, b)	(((a) < (b)) ? ((b) - (a)) : ((a) - (b)))

static int
bmp_pre_set_rate(const struct clk *clk, unsigned long rate)
{
	const struct clk *parent = clk->get_parent(clk);
	uint32_t divider;
	unsigned long match1, match2;

	if (!parent)
		return -1;

	match1 = parent->get_rate(parent) / 1;
	match2 = parent->get_rate(parent) / 12;

	divider = (ABS(match1, rate) < ABS(match2, rate)) ? 0 : 0xB;

	write_div(divider, clk);

	return 0;
}

struct fixed_rate {
	char *name;
	unsigned long rate;
};

static struct fixed_rate rates[] = {
	{ "ref_clk", REF_CLK_RATE },
	{ "if_clk", IF_CLK_RATE },
	{ "tdm1_ext_in", TDM1_CLK_RATE },
	{ "tdm2_ext_in", TDM2_CLK_RATE },
	{ "tdm3_ext_in", TDM3_CLK_RATE },

	{ /* sentinel */ },
};

static unsigned long
get_fixed_rate(const struct clk *clk)
{
	unsigned i;

	for (i = 0; rates[i].name && strcmp(rates[i].name, clk->name); i++)
		;

	return rates[i].rate;
}

struct clk_rate_funcs {
	struct clk *clk;

	unsigned long (*get_rate)(const struct clk *clk);
	struct clk_div div;
	int (*set_rate)(const struct clk *clk, unsigned long rate);
};

static struct clk_rate_funcs rate_funcs[] = {
	{ &ref_clk, get_fixed_rate, { }, },
	{ &if_clk, get_fixed_rate, { }, },
	{ &tdm1_ext_in, get_fixed_rate, { }, },
	{ &tdm2_ext_in, get_fixed_rate, { }, },
	{ &tdm3_ext_in, get_fixed_rate, { }, },
	{ &pll1_pre, div_get_rate, { PLL1_PRE_DIV_VAL }, },
	{ &pll2_pre, div_get_rate, { PLL2_PRE_DIV_VAL }, },
	{ &pll3_pre, div_get_rate, { PLL3_PRE_DIV_VAL }, },
	{ &pll1, get_pll_rate, { }, set_pll_rate },
	{ &pll2, get_pll_rate, { }, set_pll_rate },
	{ &pll3, get_pll_rate, { }, set_pll_rate },
	{ &pll4, get_pll_rate, { }, set_pll_rate },
	{ &pll1_post, div_get_rate, { PLL1_POST_DIV_VAL }, },
	{ &pll2_post, div_get_rate, { PLL2_POST_DIV_VAL }, },
	{ &pll3_post, div_get_rate, { PLL3_POST_DIV_VAL }, },
	{ &rtc_clk, div_get_rate, { RTC_DIV_VAL }, },
	{ &wd_clk, div_get_rate, { WD_DIV_VAL }, },
	{ &cs_clk, div_get_rate, { CS_DIV_VAL }, },
	{ &mcu_div, div_get_rate, { MCU_MAIN_DIV_VAL }, },
	{ &axi_sys_div, div_get_rate, { MCU_AXI_SYS_DIV_VAL }, },
	{ &sys_rom_clk, fixed_divider_2, { }, },
	{ &axi_ddr_div, div_get_rate, { MCU_AXI_DDR_DIV_VAL }, },
	{ &clk_out_clk, div_get_rate, { CLK_OUT_DIV_VAL }, },
	{ &bmp_pre, bmp_pre_get_rate, { BMP_DIV_VAL }, bmp_pre_set_rate },
	{ &apu_pre, div_get_rate, { APU_DIV_VAL }, },

	{ &css_pre, div_get_rate, { CSS_MAIN_DIV_VAL }, },
	{ &css_adpcm, div_get_rate, { ADPCM_DIV_VAL }, },
	{ &css_ahb_div, div_get_rate, { CSS_AHB_DIV_VAL }, },
	/************************* 'regular' blocks **************************/
	{ &dcls_clk, div_get_rate, { DCLS_DIV_VAL }, },
	{ &tdm1_clk, div_get_rate, { TDM1_DIV_VAL }, },
	{ &tdm2_clk, div_get_rate, { TDM2_DIV_VAL }, },
	{ &tdm3_clk, div_get_rate, { TDM3_DIV_VAL }, },
	{ &sdio_int, div_get_rate, { SDIO_INT_DIV_VAL }, },
	{ &sdio_pre, div_get_rate, { SDIO_MAIN_DIV_VAL }, },
	{ &emmc_int, div_get_rate, { EMMC_INT_DIV_VAL }, },
	{ &emmc_pre, div_get_rate, { EMMC_MAIN_DIV_VAL }, },
	{ &refdiv_clk, div_get_rate, { REFDIV_DIV_VAL }, },
	{ &eth_pre, div_get_rate, { ETH_MAIN_DIV_VAL }, },
	{ &gpu_clk, div_get_rate, { GPU_DIV_VAL }, },
	{ &lcdc_clk, div_get_rate, { LCDC_DIV_VAL }, },
	{ &spi1_clk, div_get_rate, { SPI1_DIV_VAL }, },
	{ &spi3_clk, div_get_rate, { SPI3_DIV_VAL }, },
	/*********************************************************************/

	{ &apu_ripple_div, fixed_divider_16, { }, },
	{ &rom_ahb_fixed, fixed_divider_2, { }, },

	{ /* sentinel */ },
};

static void
clk_init_rate_funcs(void)
{
	unsigned i = 0;
	struct clk_rate_funcs *funcs = &rate_funcs[i];
	struct clk *clk = funcs->clk;

	while (clk) {
		if (funcs->get_rate)
			clk->get_rate = funcs->get_rate;
		clk->div = funcs->div;
		if (funcs->set_rate)
			clk->set_rate = funcs->set_rate;

		funcs = &rate_funcs[++i];
		clk = funcs->clk;
	}
}

static int
clk_is_enabled(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);

	return (clk->en_stat_mask & readl_cmu(clk->en_reg)) ==
	        clk->en_stat_mask &&
	       (parent ? parent->is_enabled(parent) : 1);
}

static int
clk_en(const struct clk *clk, int en)
{
	const struct clk *parent = clk->get_parent(clk);

	if (parent && en && !parent->is_enabled(parent))
		parent->enable(parent, 1);

	if (((clk->en_stat_mask & readl_cmu(clk->en_reg)) ==
	     clk->en_stat_mask) != !!en) {
		if (en)
			writel_cmu(readl_cmu(clk->en_reg) | clk->en_mask,
				   clk->en_reg);
		else
			writel_cmu(readl_cmu(clk->en_reg) & ~clk->en_mask,
				   clk->en_reg);
	}

	return 0;
}

struct clk_data {
	struct clk *clk;

	uint32_t	en_reg;
	uint32_t	en_mask;
	uint32_t	en_stat_mask;
	int (*is_enabled)(const struct clk *clk);
	int (*enable)(const struct clk *clk, int en);
};

static struct clk_data clk_config[] = {
	{ &pll1_pre, PLL_BASE(1), (1 << 0), (1 << 14), },
	{ &pll2_pre, PLL_BASE(2), (1 << 0), (1 << 14), },
	{ &pll3_pre, PLL_BASE(3), (1 << 0), (1 << 14), },
	{ &pll1, 0, 0, 0, pll_is_enabled, pll_enable, },
	{ &pll2, 0, 0, 0, pll_is_enabled, pll_enable, },
	{ &pll3, 0, 0, 0, pll_is_enabled, pll_enable, },
	{ &pll1_post, PLL_BASE(1), (1 << 1), (1 << 15), },
	{ &pll2_post, PLL_BASE(2), (1 << 1), (1 << 15), },
	{ &pll3_post, PLL_BASE(3), (1 << 1), (1 << 15), },
	{ &mcu_div, GCU_MCU_CTRL, 0/*FIXME*/, (1 << 17), },
	{ &sys_rom_clk, GCU_MCU_CTRL, (1 << 11), (1 << 11), },
	{ &cs_clk, GCU_CS_CTRL, (1 << 4), (1 << 4), },
	{ &axi_sys_div, GCU_MCU_CTRL, 0/*FIXME*/, (1 << 21), },
	{ &axi_ddr_div, GCU_MCU_CTRL, 0/*FIXME*/, (1 << 20), },
	{ &axi_ddr_clk, GCU_MCU_CTRL, 0/*FIXME*/, (1 << 4), },
	{ &apu_ripple_div, GCU_APU_CTRL, (1 << 11), (1 << 11), },
	{ &clk_out_clk, GCU_CLK_OUT_CTRL, (1 << 6), (1 << 7), },
	{ &bmp_pre, GCU_BMP_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 7), },
	{ &css_cpu_clk, GCU_CSS_CTRL, (1 << 1) | (1 << 6), (1 << 1) | (1 << 6), },
	{ &css_ahb_div, GCU_CSS_CTRL, 0/*FIXME*/, (1 << 22), },
	{ &css_pre, GCU_CSS_CTRL, 0/*FIXME*/, (1 << 17), },
	{ &css_adpcm, GCU_ADPCM_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 7), },
	{ &dcls_clk, GCU_DCLS_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 7), },
	{ &tdm1_clk, GCU_TDM1_CTRL, (1 << 4) | (1 << 6), (1 << 4) | (1 << 7), },
	{ &tdm2_clk, GCU_TDM2_CTRL, (1 << 4) | (1 << 6), (1 << 4) | (1 << 7), },
	{ &tdm3_clk, GCU_TDM3_CTRL, (1 << 4) | (1 << 6), (1 << 4) | (1 << 7), },
	{ &apu_pre, GCU_APU_CTRL, (1 << 6), (1 << 7), },
	{ &refdiv_clk, GCU_REFDIV_CTRL, (1 << 6), (1 << 7), },
	{ &eth_pre, GCU_ETH_CTRL, 0/*FIXME*/, (1 << 7), },
	{ &eth_chip_clk, GCU_ETH_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 6), },
	{ &eth_mac_clk, GCU_ETH_CTRL, (1 << 19), (1 << 19), },
	{ &gpu_clk, GCU_GPU_CTRL, (1 << 6), (1 << 7), },
	{ &lcdc_clk, GCU_LCDC_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 7), },
	{ &spi1_clk, GCU_SPI1_CTRL, (1 << 6), (1 << 7), },
	{ &spi2_clk, GCU_SPI2_CTRL, (1 << 4), (1 << 4), },
	{ &spi3_clk, GCU_SPI3_CTRL, (1 << 6), (1 << 7), },
	{ &sdio_clk, GCU_SDIO_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 6), },
	{ &sdio_int, GCU_SDIO_CTRL, (1 << 4), (1 << 4), },
	{ &sdio_pre, GCU_SDIO_CTRL, 0/*FIXME*/, (1 << 7), },
	{ &emmc_clk, GCU_EMMC_CTRL, (1 << 5) | (1 << 6), (1 << 5) | (1 << 6), },
	{ &emmc_int, GCU_EMMC_CTRL, (1 << 4), (1 << 4), },
	{ &emmc_pre, GCU_EMMC_CTRL, 0/*FIXME*/, (1 << 7), },
	{ &rtc_clk, GCU_RTC_CTRL, (1 << 4) | (1 << 6), (1 << 4) | (1 << 7), },
	{ &wd_clk, GCU_WD_CTRL, (1 << 4) | (1 << 6), (1 << 4) | (1 << 7), },
	{ &nfc_clk, GCU_NFC_CTRL, (1 << 4) | (1 << 5), (1 << 4) | (1 << 5), },
	{ &daif_clk, GCU_DAIF_CTRL, (1 << 4), (1 << 4) },
	{ &timer12_apb, GCU_MCU_TIMER12_CTRL, (1 << 4), (1 << 4), },
	{ &timer34_apb, GCU_MCU_TIMER34_CTRL, (1 << 4), (1 << 4), },
	{ &timer1, GCU_MCU_TIMER12_CTRL, (1 << 10), (1 << 10), },
	{ &timer2, GCU_MCU_TIMER12_CTRL, (1 << 11), (1 << 11), },
	{ &timer3, GCU_MCU_TIMER34_CTRL, (1 << 10), (1 << 10), },
	{ &timer4, GCU_MCU_TIMER34_CTRL, (1 << 11), (1 << 11), },
	{ &uart1_clk, GCU_UART1_CTRL, (1 << 4), (1 << 4), },
	{ &uart2_clk, GCU_UART2_CTRL, (1 << 4), (1 << 4), },
	{ &uart3_clk, GCU_UART3_CTRL, (1 << 4), (1 << 4), },
	{ &uart4_clk, GCU_UART4_CTRL, (1 << 4), (1 << 4), },

	{ /* sentinel */ },
};

static void
clk_init_enable_funcs(void)
{
	unsigned i = 0;
	struct clk_data *data = &clk_config[i];
	struct clk *clk = data->clk;

	while (clk) {
		if (data->is_enabled)
			clk->is_enabled = data->is_enabled;
		if (data->enable)
			clk->enable = data->enable;

		clk->en_reg = data->en_reg;
		clk->en_mask = data->en_mask;
		clk->en_stat_mask = data->en_stat_mask;

		data = &clk_config[++i];
		clk = data->clk;
	}
}

uint32_t
sdmmc_get_phase(int have_tx, int is_sd)
{
	uint32_t addr = is_sd ? GCU_SDIO_CTRL : GCU_EMMC_CTRL;

	return (readl_cmu(addr) >> (have_tx ? 16 : 21)) & 0x1F;
}

void
sdmmc_set_phase(int have_tx, int is_sd, uint32_t shift)
{
	uint32_t addr = is_sd ? GCU_SDIO_CTRL : GCU_EMMC_CTRL;
	/* struct clk *clk = is_sd ? &sdio_clk : &emmc_clk; */ /* D:[Grandstream]:fix warning */
	/* unsigned long clk_rate = clk->get_rate(clk); */ /* D:[Grandstream]:fix warning */
	u32 reg;

	/* disable phase clock */
	writel_cmu(readl_cmu(addr) & ~(1 << (have_tx ? 11 : 12)), addr);
	udelay(1000);

	reg = readl_cmu(addr);
	reg &= ~(0x1F << (have_tx ? 16 : 21));
	reg |= (shift & 0x1F) << (have_tx ? 16 : 21);
	writel_cmu(reg, addr);

	udelay(1000);
	/* enable phase clock */
	writel_cmu(readl_cmu(addr) | (1 << (have_tx ? 11 : 12)), addr);
}

static int
check_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	switch (mode) {
	case ETH_MII:
	case ETH_MII_PHY:
	case ETH_RMII:
	case ETH_RMII_EXT:
		if (speed == ETH_SPEED_10 || speed == ETH_SPEED_100)
			return 0;
		break;
	case ETH_RGMII:
		if (speed == ETH_SPEED_1000)
			return 0;
		break;
	default:
		break;
	};

	return -1;
}

struct eth_mode_settings {
	uint8_t tx_clk_sel;
	uint8_t rx_clk_sel;
	uint8_t tx_pad_src_sel;
	uint8_t tx_pad_oen;
	uint8_t rx_pad_oen;
	uint8_t int_eth_div_en;
	uint8_t int_div_src_sel;
	uint8_t phase_shift_en;
	uint8_t tx_clk_phase_sel;
	uint8_t rmii_if_clk_en;
	uint8_t int_div_pol;
	int ethint_rate[NUM_ETH_SPEEDS];
};

/* Table 2-11 from CMU specification */
struct eth_mode_settings eth_clk_settings[NUM_ETH_MODES] = {
	/* MII (MAC mode) */
	{0, 0, 00, 1,  1, 0, 00, 0, 00, 0, 0, {2500000, 25000000, -1} },
	/* MII (PHY mode) */
	{0, 0,  1, 0,  0, 1,  0, 0, 00, 0, 0, {2500000, 25000000, -1} },
	/* RMII (internal) */
	{1, 1,  0, 0, 00, 1,  1, 1,  0, 1, 1, {2500000, 25000000, -1} },
	/* RMII (external) */
	{1, 1, 00, 1, 00, 1,  1, 0, 00, 1, 1, {2500000, 25000000, -1} },
	/* RGMII */
	{1, 0,  0, 0,  1, 1,  0, 1,  5, 0, 0, {-1, -1, 125000000} },
};

unsigned long eth_div_rates[] = {
	25000000,	/* MII (MAC mode) */
	25000000,	/* MII (PHY mode) */
	50000000,	/* RMII (internal) */
	50000000,	/* RMII (external) */
	125000000,	/* RGMII */
};

int
clk_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	int ret;
	unsigned long rate = eth_div_rates[mode];
	struct eth_mode_settings *settings;

	if (check_eth_config(mode, speed))
		return -1;

	settings = &eth_clk_settings[mode];

	reset_release("eth_chip_clk");

	if (!settings->int_div_src_sel || !settings->tx_pad_src_sel) {
		INSERT_BIT(GCU_ETH_CTRL, 11, settings->int_div_pol);
		INSERT_BIT(GCU_ETH_CTRL, 10, settings->int_div_src_sel);
		clk_disable("eth_chip_clk");

		ret = clk_set_rate("eth_chip_clk", rate);
		if (ret) {
			printf("error: cannot set eth chip clock rate to %lu\n",
			       rate);
			return ret;
		}

		clk_enable("eth_chip_clk");
	}

	INSERT_BIT(GCU_ETH_CTRL, 20, settings->rmii_if_clk_en);
	INSERT_BIT(GCU_ETH_CTRL, 13, settings->phase_shift_en);
	INSERT_BIT(GCU_ETH_CTRL, 16, settings->tx_clk_sel);
	INSERT_BIT(GCU_ETH_CTRL, 17, settings->rx_clk_sel);
	INSERT_BIT(GCU_ETH_CTRL, 18, settings->tx_pad_src_sel);
	INSERT_BIT(GCU_ETH_CTRL, 24, settings->tx_pad_oen);
	INSERT_BIT(GCU_ETH_CTRL, 25, settings->rx_pad_oen);

	write_div(settings->tx_clk_phase_sel,
		  &((const struct clk){"eth_tx_clk_phase_sel", 0,
				       {ETH_TX_CLK_PHASE_SEL}}));

	if (settings->int_eth_div_en) {
		uint32_t div = rate / settings->ethint_rate[speed];

		/* only divider values 1, 2, 20, or 50 are allowed */
		if ((div != 1 && div != 2 && div != 20 && div != 50) ||
		    div * settings->ethint_rate[speed] != rate) {
			printf("error: can't set etherintdiv clock rate to %d\n",
			       settings->ethint_rate[speed]);
			return -1;
		}

		write_div(div - 1, &((const struct clk){"eth_int_div", 0,
							{ETH_INT_DIV_VAL}}));
		SET_BIT(GCU_ETH_CTRL, 4);	/* ETH_INT_DIV_EN */
	}

	clk_enable("eth_mac_clk");

	return 0;
}

static struct clk *
clk_get(const char *id)
{
	unsigned i;

	for (i = 0; clocks[i] && strcmp(clocks[i]->name, id); i++)
		;

	return clocks[i];
}

void
clk_init(void)
{
	clk_init_parent_funcs();
	clk_init_rate_funcs();
	clk_init_enable_funcs();

	clk_set_rate("rtc_clk", 1000000);
	reset_release("rtc_clk");
	clk_enable("rtc_clk");
	clk_enable("timer2");
}

unsigned long
clk_get_rate(const char *id)
{
	struct clk *clk = clk_get(id);

	if (!clk)
		return 0;

	return clk->get_rate(clk);
}

int
clk_set_rate(char *id, unsigned long rate)
{
	struct clk *clk = clk_get(id);

	if (!clk)
		return -1;

	return clk->set_rate(clk, rate);
}

void
clk_enable(char *id)
{
	const struct clk *clk = clk_get(id);

	if (clk)
		clk->enable(clk, 1);
	else
		printf("clock \"%s\" not found\n", id);
}

void
clk_disable(char *id)
{
	const struct clk *clk = clk_get(id);

	if (clk)
		clk->enable(clk, 0);
	else
		printf("clock \"%s\" not found\n", id);
}

void
reset_set(char *id)
{
	const struct clk *clk = clk_get(id);

	if (!clk) {
		printf("clock \"%s\" not found\n", id);
		return;
	}

	if (clk->en_reg)
		SET_BIT(clk->en_reg, 0);
	else
		printf("don't know which register for reset set\n");
}

void
reset_release(char *id)
{
	const struct clk *clk = clk_get(id);

	if (!clk) {
		printf("clock \"%s\" not found\n", id);
		return;
	}

	if (GET_BIT(clk->en_reg, 0) && clk->is_enabled(clk))
		printf("cannot release %s from reset: clock is enabled\n",
		       clk->name);
	else if (clk->en_reg)
		CLEAR_BIT(clk->en_reg, 0);
	else
		printf("don't know which register for reset release\n");
}

#ifdef DVF101_CLOCK_DEBUG_STANDALONE

#define ANSI_COLOR_RED		"\x1b[31m"
#define ANSI_COLOR_GREEN	"\x1b[32m"
#define ANSI_COLOR_YELLOW	"\x1b[33m"
#define ANSI_COLOR_BLUE		"\x1b[34m"
#define ANSI_COLOR_MAGENTA	"\x1b[35m"
#define ANSI_COLOR_CYAN		"\x1b[36m"
#define ANSI_COLOR_RESET	"\x1b[0m"

#define printcolored(color, args...)					\
	do {								\
		printf(color args);					\
		printf(ANSI_COLOR_RESET);				\
	} while (0)
#define printred(args...)	printcolored(ANSI_COLOR_RED,	args)
#define printgreen(args...)	printcolored(ANSI_COLOR_GREEN,	args)
#define printyellow(args...)	printcolored(ANSI_COLOR_YELLOW,	args)
#define printblue(args...)	printcolored(ANSI_COLOR_BLUE,	args)
#define printcyan(args...)	printcolored(ANSI_COLOR_CYAN,	args)

#define NODE_BUFF	128

static uint32_t reg_buff_addr[REG_BUFF_SIZE];
static uint32_t reg_buff_val[REG_BUFF_SIZE];
static int reg_buff_used[REG_BUFF_SIZE];

static int dbg_reg_access;

void
writel(uint32_t val, uint32_t addr)
{
	unsigned i, last_unused, changed = 0;

	if (dbg_reg_access)
		printf("write %8.8X to   %8.8X\n", val, addr);

	for (i = 0, last_unused = 0; i < REG_BUFF_SIZE; i++) {
		if (!reg_buff_used[i]) {
			last_unused = i;
		} else if (reg_buff_addr[i] == addr) {
			reg_buff_val[i] = val;
			changed = 1;
			break;
		}
	}

	if (!changed) {
		reg_buff_addr[last_unused] = addr;
		reg_buff_val[last_unused] = val;
		reg_buff_used[last_unused] = 1;
	}
};

uint32_t
readl(uint32_t addr)
{
	unsigned i, found = 0;
	uint32_t val;

	for (i = 0; i < REG_BUFF_SIZE; i++) {
		if (reg_buff_used[i] && reg_buff_addr[i] == addr) {
			val = reg_buff_val[i];
			found = 1;
			break;
		}
	}

	if (dbg_reg_access)
		printf("read  %8.8X from %8.8X\n", found ? val : 0, addr);

	return found ? val : 0;
};

static void
clk_show_path_rec(const struct clk *clk)
{
	const struct clk *parent = clk->get_parent(clk);

	if (parent) {
		clk_show_path_rec(parent);
		printf("->");
	}
	printf("%s", clk->name);
}

static void
clk_show_possible_paths_rec(const struct clk *clk, unsigned indent)
{
	unsigned i;
	const struct clk *parent;

	if (!clk)
		return;

	if (indent > 0) {
		for (i = 0; i < indent - 1; i++)
			printf("\t");
		printf("`-------");
	}
	printf("%s\n", clk->name);

	indent++;
	for (i = 0, parent = clk->parents[i]; i < 4;
	     parent = clk->parents[++i])
		clk_show_possible_paths_rec(parent, indent);
}

static void
clk_show_path(char *id)
{
	const struct clk *clk = clk_get(id);

	if (!clk)
		return;

	clk_show_path_rec(clk);
	printf("\n");
}

static void
clk_show_possible_paths(const char *id)
{
	const struct clk *clk = clk_get(id);

	if (!clk)
		return;

	clk_show_possible_paths_rec(clk, 0);
	printf("\n");
}

struct clk_node {
	const struct clk *clk;
	const struct clk *children[32];
};

static struct clk_node nodes[NODE_BUFF];

static int
add_child(struct clk_node *node, const struct clk *child)
{
	unsigned i;
	const struct clk *cur;
	int ret = 0;

	for (i = 0, cur = node->children[i];
	     cur && cur != child; cur = node->children[++i])
		;

	if (cur != child)
		/* new child */
		ret = 1;

	node->children[i] = child;

	return ret;
}

void
build_tree(struct clk_node nodes[])
{
	unsigned i, j, k, change;
	const struct clk *clk;
	const struct clk *parent;
	struct clk_node *node;
	struct clk_node *node2;

	for (i = 0; i < 128; i++) {
		nodes[i].clk = NULL;
		for (j = 0; j < 32; j++)
			nodes[i].children[j] = NULL;
	}

	for (i = 0, clk = clocks[i]; clk; clk = clocks[++i])
		nodes[i].clk = clk;

	do {
		change = 0;
		for (i = 0, node = &nodes[i]; node->clk && i < 128;
		     node = &nodes[++i]) {
			for (j = 0; j < 4; j++) {
				parent = node->clk->parents[j];
				if (!parent)
					continue;

				for (k = 0, node2 = &nodes[k];
				     node2->clk && node2->clk != parent;
				     node2 = &nodes[++k])
					;
				if (node2->clk != parent)
					/* new parent */
					change = 1;
				node2->clk = parent;
				change |= add_child(node2, node->clk);
			}
		}
	} while (change);
}

static const struct clk *cur_path[128];
static int next_clk = 0;

static void
print_edge(const struct clk *clock)
{
	if (clock->flags & FIXED_DIV)
		printgreen("`-------");
	else if (clock->get_rate == fixed_divider_2)
		printyellow("`-[/2]--");
	else if (clock->get_rate == fixed_divider_16)
		printyellow("`-[/16]-");
	else if (clock->flags & EVEN_DIV)
		printcyan("`-even--");
	else
		printf("`-------");
}

static void
print_name(const struct clk *clock)
{
	if (clock->flags & OSC_IN)
		printred("%s", clock->name);
	else
		printf("%s", clock->name);
}

static void
print_bus(const struct clk *clock)
{
	int print = 0;

	if (!(clock->flags & HAS_APB || clock->flags & HAS_AHB ||
	      clock->flags & HAS_AXI))
		return;

	printred("[");
	if (clock->flags & HAS_APB) {
		printred("APB");
		print = 1;
	}
	if (clock->flags & HAS_AHB) {
		if (print)
			printred(",");
		printred("AHB");
		print = 1;
	}
	if (clock->flags & HAS_AXI) {
		if (print)
			printred(",");
		printred("AXI");
		print = 1;
	}

	printred("]");
}

void
print_tree(struct clk_node nodes[], const struct clk *root, unsigned indent)
{
	unsigned i, j;
	struct clk_node *node;
	const struct clk *child;

	if (!root)
		return;

	for (i = 0; i < next_clk; i++)
		if (cur_path[i] == root) {
			printf("loop detected:\n");
			for (j = 0; j < i; j++)
				printf("%s -> ", cur_path[j]->name);
			printf("%s\n", root->name);
			return;
		}

	if (indent > 0) {
		for (i = 0; i < indent - 1; i++)
			printf("\t");
		print_edge(root);
	}
	print_name(root);
	print_bus(root);
	printf("\n");

	cur_path[next_clk++] = root;

	indent++;
	for (i = 0; i < 128; i++)
		if (nodes[i].clk == root)
			node = &nodes[i];
	if (node == NULL) {
		printred("THIS SHOULD NOT HAVE HAPPENED\n");
		return;
	}

	for (i = 0, child = node->children[i]; i < 32;
	     child = node->children[++i])
		print_tree(nodes, child, indent);

	cur_path[--next_clk] = NULL;
}

static void
print_path(const struct clk *clock)
{
	if (!clock)
		return;

	if (!clock->get_parent(clock)) {
		printf("%s\n", clock->name);
		return;
	}

	printf("%s<--", clock->name);
	print_path(clock->get_parent(clock));
}

static void
print_path_clks(const struct clk *clock)
{
	if (!clock)
		return;

	if (!clock->get_parent(clock)) {
		printf("%s (%lu)\n", clock->name, clock->get_rate(clock));
		return;
	}

	printf("%s (%lu)<--", clock->name, clock->get_rate(clock));
	print_path_clks(clock->get_parent(clock));
}

#define CLK_PROMPT	"clk-debug> "
typedef enum {
	DBG_SET_RATE = 0,
	DBG_GET_RATE,
	DBG_ENABLE,
	DBG_DISABLE,
	DBG_IS_ENABLED,
	DBG_REG_READ,
	DBG_REG_WRITE,
	DBG_ETH_CONFIG,
	DBG_INFO,
	DBG_DEBUG,

	DBG_CMDS,
} clk_dbg_cmd;

static char *dbg_cmd_names[] = {
	"set_rate", "get_rate", "enable", "disable", "is_enabled", "read",
	"write", "eth_config", "info", "debug",
};

static int
dbg_execute(clk_dbg_cmd cmd, const struct clk *clk, unsigned long rate,
	    uint32_t reg1, uint32_t reg2)
{
	int ret;

/*
	printf("cmd: %d, clk: %s, rate: %lu, reg1: 0x%8.8X, reg2: 0x%8.8X\n",
	       cmd, clk ? clk->name : "unknown", rate, reg1, reg2);
*/
	switch (cmd) {
	case DBG_GET_RATE:
		printf("%lu\n", clk->get_rate(clk));
		break;
	case DBG_SET_RATE:
		printf("%s\n",
		       clk->set_rate(clk, rate) ? "ERR" : "OK");
		break;
	case DBG_ENABLE:
		printf("%s\n",
		       clk->enable(clk, 1) ? "ERR" : "OK");
		break;
	case DBG_DISABLE:
		printf("%s\n",
		       clk->enable(clk, 0) ? "ERR" : "OK");
		break;
	case DBG_IS_ENABLED:
		printf("%s is %sabled\n", clk->name,
		       clk->is_enabled(clk) ? "en" : "dis");
		break;
	case DBG_INFO:
		clk_show_possible_paths(clk->name);
		print_tree(nodes, clk, 0);
		print_path(clk);
		print_path_clks(clk);
		break;
	case DBG_REG_READ:
		printf("0x%8.8X\n", readl_cmu(reg1));
		break;
	case DBG_REG_WRITE:
		writel_cmu(reg1, reg2);
		printf("OK\n");
		break;
	case DBG_ETH_CONFIG:
		ret = clk_eth_config((enum eth_mode)reg1,
				     (enum eth_speed)reg2);
		if (ret)
			printf("ERR: %d\n", ret);
		else
			printf("OK\n");
		break;
	case DBG_DEBUG:
		dbg_reg_access ^= 1;
		break;
	default:
		printf("command not implemented\n");
		break;
	};

	return 0;
}

static int
dbg_parse_input(char *input)
{
	char *cur_word;
	char *ptr;
	clk_dbg_cmd cmd;
	const struct clk *clk;
	unsigned long i, rate = 0;
	uint32_t reg1 = 0, reg2 = 0;

	if (!strcmp("\n", input))
		return -1;
	strtok(input, "\n");

	cur_word = strtok(input, " ");
	if (cur_word) {
		for (i = 0; i < DBG_CMDS; i++) {
			if (!strcmp(dbg_cmd_names[i], cur_word)) {
				cmd = (clk_dbg_cmd)i;
				break;
			}
		}

		if (i >= DBG_CMDS) {
			printf("command \"%s\" not found\n", cur_word);
			return -1;
		}
	} else {
		printf("no command given\n");
		return -1;
	}

	switch (cmd) {
	case DBG_DEBUG:
		dbg_execute(cmd, NULL, 0, 0, 0);
		break;
	case DBG_SET_RATE:
	case DBG_GET_RATE:
	case DBG_ENABLE:
	case DBG_DISABLE:
	case DBG_IS_ENABLED:
	case DBG_INFO:
		cur_word = strtok(NULL, " ");
		if (cur_word) {
			clk = clk_get(cur_word);
			if (!clk) {
				printf("clock \"%s\" not found\n", cur_word);
				return -1;
			}
		} else {
			printf("no clock given\n");
			return -1;
		}

		if (cmd == DBG_SET_RATE) {
			cur_word = strtok(NULL, " ");
			if (cur_word) {
				rate = strtoul(cur_word, &ptr, 0);
			} else {
				printf("no rate given\n");
				return -1;
			}
		}

		dbg_execute(cmd, clk, rate, 0, 0);
		break;
	case DBG_ETH_CONFIG:
	case DBG_REG_READ:
	case DBG_REG_WRITE:
		cur_word = strtok(NULL, " ");
		if (cur_word) {
			reg1 = strtoul(cur_word, &ptr, 0);
		} else {
			printf("missing value/register\n");
			return -1;
		}

		if (cmd == DBG_REG_WRITE || cmd == DBG_ETH_CONFIG) {
			cur_word = strtok(NULL, " ");
			if (cur_word) {
				reg2 = strtoul(cur_word, &ptr, 0);
			} else {
				printf("no register given\n");
				return -1;
			}
		}

		dbg_execute(cmd, clk, 0, reg1, reg2);
		break;
	default:
		printf("ERR: unhandled command\n");
		return -1;
	};

	return 0;
}

/*
     pre-div  mult  OD  clk_out
PLL1       1    40   1     1000
PLL2      25   734   1      734
PLL3       3   146   1     1216
PLL4      25  1066   2      533
*/
void
dbg_init_plls(void)
{
	writel_cmu(0x0000C093, PLL_BASE(1) + PLL_UNIT_CTRL);
	writel_cmu(0x1000000, PLL_BASE(1) + PLL_CFG1);
	writel_cmu(0x11000000 + ( 40 << 8) +  1, PLL_BASE(1) + PLL_CFG2);

	writel_cmu(0x0000C093, PLL_BASE(2) + PLL_UNIT_CTRL);
	writel_cmu(0x1000000, PLL_BASE(2) + PLL_CFG1);
	writel_cmu(0x11000000 + (734 << 8) + 25, PLL_BASE(2) + PLL_CFG2);

	writel_cmu(0x0000C093, PLL_BASE(3) + PLL_UNIT_CTRL);
	writel_cmu(0x1000000, PLL_BASE(3) + PLL_CFG1);
	writel_cmu(0x11000000 + (146 << 8) +  3, PLL_BASE(3) + PLL_CFG2);
}

void
main(int argc, const char *argv[])
{
	int i;
	char user_input[1024];
	clk_dbg_cmd cmd;
	const struct clk *clk;
	unsigned long rate = 0;

	clk_init();

	build_tree(nodes);
	for (i = 0; i < NODE_BUFF; i++)
		if (nodes[i].clk)
			printf("clk #%d: \t%s\n", i,
			       nodes[i].clk->name);

	dbg_init_plls();

	while (1) {
		printf(CLK_PROMPT);
		fgets(user_input, 1024, stdin);
		dbg_parse_input(user_input);
	}
}
#endif	/* DVF101_CLOCK_DEBUG_STANDALONE */
