/*
 * (C) Copyright 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/dvf10a.h>
#include <asm/arch/hardware.h>
#include <linux/string.h>

#define DVF10A				"dvf10a"

#define DAIF_WAPU_TRAN_MASK		(1<<15)
#define DAIF_RAPU_NEW_VALUE		(1<<15)

#define DAIF_DIRECT_MIRRORED_START	0
#define DAIF_DIRECT_MIRRORED_END	64
#define DAIF_DIRECT_RAPU_WAPU_START	65
#define DAIF_DIRECT_RAPU_WAPU_END	114

#define DAIF_RAPU			41  // 0x00a4
#define DAIF_WAPU			64  // 0x0100
#define DAIF_MODE			117 // 0x01d4
#define DAIF_PEND			118 // 0x01d8
#define DAIF_PENDB0			119 // 0x01dc
#define DAIF_PENDB1			120 // 0x01e0
#define DAIF_PENDB2			121 // 0x01e4
#define DAIF_PENDB3			122 // 0x01e8

#define DAIF_ANA_DATA_WR_LSB		65
#define DAIF_ANA_DATA_WR_MSB		66
#define DAIF_ANA_ADDR_WR		67
#define DAIF_ANA_ADDR_RD		68
#define DAIF_ANA_DATA_RD_LSB		69
#define DAIF_ANA_DATA_RD_MSB		70

#define DAIF_MODE_RAPU_PRIO		(1<<2)
#define DAIF_MODE_WAPU_PRIO		(1<<3)

#define CLK_MIN			0x0
#define CLK_MAX			0xE0
#define RX_MIN			0x0
#define RX_MAX			0xFC

static int
daif_is_direct_mirrored(unsigned short reg)
{
	if ((reg <= DAIF_DIRECT_MIRRORED_END) ||
	   ((reg >= DAIF_MODE) && (reg <= DAIF_PENDB3)))
		return 1;
	return 0;
}

static int
daif_is_direct_rapu_wapu(unsigned short reg)
{
	if ((reg > DAIF_DIRECT_MIRRORED_END) &&
	    (reg - DAIF_DIRECT_RAPU_WAPU_START <= DAIF_DIRECT_RAPU_WAPU_END -
	            DAIF_DIRECT_RAPU_WAPU_START))
		return 1;
	return 0;
}

static int
daif_is_indirect_analog(unsigned short reg)
{
	if (reg >= DAIF_INDIRECT_START)
		return 1;
	return 0;
}

static unsigned short
daif_readw(unsigned short reg)
{
	return readw(DVF101_PMU_BASE + (reg * 4));
}

static void
daif_writew(unsigned short val, unsigned short reg)
{
	writew(val, DVF101_PMU_BASE + (reg * 4));
}

static void
daif_wait_wapu_tran(void)
{
	while (daif_readw(DAIF_PEND))
		;
}

unsigned short
daif_read_via_rapu(unsigned short reg)
{
	unsigned short val;

	daif_writew(reg, DAIF_RAPU);

	while (!((val = daif_readw(DAIF_RAPU)) & DAIF_RAPU_NEW_VALUE))
		;

	/* Check if indicated register address (RAPU[14:8]) matches */
	if (((val & 0x7f00) >> 8) == reg)
		return val & 0xff;
	else
		return daif_read_via_rapu(reg); /* repeat request */
}

void
daif_write_via_wapu(unsigned short reg, unsigned short value)
{
	daif_writew(((reg & 0x7f) << 8) | (value & 0xff), DAIF_WAPU);
	daif_wait_wapu_tran();
}

static void
daif_direct_mirrored_init(void)
{
	unsigned short val, reg;

	/* enable "prev_replica" */
	daif_writew(2, DAIF_MODE);

	for (reg = 0; reg < DAIF_DIRECT_MIRRORED_END; reg++) {
		if (reg == DAIF_RAPU)
			continue;
		val = daif_read_via_rapu(reg);
		udelay(10);
		daif_writew(val & 0xff, reg);
	}

	/* disable "prev_replica" */
	daif_writew(0, DAIF_MODE);
}

unsigned short
daif_read(unsigned short reg)
{
	unsigned short ret = 0;

	if (daif_is_direct_rapu_wapu(reg) ||
	    (reg == DAIF_RAADOMSB) || (reg == DAIF_LQSD_CFG1) ||
	    (reg == DAIF_PORCTL) || (reg == DAIF_BASECTRL)) {
		ret = daif_read_via_rapu(reg);
	} else if (daif_is_direct_mirrored(reg)) {
		ret = daif_readw(reg); /* immediate access */
	} else if (daif_is_indirect_analog(reg)) {
		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(DAIF_ANA_ADDR_RD, reg >> 1);
		if (daif_readw(DAIF_MODE) & DAIF_MODE_RAPU_PRIO)
			daif_wait_wapu_tran();

		/* read value */
		ret  = daif_read_via_rapu(DAIF_ANA_DATA_RD_MSB) << 8;
		udelay(10);
		ret |= daif_read_via_rapu(DAIF_ANA_DATA_RD_LSB) & 0xff;
	}

	return ret;
}

void
daif_write(unsigned short reg, unsigned short value)
{
	if (daif_is_direct_mirrored(reg)) {
		unsigned short pend_reg = reg >> 4; /* register number 0..3 */
		unsigned short pend_mask = reg - (pend_reg << 4);   /* 0..15 */
		pend_mask = 1 << pend_mask; /* bit mask */

		/* immediate access */
		while (daif_readw(DAIF_PENDB0 + pend_reg) & pend_mask)
			;
		daif_writew(value, reg);
		while (daif_readw(DAIF_PENDB0 + pend_reg) & pend_mask)
			;
	} else if (daif_is_direct_rapu_wapu(reg)) {
		daif_write_via_wapu(reg, value);
	} else if (daif_is_indirect_analog(reg)) {
		/* write value msb, lsb */
		daif_write_via_wapu(DAIF_ANA_DATA_WR_MSB, value >> 8);
		daif_write_via_wapu(DAIF_ANA_DATA_WR_LSB, value & 0xff);

		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(DAIF_ANA_ADDR_WR, reg >> 1);
	}
}

void dvf10a_init(void)
{
	clk_set_rate("apu_clk", 13824000);
	clk_enable("apu_clk");
	reset_release("daif_clk");
	clk_enable("daif_clk");

	daif_direct_mirrored_init();

	/* Disable IRQs */
	daif_write(DAIF_ICU_INT_POL,      0x0000);
	daif_write(DAIF_ICU_ANA_EN1,      0x0000);
	daif_write(DAIF_ICU_ANA_EN2,      0x0000);
	daif_write(DAIF_ICU_ANA_STAT_EN1, 0x0000);
	daif_write(DAIF_ICU_ANA_STAT1,    0x0000);
	daif_write(DAIF_ICU_ANA_STAT2,    0x0000);
	daif_write(DAIF_ICU_PWR_EN,       0x0000);
	daif_write(DAIF_ICU_PWR_STAT_EN,  0x0000);
	daif_write(DAIF_ICU_PWR_STAT,     0x0000);
}

int do_daif(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int reg, val;

	if (argc < 3)
		return CMD_RET_USAGE;

	if (strncmp(argv[1], "read", 4) == 0) {
		reg = simple_strtoul(argv[2], NULL, 0);
		printf("0x%x\n", daif_read(reg));
	} else if ((strncmp(argv[1], "write", 5) == 0) && (argc == 4)) {
		reg = simple_strtoul(argv[2], NULL, 0);
		val = simple_strtoul(argv[3], NULL, 0);
		daif_write(reg, val);
	} else {
		return CMD_RET_USAGE;
	}

	return 0;
}

U_BOOT_CMD(
	daif, 4, 1, do_daif,
	"DVF10A (DAIF) sub-system",
	"read <register>\n"
	"daif write <register> <value>\n"
);

static void
do_calibrate(void)
{
	unsigned char quit = 0;
	unsigned int rxtune = 0, clk = 0;
	char buf[3];

	rxtune = daif_read(DAIF_RXTUN);
	clk = daif_read(DAIF_CLKCTRL);

	printf("start values:\n\tDAIF_CLKCTRL = %x\n\tDAIF_RXTUN = %x\n",
	       clk, rxtune);

	if ((rxtune % 0x4) != 0)
		rxtune = 0x0;
	if ((clk % 0x20) != 0)
		clk = 0x0;

	printf("Enter < or > to tune RXTUN, q to save, a to abort\n");
	do {
		switch (getc()) {
		case '<':
			if (clk <= CLK_MIN)
				if (rxtune <= RX_MIN) {
					rxtune = RX_MIN;
					printf("\nalready at min value\n");
				} else {
					clk = CLK_MAX;
					rxtune -= 0x4;
				}
			else
				clk -= 32;
			break;

		case '>':
			if (clk >= CLK_MAX)
				if (rxtune >= RX_MAX) {
					rxtune = RX_MAX;
					printf("\nalready at max value\n");
				} else {
					clk = CLK_MIN;
					rxtune += 0x4;
				}
			else
				clk += 32;
			break;
		case 'a':
			printf("\n");
			return;
		case 'q':
			printf("\n");
			quit = 1;
			break;

		default:
			break;
		}

		if (!quit) {
			printf("\r\tDAIF_CLKCTRL = %2.2x, DAIF_RXTUN = %2.2x",
			       clk, rxtune);
			daif_write(DAIF_CLKCTRL, clk);
			daif_write(DAIF_RXTUN, rxtune);
		}
	} while (!quit);

	memset(&buf[0], 0, sizeof(buf));
	sprintf(buf, "%x", rxtune);
	setenv("rxtun", buf);

	memset(&buf[0], 0, sizeof(buf));
	sprintf(buf, "%x", clk);
	setenv("clkctrl", buf);

	saveenv();
}

void
apply_rxtun(void)
{
	char *string_clkctl = getenv("clkctrl");
	char *string_rxtun = getenv("rxtun");
	unsigned clkctl = simple_strtoul(string_clkctl, NULL, 16);
	unsigned rxtun = simple_strtoul(string_rxtun, NULL, 16);
	unsigned clk_reg = 0;

	if (clkctl > CLK_MAX)
		clkctl = CLK_MAX;
	else if (clkctl < CLK_MIN)
		clkctl = CLK_MIN;

	if (rxtun > RX_MAX)
		rxtun = RX_MAX;
	else if (rxtun < RX_MIN)
		rxtun = RX_MIN;

	clk_reg= daif_read(DAIF_CLKCTRL);
	clk_reg&= ~(0xf0);
	clk_reg |= clkctl;
	daif_write(DAIF_CLKCTRL, clk_reg);

	clk_reg = daif_read(DAIF_RXTUN);
	clk_reg &= ~(0xfc);
	clk_reg |= rxtun;
	daif_write(DAIF_RXTUN, clk_reg);

	printf("set values:\n\tDAIF_CLKCTRL = %x\n\tDAIF_RXTUN = %x\n",
	       clkctl, rxtun);
}

static int
do_rxtune(cmd_tbl_t *cmdtp, int flag, int argc, char *const argv[])
{
	if (argc != 2)
		goto usage;

	if (strcmp(argv[1], "calibrate") == 0) {
		do_calibrate();
		return 0;
	} else if (strcmp(argv[1], "apply") == 0) {
		apply_rxtun();
		return 0;
	}

usage:
	return CMD_RET_USAGE;
}

U_BOOT_CMD(rxtune, 2, 0, do_rxtune,
	   "Tune DAIF_RXTUN and DAIF_CLKCTRL values.",
	   "calibrate|apply\n"
	   "    - calibrate: save new values to environment\n"
	   "    - apply: apply values from environment"
);
