/*
 * (C) Copyright 2012 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <stdio_dev.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/hardware.h>
#include <asm/arch/keypad.h>

#define KBD_CFG 0x2B0
#define KBD_ROW 0x2B4
#define KBD_COL 0x2B8

#define DEVNAME "keypad"

#define KEYPAD_DISCHARGE_DELAY 20 /* us */
#define KEYPAD_ACTIVATE_DELAY  20 /* us */

#ifdef CONFIG_SYS_CONSOLE_OVERWRITE_ROUTINE
extern int overwrite_console (void);
#define OVERWRITE_CONSOLE overwrite_console ()
#else
#define OVERWRITE_CONSOLE 0
#endif /* CONFIG_SYS_CONSOLE_OVERWRITE_ROUTINE */

static int dvf99_keypad_rows = 0;
static int dvf99_keypad_cols = 0;

static int debug_on = 0;

static inline void
dvf99_keypad_discharge_all(void)
{
	/* STROBE All HiZ */
	writel(0, DVF99_GPIO_BASE + KBD_COL);
	udelay(KEYPAD_DISCHARGE_DELAY);
}

static inline void
dvf99_keypad_activate_col(int col)
{
	unsigned long val;

	dvf99_keypad_discharge_all();

	/* STROBE col -> High, not col -> HiZ */
	val = readl(DVF99_GPIO_BASE + KBD_COL);
	val |= (1 << col);
	writel(val, DVF99_GPIO_BASE + KBD_COL);
	udelay(KEYPAD_ACTIVATE_DELAY);
}

static inline void
dvf99_keypad_activate_all(void)
{
	dvf99_keypad_discharge_all();

	/* STROBE ALL -> High */
	writel((1 << dvf99_keypad_cols) - 1, DVF99_GPIO_BASE + KBD_COL);
	udelay(KEYPAD_ACTIVATE_DELAY);
}

static int
dvf99_keypad_key_pressed(void)
{
	return (readl(DVF99_GPIO_BASE + KBD_ROW) &
	       ((1 << dvf99_keypad_rows) - 1));
}

int dvf99_keypad_scan(int row, int col)
{
	if ((row < 0) || (row >= dvf99_keypad_rows) ||
	    (col < 0) || (col >= dvf99_keypad_cols))
		return 0;

	/* Discharge the output driver capacitance in the keypad matrix. */
	dvf99_keypad_activate_col(col);

	if (dvf99_keypad_key_pressed() & (1<<row))
		return 1;

	return 0;
}

void dvf99_keypad_debug(void)
{
	int c, r, row_mask;

	if (!debug_on)
		return;

	for (c = 0; c < dvf99_keypad_cols; c++) {
		dvf99_keypad_discharge_all();

		/* Discharge the output driver capacitance in the keypad matrix. */
		dvf99_keypad_activate_col(c);

		row_mask = dvf99_keypad_key_pressed();

		for (r = 0; r < dvf99_keypad_rows; r++) {
			if (row_mask & (1 << r))
				printf("row %d col %d\n", r, c);
		}
	}
}

static int dvf99_keypad_tstc(struct stdio_dev *dev)
{
	/* enable all colums */
	dvf99_keypad_activate_all();

	/* check for any active rows */
	return !!dvf99_keypad_key_pressed();
}

static int dvf99_keypad_getc(struct stdio_dev *dev)
{
	return 0;
}

void dvf99_keypad_init(int nr_rows, int nr_cols)
{
	dvf99_keypad_rows = nr_rows;
	dvf99_keypad_cols = nr_cols;

	writel( ((1 << dvf99_keypad_cols) - 1) | (((1 << dvf99_keypad_rows) - 1) << 8),
	       DVF99_GPIO_BASE + KBD_CFG);

	writel( (1 << dvf99_keypad_cols) - 1, DVF99_GPIO_BASE + KBD_COL);
}

void dvf99_keypad_late_init(void)
{
	char *tmp = NULL;
	int val = 0;
	struct stdio_dev keypaddev;
	char *stdinname = getenv ("stdin");

	/* register as input device */
	memset (&keypaddev, 0, sizeof(keypaddev));
	strcpy(keypaddev.name, DEVNAME);
	keypaddev.flags =  DEV_FLAGS_INPUT | DEV_FLAGS_SYSTEM;
	keypaddev.putc = NULL;
	keypaddev.puts = NULL;
	keypaddev.getc = dvf99_keypad_getc;
	keypaddev.tstc = dvf99_keypad_tstc;

	val = stdio_register (&keypaddev);
	if (val == 0) {
		/* check if this is the standard input device */
		if(strcmp(stdinname,DEVNAME) == 0) {
			/* reassign the console */
			if(OVERWRITE_CONSOLE == 0) {
				val = console_assign(stdin, DEVNAME);
				if (val != 0) {
					/* XXX */
				}
			}
		}
	}

	tmp = getenv("keypad_debug_on");
	debug_on = tmp ? simple_strtoul(tmp, NULL, 10) : 0;
}
