/*
 * (C) Copyright 2015 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>

#include "pll.h"
#include "clock_regs.h"

#define ABSDIFF(x, y)	((x) > (y) ? ((x) - (y)) : ((y) - (x)))

static long
round_rate_pll(unsigned long in_rate, unsigned long rate,
	       unsigned long *rprediv, unsigned long *rmult,
	       unsigned long *rpostdiv1, unsigned long *rpostdiv2)
{
	unsigned long prediv, mult, postdiv1, postdiv2;
	unsigned long match_prediv = 0, match_mult = 0, match_postdiv1 = 0,
		      match_postdiv2 = 0;
	unsigned long match_rate = 0, match_diff = ~0ul, match_vco = ~0ul;

	if (!(rate >> 10))
		return -1;

	for (prediv = 1; prediv <= 63; prediv++) {
		unsigned long phaseclk = (in_rate >> 10) / prediv;

		if (phaseclk < (1000000 >> 10))
			break;

		for (mult = 16; mult <= 2400; mult++) {
			unsigned long vcoclk, postdiv;

			vcoclk = phaseclk * mult;

			/* keep FVCO and FDIVREF in valid range */
			if (phaseclk > vcoclk / 16 ||
			    vcoclk < (600000000 >> 10))
				continue;
			else if (vcoclk > (1600000000 >> 10))
				break;

			postdiv = round_div(vcoclk, rate >> 10);

			if (postdiv < 1)
				postdiv = 1;
			else if (postdiv > 49)
				break;

			for (postdiv1 = 7; postdiv1 >= 1; postdiv1--) {
				unsigned long diff, tmp, out;

				postdiv2 = round_div(postdiv, postdiv1);

				if (postdiv2 < 1)
					postdiv2 = 1;
				else if (postdiv2 > postdiv1)
					break;

				out = vcoclk / postdiv1 / postdiv2;

				/* quick check if in target range */
				if (out - 10000 > (rate >> 10) ||
				    out + 10000 < (rate >> 10))
					continue;

				/* ok, do the precise calculation */
				tmp = in_rate * mult;
				tmp /= prediv;
				out = tmp / postdiv1 / postdiv2;

				/* better match? */
				diff = ABSDIFF(rate, out);
				if ((diff <  match_diff) ||
				    (diff == match_diff &&
				     vcoclk < match_vco)) {
					match_diff     = diff;
					match_vco      = vcoclk;
					match_rate     = out;
					match_prediv   = prediv;
					match_mult     = mult;
					match_postdiv1 = postdiv1;
					match_postdiv2 = postdiv2;
				}
			}
		}
	}

	if (!match_rate)
		return -1;

	*rprediv   = match_prediv;
	*rmult     = match_mult;
	*rpostdiv1 = match_postdiv1;
	*rpostdiv2 = match_postdiv2;

	return match_rate;
}

static unsigned long
pll_get_input(uint32_t addr)
{
	uint32_t pll_ctrl = readl_cmu(addr);

	switch ((pll_ctrl >> 8) & 0x3) {
	case 0:
		return 25000000ul;
	case 2:
		return 13824000ul;
	case 3:
		/* TODO TDM clocks for PLL3 */
	default:
		return 0;
	}
}

static int
pll_deactivate(uint32_t addr)
{
	uint32_t ctrl = readl_cmu(addr);
	int automatic = !(ctrl & (1 << 5));

	if (!(ctrl & (1 << 7)))
		/* pll already powered down */
		return -1;

	if (automatic) {
		ctrl |= 1 << 2;
		writel_cmu(ctrl, addr);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (ctrl & (1 << 7));
#endif
	} else {
		/* set mux to bypass */
		ctrl &= ~(1 << 4);
		writel_cmu(ctrl, addr);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (ctrl & (1 << 7));
#endif
		/* power down */
		ctrl |= 1 << 2;
		writel_cmu(ctrl, addr);
	}

	return 0;
}

static int
pll_activate(uint32_t addr, uint32_t addr_cfg1)
{
	uint32_t ctrl = readl_cmu(addr);
	int automatic = !(ctrl & (1 << 5));

	if (ctrl & (1 << 7))
		/* pll already powered up */
		return -1;

	/* make sure pre and post dividers are enabled */
	if (!(ctrl & (1 << 14))) {
		ctrl |= 1 << 0;
		writel_cmu(addr, ctrl);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (!(ctrl & (1 << 14)));
#endif
	}
	if (!(ctrl & (1 << 15))) {
		ctrl |= 1 << 1;
		writel_cmu(addr, ctrl);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (!(ctrl & (1 << 15)));
#endif
	}

	/* disable power down */
	ctrl &= ~(1 << 2);
	writel_cmu(ctrl, addr);

	if (automatic) {
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (!(ctrl & (1 << 7)));
#endif
	} else {
		writel_cmu(ctrl, addr);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr_cfg1);
		} while (!(ctrl & (1 << 24)));
#endif
		/* disable bypass */
		ctrl &= 1 << 4;
		writel_cmu(ctrl, addr);
#ifndef DVF101_CLOCK_DEBUG_STANDALONE
		do {
			ctrl = readl_cmu(addr);
		} while (!(ctrl & (1 << 7)));
#endif
	}

	return 0;
}

static inline unsigned
pll_name_to_num(const struct clk *clk)
{
	unsigned pll = 0;

	if (!strcmp(clk->name, "pll1"))
		pll = 1;
	else if (!strcmp(clk->name, "pll2"))
		pll = 2;
	else if (!strcmp(clk->name, "pll3"))
		pll = 3;
	else if (!strcmp(clk->name, "pll4"))
		pll = 4;

	return pll;
}

static int
clk_has_parent(const struct clk *child, const struct clk *parent)
{
	if (child == parent)
		return 1;

	if ((child = child->get_parent(child)))
		return clk_has_parent(child, parent);

	return 0;
}

static void
fix_children(const struct clk *pll,
	     unsigned long before, unsigned long after)
{
	int i;
	struct clk *clk;
	float factor;

	if (after == before)
		return;

	for (i = 0, clk = clocks[i]; clk; ++i, clk = clocks[i]) {
		if (clk->flags & IS_LEAF && clk_has_parent(clk, pll)) {
			printf("found child: %s\n", clk->name);
			factor = (float)before / after;
			clk->set_rate(clk,
				      clk->get_rate(clk) * factor);
		}
	}
}

int
pll_is_enabled(const struct clk *clk)
{
	uint32_t ctrl = readl_cmu(PLL_BASE(pll_name_to_num(clk)) +
				  PLL_UNIT_CTRL);

	if (ctrl & (1 << 7))
		return 1;
	else
		return 0;
}

int
pll_enable(const struct clk *clk, int en)
{
	unsigned pll = pll_name_to_num(clk);
	uint32_t pll_base = PLL_BASE(pll);

	if (en)
		return pll_activate(pll_base + PLL_UNIT_CTRL,
				    pll_base + PLL_CFG1);
	else
		return pll_deactivate(pll_base + PLL_UNIT_CTRL);
}

int
set_pll_rate(const struct clk *clk, unsigned long rate)
{
	unsigned pll = pll_name_to_num(clk);
	uint32_t reg, pll_base = PLL_BASE(pll);
	unsigned long refdiv = 0, fbdiv = 0, postdiv1 = 0, postdiv2 = 0;
	unsigned long before, after;

	before = clk->get_rate(clk);

	pll_deactivate(pll_base + PLL_UNIT_CTRL);

	after = round_rate_pll(pll_get_input(pll_base + PLL_UNIT_CTRL), rate,
			       &refdiv, &fbdiv, &postdiv1, &postdiv2);

	fix_children(clk, before, after);

	/* set PLL1 pre, post and MCU_MAIN dividers to bypass */
	reg = readl_cmu(pll_base + PLL_UNIT_CTRL);
	reg &= ~(0x3F << 24);
	reg &= ~( 0x3 << 16);
	writel_cmu(reg, pll_base + PLL_UNIT_CTRL);
	writel_cmu(GCU_MCU_DIV_VAL, 0x0);

	reg = ( refdiv   <<  0 ) |
	      ( fbdiv    <<  8 ) |
	      ( postdiv1 << 24 ) |
	      ( postdiv2 << 28 );
	writel_cmu(reg, pll_base + PLL_CFG2);

	pll_activate(pll_base + PLL_UNIT_CTRL, pll_base + PLL_CFG1);

	return 0;
}

unsigned long
get_pll_rate(const struct clk *clk)
{
	uint32_t reg, refdiv, fbdiv, postdiv1, postdiv2;
	unsigned pll;
	unsigned long rate;
	const struct clk *parent;

	pll = pll_name_to_num(clk);

	parent = clk->get_parent(clk);
	if (!parent)
		return 0;
	reg = readl_cmu(PLL_BASE(pll) + PLL_UNIT_CTRL);
	if (reg & (1 << 2))
		/* PLL powered down (or bypassed) */
		return 0;

	rate = parent->get_rate(clk);

	reg = readl_cmu(PLL_BASE(pll) + PLL_CFG2);
	refdiv   = (reg >>  0) &  0x3F;
	fbdiv    = (reg >>  8) & 0xFFF;
	postdiv1 = (reg >> 24) &   0x7;
	postdiv2 = (reg >> 28) &   0x7;

	if (!(refdiv * postdiv1 * postdiv2))
		return 0;

	return rate / refdiv * fbdiv / (postdiv1 * postdiv2);
}
