/*
 * (C) Copyright 2015 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/clock.h>
#include <asm/arch/pads.h>
#include <asm/arch/hardware.h>
#include <netdev.h>
#ifdef CONFIG_RTL8309_SWITCH
#include <rtk_api.h>
#include <rtk_api_ext.h>
#include <rtl8309n_asictypes.h>
#endif
#ifdef CONFIG_QCA833X_SWITCH
#include <qca/athrs17_phy.h>
#endif

#if defined(CONFIG_PHY8710) && defined (CONFIG_RTL8309_SWITCH)
	#error Both switch and PHY are enabled!
#endif

#define ETH_CLK_MODE		ETH_RMII
#define ETH_CLK_SPEED		ETH_SPEED_100
#define ETH_MODE		"RMII"

/* optional: override default ETH speed */
#ifdef CONFIG_ETH_SPEED
	#undef ETH_CLK_SPEED
	#undef RTK_SPEED
	#if (CONFIG_ETH_SPEED == 10)
		#define ETH_CLK_SPEED	ETH_SPEED_10
	#elif (CONFIG_ETH_SPEED == 100)
		#define ETH_CLK_SPEED	ETH_SPEED_100
	#else
		#error Wrong CONFIG_ETH_SPEED value
	#endif
#endif

int board_eth_init(bd_t *bis)
{
#ifdef CONFIG_RTL8309_SWITCH
	rtk_api_ret_t rtk_ret;
#endif
	int rc = 0;

	/* configure clocks for requested mode/speed */
	rc = clk_eth_config(ETH_CLK_MODE, ETH_CLK_SPEED);
	if (rc) {
		printf("error: can't configure ethernet clocks\n");
		return -1;
	}

	puts(ETH_MODE);
	puts(", ");

	clk_enable("rmii_if");

	reset_release("ethmac");
	clk_enable("ethmac");

	/* Reset the switch (nRESET) */
	/* According to the programming guide we must wait 1s after reset */
	io_mux("gpio7");
	gpio_direction_output(AGPIO(7), 0);
	mdelay(100);
	gpio_set_value(AGPIO(7), 1);
	mdelay(1000);

#ifdef CONFIG_RTL8309_SWITCH
	/* Initialize Realtek Ethernet switch */
	rtk_ret = rtk_switch_init();
	if (rtk_ret != RT_ERR_OK) {
		printf("error: rtk_switch_init() returned %d\n", (int)rtk_ret);
		return -1;
	}
#endif

#ifdef CONFIG_QCA833X_SWITCH
	/* Initialize ethernet switch */
	athrs17_reg_init();
#endif

	/* The PHY will be probed dynamically */
	rc = designware_initialize(DVF99_ETHERMAC_BASE, 0);
	if (rc < 0) {
		printf("error: can't register MAC\n");
		return -1;
	}

	{
		rtl8309n_rldp_controlAttr_t rldp;

		/* set MAC address for RLDP */
		rtl8309n_rldp_controlAttr_get(&rldp);
		eth_getenv_enetaddr_by_index("eth", 0, &rldp.mac.octet);
		rtl8309n_rldp_controlAttr_set(&rldp);
		/* disable RLDP */
		rtl8309n_rldp_enable_set(1);
	}

	return 0;
}

#ifdef CONFIG_RTL8309_SWITCH
int do_print_phy_status(void)
{
	rtk_api_ret_t ret;
	int i;
	rtk_port_linkStatus_t stat;
	rtk_port_speed_t speed;
	rtk_port_duplex_t dup;

	for (i = 0; i < 5; i++) {
		if (i == 1)
			i = 3;
		ret = rtk_port_phyStatus_get(i, &stat, &speed, &dup);
		if (ret != RT_ERR_OK) {
			printf("error: rtk_port_phyStatus_get() for port %d returned %d\n", i, (int)ret);
		}

		printf("port %d:\n	Link is %s\n	Speed = %d\n	%s duplex\n\n", i,
			stat == PORT_LINKUP ? "UP" : "DOWN",
			speed == PORT_SPEED_10M ? 10 : (speed == PORT_SPEED_100M ? 100 : 1000),
			dup == PORT_HALF_DUPLEX ? "HALF" : "FULL");
	}

	return 0;
}

int do_dump_phy_regs(void)
{
	rtk_api_ret_t rtk_ret;
	rtk_port_phy_data_t d;
	int i, j;

	for (i = 0; i < 5; i++) {
		if (i == 1)
			i = 3;
		for (j=0 ; j<16 ; j++) {
			rtk_ret = rtk_port_phyReg_get(i, j, &d);
			if (rtk_ret != RT_ERR_OK) {
				printf("error: rtk_port_phyReg_get() for port,reg %d,%d returned %d\n", i, j, (int)rtk_ret);
			}
			printf("phyReg_get: port=%d reg=%d  ==> 0x%x\n", i, j, (int)d);
		}
		printf("\n");
	}
	return 0;
}

int do_read_phy(int argc, char * const argv[])
{
	rtk_api_ret_t rtk_ret;
	rtk_port_phy_data_t d;
	int i, j;

	if (argc != 4)
		return CMD_RET_USAGE;

	i = simple_strtoul(argv[2], NULL, 10);
	j = simple_strtoul(argv[3], NULL, 10);

	rtk_ret = rtk_port_phyReg_get(i, j, &d);

	printf("phyReg_get: port=%d reg=%d ==> 0x%x\n", i, j, &d);

	return 0;
}

int do_write_phy(int argc, char * const argv[])
{
	rtk_api_ret_t rtk_ret;
	rtk_port_phy_data_t d;
	int i, j;

	if (argc != 5)
		return CMD_RET_USAGE;

	i = simple_strtoul(argv[2], NULL, 10);
	j = simple_strtoul(argv[3], NULL, 10);
	d = simple_strtoul(argv[4], NULL, 10);

	rtk_ret = rtk_port_phyReg_set(i, j, d);

	return 0;
}

int do_phy(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	char *cmd = argv[1];

	if (strcmp(cmd, "regs") == 0)
		return do_dump_phy_regs();

	if (strcmp(cmd, "status") == 0)
		return do_print_phy_status();

	if (strcmp(cmd, "read") == 0)
		return do_read_phy(argc, argv);

	if (strcmp(cmd, "write") == 0)
		return do_write_phy(argc, argv);

	return CMD_RET_USAGE;
}

U_BOOT_CMD(
	phy, 5, 1, do_phy,
	"PHY sub-system",
	"regs   - dump registers\n"
	"phy status - print status\n"
	"phy read <port> <reg> - read register\n"
	"phy write <port> <reg> <value> - write register\n"
);
#endif

#ifdef CONFIG_QCA833X_SWITCH
int do_read_ath_reg(int argc, char * const argv[])
{
	int reg;

	if (argc != 3)
		return CMD_RET_USAGE;

	reg = simple_strtoul(argv[2], NULL, 16);
	printf("0x%x\n", athrs17_reg_read(reg));

	return 0;
}

int do_write_ath_reg(int argc, char * const argv[])
{
	int reg, val;

	if (argc != 4)
		return CMD_RET_USAGE;

	reg = simple_strtoul(argv[2], NULL, 16);
	val = simple_strtoul(argv[3], NULL, 16);
	athrs17_reg_write(reg, val);

	return 0;
}

int do_read_phy_reg(int argc, char * const argv[])
{
	int phy, reg;

	if (argc != 4)
		return CMD_RET_USAGE;

	phy = simple_strtoul(argv[2], NULL, 16);
	reg = simple_strtoul(argv[3], NULL, 16);
	printf("0x%x\n", phy_reg_read(0, phy, reg));

	return 0;
}

int do_write_phy_reg(int argc, char * const argv[])
{
	int phy, reg, val;

	if (argc != 5)
		return CMD_RET_USAGE;

	phy = simple_strtoul(argv[2], NULL, 16);
	reg = simple_strtoul(argv[3], NULL, 16);
	val = simple_strtoul(argv[4], NULL, 16);
	phy_reg_write(0, phy, reg, val);

	return 0;
}

int do_atheros(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	char *cmd = argv[1];

	if (strcmp(cmd, "phy_read") == 0)
		return do_read_phy_reg(argc, argv);

	if (strcmp(cmd, "phy_write") == 0)
		return do_write_phy_reg(argc, argv);

	if (strcmp(cmd, "read") == 0)
		return do_read_ath_reg(argc, argv);

	if (strcmp(cmd, "write") == 0)
		return do_write_ath_reg(argc, argv);

	return CMD_RET_USAGE;
}

U_BOOT_CMD(
	atheros, 5, 1, do_atheros,
	"Atheros switch sub-system",
	"read      - read register\n"
	"atheros write     - write register\n"
	"atheros phy_read  - read PHY register\n"
	"atheros phy_write - write PHY register\n"
);
#endif
