/*
 * (C) Copyright 2014 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <spi.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/clock.h>
#include <asm/arch/pads.h>
#include <asm/arch/hardware.h>
#include <asm/arch/spi.h>
#include "tulips.h"

#define LCD_WIDTH		128
#define LCD_HEIGHT		96

#define CONFIG_SPI_LCD_RSTN	AGPIO(26)
#define CONFIG_SPI_LCD_CS	AGPIO(27)

static unsigned char image_buffer[LCD_WIDTH * LCD_HEIGHT];
static unsigned char display_buffer[LCD_WIDTH * LCD_HEIGHT / 2];
static unsigned char scale[LCD_WIDTH * 4];
static struct spi_slave *slave;

static void lcd_send_command(struct spi_slave *slave, unsigned char cmd)
{
	unsigned char tmp[2];

	tmp[0] = cmd;
	if (spi_xfer(slave, 8, tmp, NULL, SPI_XFER_BEGIN | SPI_XFER_END) != 0)
		printf("Error during SPI transaction\n");
}

static void lcd_send_2byte_command(struct spi_slave *slave,
				   unsigned char cmd1, unsigned char cmd2)
{
	unsigned char tmp[3];

	tmp[0] = cmd1;
	tmp[1] = cmd2;
	if (spi_xfer(slave, 8*2, tmp, NULL, SPI_XFER_BEGIN | SPI_XFER_END) != 0)
		printf("Error during SPI transaction\n");
}

static void lcd_send_data(struct spi_slave *slave, int page, int column,
			  int len, unsigned char *buf)
{
	unsigned char tmp[4];

	tmp[0] = 0xb0 | (page & 0xf);
	tmp[1] = 0x10 | ((column >> 4) & 0xf);
	tmp[2] = 0x00 | (column & 0xf);
	if (spi_xfer(slave, 8*3, tmp, NULL, SPI_XFER_BEGIN) != 0)
		printf("Error during SPI transaction\n");

	tmp[0] = 0xe8; /* set data direction command */
	tmp[1] = len - 1;
	if (spi_xfer(slave, 8*2, tmp, NULL, 0) != 0)
		printf("Error during SPI transaction\n");

	if (spi_xfer(slave, 8*len, buf, NULL, SPI_XFER_END) != 0)
		printf("Error during SPI transaction\n");
}

void
board_spi_lcd_init(int internal)
{
	int i, frc, pwm;
	unsigned char cmd;

	slave = spi_setup_slave(1, 0, 20000000, SPI_MODE_3);
	if (!slave)
		return;

	spi_claim_bus(slave);

	/* reset pin */
	gpio_pull_ctrl(CONFIG_SPI_LCD_RSTN, 1, 1);
	gpio_direction_output(CONFIG_SPI_LCD_RSTN, 1);
	gpio_enable(CONFIG_SPI_LCD_RSTN);

	mdelay(10);
	gpio_set_value(CONFIG_SPI_LCD_RSTN, 0);
	mdelay(10);
	gpio_set_value(CONFIG_SPI_LCD_RSTN, 1);
	mdelay(20);

	lcd_send_command(slave, 0xa2); /* icon off */
	lcd_send_command(slave, 0xae); /* display off */

	if (!internal)
		lcd_send_command(slave, 0xa9); /* set power save */

	/* partial/display duty */
	lcd_send_2byte_command(slave, 0x48, LCD_HEIGHT);
	/* ADC select: normal direction*/
	lcd_send_command(slave, 0xa0);
	/* SHL select: normal direction */
	lcd_send_command(slave, 0xc0);
	/* COM0 register select */
	lcd_send_2byte_command(slave, 0x44, 16);
	/* initial display line */
	lcd_send_2byte_command(slave, 0x40, 0x00);

	/* mode switch: 70Hz +/- 20%, BE=1, EXT=0 */
	lcd_send_2byte_command(slave, 0x38, 0x74);

	lcd_send_command(slave, 0xab); /* oscillator on */

	/* DC-DC step-up register select */
	lcd_send_command(slave, 0x64); /* booster 3x */
	mdelay(20);
	lcd_send_command(slave, 0x64+1); /* booster 4x */
	mdelay(20);
	lcd_send_command(slave, 0x64+2); /* booster 5x */
	mdelay(20);

	lcd_send_command(slave, 0x25); /* select regulator resistor */

	/* set electronic volume/reference voltage */
	lcd_send_2byte_command(slave, 0x81, 32);

	lcd_send_command(slave, 0x50); /* LCD bias: 1/5 */

	if (internal) {
		lcd_send_command(slave, 0x2c); /* VC */
		mdelay(200);
		lcd_send_command(slave, 0x2e); /* VR */
		mdelay(200);
		lcd_send_command(slave, 0x2f); /* VF */
		mdelay(200);
	}

	cmd = 0x90;
	frc = 4;
	pwm = 60;
	if (frc == 3)
		cmd |= 0x4;
	if (pwm == 60)
		cmd |= 0x2;
	lcd_send_command(slave, cmd); /* 3FRC, 45PWM */

	/* mode switch: 70Hz +/- 20%, BE=1, EXT=1 */
	lcd_send_2byte_command(slave, 0x38, 0x75);

	if (pwm == 45) {
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x80+i, 0); /* Set white mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x84+i, 3); /* Set gray level 1 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x88+i, 6); /* Set gray level 2 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x8c+i, 9); /* Set gray level 3 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x90+i, 12); /* Set gray level 4 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x94+i, 15); /* Set gray level 5 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x98+i, 18); /* Set gray level 6 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x9c+i, 21); /* Set gray level 7 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa0+i, 24); /* Set gray level 8 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa4+i, 27); /* Set gray level 9 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa8+i, 30); /* Set gray level 10 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xac+i, 33); /* Set gray level 11 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb0+i, 36); /* Set gray level 12 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb4+i, 39); /* Set gray level 13 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb8+i, 42); /* Set gray level 14 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xbc+i, 45); /* Set dark mode */
	} else {
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x80+i, 0x00); /* Set white mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x84+i, 0x06); /* Set gray level 1 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x88+i, 0x0b); /* Set gray level 2 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x8c+i, 0x10); /* Set gray level 3 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x90+i, 0x15); /* Set gray level 4 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x94+i, 0x1a); /* Set gray level 5 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x98+i, 0x1e); /* Set gray level 6 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0x9c+i, 0x23); /* Set gray level 7 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa0+i, 0x27); /* Set gray level 8 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa4+i, 0x2b); /* Set gray level 9 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xa8+i, 0x2f); /* Set gray level 10 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xac+i, 0x32); /* Set gray level 11 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb0+i, 0x35); /* Set gray level 12 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb4+i, 0x38); /* Set gray level 13 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xb8+i, 0x3a); /* Set gray level 14 mode */
		for (i = 0; i < frc; i++)
			lcd_send_2byte_command(slave, 0xbc+i, 0x3c); /* Set dark mode */
	}

	/* mode switch: 70Hz +/- 20%, BE=1, EXT=0 */
	lcd_send_2byte_command(slave, 0x38, 0x74);

	if (!internal)
		lcd_send_command(slave, 0xe1); /* release power save */

	lcd_send_command(slave, 0xaf); /* display on */

	//spi_release_bus(slave);
	//spi_free_slave(slave);
}

static void
convert_raw_image(unsigned char *out, unsigned char *in)
{
	int i, j, page, col, row, i_pos;
	int xres = LCD_WIDTH, yres = LCD_HEIGHT;

	for (i = 0; i < (xres * yres) / 2; i++)
		out[i] = 0xff;

	for (i = 0; i < yres; i++) {
		page = (i/8) * xres * 4;
		row = i & 0x7;
		col = 0;
		for (j = 0; j < xres; j++) {
			i_pos = (i * xres) + j;
			if (in[i_pos] & 8)
				out[page+col+0] /*|= (1 << row);*/ &= ~(1 << row);
			if (in[i_pos] & 4)
				out[page+col+1] /*|= (1 << row);*/ &= ~(1 << row);
			if (in[i_pos] & 2)
				out[page+col+2] /*|= (1 << row);*/ &= ~(1 << row);
			if (in[i_pos] & 1)
				out[page+col+3] /*|= (1 << row);*/ &= ~(1 << row);
			col += 4;
		}
	}
}

static void
display_fill(void)
{
	int i, off, page;
	struct spi_slave *spi = slave;

	for (page = 0; page < LCD_HEIGHT / 8; page++) {
		off = 0;
		for (i = 0; i < LCD_WIDTH * 4; i += 256) {
			lcd_send_data(spi, page, off, 256,
				      display_buffer + (page * LCD_WIDTH * 4) + i);
			off += 64;
		}
	}
}

static int display_tulips(void)
{
	convert_raw_image(display_buffer, header_data);

	display_fill();

	return 0;
}

void
board_spi_init(void)
{
	reset_release("spi2");
	clk_enable("spi2");

	gpio_disable(BGPIO(13)); /* SPI2DO */
	gpio_pull_ctrl(BGPIO(13), 1, 1);
	gpio_disable(BGPIO(14)); /* SPI2DI */
	gpio_pull_ctrl(BGPIO(14), 1, 0);
	gpio_disable(BGPIO(16)); /* SPI2CK */
	gpio_pull_ctrl(BGPIO(16), 1, 1);

	/* Reset_N */
	gpio_direction_output(AGPIO(29), 0);
	gpio_enable(AGPIO(29)); /* FXS1 */

	gpio_direction_output(AGPIO(28), 0);
	gpio_enable(AGPIO(28)); /* FXS2 */

	/* CS_N */
	gpio_direction_output(EGPIO(30), 1);
	gpio_enable(EGPIO(30)); /* FXS1 */

	gpio_direction_output(EGPIO(31), 1);
	gpio_enable(EGPIO(31)); /* FXS2 */

	/* INT_N */
	gpio_direction_input(AGPIO(17));
	gpio_disable(AGPIO(17)); /* FXS1 (EXTINT1) */

	gpio_direction_input(AGPIO(16));
	gpio_disable(AGPIO(16)); /* FXS2 (EXTINT0) */

	/* LCD-CS */
	gpio_direction_output(CONFIG_SPI_LCD_CS, 1);
	gpio_enable(CONFIG_SPI_LCD_CS);

	pads_enable("spi2");

	clk_set_rate("spi2", 20000000);

	board_spi_lcd_init(1);
	(void)display_tulips();
}

void spi_3wire(unsigned int bus, int read) { }

void spi_cs_activate(struct spi_slave *slave)
{
	if (slave->bus == 1)
		gpio_set_value(CONFIG_SPI_LCD_CS, 0);
	udelay(1);
}

void spi_cs_deactivate(struct spi_slave *slave)
{
	udelay(1);
	if (slave->bus == 1)
		gpio_set_value(CONFIG_SPI_LCD_CS, 1);
}

int do_display_cmd(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	struct spi_slave *spi = slave;
	unsigned char tmp[2];
	int len = 1;

	if (argc < 2)
		return CMD_RET_USAGE;

	tmp[0] = (unsigned char)simple_strtoul(argv[1], NULL, 16);

	if (argc > 2) {
		tmp[1] = (unsigned char)simple_strtoul(argv[1], NULL, 16);
		len = 2;
	}

	/* spi = spi_setup_slave(1, 0, 25000000, SPI_MODE_3);
	if (!spi)
		return -1;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("unable to claim SPI bus\n");
		return ret;
	} */

	ret = spi_xfer(spi, 8*len, tmp, NULL, SPI_XFER_BEGIN | SPI_XFER_END);
	if (ret)
		return -1;

	//spi_release_bus(spi);

	return 0;
}

int do_display_init_int(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	board_spi_lcd_init(1);

	return 0;
}

int do_display_init_ext(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	board_spi_lcd_init(0);

	return 0;
}

int do_display_clear(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int i, j;
	unsigned char value = 0xff;

	if (argc >= 2)
		value = simple_strtoul(argv[1], NULL, 10);

	for (i = 0; i < LCD_HEIGHT; i++) {
		for (j = 0; j < LCD_WIDTH; j++)
			image_buffer[i * LCD_WIDTH + j] = value;
		convert_raw_image(display_buffer, image_buffer);
		display_fill();
		mdelay(10);
	}

	/* for (i = 0; i < LCD_WIDTH * LCD_HEIGHT; i++)
		image_buffer[i] = value;

	convert_raw_image(display_buffer, image_buffer);

	display_fill(); */

	return 0;
}

int do_display_scale(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int i, j, dir = 0, val, start = 0;

	if (argc >= 2)
		dir = simple_strtoul(argv[1], NULL, 10);

	if (argc >= 3)
		start = simple_strtoul(argv[2], NULL, 10);

	for (i = 0; i < LCD_HEIGHT; i++) {
		val = i/6;
		if (dir)
			val = 15 - i/6;
		val -= start;
		if (val < 0)
			val = 0;
		for (j = 0; j < LCD_WIDTH; j++)
			image_buffer[i * LCD_WIDTH + j] = val;
	}

	convert_raw_image(display_buffer, image_buffer);

	display_fill();

	return 0;
}

int do_display_image(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	return display_tulips();
}

int do_display_data(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int i, mode = 0, len = LCD_WIDTH * 4, page = 0, off = 0;
	struct spi_slave *spi = slave;

	if (argc >= 2)
		mode = simple_strtoul(argv[1], NULL, 10);

	if (argc >= 3)
		page = simple_strtoul(argv[2], NULL, 10);

	if (argc >= 4)
		len = simple_strtoul(argv[3], NULL, 10);

	/* spi = spi_setup_slave(1, 0, 25000000, SPI_MODE_3);
	if (!spi)
		return -1;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("unable to claim SPI bus\n");
		return ret;
	} */

	if (len > LCD_WIDTH * 4)
		len = LCD_WIDTH * 4;

	if (mode == 0) {
		for (i = 0; i < LCD_WIDTH * 4; i += 4) {
			if (i & 32)
				scale[i]   = 0xff;
			else
				scale[i]   = 0x00;
			if (i & 16)
				scale[i+1] = 0xff;
			else
				scale[i+1] = 0x00;
			if (i & 8)
				scale[i+2] = 0xff;
			else
				scale[i+2] = 0x00;
			if (i & 4)
				scale[i+3] = 0xff;
			else
				scale[i+3] = 0x00;
		}
	} else if (mode == 1) {
		for (i = 0; i < LCD_WIDTH * 4; i += 4) {
			if (i & 7) {
				scale[i]   = 0;
				scale[i+1] = 0;
				scale[i+2] = 0;
				scale[i+3] = 0;
			} else {
				scale[i]   = 0xff;
				scale[i+1] = 0xff;
				scale[i+2] = 0xff;
				scale[i+3] = 0xff;
			}
		}
	} else if (mode == 2) {
		for (i = 0; i < LCD_WIDTH * 4; i += 4) {
			scale[i]   = 0xff;
			scale[i+1] = 0xff;
			scale[i+2] = 0xff;
			scale[i+3] = 0xff;
		}
	} else if (mode == 3) {
		for (i = 0; i < LCD_WIDTH * 4; i += 4) {
			scale[i]   = 0x00;
			scale[i+1] = 0x00;
			scale[i+2] = 0x00;
			scale[i+3] = 0x00;
		}
	}

	i = 0;
	while (i < len) {
		if ((len - i) >= 256) {
			lcd_send_data(spi, page, off, 256, scale + i);
			i += 256;
			off += 64;
		} else {
			lcd_send_data(spi, page, off, len - i, scale + i);
			break;
		}
	}

	//spi_release_bus(spi);

	return 0;
}

U_BOOT_CMD(display_cmd, 3, 0, do_display_cmd,
		"send display command",
		"<value>"
	  );

U_BOOT_CMD(display_data, 4, 0,  do_display_data,
		"display data on SPI-LCD",
		"[<mode>] [<page>] [<len>]\n"
		"mode 0: grayscale\n"
		"mode 1: black & white\n"
		"mode 2: all black\n"
		"mode 3: all white\n"
	  );

U_BOOT_CMD(display_image, 1, 0, do_display_image,
		"display image",
		""
	  );

U_BOOT_CMD(display_scale, 3, 0, do_display_scale,
		"display grayscale",
		"[direction] [offset]"
	  );

U_BOOT_CMD(display_clear, 2, 0, do_display_clear,
		"display clear",
		""
	  );

U_BOOT_CMD(display_init_int, 1, 0, do_display_init_int,
		"display init (internal voltage)",
		""
	  );

U_BOOT_CMD(display_init_ext, 1, 0, do_display_init_ext,
		"display init (external voltage)",
		""
	  );
