/*
 * (C) Copyright 2012, 2016 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <spi.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/clock.h>
#include <asm/arch/pads.h>
#include <asm/arch/hardware.h>
#include <asm/arch/dvf9a.h>

#define AUXBGEN			0x01
#define AUXADEN			0x02
#define AUXADMUX_SEL_DCIN3	0x0E
#define AUXADMUX_SEL_DCIN1	0x0A
#define SAMPLE_WINDOW_15BIT	0x07
#define SAMPLE_WINDOW_8BIT	0x00
#define START_SAMPLE		0x01
#define ADC_SAMPLING		0x01
#define R1			330
#define R2			1000

static int
aux_adc_init(int bg_val)
{
	/* Enable the Analog auxilary A/D and band gap generation */
	daif_write(DAIF_AUXEN, AUXADEN | AUXBGEN | 0xa0);

	/* Select the input DCIN3 to the Mux */
	daif_write(DAIF_AUXADCFG, AUXADMUX_SEL_DCIN3);
	daif_write(DAIF_RAADC, SAMPLE_WINDOW_15BIT << 1);

	return 0;
}

static void
aux_adc_exit(void)
{
	/* Disable the Analog auxilary A/D and band gap generation */
	daif_write(DAIF_AUXEN, 0);
}

static int
read_adc_val(void)
{
	int adc_val = daif_read(DAIF_RAADC);

	daif_write(DAIF_RAADC, adc_val | START_SAMPLE);
	while (daif_read(DAIF_RAADC) & ADC_SAMPLING)
		;

	/* Read the RAADO & DAIF_RAADOMSB register via RAPU
	   instead from the mirror */
	return (((daif_read_via_rapu(DAIF_RAADOMSB) & 0x7f) << 8) |
	       (daif_read_via_rapu(DAIF_RAADO) & 0xff));
}

static int
measure_voltage(int bgp_val)
{
	int adc_val;

	daif_write(DAIF_AUXBGPROG, bgp_val);
	adc_val = read_adc_val();

	return (adc_val * 2000) / ((1 << 15) - 1);
}

static void
bandgap_calibrate(unsigned short ref_voltage)
{
	unsigned short auxbgp;
	int min_bandgap = 0;
	int max_bandgap = 255;
	int delta;
	int current_reading;
	char str[10];

	aux_adc_init(ref_voltage);

	/* binary search */
	do {
		auxbgp = (min_bandgap + max_bandgap) / 2;

		/* read voltage */
		current_reading = measure_voltage(auxbgp);

		if (ref_voltage < current_reading)
			min_bandgap = auxbgp + 1;
		else
			max_bandgap = auxbgp - 1;
	} while ((ref_voltage != current_reading) && (min_bandgap <= max_bandgap));

	delta = ref_voltage - current_reading;
	if (delta > 20 || delta < -20) {
		/* the delta is too big */
		printf("calibration error: could not perform bandgap "
		       "calibration, wrong reading from DCIN3\n");
		goto out;
	}

	snprintf(str, 10, "%i", auxbgp);
	setenv("bgvalue", str);
	printf("Setting auxbgp to %s.\nTo save to uboot environment type \"saveenv\"\n",
	       str);

out:
	aux_adc_exit();
}

int do_calibrate(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned long vref = 0;

	//dvf9a_init();

	if (argc < 2) {
		printf("Not sufficient arguments!! "
		       "Please provide the applied voltage in mV\n");
		return 0;
	}

	vref = simple_strtoul(argv[1], NULL, 10);
	printf("Assuming external power is %lu mV\n", vref);
	bandgap_calibrate(vref);

	return 0;
}

U_BOOT_CMD(
	bg_calibration, 2, 1, do_calibrate,
	"calibrate bandgap",
	"reference_voltage"
);

static void
dcdc_calibrate(int dcin2_voltage, int ref_voltage)
{
	char str[10];
	int dcin2_measured = 0, dcdc_measured = 0, dcdc_voltage;
	int i, dpu_ctl1, dcdc = 0;

	/* Select the mux VSENS_pad (VSENS pin of the VDDC_core DCDC) */
	daif_write(DAIF_DPUCTL4, 0x40);

	/* Enable the Analog auxilary A/D */
	daif_write(DAIF_AUXEN, AUXADEN | AUXBGEN);

	/* Select the DCIN2 input */
	daif_write(DAIF_AUXADCFG, 0x0c);
	daif_write(DAIF_RAADC, SAMPLE_WINDOW_15BIT << 1);

	for (i = 0; i < 8; i++)
		dcin2_measured += read_adc_val();

	/* Select the DCDC input */
	daif_write(DAIF_AUXADCFG, 0x1d);
	daif_write(DAIF_RAADC, SAMPLE_WINDOW_15BIT << 1);

	for (i = 0; i < 8; i++)
		dcdc_measured += read_adc_val();

	dcdc_voltage = (dcdc_measured * dcin2_voltage) / dcin2_measured;

	dpu_ctl1 = daif_read(DAIF_DPUCTL1);
	dcdc = (dpu_ctl1 & 0x1f);
	dpu_ctl1 &= ~0x1f;

	while ((dcdc_voltage < ref_voltage) && (dcdc < 31)) {
		dcdc++;
		daif_write(DAIF_DPUCTL1, dpu_ctl1 | dcdc);

		mdelay(10);

		dcdc_measured = 0;
		for (i = 0; i < 8; i++)
			dcdc_measured += read_adc_val();

		dcdc_voltage = (dcdc_measured * dcin2_voltage) / dcin2_measured;
	};

	while ((dcdc_voltage > ref_voltage) && (dcdc > 0)) {
		dcdc--;
		daif_write(DAIF_DPUCTL1, dpu_ctl1 | dcdc);

		mdelay(10);

		dcdc_measured = 0;
		for (i = 0; i < 8; i++)
			dcdc_measured += read_adc_val();

		dcdc_voltage = (dcdc_measured * dcin2_voltage) / dcin2_measured;
	};

	snprintf(str, 10, "%i", dcdc);
	setenv("dcdc", str);
	printf("Setting DCDC to %s.\nTo save to uboot environment type \"saveenv\"\n",
	       str);

	aux_adc_exit();
}

int do_dcdc(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned long dcin2 = 0, vref = 0;

	if (argc < 3) {
		printf("Missing arguments!\n"
		       "Please provide the reference voltage at DCIN2 (in mV) "
		       "and the target voltage (in mV)\n");
		return 0;
	}

	dcin2 = simple_strtoul(argv[1], NULL, 10);
	vref = simple_strtoul(argv[2], NULL, 10);
	dcdc_calibrate(dcin2, vref);

	return 0;
}

U_BOOT_CMD(
	dcdc_calibration, 3, 1, do_dcdc,
	"calibrate DCDC",
	"reference_voltage"
);
