#include <common.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/clock.h>
#include <asm/arch/hardware.h>

#define CSS_MAGIC			0xDADADADA
#define CSS_IMG_VERSION			0x0100
#define CSS_ABI_VERSION			0x0203

/* CSS image headers*/

struct boot_config {
	unsigned long dram_addr;
	unsigned long dram_size;
	int uart;
	unsigned long cfifo_l2c;
	unsigned long cfifo_c2l;
};

struct css_img_hdr {
	unsigned long magic;
	unsigned long parts;
	unsigned long dc_addr; /* unused */
	unsigned long boot_config;
	unsigned long elf_md5_addr;
	u16 abi_version;
	u16 img_version;
	char info[52];
} __attribute__((__packed__));

struct css_part_hdr {
	unsigned long size;
	unsigned long rsize;
	unsigned long dest_addr;
	unsigned char md5[16];
} __attribute__((__packed__));



static int css_state = 0;

/* MUST be 1M aligned */
unsigned long css_dram = CONFIG_CSS_DRAM_ADDR;

static void css_power_up(void)
{
    reset_release("css");
    clk_enable("css");

    reset_release("css_etm");
#ifdef CONFIG_CSS_DEBUG
    clk_enable("css_etm");
#endif /* CONFIG_CSS_DEBUG */

    clk_enable("css_arm926");

    /* release RF from reset */
    //gpio_set_value(CONFIG_CSS_RF_RESET_GPIO, 1);
}

static void css_power_down(void)
{
    /* Shut down execution on the CSS */
    clk_disable("css_arm926");
#ifdef CONFIG_CSS_DEBUG
    clk_disable("css_etm");
#endif /* CONFIG_CSS_DEBUG */
    reset_set("css_arm926");
    reset_set("css_etm");
    clk_enable("css_etm");
    clk_enable("css_arm926");
    mdelay(1);
    clk_disable("css_arm926");
    clk_disable("css_etm");

    /* Put all devices back into reset */
    reset_set("css_adpcm");
    reset_set("css_bmp");
    reset_set("css_gdmac");
    reset_set("css_timer1");
    reset_set("css_timer2");
    reset_set("uart2");
    reset_set("dclass");

    /* correctly reset BMP, ADPCMs, DRT, RFAPU (need clock during reset) */
    clk_enable("css_adpcm");
    clk_enable("css_bmp");
    clk_enable("drt_com");
    clk_enable("rfapu");
    mdelay(1);
    clk_disable("css_adpcm");
    clk_disable("css_bmp");
    clk_disable("drt_com");
    clk_disable("rfapu");
    
    /* make sure RF is in reset */
    //gpio_set_value(CONFIG_CSS_RF_RESET_GPIO, 0);

    /* Shutdown rest of the CSS */
    clk_disable("css");
    reset_set("css");
    clk_enable("css");
    mdelay(2);
    clk_disable("css");
}

static int got_firmware(const void *css, size_t size)
{
    unsigned long csize;
    void *virt_elf_md5;
    void *virt_boot_config;
    struct css_img_hdr *img_hdr;
    struct css_part_hdr *itcm_hdr;
#if (CONFIG_CSS_AHBRAM > 0)
    struct css_part_hdr *ahbram_hdr;
#endif /* CONFIG_CSS_AHBRAM */
    struct css_part_hdr *dram_hdr;
    struct css_part_hdr *elf_hdr = NULL;
    struct boot_config bootcfg;

    if (css_state != 0) {
	printf("CSS firmware already loaded!\n");
	return 0;
    }

    if (css == 0) {
	printf("loading CSS firmware failed\n");
	return -1;
    }

    css_power_up();

    /* check image header */
    img_hdr = (struct css_img_hdr *)css;

    if (img_hdr->parts <
#if (CONFIG_CSS_AHBRAM > 0)
	3
#else /* CONFIG_CSS_AHBRAM */	 
	2
#endif /* CONFIG_CSS_AHBRAM */	 
	) {
	printf("invalid number of parts: %lu\n", img_hdr->parts);
	goto out_err;
    }

    if (img_hdr->magic != CSS_MAGIC) {
	printf("invalid image magic\n");
	goto out_err;
    }

    if (img_hdr->img_version != CSS_IMG_VERSION) {
	printf("image version mismatch (got: v%x, ""expected: v%x)\n", img_hdr->img_version, CSS_IMG_VERSION);
	goto out_err;
    }

    if (img_hdr->abi_version != CSS_ABI_VERSION) {
	printf("ABI mismatch (got: v%x, expected: v%x)\n",img_hdr->abi_version, CSS_ABI_VERSION);
	goto out_err;
    }

    debug("loading %s\n", img_hdr->info);

    csize = sizeof(*img_hdr);
    itcm_hdr   = (struct css_part_hdr *)(css + csize);
    csize += sizeof(*itcm_hdr) + itcm_hdr->size;
    if ((csize > size) || (itcm_hdr->size > CONFIG_CSS_ITCM_SIZE)) {
	printf("ITCM size too big! sz=0x%08lX, max=0x%08X\n", itcm_hdr->size, CONFIG_CSS_ITCM_SIZE);
	goto out_err;
    }

#if (CONFIG_CSS_AHBRAM > 0)
    ahbram_hdr = (struct css_part_hdr *)(css + csize);
    csize += sizeof(*ahbram_hdr) + ahbram_hdr->size;
    if ((csize > size) || (ahbram_hdr->size > CONFIG_CSS_AHBRAM_SIZE)) {
	printf("AHBRAM size too big! sz=0x%08lX, max=0x%08X\n", ahbram_hdr->size, CONFIG_CSS_AHBRAM_SIZE);
	goto out_err;
    }
#endif /* CONFIG_CSS_AHBRAM */

    dram_hdr   = (struct css_part_hdr *)(css + csize);
    csize += sizeof(*dram_hdr) + dram_hdr->size;
    if ((csize > size) || (dram_hdr->size > dram_hdr->rsize)) {
	printf("DRAM size too big! sz=0x%08lX, max=0x%08X\n", dram_hdr->size, CONFIG_CSS_DRAM_SIZE);
	goto out_err;
    }

    if (img_hdr->parts == 
#if (CONFIG_CSS_AHBRAM > 0)
	4
#else /* CONFIG_CSS_AHBRAM */
	3
#endif /* CONFIG_CSS_AHBRAM */	
	) {
	elf_hdr = (struct css_part_hdr *)(css + csize);
	csize += sizeof(*dram_hdr) + elf_hdr->size;
	if (csize > size) {
	    printf("ELF size too big! sz=0x%08lX\n", elf_hdr->size);
	    goto out_err;
	}

	debug("MD5 sum of ELF file: 0x%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X"
	      "%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X\n",
	      elf_hdr->md5[0],
	      elf_hdr->md5[1],
	      elf_hdr->md5[2],
	      elf_hdr->md5[3],
	      elf_hdr->md5[4],
	      elf_hdr->md5[5],
	      elf_hdr->md5[6],
	      elf_hdr->md5[7],
	      elf_hdr->md5[8],
	      elf_hdr->md5[9],
	      elf_hdr->md5[10],
	      elf_hdr->md5[11],
	      elf_hdr->md5[12],
	      elf_hdr->md5[13],
	      elf_hdr->md5[14],
	      elf_hdr->md5[15]);
    }

    if ((itcm_hdr->dest_addr < CONFIG_CSS_ITCM_BASE) ||
	(itcm_hdr->dest_addr > (CONFIG_CSS_ITCM_BASE + CONFIG_CSS_ITCM_SIZE)) ||
	((itcm_hdr->dest_addr + itcm_hdr->size) > (CONFIG_CSS_ITCM_BASE + CONFIG_CSS_ITCM_SIZE))) {
	printf("ITCM destination pointer out of range! ptr=0x%08lX\n", itcm_hdr->dest_addr);
	goto out_err;
    }

#if (CONFIG_CSS_AHBRAM > 0)
    if ((ahbram_hdr->dest_addr < CONFIG_CSS_AHBRAM_BASE) ||
	(ahbram_hdr->dest_addr > (CONFIG_CSS_AHBRAM_BASE + CONFIG_CSS_AHBRAM_SIZE)) ||
	((ahbram_hdr->dest_addr + ahbram_hdr->size) > (CONFIG_CSS_AHBRAM_BASE + CONFIG_CSS_AHBRAM_SIZE))) {
	printf("AHBRAM destination pointer out of range! ptr=0x%08lX\n", ahbram_hdr->dest_addr);
	goto out_err;
    }
#endif /* CONFIG_CSS_AHBRAM */

#if 0
    if ((img_hdr->dc_addr > dram_hdr->size) ||
	((img_hdr->dc_addr + 4) > dram_hdr->size)) {
	printf("DC pointer out of range! ptr=0x%08lX\n", img_hdr->dc_addr);
	goto out_err;
    }
#endif

    if ((img_hdr->elf_md5_addr > dram_hdr->size) ||
	((img_hdr->elf_md5_addr + 4) > dram_hdr->size)) {
	printf("Elf MD5 pointer out of range! ptr=0x%08lX\n", img_hdr->elf_md5_addr);
	goto out_err;
    }

    /* clear memory */
    memset((void *)CONFIG_CSS_ITCM_BASE, 0, CONFIG_CSS_ITCM_SIZE);
#if (CONFIG_CSS_AHBRAM > 0)
    if (ahbram_hdr->size) {
	memset((void *)CONFIG_CSS_AHBRAM_BASE, 0, CONFIG_CSS_AHBRAM_SIZE); 
    }
#endif /* CONFIG_CSS_AHBRAM */
    memset((void *)css_dram, 0, CONFIG_CSS_DRAM_SIZE);

    /* copy CSS code and data */
    memcpy((void *)itcm_hdr->dest_addr, (void *)itcm_hdr + sizeof(*itcm_hdr), itcm_hdr->size);
#if (CONFIG_CSS_AHBRAM > 0)
    if (ahbram_hdr->size > 0) {
	memcpy((void *)ahbram_hdr->dest_addr, (void *)ahbram_hdr + sizeof(*ahbram_hdr), ahbram_hdr->size); 
    }
#endif /* CONFIG_CSS_AHBRAM */
    memcpy((void *)(css_dram + dram_hdr->dest_addr), (void *)dram_hdr + sizeof(*dram_hdr), dram_hdr->size);

    debug("DRAM size: %lu\n", dram_hdr->size);

    if (img_hdr->parts == 
#if (CONFIG_CSS_AHBRAM > 0)
	4
#else /* CONFIG_CSS_AHBRAM */
	3
#endif /* CONFIG_CSS_AHBRAM */
	) {
		/* fill elf md5 sum of CSS */
		virt_elf_md5 = (void *)(css_dram + img_hdr->elf_md5_addr);
		memcpy(virt_elf_md5, elf_hdr->md5, 16);
	}

    /* fill the boot config structure for the CSS */
    bootcfg.dram_addr = (uint32_t)css_dram;
    bootcfg.dram_size = CONFIG_CSS_DRAM_SIZE;
    bootcfg.uart = 1;
    bootcfg.cfifo_l2c = 0;
    bootcfg.cfifo_c2l = 0;
    virt_boot_config = (void *)(CONFIG_CSS_ITCM_BASE + img_hdr->boot_config);
    memcpy(virt_boot_config, (void *)&bootcfg, sizeof(bootcfg));

    /* release ARM from reset */
    clk_disable("css_arm926");
    reset_release("css_arm926");
    clk_enable("css_arm926");
    mdelay(2);

    css_state = 1;

    printf("CSS firmware successfully loaded\n");

    return 0;

out_err:

    printf("failed to load CSS firmware\n");
    css_power_down();

    return -1;
}

int css_load(const void *css, size_t size)
{
    return got_firmware(css, size);
}


