/*
 * (C) Copyright 2012, DSP Group
 *
 * based on spr_smi.c
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <flash.h>
#include <linux/err.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/concat.h>

#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/clock.h>

#if !defined(CONFIG_SYS_NO_FLASH)

/* common registers */
#define QSPIC_CFG		0x000
#define QSPIC_STS		0x008
#define QSPIC_ISR		0x00c
#define QSPIC_IER		0x010
#define QSPIC_ICR		0x014

#define QSPIC_STS_BUSY		(1UL << 31)
#define QSPIC_STS_IDLE		(1UL <<  0)

#define QSPIC_INT_APBTRSFR_DONE   (1 << 6)

/* write and read buffer exposed via register */
#define QSPIC_FWD		0x020
#define QSPIC_FRD		0x040
#define QSPIC_DATASIZE		0x020

/* per-chipselect register offsets */
#define QSPIC_RDTRSTP		0x03c
#define QSPIC_RDCMD		0x040
#define QSPIC_PPTRSTP		0x044
#define QSPIC_PPCMD		0x048
#define QSPIC_MEMADD1		0x054
#define QSPIC_MEMADD2		0x050

#define QSPIC_RDTRSTP_OPCODE(x)		((x) <<  0)
#define QSPIC_RDTRSTP_ADD(x)		((x) <<  2)
#define QSPIC_RDTRSTP_DUMMY(x)		((x) <<  4)
#define QSPIC_RDTRSTP_DIN(x)		((x) <<  8)
#define QSPIC_RDTRSTP_ADDMODE(x)	((x) << 10)
#define QSPIC_RDTRSTP_DUMMY_SIZE(x)	((x) << 12)
#define QSPIC_RDTRSTP_HAS_OPCODE	(1   << 19)
#define QSPIC_RDTRSTP_HAS_MODE		(1   << 20)

#define QSPIC_PPTRSTP_OPCODE(x)		((x) <<  0)
#define QSPIC_PPTRSTP_ADD(x)		((x) <<  2)
#define QSPIC_PPTRSTP_DUMMY(x)		((x) <<  4)
#define QSPIC_PPTRSTP_DOUT(x)		((x) <<  6)
#define QSPIC_PPTRSTP_ADDMODE(x)	((x) << 10)
#define QSPIC_PPTRSTP_HAS_OPCODE	(1   << 19)

/* common chip select registers */
#define QSPIC_TRSTP		0x300
#define QSPIC_CMD		0x304
#define QSPIC_ADD		0x308
#define QSPIC_CSTIMING		0x408

#define QSPIC_TRSTP_OPCODE(x)	((x) <<  0)
#define QSPIC_TRSTP_ADD(x)	((x) <<  2)
#define QSPIC_TRSTP_DUMMY(x)	((x) <<  4)
#define QSPIC_TRSTP_DOUT(x)	((x) <<  6)
#define QSPIC_TRSTP_DIN(x)	((x) <<  8)
#define QSPIC_TRSTP_ADDMODE(x)	((x) << 10)
#define QSPIC_TRSTP_HAS_OPCODE	(1   << 19)
#define QSPIC_TRSTP_SIZE(x)	((x) << 20)

#define TRSTP_SERIAL		0
#define TRSTP_DUAL		1
#define TRSTP_QUAD		2

#define TRSTP_ADD_3B		1
#define TRSTP_ADD_4B		2

#define QSPIC_CMD_OPCODE(x)	((x) <<  0)
#define QSPIC_CMD_TYPE(x)	((x) <<  8)
#define QSPIC_CMD_READ		(1   << 13)
#define QSPIC_CMD_WRITE		(1   << 14)
#define QSPIC_CMD_CS(x)		((x) << 30)
#define QSPIC_CMD_START		(1   << 31)

#define QSPIC_CMD_TYPE_SECT_ERASE	QSPIC_CMD_TYPE(0)
#define QSPIC_CMD_TYPE_CHIP_ERASE	QSPIC_CMD_TYPE(1)
#define QSPIC_CMD_TYPE_READ_STATUS	QSPIC_CMD_TYPE(5)
#define QSPIC_CMD_TYPE_WRITE_STATUS	QSPIC_CMD_TYPE(6)
#define QSPIC_CMD_TYPE_OTHER		QSPIC_CMD_TYPE(9)

/* Common SPI commands */
#define CMD_RDSR	(QSPIC_CMD_OPCODE(0x05) | \
			 QSPIC_CMD_TYPE_READ_STATUS | \
			 QSPIC_CMD_READ)
#define CMD_WRSR	(QSPIC_CMD_OPCODE(0x01) | \
			 QSPIC_CMD_TYPE_WRITE_STATUS | \
			 QSPIC_CMD_WRITE)
#define CMD_RDID	(QSPIC_CMD_OPCODE(0x9f) | \
			 QSPIC_CMD_TYPE_OTHER | \
			 QSPIC_CMD_READ)
#define CMD_CE		(QSPIC_CMD_OPCODE(0xc7) | \
			 QSPIC_CMD_TYPE_CHIP_ERASE)
#define CMD_SE		(QSPIC_CMD_OPCODE(0x20) | \
			 QSPIC_CMD_TYPE_SECT_ERASE)
#define CMD_WREN	(QSPIC_CMD_OPCODE(0x06) | \
			 QSPIC_CMD_TYPE_OTHER)
#define CMD_RSTEN	(QSPIC_CMD_OPCODE(0x66) | \
			 QSPIC_CMD_TYPE_OTHER)
#define CMD_RST		(QSPIC_CMD_OPCODE(0x99) | \
			 QSPIC_CMD_TYPE_OTHER)

#define BUSY	0x01
#define WEL	0x02

flash_info_t flash_info[CONFIG_SYS_MAX_FLASH_BANKS];

struct chipsel {
	flash_info_t	*info;
	int		cs;
	char		*name;

	void		*shadow_buf;

	int (*wait)(struct chipsel *flash);
	void (*exit)(struct chipsel *flash);

	unsigned	sector_size;	/* erasable unit size */
	unsigned	n_sectors;	/* number of erase sectors */
	unsigned short	page_size;	/* programmable unit size */
	u32		aux_cfg;

	unsigned short	read_freq;
	unsigned short	write_freq;
};

static int dspg_qspic_write(struct chipsel *flash, uchar *src, ulong dest_addr, ulong length);

static struct chipsel qspic_cs[CONFIG_SYS_MAX_FLASH_BANKS];
static void *regs;

struct qspi_info {
	const char	*name;
	u32		jedec_id;
	u16		ext_id;

	unsigned short	read_freq;	/* maximum read frequency in MHz */
	unsigned short	write_freq;	/* maximum write frequency in MHz */
	unsigned short	page_size;	/* programmable unit size */
	unsigned	sector_size;	/* erasable unit size */
	unsigned	n_sectors;	/* number of erase sectors */

	int (*init)(struct chipsel *flash, const struct qspi_info *info);

	int (*wait)(struct chipsel *flash);
	void (*exit)(struct chipsel *flash);

	u32	read_cfg;
	u32	read_cmd;
	u32	write_cfg;
	u32	write_cmd;
	u32	aux_cfg;	/* ADD_MODE and *_TRNSFR_TYPE fields */
};

static inline struct chipsel *info_to_chipsel(flash_info_t *flash)
{
	if (flash == &flash_info[0])
		return &qspic_cs[0];
#if CONFIG_SYS_MAX_FLASH_BANKS > 1
	return &qspic_cs[1];
#else
	return NULL;
#endif
}

int write_buff(flash_info_t *info, uchar *src, ulong dest_addr, ulong length)
{
	struct chipsel *flash = info_to_chipsel(info);

	return dspg_qspic_write(flash, src, dest_addr, length);
}

void flash_print_info(flash_info_t *info)
{
	int i;
	if (info->flash_id == FLASH_UNKNOWN) {
		puts("missing or unknown FLASH type\n");
		return;
	}
	printf("  Size: %ld MB in %d Sectors\n",
	       info->size >> 20, info->sector_count);

	puts("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; ++i) {
#ifdef CONFIG_SYS_FLASH_EMPTY_INFO
		int size;
		int erased;
		u32 *flash;

		/*
		 * Check if whole sector is erased
		 */
		size = (info->size) / (info->sector_count);
		flash = (u32 *) info->start[i];
		size = size / sizeof(int);

		while ((size--) && (*flash++ == ~0))
			;

		size++;
		if (size)
			erased = 0;
		else
			erased = 1;

		if ((i % 5) == 0)
			printf("\n");

		printf(" %08lX%s",
		       info->start[i], erased ? " E" : "  ");
#else
		if ((i % 5) == 0)
			printf("\n   ");
		printf(" %08lX", info->start[i]);
#endif
	}
	putc('\n');
	return;
}

static void dspg_qspic_cmd(int cs, u32 cmd, u32 cfg)
{
	u32 reg;

	/* wait until controller is idle */
	reg = readl(regs + QSPIC_STS);
	while ((reg & QSPIC_STS_BUSY) || !(reg & QSPIC_STS_IDLE))
		reg = readl(regs + QSPIC_STS);

	cfg |= QSPIC_TRSTP_HAS_OPCODE;
	cmd |= QSPIC_CMD_CS(cs) |
	       QSPIC_CMD_START;

	writel(cfg, regs + QSPIC_TRSTP);
	writel(cmd, regs + QSPIC_CMD);

	while (!(readl(regs + QSPIC_ISR) & QSPIC_INT_APBTRSFR_DONE))
		;
	writel(QSPIC_INT_APBTRSFR_DONE, regs + QSPIC_ICR);
	while (readl(regs + QSPIC_ISR) & QSPIC_INT_APBTRSFR_DONE)
		;
	writel(0, regs + QSPIC_ICR); /* not self clearing */
}

static void dspg_qspic_readbuf(void *dst, unsigned len)
{
	const u8 *p = (const u8 *)(regs + QSPIC_FRD);
	u8 *buf = dst;

	/* the data is backwards, so move to the end of the buffer */
	p += QSPIC_DATASIZE;

	while (len--)
		*buf++ = *--p;
}

static void dspg_qspic_writebuf(const void *src, unsigned len)
{
	u32 *p = (u32 *)(regs + QSPIC_FWD + QSPIC_DATASIZE);
	const u8 *buf = src;
	u32 data;
	u8 *tmp = (u8 *)&data;

	if (len > QSPIC_DATASIZE)
		len = QSPIC_DATASIZE;

	/* the data is backwards, so move to the end of the buffer */
	while (len) {
		data = 0xffffffff;
		if (len) { len--; tmp[3] = *buf++; }
		if (len) { len--; tmp[2] = *buf++; }
		if (len) { len--; tmp[1] = *buf++; }
		if (len) { len--; tmp[0] = *buf++; }
		*--p = data;
	}
}

static void dspg_qspic_aux_cmd(struct chipsel *flash, u32 cmd)
{
	u32 cfg = flash->aux_cfg;

	cfg &= ~QSPIC_TRSTP_ADDMODE(3);	/* no address required */

	dspg_qspic_cmd(flash->cs, cmd, cfg);
}

static void dspg_qspic_aux_read(struct chipsel *flash, u32 cmd, void *buf, unsigned len)
{
	u32 cfg = flash->aux_cfg;

	cfg &= ~QSPIC_TRSTP_ADDMODE(3);	/* no address required */
	cfg |= QSPIC_TRSTP_SIZE(len);

	dspg_qspic_cmd(flash->cs, cmd, cfg);
	dspg_qspic_readbuf(buf, len);
}

static void dspg_qspic_aux_write(struct chipsel *flash, u32 cmd, const void *buf, unsigned len)
{
	u32 cfg = flash->aux_cfg;

	cfg &= ~QSPIC_TRSTP_ADDMODE(3);	/* no address required */
	cfg |= QSPIC_TRSTP_SIZE(len);

	dspg_qspic_writebuf(buf, len);
	dspg_qspic_cmd(flash->cs, cmd, cfg);
}

#define MAX_READY_WAIT 400

static int dspg_qspic_wait(struct chipsel *flash)
{
	unsigned long deadline = 0;
	int ret;

	while (deadline < MAX_READY_WAIT) {
		ret = flash->wait(flash);
		if (ret <= 0)
			return ret;
		udelay(1000);
		deadline++;
	}

	return -ETIMEDOUT;
}

static void dspg_qspic_write_enable(struct chipsel *flash)
{
	u8 sr1;

	dspg_qspic_aux_cmd(flash, CMD_WREN);

	do {
		dspg_qspic_aux_read(flash, CMD_RDSR, &sr1, 1);
	} while ((sr1 & WEL) == 0);
}

static uchar page_buf[256] __attribute__ ((aligned(4)));

static int dspg_qspic_write_sect(struct chipsel *flash, u32 off, u32 len, uchar *buf)
{
	unsigned long src = (unsigned long)0x00200000;
	unsigned long dest = (unsigned long)off;
	u32 real_len = len;
	int i;

	if (is_dvf97())
		src = (unsigned long)page_buf;

	if (real_len & 3)
		real_len = (real_len + 3) & ~3;

	memcpy((void *)src, buf, len);
	memcpy((void *)(src + len), (void *)(off + len), real_len - len);

	dspg_qspic_write_enable(flash);

	for (i = 0; i < real_len; i += 4)
		*(unsigned long *)(dest + i) = *(unsigned long *)(src + i);

	/* drain write buffer */
	i = 0;
	__asm__ __volatile__("mcr p15, 0, %0, c7, c10, 4" : "=r"(i));

	return dspg_qspic_wait(flash);
}

static int dspg_qspic_write(struct chipsel *flash, uchar *src, ulong dest_addr, ulong length)
{
	u32 page_offset, page_size;
	int ret;

	if (length == 0)
		return 0;

	/* do all the bytes fit onto one page? */
	page_offset = dest_addr & (flash->page_size - 1);
	if (page_offset + length <= flash->page_size) {
		ret = dspg_qspic_write_sect(flash, dest_addr, length, src);
	} else {
		page_size = flash->page_size - page_offset;
		while (length) {
			ret = dspg_qspic_write_sect(flash, dest_addr, page_size, src);
			if (ret)
				return ret;

			length -= page_size;
			dest_addr += page_size;
			src += page_size;

			page_size = (length > flash->page_size) ?
				flash->page_size : length;
		}
	}

	return 0;
}

static int dspg_qspic_erase_sector(flash_info_t *info, flash_sect_t sect)
{
	struct chipsel *flash = info_to_chipsel(info);

	dspg_qspic_write_enable(flash);

	writel(sect * (info->size / info->sector_count), regs + QSPIC_ADD);
	dspg_qspic_cmd(flash->cs, CMD_SE, flash->aux_cfg);

	return dspg_qspic_wait(flash);
}

static int dspg_qspic_wait_generic(struct chipsel *flash)
{
	u8 status;

	dspg_qspic_aux_read(flash, CMD_RDSR, &status, 1);

	return status & (WEL | BUSY);
}

static void dspg_qspic_wait_write_disable(struct chipsel *flash)
{
	do {
		udelay(1000);
	} while (dspg_qspic_wait_generic(flash));
}

static void dspg_qspic_reset_generic(struct chipsel *flash)
{
	dspg_qspic_aux_cmd(flash, CMD_RSTEN);
	dspg_qspic_aux_cmd(flash, CMD_RST);
}

/*
 * ATMEL specific functions...
 */

static int dspg_qspic_init_atmel(struct chipsel *flash, const struct qspi_info *info)
{
	u8 status = 0;

	/* un-protect all sectors */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, CMD_WRSR, &status, 1);

	return 0;
}

static int dspg_qspic_wait_atmel(struct chipsel *flash)
{
	u8 status;

	dspg_qspic_aux_read(flash, CMD_RDSR, &status, 1);

	/* first check ready/busy */
	if (status & BUSY)
		return 1;
	else
		/* if ready, check Erase/Program Error */
		return (status & (1 << 5)) ? -EIO : 0;
}

static int dspg_qspic_init_sst16_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 unlock[6] = { 0, 0, 0, 0, 0, 0 };

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x38) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	/* un-protect all sectors */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, 0x42, &unlock, 6);

	return 0;
}

static int dspg_qspic_init_sst32_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 unlock[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x38) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	/* un-protect all sectors */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, 0x42, &unlock, 10);

	return 0;
}

static int dspg_qspic_wait_sst(struct chipsel *flash)
{
	u8 status;

	dspg_qspic_aux_read(flash, CMD_RDSR, &status, 1);

	return status & (1<<7); /* check WIP */
}

/*
 * WINBOND specific functions...
 */

static int dspg_qspic_init_winbond_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr2, params;

	/* volatile set QE in status register 2 */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x35) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr2, 1);
	sr2 |= 1 << 1;
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x50) | QSPIC_CMD_TYPE_OTHER);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x31) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, &sr2, 1);

	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x98) | QSPIC_CMD_TYPE_OTHER);

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x38) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	/* set read parameters: 6 dummy cyles; must be done in QPI mode */
	params = 0x2 << 4;
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0xc0) | QSPIC_CMD_WRITE |
		QSPIC_CMD_TYPE_OTHER, &params, 1);

	return 0;
}

static int dspg_qspic_init_winbond16_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr[2];

	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x05) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[0], 1);

	/* volatile set QE in status register 2 */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x35) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[1], 1);
	sr[1] |= 1 << 1;
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x50) | QSPIC_CMD_TYPE_OTHER);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x01) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, sr, 2);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	return 0;
}

static int dspg_qspic_init_winbond64_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr[2], params;

	/* volatile set QE in status register 2 */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x05) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[0], 1);
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x35) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[1], 1);
	sr[1] |= 1 << 1;
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x50) | QSPIC_CMD_TYPE_OTHER);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x01) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, sr, 2);

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x38) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	/* set read parameters: 6 dummy cyles; must be done in QPI mode */
	params = 0x2 << 4;
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0xc0) | QSPIC_CMD_WRITE |
		QSPIC_CMD_TYPE_OTHER, &params, 1);

	return 0;
}

static int dspg_qspic_init_winbond256_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr2, sr3, params;

	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xe9) | QSPIC_CMD_TYPE_OTHER);

	/* volatile set QE in status register 2 */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x35) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr2, 1);
	sr2 |= 1 << 1;
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x50) | QSPIC_CMD_TYPE_OTHER);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x31) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, &sr2, 1);

	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x98) | QSPIC_CMD_TYPE_OTHER);

	if (info->read_cfg & QSPIC_RDTRSTP_ADDMODE(TRSTP_ADD_4B)) {
		/* enter 4B mode */
		do {
			dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xb7) | QSPIC_CMD_TYPE_OTHER);
			dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x15) |
				QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr3, 1);
		} while ((sr3 & 0x1) != 0x1);
	}

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x38) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QSPI & 4B mode! */
	flash->aux_cfg = info->aux_cfg;

	/* set read parameters: 6 dummy cyles; must be done in QPI mode */
	params = 0x2 << 4;
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0xc0) | QSPIC_CMD_WRITE |
		QSPIC_CMD_TYPE_OTHER, &params, 1);

	return 0;
}

static int dspg_qspic_init_mx66l_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr[2];

	dspg_qspic_aux_read(flash, CMD_RDSR, &sr[0], 1);
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x15) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[1], 1);
	sr[0] &= ~(0x3c); /* disable block protection */
	sr[0] |= 1 << 6;  /* set QE */
	sr[1] &= 0x3f;    /* set 6 dummy cycles */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, CMD_WRSR, sr, 2);
	dspg_qspic_wait_write_disable(flash);

	if (info->read_cfg & QSPIC_RDTRSTP_ADDMODE(TRSTP_ADD_4B)) {
		/* enter 4B mode */
		dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xb7) | QSPIC_CMD_TYPE_OTHER);
		do {
			dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x15) |
				QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, sr, 1);
		} while ((sr[0] & 0x20) != 0x20);
	}

	/* enter QPI mode */
	dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0x35) | QSPIC_CMD_TYPE_OTHER);

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	return 0;
}

static int dspg_qspic_init_n25q_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	unsigned char sr;

	/* status register */
	dspg_qspic_aux_read(flash, CMD_RDSR, &sr, 1);
	sr &= ~(0x5c); /* disable block protection */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, CMD_WRSR, &sr, 1);
	dspg_qspic_wait_write_disable(flash);

	/* volatile configuration register */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x85) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr, 1);
	sr &= ~(0xf0);
	sr |= 6 << 4; /* set 6 dummy cycles */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x81) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, &sr, 1);
	dspg_qspic_wait_write_disable(flash);

	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x85) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr, 1);

	/* enhanced volatile configuration register */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x65) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr, 1);
	sr &= ~(1 << 7);  /* set QE */
	sr |=   1 << 6;   /* disable dual mode */
	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x61) |
		QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, &sr, 1);

	flash->aux_cfg = info->aux_cfg;

	dspg_qspic_wait_write_disable(flash);

	if (info->read_cfg & QSPIC_RDTRSTP_ADDMODE(TRSTP_ADD_4B)) {
		/* enter 4B mode */
		dspg_qspic_write_enable(flash);
		dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xb7) | QSPIC_CMD_TYPE_OTHER);
		dspg_qspic_wait_write_disable(flash);
	}

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	return 0;
}

static int dspg_qspic_init_gd25q_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr;

	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x05) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr, 1);

	/* non-volatile set QE in status register 1 */
	if (!(sr & (1 << 6)) || (sr & 0x3C)) {
		sr |= 1 << 6;
		sr &= ~0x3C;
		dspg_qspic_write_enable(flash);
		udelay(100);
		dspg_qspic_aux_write(flash, QSPIC_CMD_OPCODE(0x01) |
			QSPIC_CMD_TYPE_WRITE_STATUS | QSPIC_CMD_WRITE, &sr, 1);
		dspg_qspic_wait_write_disable(flash);
	}

	if (info->read_cfg & QSPIC_RDTRSTP_ADDMODE(TRSTP_ADD_4B)) {
		/* enter 4B mode */
		dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xb7) | QSPIC_CMD_TYPE_OTHER);
		do {
			dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x35) |
				QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr, 1);
		} while ((sr & 0x20) != 0x20);
	}

	/* we are now in QPI mode! */
	flash->aux_cfg = info->aux_cfg;

	return 0;
}
static int dspg_qspic_init_mx25_qpi(struct chipsel *flash, const struct qspi_info *info)
{
	u8 sr[3];

	/* status register */
	dspg_qspic_aux_read(flash, CMD_RDSR, &sr[0], 1);
	/* configuration register */
	dspg_qspic_aux_read(flash, QSPIC_CMD_OPCODE(0x15) |
		QSPIC_CMD_TYPE_READ_STATUS | QSPIC_CMD_READ, &sr[1], 2);

	sr[0] &= ~0xbc;   /* disable SRWD and block protection */
	sr[0] |= 1 << 6;  /* QE enable */
	sr[2] = 2;        /* high-performance mode */

	dspg_qspic_write_enable(flash);
	dspg_qspic_aux_write(flash, CMD_WRSR, &sr[0], 3);
	dspg_qspic_wait_write_disable(flash);

	/* we are now in 4IO mode! */
	flash->aux_cfg = info->aux_cfg;

	return 0;
}

#define QSPI_INFO(_name, id, ext, fr, fw, sects)	\
	.name = (_name),				\
	.jedec_id = (id),				\
	.ext_id = (ext),				\
	.read_freq = (fr),				\
	.write_freq = (fw),				\
	.page_size = 256,				\
	.sector_size = 4096,				\
	.n_sectors = (sects)

#define FLASH_QSPI(read, write, addr, dummy)			\
	.read_cfg = QSPIC_RDTRSTP_OPCODE(TRSTP_SERIAL) |	\
		    QSPIC_RDTRSTP_ADD(TRSTP_SERIAL) |		\
		    QSPIC_RDTRSTP_DUMMY(TRSTP_SERIAL) |		\
		    QSPIC_RDTRSTP_DIN(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_ADDMODE(addr)	|		\
		    QSPIC_RDTRSTP_DUMMY_SIZE(dummy) |		\
		    QSPIC_RDTRSTP_HAS_OPCODE,			\
	.read_cmd = (read),					\
	.write_cfg = QSPIC_PPTRSTP_OPCODE(TRSTP_SERIAL) |	\
		     QSPIC_PPTRSTP_ADD(TRSTP_SERIAL) |		\
		     QSPIC_PPTRSTP_DOUT(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_ADDMODE(addr) |		\
		     QSPIC_PPTRSTP_HAS_OPCODE,			\
	.write_cmd = (write),					\
	.aux_cfg = QSPIC_TRSTP_OPCODE(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_ADD(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_DOUT(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_DIN(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_ADDMODE(addr)

#define FLASH_4IO(read, write, addr, dummy)			\
	.read_cfg = QSPIC_RDTRSTP_OPCODE(TRSTP_SERIAL) |	\
		    QSPIC_RDTRSTP_ADD(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_DUMMY(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_DIN(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_ADDMODE(addr)	|		\
		    QSPIC_RDTRSTP_DUMMY_SIZE(dummy) |		\
		    QSPIC_RDTRSTP_HAS_OPCODE,			\
	.read_cmd = (read),					\
	.write_cfg = QSPIC_PPTRSTP_OPCODE(TRSTP_SERIAL) |	\
		     QSPIC_PPTRSTP_ADD(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_DOUT(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_ADDMODE(addr) |		\
		     QSPIC_PPTRSTP_HAS_OPCODE,			\
	.write_cmd = (write),					\
	.aux_cfg = QSPIC_TRSTP_OPCODE(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_ADD(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_DOUT(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_DIN(TRSTP_SERIAL) |		\
		   QSPIC_TRSTP_ADDMODE(addr)

#define FLASH_QPI(read, write, addr, dummy)			\
	.read_cfg = QSPIC_RDTRSTP_OPCODE(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_ADD(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_DUMMY(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_DIN(TRSTP_QUAD) |		\
		    QSPIC_RDTRSTP_ADDMODE(addr)	|		\
		    QSPIC_RDTRSTP_DUMMY_SIZE(dummy) |		\
		    QSPIC_RDTRSTP_HAS_OPCODE,			\
	.read_cmd = (read),					\
	.write_cfg = QSPIC_PPTRSTP_OPCODE(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_ADD(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_DOUT(TRSTP_QUAD) |		\
		     QSPIC_PPTRSTP_ADDMODE(addr) |		\
		     QSPIC_PPTRSTP_HAS_OPCODE,			\
	.write_cmd = (write),					\
	.aux_cfg = QSPIC_TRSTP_OPCODE(TRSTP_QUAD) |		\
		   QSPIC_TRSTP_ADD(TRSTP_QUAD) |		\
		   QSPIC_TRSTP_DOUT(TRSTP_QUAD) |		\
		   QSPIC_TRSTP_DIN(TRSTP_QUAD) |		\
		   QSPIC_TRSTP_ADDMODE(addr)

#define TO4KSEC(mbit)	((mbit) * (1024*1024/8) / 4096)

/* should always contain maximum flash size from the table below */
#define MAX_FLASH_SIZE_MBITS	512

#if TO4KSEC(MAX_FLASH_SIZE_MBITS) > CONFIG_SYS_MAX_FLASH_SECT
#error CONFIG_SYS_MAX_FLASH_SECT is too small!
#endif

static const struct qspi_info known_flashes[] = {
	{
		QSPI_INFO("at25dq321a", 0x1f8800, 0, 62, 33, TO4KSEC(32)),
		FLASH_QSPI(0x6b, 0x32, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_atmel,
		.wait = dspg_qspic_wait_atmel,
	},
	{
		QSPI_INFO("w25q16dv", 0xef4015, 0, 62, 33, TO4KSEC(16)),
		FLASH_QSPI(0x6b, 0x32, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_winbond16_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("w25q64fv", 0xef4017, 0, 62, 33, TO4KSEC(64)),
		FLASH_QSPI(0x6b, 0x32, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_winbond16_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("w25q64fv", 0xef6017, 0, 62, 33, TO4KSEC(64)),
		FLASH_QSPI(0x6b, 0x32, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_winbond16_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("w25q128fv", 0xef4018, 0, 62, 33, TO4KSEC(128)),
		FLASH_QPI(0x0b, 0x02, TRSTP_ADD_3B, 3),
		.init = dspg_qspic_init_winbond_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("w25q256fv", 0xef4019, 0, 62, 33, TO4KSEC(256)),
		FLASH_QPI(0xeb, 0x02, TRSTP_ADD_4B, 3),
		.init = dspg_qspic_init_winbond256_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("w25q256fv", 0xef6019, 0, 62, 33, TO4KSEC(256)),
		FLASH_QPI(0xeb, 0x02, TRSTP_ADD_4B, 3),
		.init = dspg_qspic_init_winbond256_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("sst26vf016", 0xbf2601, 0, 62, 33, TO4KSEC(16)),
		FLASH_QPI(0x0b, 0x02, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_sst16_qpi,
		.wait = dspg_qspic_wait_sst,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("sst26vf032", 0xbf2602, 0, 62, 33, TO4KSEC(32)),
		FLASH_QPI(0x0b, 0x02, TRSTP_ADD_3B, 1),
		.init = dspg_qspic_init_sst32_qpi,
		.wait = dspg_qspic_wait_sst,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("mx66l51235f", 0xc2201a, 0, 62, 33, TO4KSEC(512)),
		FLASH_QPI(0xeb, 0x02, TRSTP_ADD_4B, 3),
		.init = dspg_qspic_init_mx66l_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("n25q128a", 0x20ba18, 0, 62, 33, TO4KSEC(128)),
		FLASH_QPI(0x0b, 0x02, TRSTP_ADD_3B, 3),
		.init = dspg_qspic_init_n25q_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("n25q256a", 0x20ba19, 0, 62, 33, TO4KSEC(256)),
		FLASH_QPI(0x0b, 0x02, TRSTP_ADD_4B, 3),
		.init = dspg_qspic_init_n25q_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("gd25q256c", 0xc84019, 0, 62, 33, TO4KSEC(256)),
		FLASH_QSPI(0x6b, 0x32, TRSTP_ADD_4B, 1),
		.init = dspg_qspic_init_gd25q_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
	{
		QSPI_INFO("mx25r6435f", 0xc22817, 0, 60, 33, TO4KSEC(64)),
		FLASH_4IO(0xeb, 0x38, TRSTP_ADD_3B, 3),
		.init = dspg_qspic_init_mx25_qpi,
		.wait = dspg_qspic_wait_generic,
		.exit = dspg_qspic_reset_generic,
	},
};

static const struct qspi_info *dspg_qspic_identify(struct chipsel *flash)
{
	u8 buffer[5];
	u32 jedec_id;
	u16 ext_id;
	int i;
	const struct qspi_info *info;

	dspg_qspic_aux_read(flash, CMD_RDID, &buffer, sizeof(buffer));

	jedec_id = ((u32)buffer[0] << 16) |
		   ((u32)buffer[1] <<  8) |
		    (u32)buffer[2];
	ext_id = ((u16)buffer[3] << 8) | (u16)buffer[4];

	/* search for matching flash in table */
	for (i = ARRAY_SIZE(known_flashes), info = known_flashes; i > 0; i--, info++) {
		if (info->jedec_id == jedec_id) {
			if (info->ext_id != 0 && info->ext_id != ext_id)
				continue;
			return info;
		}
	}

	return NULL;
}

static struct mtd_info qspi_mtd[2];
static char qspi_dev_name[2][8];

static int qspi_mtd_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	flash_info_t *fi = mtd->priv;
	size_t a_start = fi->start[0] + instr->addr;
	size_t a_end = a_start + instr->len;
	int s_first = -1;
	int s_last = -1;
	int error, sect;

	for (sect = 0; sect < fi->sector_count; sect++) {
		if (a_start == fi->start[sect])
			s_first = sect;

		if (sect < fi->sector_count - 1) {
			if (a_end == fi->start[sect + 1]) {
				s_last = sect;
				break;
			}
		} else {
			s_last = sect;
			break;
		}
	}

	if (s_first >= 0 && s_first <= s_last) {
		instr->state = MTD_ERASING;

		error = flash_erase(fi, s_first, s_last);

		if (error) {
			instr->state = MTD_ERASE_FAILED;
			return -EIO;
		}

		instr->state = MTD_ERASE_DONE;
		mtd_erase_callback(instr);
		return 0;
	}

	return -EINVAL;
}

static int qspi_mtd_read(struct mtd_info *mtd, loff_t from, size_t len,
	size_t *retlen, u_char *buf)
{
	flash_info_t *fi = mtd->priv;
	u_char *f = (u_char*)(fi->start[0]) + from;

	memcpy(buf, f, len);
	*retlen = len;

	return 0;
}

static int qspi_mtd_write(struct mtd_info *mtd, loff_t to, size_t len,
	size_t *retlen, const u_char *buf)
{
	flash_info_t *fi = mtd->priv;
	u_long t = fi->start[0] + to;
	int error;

	error = write_buff(fi, (u_char*)buf, t, len);
	if (!error) {
		*retlen = len;
		return 0;
	}

	return -EIO;
}

static void qspi_mtd_sync(struct mtd_info *mtd)
{
	/*
	 * This function should wait until all pending operations
	 * finish. However this driver is fully synchronous, so
	 * this function returns immediately
	 */
}

static int dspg_qspic_probe_cs(int cs, void *base)
{
	struct chipsel *flash;
	const struct qspi_info *info;
	int ret;
	void *cs_regs;
	unsigned long size;
	int i;
	struct mtd_info *mtd;

	flash = &qspic_cs[cs];

	/* Start up in standard SPI mode */
	flash->cs = cs;
	flash->aux_cfg = QSPIC_TRSTP_OPCODE(TRSTP_SERIAL) |
			   QSPIC_TRSTP_ADD(TRSTP_SERIAL) |
			   QSPIC_TRSTP_DOUT(TRSTP_SERIAL) |
			   QSPIC_TRSTP_DIN(TRSTP_SERIAL) |
			   QSPIC_TRSTP_ADDMODE(TRSTP_ADD_3B);

	info = dspg_qspic_identify(flash);
	if (!info) {
		/* Try QPI with 4B mode (Winbond flashes) */
		dspg_qspic_reset_generic(flash);

		udelay(1000);

		/* Exit 4-byte mode */
		dspg_qspic_aux_cmd(flash, QSPIC_CMD_OPCODE(0xe9) | QSPIC_CMD_TYPE_OTHER);

		flash->aux_cfg = QSPIC_TRSTP_OPCODE(TRSTP_QUAD) |
				 QSPIC_TRSTP_ADD(TRSTP_QUAD) |
				 QSPIC_TRSTP_DOUT(TRSTP_QUAD) |
				 QSPIC_TRSTP_DIN(TRSTP_QUAD) |
				 QSPIC_TRSTP_ADDMODE(TRSTP_ADD_3B);

		info = dspg_qspic_identify(flash);
	}

	if (!info)
		return 0; /* no error */

	size = info->sector_size * info->n_sectors;

	flash->name = info->name;
	flash->wait = info->wait;
	flash->sector_size = info->sector_size;
	flash->n_sectors = info->n_sectors;
	flash->page_size = info->page_size;
	flash->read_freq = info->read_freq;
	flash->write_freq = info->write_freq;

	flash_info[cs].size = size;
	flash_info[cs].sector_count = size / info->sector_size;
	flash_info[cs].flash_id = info->jedec_id;

	flash_info[cs].start[0] = (ulong)base;
	for (i = 1; i < flash_info[cs].sector_count; i++)
		flash_info[cs].start[i] =
		    flash_info[cs].start[i - 1] + info->sector_size;

	cs_regs = regs + (cs ? 0x200 : 0x100);
	writel(info->read_cfg, cs_regs + QSPIC_RDTRSTP);
	writel(info->read_cmd, cs_regs + QSPIC_RDCMD);
	writel(info->write_cfg, cs_regs + QSPIC_PPTRSTP);
	writel(info->write_cmd, cs_regs + QSPIC_PPCMD);
	if (!cs)
		writel(0xFFFF0000, cs_regs + QSPIC_MEMADD1); /* we verify ourselves */
	else
		writel(0xFFFF0000, cs_regs + QSPIC_MEMADD2); /* we verify ourselves */

	flash->shadow_buf = base; //CONFIG_SYS_FLASH_ADDR_BASE;

	if (info->init) {
		ret = info->init(flash, info);
		if (ret)
			return ret;
	}

	/*
	 * Now send auxiliary commands with flash specific settings. Some init
	 * callbacks may have already done this.
	 */
	flash->aux_cfg = info->aux_cfg;
	flash->exit = info->exit;

	if (flash->write_freq > flash->read_freq)
		ret = clk_set_rate("qspi", flash->read_freq*1000000);
	else
		ret = clk_set_rate("qspi", flash->write_freq*1000000);
	if (ret)
		return ret;


#ifdef CONFIG_MTD_DEVICE
	mtd = &qspi_mtd[cs];
	sprintf(qspi_dev_name[cs], "nor%d", cs);
	mtd->name = qspi_dev_name[cs];
	mtd->type = MTD_NORFLASH;
	mtd->flags = MTD_WRITEABLE; //MTD_CAP_NORFLASH;
	mtd->size = size;
	mtd->writesize = info->page_size;
	mtd->numeraseregions = 0;

	/* Do not report "info->sector_size" (4KB) here as filesystems like
	 * jffs2 cannot cope with such small erase block sizes. On the other
	 * hand, 8/32/64KB erase commands differ from flash to flash so we
	 * have to stay with the 4KB erase commands.
	 */
	mtd->erasesize = 64 * 1024;

	mtd->_erase	= qspi_mtd_erase;
	mtd->_read	= qspi_mtd_read;
	mtd->_write	= qspi_mtd_write;
	mtd->_sync	= qspi_mtd_sync;
	mtd->priv	= &flash_info[cs];

	if (add_mtd_device(mtd))
		return -ENOMEM;
#endif

	return 0;
}

/*
 * flash_erase - Erase SMI flash
 *
 * Erase SMI flash
 */
int flash_erase(flash_info_t *info, int s_first, int s_last)
{
	int ret = 0;
	flash_sect_t sect;

	if ((s_first < 0) || (s_first > s_last)) {
		puts("- no sectors to erase\n");
		return 1;
	}

	for (sect = s_first; sect <= s_last; sect++) {
		if (dspg_qspic_erase_sector(info, sect))
			ret = 1;
	}
	puts(" done\n");

	return ret;
}

void qspi_init(void)
{
	int ret;
	void *base;

	ret = clk_set_rate("qspi", 33000000);
	if (ret)
		return;

	clk_enable("qspi");

	base = (void*)DVF99_QSPI_BASE;

	writel(0x300, base + QSPIC_CSTIMING); /* enlarged cs timing */
	writel(0x10009, base + QSPIC_CFG); /* long CS deselect time, spi mode 3, qspic enable */

	/* clear all interrupts */
	writel(~0UL, base + QSPIC_ICR);
	while (readl(base + QSPIC_ISR))
		;
	writel(0, base + QSPIC_ICR); /* not self clearing */

	/* start with sane settings (cmd 0x03 for read) */
	writel(0x00080400, base + 0x100 + QSPIC_RDTRSTP);
	writel(0x00000003, base + 0x100 + QSPIC_RDCMD);
	writel(0x00080400, base + 0x200 + QSPIC_RDTRSTP);
	writel(0x00000003, base + 0x200 + QSPIC_RDCMD);
	writel(0xFFFF0000, base + 0x100 + QSPIC_MEMADD1); /* we verify ourselves */
	writel(0xFFFF0000, base + 0x200 + QSPIC_MEMADD1); /* we verify ourselves */
}

unsigned long flash_init(void)
{
	unsigned long size = 0;
	int ret;

	regs = (void*)DVF99_QSPI_BASE;

	flash_info[0].flash_id = 0xffff;
	flash_info[1].flash_id = 0xffff;
	flash_info[0].size = 0;
	flash_info[1].size = 0;

	/* detect flashes */
	ret = dspg_qspic_probe_cs(0, (void *)0x80000000);
	if (ret)
		return 0;
	size += flash_info[0].size;

#if CONFIG_SYS_MAX_FLASH_BANKS > 1
	ret = dspg_qspic_probe_cs(1, (void *)0xa0000000);
	if (ret)
		return size;
	size += flash_info[1].size;
#endif

	return size;
}

void flash_exit(void)
{
	if (qspic_cs[0].exit)
		qspic_cs[0].exit(&qspic_cs[0]);

#if CONFIG_SYS_MAX_FLASH_BANKS > 1
	if (qspic_cs[1].exit)
		qspic_cs[1].exit(&qspic_cs[1]);
#endif
}

int do_qspi(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	if (argc < 2)
		return CMD_RET_USAGE;

	if (!strcmp(argv[1], "info")) {
		if (qspic_cs[0].name)
			printf("CS 0: %s, base 0x80000000\n", qspic_cs[0].name);
#if CONFIG_SYS_MAX_FLASH_BANKS > 1
		if (qspic_cs[1].name)
			printf("CS 1: %s, base 0xa0000000\n", qspic_cs[1].name);
#endif
	}

	return 0;
}

U_BOOT_CMD(qspi, 2, 0, do_qspi,
	"QSPI support\n",
	"info - print information on QSPI devices\n"
);
#endif

