#include <common.h>
#include <asm/io.h>
#include <miiphy.h>
#include <malloc.h>
#include "synop_gmac_dev.h"

#define MACG_TX_TIMEOUT 30000

#define to_gmac(_nd) container_of(_nd, struct synopGMACDeviceStruct, netdev)

#if defined(CONFIG_CMD_NET)
static int gmac_send(struct eth_device *netdev, void *packet, int length)
{
	s32 status = 0;
	u32 offload_needed = 0;
	synop_gmac_device *gmacdev = to_gmac(netdev);
	int i = 0;
	unsigned char *packet_on_sdram = 0;
	unsigned int timeout = MACG_TX_TIMEOUT;

	TR2("\n");
	if (packet == NULL) {
		TR0("Packet is NULL. What happened to u-boot?\n");
		return -1;
	}

	packet_on_sdram = (unsigned char *)(virt_to_bus((u32)packet)); // - 0x80000000);

	TR2("packet = 0x%x packet on sdram = 0x%x length = %d \n",
	    (unsigned int)packet, (unsigned int)packet_on_sdram , length);
	for (i = 0; i < length; i++)
	 	TR2("packet[%d] = 0x%x\n",i , packet_on_sdram[i]);

	status = synop_gmac_set_tx_qptr(gmacdev, (u32)packet_on_sdram, length,
	                                (u32)packet, 0, 0, 0, offload_needed);
	if (status < 0) {
		TR0("No More Free Tx Descriptors\n");
		return -1;
	}

	/* Now force the DMA to start transmission */
	synop_gmac_resume_dma_tx(gmacdev);

	while (synop_gmac_intr_tx_handler(gmacdev) != 0 && --timeout > 0);

	if (!timeout)
		TR0("Timeout\n");
	
	return 0;
}

static int gmac_recv(struct eth_device *netdev)
{
	synop_gmac_device *gmacdev = to_gmac(netdev);

	synop_gmac_intr_rx_handler(gmacdev);

	return 0;
}

static int gmac_init(struct eth_device *netdev, bd_t *bd)
{
	synop_gmac_device *gmacdev = to_gmac(netdev);
	TR2("\n");

	synop_gmac_setup_gmac(gmacdev);
	synop_gmac_enable_rx_chksum_offload(gmacdev);

	return 0;
}

static void gmac_halt(struct eth_device *netdev)
{
	synop_gmac_device *gmacdev = to_gmac(netdev);
	TR2("\n");

	synop_gmac_tx_disable(gmacdev);
	synop_gmac_rx_disable(gmacdev);
	/* Disable TX and RX */
	synop_gmac_disable_dma_tx(gmacdev);
	synop_gmac_disable_dma_rx(gmacdev);
}

static int gmac_write_hwaddr(struct eth_device *netdev)
{
	synop_gmac_device *gmacdev = to_gmac(netdev);

	return synop_gmac_set_mac_addr(gmacdev, netdev->enetaddr);
}

#if defined(CONFIG_MII) || defined(CONFIG_CMD_MII) || defined(CONFIG_PHYLIB)
int gmac_miiphy_read(struct mii_dev *bus, int addr, int devad, int reg)
{
	u16 data;
	int ret;

	ret = (int)synop_gmac_read_phy_reg(bus->priv, (u32)addr, (u32)reg, &data);
	if (ret)
		return -1;

	return (int)data;
}

int gmac_miiphy_write(struct mii_dev *bus, int addr, int devad, int reg, u16 val)
{
	return synop_gmac_write_phy_reg(bus->priv, addr, reg, val);
}
#endif

int gmac_eth_register(int id, void *regs, int ignore_phy, u32 speed, u32 clk_csr_i)
{
	synop_gmac_device *gmacdev;
	struct eth_device *netdev;
#if defined(CONFIG_MII) || defined(CONFIG_CMD_MII) || defined(CONFIG_PHYLIB)
	struct mii_dev *miidev;
	int ret;
#endif

	gmacdev = malloc(sizeof(synop_gmac_device));
	if (!gmacdev) {
		printf("Error: Failed to allocate memory for GMAC%d\n", id);
		return -1;
	}

	memset(gmacdev, 0, sizeof(synop_gmac_device));

	netdev = &(gmacdev->netdev);

	sprintf(netdev->name,"gmac%d@%uMbps", id, speed);
	netdev->init = gmac_init;
	netdev->halt = gmac_halt;
	netdev->send = gmac_send;
	netdev->recv = gmac_recv;
	netdev->write_hwaddr = gmac_write_hwaddr;

	gmacdev->Speed = speed;
	synop_gmac_reset(gmacdev);

	/* Attach the device to the MAC struct. This will configure all the
	 * required base addresses such as the Mac base, configuration base,
	 * phy base address (out of 32 possible phys). */
	synop_gmac_attach(	gmacdev,
						(u32)(regs + MACBASE),
						(u32)(regs + DMABASE),
						ignore_phy ? DEFAULT_PHY_IGNORE : DEFAULT_PHY_UNDEFINE,
						clk_csr_i);

	/*
	 * Do some basic initialization so that we at least can talk
	 * to the PHY
	 */
	eth_register(netdev);

#if defined(CONFIG_MII) || defined(CONFIG_CMD_MII) || defined(CONFIG_PHYLIB)
	miidev = mdio_alloc();
	if (!miidev) {
		printf("Error: mdio_alloc() failed\n");
		return -1;
	}

	strcpy(miidev->name, netdev->name);
	miidev->priv = gmacdev;
	miidev->read = gmac_miiphy_read;
	miidev->write = gmac_miiphy_write;
	ret = mdio_register(miidev);
	if (ret) {
		printf("Error: mdio_register() failed\n");
		return -1;
	}
#endif

	/* return number of interfaces detected */
	return 1;
}
#endif

