/*
 * (C) Copyright 2014 DSP Group
 *
 * Modified by Grandstream for dvf99
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <qca/athrs17_phy.h>
#include <miiphy.h>

#define MODULE_NAME "ATHRS17"

/* for PHY related debugging messages */
#undef S17_PHY_DEBUG
#ifdef S17_PHY_DEBUG
#define DPRINTF_PHY(_fmt,...) do { \
	printf(MODULE_NAME":"_fmt, __VA_ARGS__); \
} while (0)
#define DPRINTF_PHY2(_fmt,...)
#else
#define DPRINTF_PHY(_fmt,...)
#define DPRINTF_PHY2(_fmt,...)
#endif

#define S17_PHY_MAX 5
#define S17_PHY0_ADDR   0x0
#define S17_PHY1_ADDR   0x1
#define S17_PHY2_ADDR   0x2
#define S17_PHY3_ADDR   0x3
#define S17_PHY4_ADDR   0x4
#define S17_IND_PHY 4

typedef struct {
	uint32_t phy_base;
	uint32_t phy_addr;
} athr_phyinfo_t;

static athr_phyinfo_t athphyinfo[] =
{
	{0, S17_PHY0_ADDR},
	{0, S17_PHY1_ADDR},
	{0, S17_PHY2_ADDR},
	{0, S17_PHY3_ADDR},
	{0, S17_PHY4_ADDR},
	{0, 0x00},
};

#define S17_PHYBASE(port) (athphyinfo[port].phy_base)
#define S17_PHYADDR(port) (athphyinfo[port].phy_addr)

uint16_t phy_reg_read(int dev_id, uint32_t phy_addr,
	uint8_t reg);
void phy_reg_write(int dev_id, uint32_t phy_addr,
	uint8_t reg, uint16_t data);
static void athrs17_reg_rmw(unsigned int reg_addr,
	unsigned int reg_val);
static int athrs17_is_az_fixed(void);

void athrs17_port_isolation(int isolat)
{
	/*For *8334 */
	u32 p0reg, p2reg, p3reg, p6reg;
	p0reg = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	p2reg = athrs17_reg_read(S17_P2LOOKUP_CTRL_REG);
	p3reg = athrs17_reg_read(S17_P3LOOKUP_CTRL_REG);
	p6reg = athrs17_reg_read(S17_P6LOOKUP_CTRL_REG);

	if (isolat) {
		p0reg = (p0reg & 0xffffff80) | PORT2_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p2reg = (p2reg & 0xffffff80) | PORT0_BITMAP;
		p3reg = (p3reg & 0xffffff80) | PORT0_BITMAP;
		p6reg = (p6reg & 0xffffff80) | PORT0_BITMAP;
	} else {
		p0reg = (p0reg & 0xffffff80) | PORT2_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p2reg = (p2reg & 0xffffff80) | PORT0_BITMAP |
			PORT3_BITMAP | PORT6_BITMAP;
		p3reg = (p3reg & 0xffffff80) | PORT0_BITMAP |
			PORT2_BITMAP | PORT6_BITMAP;
		p6reg = (p6reg & 0xffffff80) | PORT0_BITMAP |
			PORT2_BITMAP | PORT3_BITMAP;
	}

	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG, p0reg);
	athrs17_reg_write(S17_P2LOOKUP_CTRL_REG, p2reg);
	athrs17_reg_write(S17_P3LOOKUP_CTRL_REG, p3reg);
	athrs17_reg_write(S17_P6LOOKUP_CTRL_REG, p6reg);
}

void athrs17_reg_init(athr_switch_intf_type_t interface, int master)
{

	int phy_addr;

	if (interface == RGMII) {
		/* configure for RGMII */
		athrs17_reg_write(S17_GLOFW_CTRL1_REG,
			S17_BROAD_DPALL | S17_MULTI_FLOOD_DPALL |
			S17_UNI_FLOOD_DPALL);
		athrs17_reg_write(S17_PWS_REG, S17_PWS_CHIP_AR8327);

		/* QCA8334 MAC0 RXCLK_DELAY configure by S17 P5PAD_MODE_REG */
		athrs17_reg_write(S17_P0PAD_MODE_REG, S17_MAC0_RGMII_EN |
			(1 << S17_MAC0_RGMII_RXCLK_SHIFT));
		athrs17_reg_write(S17_P5PAD_MODE_REG, S17_MAC5_RGMII_RXCLK_DELAY);
		/* Disable forwarding from LAN to WAN during booting */
		athrs17_port_isolation(1);

		athrs17_reg_write(S17_P0STATUS_REG, S17_PORT_STATUS_DEFAULT_1000M);

#if HYBRID_SWITCH_PORT6_USED
		/* configure AR8327 mac6 as RGMII mode */
		athrs17_reg_write(S17_PWS_REG, S17_PWS_CHIP_AR8327);
		athrs17_reg_write(S17_P6PAD_MODE_REG, S17_MAC6_RGMII_EN |
			S17_MAC6_RGMII_TXCLK_DELAY |
			S17_MAC6_RGMII_RXCLK_DELAY |
			(2 << S17_MAC6_RGMII_RXCLK_SHIFT) |
			(1 << S17_MAC6_RGMII_TXCLK_SHIFT));
		athrs17_reg_write(S17_P6STATUS_REG, S17_PORT6_STATUS_DEFAULT_1000M);
#else
		athrs17_reg_write(S17_P6PAD_MODE_REG, S17_MAC6_RGMII_RXCLK_DELAY |
			(0 << S17_MAC6_RGMII_RXCLK_SHIFT));
#endif

		/* AR8327/AR8328 v1.0 fixup */
		if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff)== S17_CHIPID_V1_0) {
			for (phy_addr = 0x0; phy_addr <= S17_PHY_MAX; phy_addr++) {
				/* For 100M waveform */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x0);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x02ea);
				/* Turn On Gigabit Clock */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x3d);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x68a0);
			}
		}

		/* AR8337 v1.0 fixup */
		if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff) == S17C_CHIPID_V1_0) {
			for (phy_addr = 0x0; phy_addr <= S17_PHY_MAX; phy_addr++) {
				/* Turn On Gigabit Clock */
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_ADDRESS, 0x3d);
				phy_reg_write(0, phy_addr, S17_DEBUG_PORT_DATA, 0x6820);
			}
		}
	} else if (interface == RMII) {
		/* configure for RMII */
		unsigned int	value;
		value = athrs17_reg_read(0x10);
		value |= 0x10000000;
		athrs17_reg_write(0x10, value);

		value = athrs17_reg_read(0x30);
		value |= 0x00000001;

		athrs17_reg_write(0x30, value);

		if (master) /* Qualccomm switch master mode */
			athrs17_reg_write(0x4, 0x40000000);
		else /* Qualccomm switch slave mode*/
			athrs17_reg_write(0x4, 0x20000000);

		athrs17_reg_write(0x0624, 0x007f7f7f);
		/* RMII 100MB/s */
		athrs17_reg_write(0x007c, 0x007d);

		/* RMII 10MB/s
		athrs17_reg_write(0x007c, 0x007c);
		*/
	}

	athrs17_reg_write(S17_GLOFW_CTRL0_REG, athrs17_reg_read(S17_GLOFW_CTRL0_REG) | S17_CPU_PORT_EN);

	/* enable HOL by default */
	for (phy_addr = 0; phy_addr < S17_MAC_MAX; phy_addr++) {
		switch (phy_addr) {
		case 0:
		case 5:
		case 6:
			athrs17_reg_write(S17_PORT0_HOL_CTRL0 + phy_addr * 8,
				S17_HOL_CTRL0_WAN);
			break;

		default:
			athrs17_reg_write(S17_PORT0_HOL_CTRL0 + phy_addr * 8,
				S17_HOL_CTRL0_LAN);
			break;
		}

		athrs17_reg_write(S17_PORT0_HOL_CTRL1 + phy_addr * 8,
			S17_HOL_CTRL1_DEFAULT);
	}

#ifdef S17_8023AZ_FEATURE
	if (!athrs17_is_az_fixed())
		athrs17_reg_write(S17_EEE_CTRL_REG, S17_LPI_DISABLE_ALL);
#endif

	/* Turn all LEDs off .*/
	/* LED control */
	athrs17_reg_write(S17_LED_CTRL0_REG, 0);
	athrs17_reg_write(S17_LED_CTRL1_REG, 0);
	athrs17_reg_write(S17_LED_CTRL2_REG, 0);
}

uint16_t phy_reg_read(int dev_id,
	uint32_t phy_addr, uint8_t reg)
{
	uint16_t data = 0;
	int ret;
	const char* devname = miiphy_get_current_dev();

	ret = miiphy_read(devname, (unsigned char)phy_addr, reg, &data);
	if (ret) {
		DPRINTF_PHY("miiphy_read returns error %d\n", ret);
		return 0;
	} else
		return data;
}

void phy_reg_write(int dev_id, uint32_t phy_addr,
	uint8_t reg, uint16_t data)
{
	int ret;
	const char* devname = miiphy_get_current_dev();

	ret = miiphy_write(devname, (unsigned char)phy_addr, reg, data);
	if (ret)
		DPRINTF_PHY("miiphy_write returns error %d\n", ret);
}

void athrs17_pc_port_enable(void)
{
	athrs17_reg_write(S17_P2STATUS_REG,
		S17_PORT_STATUS_DEFAULT);
	printf("PC port enabled.\n");
}

void athrs17_pc_port_disable(void)
{
	athrs17_reg_write(S17_P2STATUS_REG, 0);
	printf("PC port disabled.\n");
}

void athrs17_pc_port_mirror_enable(void)
{
	unsigned int value;

	/* Set mirror port number 2. */
	value = athrs17_reg_read(S17_GLOFW_CTRL0_REG);
	value &= ~(0x0F << S17_MIRROR_PORT_SHIFT);
	athrs17_reg_write( S17_GLOFW_CTRL0_REG,
		value | (2 << S17_MIRROR_PORT_SHIFT));

	/* Enable mirror for incoming packets to port 0. */
	value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG, value | (1 << 25));

	/* Enable mirror for outgoing packets from port 0. */
	value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
	athrs17_reg_write(S17_PORT0_HOL_CTRL1, value | (1 << 16));

	printf("PC port mirror enabled.\n");
}

void athrs17_pc_port_mirror_disable(void)
{
	unsigned int value;

	/* Disable mirror for incoming packets to port 0.*/
	value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
	athrs17_reg_write(S17_P0LOOKUP_CTRL_REG,
		value & ~(1 << 25));

	/* Disable mirror for outgoing packets from port 0.*/
	value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
	athrs17_reg_write(S17_PORT0_HOL_CTRL1,
		value & ~(1 << 16));

	printf("PC port mirror disabled.\n");
}

static int athrs17_port_status(int port)
{
	static int last_status = 0;
	uint16_t phy_status, phy_control;
	uint32_t phy_base;
	uint32_t phy_addr;

	phy_base = S17_PHYBASE(port);
	phy_addr = S17_PHYADDR(port);

	if (last_status) {
		/* link was previously up */
		phy_status = phy_reg_read(phy_base, phy_addr,
			S17_PHY_SPEC_STATUS);

		/* See if we've lost link */
		if (!(phy_status & S17_STATUS_LINK_PASS)) {
			/* Flush the port from ARL entry */
			athrs17_reg_rmw(S17_ATU_FUNC_REG, ( port << 8 ));
			last_status = 0;
		}
	} else {
		/* last known link status was DEAD
		* Check for reset complete
		*/
		phy_status = phy_reg_read(phy_base, phy_addr,
			S17_PHY_STATUS);
		if (!S17_RESET_DONE(phy_status)) {
			return 0;
		}
		phy_control = phy_reg_read(phy_base, phy_addr,
			S17_PHY_CONTROL);
		/* Check for AutoNegotiation complete */
		if ((!( phy_control & S17_CTRL_AUTONEGOTIATION_ENABLE))
			|| S17_AUTONEG_DONE(phy_status)) {
			phy_status = phy_reg_read(phy_base, phy_addr,
				S17_PHY_SPEC_STATUS);
			if (phy_status & S17_STATUS_LINK_PASS)
				last_status = 1;
		}
	}

	return last_status;
}

int athrs17_lan_status(void)
{
	return athrs17_port_status(2);
}

int athrs17_pc_status(void)
{
	return athrs17_port_status(1);
}

unsigned int athrs17_reg_read(unsigned int reg_addr)
{
	uint32_t reg_word_addr;
	uint32_t phy_addr, tmp_val, reg_val;
	uint16_t phy_val;
	uint8_t phy_reg;

	/* change reg_addr to 16-bit word address, 32-bit aligned */
	reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

	/* configure register high address */
	phy_addr = 0x18;
	phy_reg = 0x0;
	/* bit16-8 of reg address */
	phy_val = (uint16_t)((reg_word_addr >> 8) & 0x1ff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* For some registers such as MIBs, since it is read/clear, we should */
	/* read the lower 16-bit register then the higher one */

	/* read register in lower address */
	/* bit7-5 of reg address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7);
	/* bit4-0 of reg address */
	phy_reg = (uint8_t)(reg_word_addr & 0x1f);
	reg_val = (uint32_t)phy_reg_read(0, phy_addr, phy_reg);

	/* read register in higher address */
	reg_word_addr++;
	/* bit7-5 of reg address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7);
	/* bit4-0 of reg address */
	phy_reg = (uint8_t)(reg_word_addr & 0x1f);
	tmp_val = (uint32_t)phy_reg_read(0, phy_addr, phy_reg);
	reg_val |= (tmp_val << 16);

	return reg_val;
}

void athrs17_reg_write(unsigned int reg_addr, unsigned int reg_val)
{
	uint32_t reg_word_addr;
	uint32_t phy_addr;
	uint16_t phy_val;
	uint8_t phy_reg;

	/* change reg_addr to 16-bit word address, 32-bit aligned */
	reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

	/* configure register high address */
	phy_addr = 0x18;
	phy_reg = 0x0;
	/* bit16-8 of reg address */
	phy_val = (uint16_t)((reg_word_addr >> 8) & 0x1ff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* For S17 registers such as ARL and VLAN, since they include BUSY bit
	* in higher address, we should write the lower 16-bit register then the
	* higher one
	*/
	/* write register in lower address */
	/* bit7-5 of reg address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7);
	/* bit4-0 of reg address */
	phy_reg = (uint8_t)(reg_word_addr & 0x1f);
	phy_val = (uint16_t)(reg_val & 0xffff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);

	/* read register in higher address */
	reg_word_addr++;
	/* bit7-5 of reg address */
	phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7);
	/* bit4-0 of reg address */
	phy_reg = (uint8_t)(reg_word_addr & 0x1f);
	phy_val = (uint16_t)((reg_val >> 16) & 0xffff);
	phy_reg_write(0, phy_addr, phy_reg, phy_val);
}

static void athrs17_reg_rmw(unsigned int reg_addr,
	unsigned int reg_val)
{
	reg_val |= athrs17_reg_read(reg_addr);
	athrs17_reg_write(reg_addr,reg_val);
}

#ifdef S17_8023AZ_FEATURE
static int athrs17_is_az_fixed(void)
{
	if ((athrs17_reg_read(S17_MASK_CTRL_REG) & 0xffff) > S17_CHIPID_V1_1)
		return 1;

	return 0;
}
#endif

/* EOF */
