/*
 * NVRAM variable manipulation (Linux kernel half)
 *
 * Copyright 2004, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram_linux.c,v 1.1 2013/02/05 23:28:00 sfrick Exp $
 *
 * SXIE - added ioctl to erase all tuples
 * SCF - rewrote _nvram_read and nvram_commit from scratch to support zlib compression
 *     - modified _nvram_free to correctly track nvram_size
 *     - modified _nvram_realloc to allow hash table to grow dynamically to any size up
       -          to max avail kmalloc'able memory
 */
#include <common.h>
#include <malloc.h>
#include <errno.h>
#include <flash.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/concat.h>
#include <jffs2/jffs2.h>
#include <u-boot/zlib.h>
#include "nvram.h"

/* Defines */
#define DEVNO_NVRAM_READ_SAFE 1
#define DEVNO_NVRAM_READ_UNSAFE 0
#define init_waitqueue_head(...)    do { } while (0)
#define wait_event_interruptible(...)   0
#define wake_up_interruptible(...)  do { } while (0)
#define yield()             do { } while (0)
#define ERASE_RETRIES 3

typedef int wait_queue_head_t;

/* In BSS to minimize text size and page aligned so it can be mmap()-ed */
static char nvram_buf[NVRAM_BUFFER_SPACE];

extern char * _nvram_get(const char *name);
extern int _nvram_set(const char *name, const char *value, int set_dirty_flag);
extern int _nvram_unset(const char *name);
extern int _nvram_getall(char *buf, int count);
extern int _nvram_commit(struct nvram_header *header);
extern int _nvram_init(void);
extern int _nvram_erase_hushtable(void);
extern int _nvram_purge_unlocked(void);
extern void _nvram_exit(void);
extern int _nvram_setacl(const char *name);

static int erase_helper( struct mtd_info* mtd, loff_t offset, size_t len );
static void erase_callback( struct erase_info* info );

/* Globals */
static unsigned long nvram_offset = 0;

unsigned int  nvram_dirty = 0;

static int nvram_space_used = 0;

static struct mtd_info* mtd;

static unsigned int current_block_index = 0;
static unsigned int current_part_index = 0;
static unsigned long current_write_counter = 0;
static unsigned int total_valid_parts = 0;
struct nvram_flash_block nvram_flash_blocks[NVRAM_NUM_BLOCKS];

#define NVRAM_ZALLOC_ALIGNMENT 16

void *nvram_zalloc(void *x, unsigned items, unsigned size)
{
	void *p;

	size *= items;
	size = (size + NVRAM_ZALLOC_ALIGNMENT - 1) &
		~(NVRAM_ZALLOC_ALIGNMENT - 1);

	p = malloc (size);

	return (p);
}

void nvram_zfree(void *x, void *addr, unsigned nb)
{
	free (addr);
}

static unsigned short calc_zbuf_chksum(int size, char *p)
{
	unsigned short sum = 0;
	unsigned short checksum = 0;
	int i;

	size = ((size + 1)>>1) << 1;
	for (i = 0; i < size; i+=2)
	{
		sum += (p[i] << 8 | p[i+1]);    // fact config use be for checksum calc
	}

	checksum = 0xffff - sum + 0x1;

	//printf("Chksum is 0x%04x\n", (unsigned short)(checksum >> 8 | checksum << 8));
	return (checksum >> 8 | checksum << 8); // be
}
static unsigned short apply_zbuf_chksum(struct nvram_header *header, char *zbuf)
{
	header->chksum = calc_zbuf_chksum(header->len_compressed, zbuf);
	return 0;
}
static unsigned short verify_zbuf_chksum(char *zbuf_with_header)
{
	struct nvram_header *header = (struct nvram_header*)zbuf_with_header;
	return (header->chksum == calc_zbuf_chksum(header->len_compressed,
		zbuf_with_header + NVRAM_HEADER_SIZE)? 0 : 1);
}

static void erase_callback(struct erase_info* info)
{
	wake_up_interruptible((wait_queue_head_t *)ei->priv);
}

static int erase_helper(struct mtd_info* mtd, loff_t offset, size_t len)
{
	struct erase_info info;
	int ret;
	wait_queue_head_t waitq;
	int retries = 0;

retry:
	init_waitqueue_head(&waitq);
	memset( &info, 0, sizeof ( info ) );
	info.mtd = mtd;
	info.addr = offset;
	info.len = len;
	info.callback = erase_callback;
	info.priv = (unsigned long)&waitq;

	ret = mtd_erase( mtd, &info );
	if ( ret )
	{
		if ( retries++ < ERASE_RETRIES )
		{
			debug( "Erase failed, %d retry\n", retries );
			yield();
			goto retry;
		}
		return -EIO;
	}

	if ( info.state != MTD_ERASE_DONE )
	return -EIO;

	return 0;
}

/*
 * This read will be called only once when nvram init.
 * Normal read error does not really indicate block is bad, so we will not mark block bad here
 */
static int nvram_flash_read_zbuf(char *zbuf, u_int32_t len)
{
	struct nvram_header *flash_header = (struct nvram_header *) zbuf;
	int block_read_err;
	int ii, jj;
	size_t read_len;

	//printf("nvram_flash_read_zbuf()\n");

	if (len != NVRAM_SPACE) //just sanity check
	printf("!!!! nvram_flash_read_zbuf: wrong size 0x%x\n", len);

	current_write_counter = 0;
	//    printf("scan nvram data blocks\n");
	/* scan all blocks to locate currently using block and part */
	for (ii = 0; ii < NVRAM_NUM_BLOCKS; ii ++)
	{
		/* we scan part even bad blocks, to maximum extension to get setting back */
		if ( !mtd->_block_isbad( mtd, nvram_flash_blocks[ii].start_addr ) )
		{
			/* increase good parts for writing */
			total_valid_parts += nvram_flash_blocks[ii].part_count;
		}

		block_read_err = 0;

		for (jj = 0; jj < nvram_flash_blocks[ii].part_count; jj ++)
		{
			if (mtd_read( mtd, nvram_flash_blocks[ii].start_addr +
				jj * NVRAM_SPACE, NVRAM_SPACE, &read_len, (u_char *)zbuf )
				|| read_len != NVRAM_SPACE)
			{
				block_read_err++;
				printf("part[%d:%d] ecc failed\n", ii, jj);
			}
			else
			{
				if (flash_header->magic != NVRAM_MAGIC || verify_zbuf_chksum(zbuf))
				{
					/* nvram data invalid */
					continue;
				}
				else
				{
					/* use = here is for reading old version  nvram data */
					if (current_write_counter <= flash_header->write_counter)
					{
						current_write_counter = flash_header->write_counter;
						current_block_index = ii;
						current_part_index = jj;
					}
				}
			}
		}
	}
	printf("total valid parts=%d\n", total_valid_parts);

	if ( mtd_read( mtd, nvram_flash_blocks[current_block_index].start_addr +
		current_part_index * NVRAM_SPACE, NVRAM_SPACE, &read_len, (u_char *)zbuf ) ||
		read_len != NVRAM_SPACE )
	{
		/* Will it fail in second read? */
		return -1;
	}
	printf("active nvram part[%d:%d], write=%ld\n",
		current_block_index, current_part_index, flash_header->write_counter);
	return 0;
}

static int nvram_flash_get_part_for_write(int *block_index, int *part_index)
{
	int ii;

	*part_index += 1;
	if (*part_index == nvram_flash_blocks[*block_index].part_count)
	{
		*block_index += 1;
		*part_index = 0;
	}

	for (ii = 0; ii < NVRAM_NUM_BLOCKS; ii ++)
	{
		if (*block_index == NVRAM_NUM_BLOCKS)
			*block_index = 0;
		if ( mtd->_block_isbad( mtd, nvram_flash_blocks[*block_index].start_addr ) )
		{
			/* try next block, first part */
			*block_index += 1;
			*part_index = 0;
			continue;
		}
		break;
	}
	if (ii == NVRAM_NUM_BLOCKS)
	{
		return -1;
	}
	return 0;
}

/*
 * We only erase block once but will write multiple times. After each write we will read it out to check
 * whether the part is able be read out correctly. If we found one write bad we will write it again to next
 * part.
 * Note we do not mark block bad any more unless all parts are not able to write
 */
static int nvram_flash_write_zbuf(char *zbuf, u_int32_t len)
{
	int ii;
	int block_index_write = current_block_index;
	int part_index_write = current_part_index;
	char *tmpbuf = NULL;
	int ret = 0;
	int write_fail = 0;
	size_t read_len;

	printf("nvram_flash_write_zbuf()\n");

	if (len != NVRAM_SPACE) //just sanity check
		printf("!!!! nvram_flash_write_zbuf: wrong size 0x%x\n", len);

	tmpbuf = malloc(NVRAM_SPACE);

	for (ii = 0; ii < total_valid_parts; ii ++)
	{
		/* we write to next part, even the current part is empty case (no nvram data at all). */
		if (nvram_flash_get_part_for_write(&block_index_write, &part_index_write))
		{
			/* no blocks found for writing */
			goto all_bad;
		}
		printf("nvram_flash_write_zbuf: block_index=%d; part_index=%d\n",
			block_index_write, part_index_write);
		if ( part_index_write == 0)
		{
			erase_helper( mtd, nvram_flash_blocks[block_index_write].start_addr,
				NVRAM_SPACE * nvram_flash_blocks[block_index_write].part_count );
			write_fail = 0; /* reset write failure count */
		}

		/* Write Flash */
		mtd_write( mtd, nvram_flash_blocks[block_index_write].start_addr +
			NVRAM_SPACE * part_index_write, NVRAM_SPACE, &read_len, (u_char *)zbuf );
		/* Check with reading back */
		{
			if (tmpbuf)
			{
				if (!mtd_read( mtd, nvram_flash_blocks[block_index_write].start_addr +
					part_index_write * NVRAM_SPACE, NVRAM_SPACE, &read_len, (u_char *)tmpbuf ))
				{
					/* read back success */
					break;
				}
			}
			else    //be lazy, if malloc failed we just skip doing read back check
			break;
			write_fail ++;

			if (write_fail == nvram_flash_blocks[block_index_write].part_count)
			{
				/* we failed to write all parts in this block, mark the block as bad block */
				mtd_block_markbad( mtd, nvram_flash_blocks[block_index_write].start_addr );
				total_valid_parts -= nvram_flash_blocks[block_index_write].part_count;
				printf("block[%d] has no good parts any more, marked bad for retirement\n",
					block_index_write);
			}
			printf("nvram part[%d:%d] failure, rewrite performed\n",
				block_index_write, part_index_write);
		}
	}
	if (ii == NVRAM_NUM_BLOCKS)
	{
		/* no good blocks any more */
		goto all_bad;
	}
	else
	{
		/* update to new current block/part index */
		current_block_index = block_index_write;
		current_part_index = part_index_write;
		ret = 0;
	}

	goto exit;

all_bad:
	printf("no good parts found for nvram settings\n");
	ret = -1;

exit:
	if (tmpbuf) free(tmpbuf);
		return ret;
}

/* read all of nvram data from flash into buffer */
int _nvram_read(char *buf)
{
	int ret = 0;
	char *zbuf = NULL;
	z_stream zlib_stream;
	struct nvram_header *flash_header;
	int header_offset = NVRAM_HEADER_SIZE;
	//u_int32_t nvram_flash_addr;

	do {
		if (!(zbuf = malloc(NVRAM_SPACE))) {
			ret = -ENOMEM;
			break;
		}

		memset(zbuf, 0x0, NVRAM_SPACE);

		nvram_flash_read_zbuf(zbuf, NVRAM_SPACE);

		/* Check magic number */
		flash_header = (struct nvram_header *) zbuf;
		{
			if (flash_header->magic != NVRAM_MAGIC)
			{
				printk("Invalid nvram data\n");
				break;
			}
		}

		/* verify chksum */
		if (((struct nvram_header *)zbuf)->chksum_protected)
		{
			if (verify_zbuf_chksum(zbuf))
			{
				printk("nvram_init: invalid chksum, corrupt nvram detected.\n");
				break;
			}
		}
		else
		{
			/* keep it back compatible */
			flash_header->len_compressed = NVRAM_SPACE - NVRAM_HEADER_SIZE;
		}

		nvram_space_used = flash_header->space_used;
		/* copy header */
		memcpy(buf, zbuf, NVRAM_HEADER_SIZE);

		zlib_stream.zalloc = nvram_zalloc;
		zlib_stream.zfree = nvram_zfree;

		ret = inflateInit2(&zlib_stream, MAX_WBITS);
		if (ret != Z_OK) {
			printf ("%s: Error: inflateInit2() returned %d\n", __func__, ret);
			ret = -1;
			break;
		}
		zlib_stream.next_in = (unsigned char*)zbuf+NVRAM_HEADER_SIZE;
		zlib_stream.avail_in = ((struct nvram_header *)zbuf)->len_compressed; //NVRAM_SPACE-NVRAM_HEADER_SIZE;
		zlib_stream.next_out = (unsigned char*)buf + header_offset; //p_inflate_stream->data;
		zlib_stream.avail_out = NVRAM_BUFFER_SPACE - header_offset;

		ret = inflate(&zlib_stream, Z_SYNC_FLUSH);
		printk("Inflate: %d, total_out=%ld\n", ret, zlib_stream.total_out);
		inflateEnd(&zlib_stream);

		if (ret != Z_STREAM_END) {
			printk("nvram_init: corrupt nvram detected when decompress\n");
			memset(buf + header_offset, 0x0, NVRAM_BUFFER_SPACE - header_offset);
			break; /* premature inflate loop termination */
		}
		printk("nvram: %d -> %d (%d%% -> %d%%)\n",
			flash_header->len_compressed,
			flash_header->len,
			(flash_header->len_compressed * 100) / NVRAM_SPACE,
			(flash_header->len * 100) / NVRAM_BUFFER_SPACE);

		ret = 0;

	} while (0);

	if (zbuf) free(zbuf);

	return ret;
}

/* return NULL pointer on failure */
struct nvram_tuple *_nvram_realloc(struct nvram_tuple *t,
	const char *name, const char *value, int *value_changed)
{
	/* Offset of 3 added to account for flag char, '=', and \0 in the commitable nvram buffer.
	 * Note that if free space tracking is off here, we will potentially add more tuples to the hash
	 * table than we have room for in the nvram buffer, causing tuples to be lost when _nvram_commit
	 * is called.
	 */
	if ( ( nvram_offset + strlen( value ) + strlen( name ) + 3 ) > NVRAM_BUFFER_SPACE )
	{
		/* We are run out of buffer space */
		return NULL;
	}

	if ( t == NULL )
	{
		if ( !( t = kmalloc( sizeof ( struct nvram_tuple ) + strlen( name ) + 1, GFP_KERNEL ) ) )
		{
			return NULL;
		}

		t->name = (char *)&t[1];
		nvram_offset += strlen( name ) + 2; /* nul + flag char */
		strcpy( t->name, name );

		t->flag = 0;
		t->value = NULL;
		if ( value_changed != NULL ) {
			*value_changed = 1;
		}
	}

	/* Copy value */
	if ( !t->value ||
		 strlen( t->value ) != strlen( value ) ||
		 strcmp( t->value, value ) )
	{
		if ( t->value && ( strlen( value ) <= strlen( t->value ) ) )
		{
			strcpy( t->value, value );
		}
		else
		{
			/* SXIE: we will get unused fragment here in the size not enough case!!! */
			/* SCF: unused fragment will be reclaimed on commit */
			t->value = &nvram_buf[ nvram_offset ];
			strcpy( t->value, value );
			nvram_offset += strlen( value ) + 1;
		}
		if ( value_changed != NULL ) {
			*value_changed = 1;
		}
	}

	return t;
}

void _nvram_free(struct nvram_tuple *t)
{
	if (!t)
	{
		//printf("_nvram_free setting nvram_offset = 0\n");
		nvram_offset = 0;
	}
	else
		free(t);
}

int nvram_set(const char *name, const char *value)
{
	int ret;
	struct nvram_header *header;

	if ((ret = _nvram_set(name, value, 1)) == -ENOMEM) {
		/* Consolidate space and try again */
		if ((header = kmalloc(NVRAM_BUFFER_SPACE, GFP_KERNEL))) {
			if (_nvram_commit(header) == 0)
				ret = _nvram_set(name, value, 1);
			free(header);
		}
	}

	return ret;
}

char *real_nvram_get(const char *name)
{
	char *value;

	value = _nvram_get(name);
	return value;
}

char *nvram_get(const char *name)
{
	return real_nvram_get(name);
}

int nvram_unset(const char *name)
{
	int ret;

	ret = _nvram_unset(name);

	return ret;
}

int nvram_setacl(const char *name)
{
	int ret;

	ret = _nvram_setacl(name);

	return ret;
}

void nvram_commit(void)
{
	char *buf;
	char *zbuf;
	int ret;
	struct nvram_header *header;
	z_stream zlib_stream;

	printf("nvram_commit()\n");

	/* buffer for uncompressed data */
	if (!(buf = malloc(NVRAM_BUFFER_SPACE))) {
		printf( "NVRAM commit out-of-memory ( %d )\n", __LINE__);
		return;
	}

	/* buffer for compressed data */
	if (!(zbuf = malloc(NVRAM_SPACE-NVRAM_HEADER_SIZE))) {
		free(buf);
		printf( "NVRAM commit out-of-memory ( %d )\n", __LINE__);
		return;
	}
	memset(zbuf, 0, NVRAM_SPACE-NVRAM_HEADER_SIZE);

	header = (struct nvram_header *)buf;

	/* Regenerate NVRAM */
	ret = _nvram_commit(header); /* create copy of nvram data in header */

	if (ret || !nvram_dirty) {
		free(buf);
		free(zbuf);
		printf("nvram_commit: commit abort, ret=%x, nvram_dirty=%x\n", ret, nvram_dirty);
		return;
	}
	nvram_dirty = 0;

	nvram_space_used = header->space_used =
		(100*(header->len))/(NVRAM_BUFFER_SPACE-NVRAM_HEADER_SIZE);

	zlib_stream.zalloc = nvram_zalloc;
	zlib_stream.zfree = nvram_zfree;
	zlib_stream.opaque = Z_NULL;

	ret = deflateInit2_(&zlib_stream, Z_BEST_COMPRESSION,
		Z_DEFLATED, MAX_WBITS, MAX_MEM_LEVEL, Z_DEFAULT_STRATEGY,
			ZLIB_VERSION, sizeof(z_stream));
	if (ret != Z_OK) {
		printf ("Error: deflateInit2_() returned %d\n", ret);
		return;
	}
	zlib_stream.next_in = (unsigned char *)buf+NVRAM_HEADER_SIZE;
	zlib_stream.avail_in = header->len;
	zlib_stream.next_out = (unsigned char *)zbuf + NVRAM_HEADER_SIZE;
	zlib_stream.avail_out = NVRAM_SPACE-NVRAM_HEADER_SIZE;

	ret = deflate(&zlib_stream, Z_FINISH);
	if (ret == Z_STREAM_ERROR) {
		printf("Error: deflate() returned %d\n", ret);
			ret = -1;
			goto done;
	}

	header->magic = NVRAM_MAGIC;
	/* SXIE: we record the bigger one between compressed and uncompressed */
	if (header->space_used < (100*(zlib_stream.total_out))/(NVRAM_SPACE-NVRAM_HEADER_SIZE))
		nvram_space_used = header->space_used =
			(100*(zlib_stream.total_out))/(NVRAM_SPACE-NVRAM_HEADER_SIZE);
	/* len_compressed is needed for calculate checksum below */
	header->len_compressed = zlib_stream.total_out;

	apply_zbuf_chksum((struct nvram_header*)buf,zbuf + NVRAM_HEADER_SIZE);
	header->chksum_protected = 1;
	header->write_counter = ++current_write_counter;

	/* add header to zbuf */
	memcpy(zbuf, header, NVRAM_HEADER_SIZE);

	nvram_flash_write_zbuf(zbuf, NVRAM_SPACE);

	ret = header->len; //success

done:
	deflateEnd(&zlib_stream);
	free(zbuf);
	free(buf);
	return;
}

int nvram_getall(char *buf, int count)
{
	int ret;

	ret = _nvram_getall(buf, count);

	return ret;
}

int nvram_eraseall(void)
{
	int ret;

	ret = _nvram_purge_unlocked();
	nvram_dirty = 1;

	return ret;
}

/* User mode interface below */
static ssize_t dev_nvram_read_safe(char * buf, size_t count)
{
	char *name, *value;
	ssize_t ret;
	unsigned long off;
	char *saved_name_p;
	int len;

	if (!(name = malloc(count)))
		return -ENOMEM;

	saved_name_p = name;

	if (memcpy(name, buf, count))
	{
		ret = -EFAULT;
		goto done;
	}

	/* sanity check input */
	if ( strnlen( name, count ) == count )
	{
		ret = -EINVAL;
		goto done;
	}

	if (*name == '\0')
	{
		/* Get all variables */
		ret = nvram_getall(name, count);
		if (ret == 0)
		{
			if (memcpy(buf, name, count))
			{
				ret = -EFAULT;
				goto done;
			}
			ret = count;
		}
	}
	else
	{
		if (!(value = nvram_get(name)))
		{
			ret = 0;
			goto unlock_done;
		}

		/* Provide the offset into mmap() space */
		off = (unsigned long) value - (unsigned long) nvram_buf;

		len = strnlen( &nvram_buf[ off ], count );
		if ( len == count )
		{
			ret = -ENOSPC;
			goto unlock_done;
		}

		if ( memcpy( buf, &nvram_buf[ off ], len+1 ) != 0 )
		{
			ret = -EFAULT;
			goto unlock_done;
		}

		ret = len+1;
	}

unlock_done:
done:
    free(saved_name_p);

	return ret;
}

ssize_t dev_nvram_read(char *buf, size_t count)
{
	return dev_nvram_read_safe( buf, count );
}

ssize_t dev_nvram_write(const char *buf, size_t count)
{
	char tmp[100], *name = tmp, *value;
	ssize_t ret;

	if (count > sizeof(tmp)) {
		if (!(name = kmalloc(count, GFP_KERNEL)))
			return -ENOMEM;
	}

	if (memcpy(name, buf, count)) {
		ret = -EFAULT;
		goto done;
	}

	/* SCF - check for NULL termination, otherwise assume malformed input and bail */
	if (name[count-1]!='\0') {
		printf("Name/Value is invalid\n");
		ret = -EINVAL;
		goto done;
	}

	value = name;
	name = strsep(&value, "=");
	if (name[0] == MARK_L1LOCKED || (name[0] == MARK_L2LOCKED))
	{
		if (value)
		{
			//printf("set value with locked api\n");
			ret = nvram_set(name, value) ? : count;
		}
		else
		{
			//printf("set locked level\n");
			ret = nvram_setacl(name) ? : count;
		}
	}
	else
	{
		if (value)
		{
			ret = nvram_set(name, value) ? : count;
		}
		else
		{
			ret = nvram_unset(name) ? : count;
		}
	}

done:
	if (name != tmp)
		free(name);

	return ret;
}

/* Initialize nvram_flash_blocks table. */
static void init_nvram_block_table(void)
{
	int i;
	for (i = 0; i < NVRAM_NUM_BLOCKS; i++) {
		nvram_flash_blocks[i].start_addr = NVRAM_BLOCK1_ADDR + mtd->erasesize * i;
		nvram_flash_blocks[i].part_count = NVRAM_NUM_SUBPARTS;
	}
}

int dev_nvram_init(void)
{
	char mtd_dev[16];
	struct mtd_device *dev;
	struct part_info *part;
	u8 pnum;
	struct mtd_partition mtd_part;
	struct mtd_info *mtd_info;
	char buffer[80];
	static unsigned int nvram_initialized = 0;

	if ( nvram_initialized )
	{
		printf("NVRAM Driver already initialized\n");
		return 0;
	}

	if (mtdparts_init() != 0) {
		printf("could not initialize device list\n");
		return 0;
	}

	if (find_dev_and_part( "nvram", &dev, &pnum, &part)) {
		printf("Partition nvram not found!\n");
		return 1;
	}
	sprintf(mtd_dev, "%s%d", MTD_DEV_TYPE(dev->id->type), dev->id->num);
	mtd_info = get_mtd_device_nm(mtd_dev);

	if ( mtd_info <= 0 )
	{
		printk("dev_nvram_init: can't get mtd device\n");
		return -ENODEV;
	}

	sprintf(buffer, "mtd=%d", pnum);
	memset(&mtd_part, 0, sizeof(mtd_part));
	mtd_part.name = buffer;
	mtd_part.size = part->size;
	mtd_part.offset = part->offset;
	add_mtd_partitions(mtd_info, &mtd_part, 1);
	mtd = get_mtd_device_nm(buffer);
	if ( mtd <= 0 )
	{
		printk("dev_nvram_init: can't get mtd device\n");
		return -ENODEV;
	}
	init_nvram_block_table();
	/* Initialize hash table */
	_nvram_init();
	nvram_initialized = 1;
	printf("NVRAM Driver Initialized\n");
	return 0;
}

